import os
import time
import warnings

import numpy as np
import tensorflow as tf
from skimage.io import imread, imsave
from skimage.transform import resize
from tensorflow.python.framework import convert_to_constants
from tensorflow.python.saved_model import tag_constants


# For one inference process
class SavedModelInferrer():
    def __init__(self, input_params):
        self.image_filenames = input_params["image_filenames"]
        self.output_dir = input_params["output_dir"]
        self.annotations = input_params["annotations"]
        self.model_dir = input_params["model_dir"]
        self.batch_size = input_params["batch_size"]

        self.batch_count = int(np.ceil(len(self.image_filenames) / float(self.batch_size)))
        self.model_func = self._load_model_func()
        pass

    ###############################################################################

    # only normalize as preprocessing option
    def predict_single_image(self, image):
        image_as_batch = np.expand_dims(image, axis=0)
        preprocessed_input = image_as_batch / 255.
        return self.model_func(tf.convert_to_tensor(preprocessed_input, dtype='float32'))

    def infer(self, target_shape):
        print("Starting Inference ...")

        for i in range(0, self.batch_count):
            print("Batch {}:".format(i))
            before_batch = time.perf_counter()

            current_batch = self.get_batch(i, target_shape)

            output = self.model_func(current_batch)

            batch_time = time.perf_counter() - before_batch
            print("Batch load and predict time: {}s -- {}s per image".format(batch_time,
                                                                             batch_time / self.batch_size))
            self._save_batch(output, i)
        pass

    def get_batch(self, index, target_shape):
        my_slice = self.image_filenames[index * self.batch_size: (1 + index) * self.batch_size]
        batch = np.array(
            [resize(imread(img), target_shape) for img in my_slice])

        preprocessed_input = batch / 255.
        return preprocessed_input

    ###############################################################################

    def _load_model_func(self):
        tf.keras.backend.set_learning_phase(0)
        print("Loading saved model...")
        saved_model_loaded = tf.saved_model.load(self.model_dir, tags=[tag_constants.SERVING])
        signature_keys = list(saved_model_loaded.signatures.keys())
        print("Signature keys:")
        print(signature_keys)
        graph_func = saved_model_loaded.signatures['serving_default']
        frozen_func = convert_to_constants.convert_variables_to_constants_v2(
            graph_func)

        return frozen_func

    ###############################################################################

    def _make_filepath(self, filenames_index, postfix=""):
        my_postfix = postfix
        if postfix != "":
            my_postfix = "_" + postfix

        only_filename = os.path.split(self.image_filenames[filenames_index])[1]
        new_filename = "{}_pred{}.jpg".format(only_filename.strip('.jpg'), my_postfix)
        img_path = os.path.join(self.output_dir, new_filename)
        return img_path

    def _imsave_no_warning(self, filename, content):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            imsave(filename, content)
        return

    def _save_batch(self, model_output, index, postfix=""):
        count = index * self.batch_size

        for element in model_output:
            new_filename = self._make_filepath(count, postfix)
            self._imsave_no_warning(new_filename, np.sum(element, axis=2))
            count += 1
        pass
