# Adapted from ImageZMQ examples (https://github.com/jeffbass/imagezmq)

import zmq

from deploy.server.base.imagezmq import SerializingContext
from misc_util.logged_object import LoggedObject
from misc_util.timer import log_time


# we use only one context and hope this leads to less problems
class ZMQHubFactory():
    context = SerializingContext()

    @staticmethod
    def get_REQ(connect_to):
        zmq_socket = ZMQHubFactory.context.socket(zmq.REQ)
        zmq_socket.connect(connect_to)
        return zmq_socket

    @staticmethod
    def get_REP(open_port):
        zmq_socket = ZMQHubFactory.context.socket(zmq.REP)
        zmq_socket.bind(open_port)
        return zmq_socket


# class to distribute data to several servers asynchronously
# using the ZMQ polling mechanism to collect answers
class WorkDistributorHub(LoggedObject):
    def __init__(self, server_list, compressed=False, do_logging=False):
        super().__init__(do_logging)
        self.server_list = server_list
        self.sockets = []
        self.compressed = compressed
        self.poller = zmq.Poller()
        self._init_sockets()
        return

    def _init_sockets(self):
        for server in self.server_list:
            new_socket = ZMQHubFactory.get_REQ(server)
            self.poller.register(new_socket, zmq.POLLIN)
            self.sockets.append(new_socket)
        return

    @log_time
    def distribute_data(self, msg, data):
        for socket in self.sockets:
            self.logger.debug("Sending data with message {}".format(msg))
            socket.send_data(data, msg, copy=False, compressed=self.compressed)

    @log_time
    def collect_replies(self):
        not_done = True
        check_sockets = {id(socket): False for socket in self.sockets}
        replies = {}
        while not_done:
            socks = dict(self.poller.poll())

            for socket in self.sockets:
                if socks.get(socket) == zmq.POLLIN:
                    msg, data = socket.recv_data()
                    check_sockets[id(socket)] = True
                    replies[msg] = data
                    self.logger.debug("Received data from {}".format(msg))

            if False in check_sockets.values():
                not_done = True
            else:
                not_done = False

        return replies
