import io

import numpy as np
import cv2
from PIL import Image


def demean_heatmap(heatmap):
    inf_mean = np.mean(heatmap)
    demeaned_heatmap = heatmap - inf_mean
    return demeaned_heatmap


def filter_three_sigma_events(heatmap):
    heatmap_in_sigmas = heatmap / np.std(heatmap)
    mask_3sigma = np.where(heatmap_in_sigmas < 3.0)
    heatmap_in_sigmas[mask_3sigma] = 0
    return heatmap_in_sigmas


def rescale_heatmap(heatmap):
    inf_max = np.max(heatmap)
    rescaled_heatmap = heatmap / inf_max * 255.
    return rescaled_heatmap


def decompress_image_pil(encoded_image) -> np.ndarray:
    image = Image.open(io.BytesIO(encoded_image))
    return np.array(image)


def decompress_image_cv2(image) -> np.ndarray:
    image = cv2.imdecode(np.fromstring(image, dtype='uint8'), -1)
    return image


def compress_image(image, quality, method='pil'):
    image_as_raw = None
    if method == 'pil':
        if type(image) is Image:
            image_as_raw = compress_image_pil(image, quality)
        elif type(image) is np.ndarray:
            image_as_raw = compress_image_array_pil(image, quality)
    else:
        raise ValueError("Unknown compression method.")
    return image_as_raw


def compress_image_pil(image: Image, quality=95) -> bytes:
    output = io.BytesIO()
    image.save(output, format="JPEG", quality=quality, subsampling=0)
    image_as_raw = output.getvalue()
    return image_as_raw


# from the pil doc:
# quality
# The image quality, on a scale from 1 (worst) to 95 (best).
# The default is 75. Values above 95 should be avoided;
# 100 disables portions of the JPEG compression algorithm,
# and results in large files with hardly any gain in image quality.
def compress_image_array_pil(image: np.ndarray, quality=95) -> bytes:
    if len(image.shape) < 3:
        pil_img = Image.fromarray(image, mode='L')
    else:
        pil_img = Image.fromarray(image, mode='RGB')
    output = io.BytesIO()
    pil_img.save(output, format="JPEG", quality=quality, subsampling=0)
    image_as_raw = output.getvalue()
    return image_as_raw


# needs the shape as column-major
def resize_image_array_pil(image: np.ndarray, target_shape: tuple) -> np.ndarray:
    pil_img = Image.fromarray(image)
    pil_img = pil_img.resize(target_shape, Image.LANCZOS)
    return np.array(pil_img)


def resize_image_pil(image: Image, target_shape: tuple) -> Image:
    pil_img = image.resize(target_shape, Image.LANCZOS)
    return pil_img


# needs the shape as column-major
def resize_image_array_pil_antialias(image: np.ndarray, target_shape: tuple) -> np.ndarray:
    pil_img = Image.fromarray(image)
    pil_img = pil_img.resize(target_shape, Image.ANTIALIAS)
    return np.array(pil_img)
