import time
from functools import wraps

import numpy as np


def log_time(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        result = func(*args, **kwargs)
        func_time = time.perf_counter() - start
        args[0].logger.warning("Time for {}: {}".format(func.__name__, func_time))
        return result

    return wrapper


class Timer:
    def __init__(self, verbose=True, keep_record=False, record_to_file=None):
        self.sum = 0
        self.verbose = verbose
        self.before = 0
        self.after = 0
        self.started = False
        self.message = ""
        self.record_array = None
        if keep_record:
            self.record_array = []
        self.record_file = None
        if record_to_file is not None:
            self.record_file = record_to_file

    def start(self, message=""):
        self.message = message
        if self.started:
            raise Exception("Started timer before stopping it")
        self.started = True
        self.before = time.perf_counter()
        return

    def stop(self):
        self.after = time.perf_counter() - self.before
        self.sum += self.after

        if self.record_array is not None:
            self.record_array.append(self.after)

        if self.verbose:
            print("Time for {}: {}".format(self.message, self.after))
        if not self.started:
            raise Exception("Stopped timer before starting it")
        self.started = False
        self.message = ""
        return self.after

    def get_sum(self):
        return self.sum

    def get_average(self):
        if self.record_array is not None:
            return np.mean(np.array(self.record_array), dtype=np.float64)
        else:
            return None

    def store(self):
        if self.record_array is None:
            print("Timer -- Error when storing to file: Not keeping records.")
            return
        if self.record_file is not None:
            with open(self.record_file, 'w') as time_file:
                for elem in self.record_array:
                    time_file.write(str(elem) + '\n')
                # time_file.writelines([str(elem) for elem in self.record_array])
                time_file.write("arithmetic avg\n")
                time_file.write(str(self.get_average()))

            time_file.close()
        else:
            print("Timer -- Error when storing to file: Invalid filename.")

    def reset(self):
        self.sum = 0
        self.before = 0
        self.after = 0
        self.message = ""
        return
