###############################################################################
# Configs for architectures
from research.ablation_studies.dataset.dataset_utils import gather_tfrecords
from research.ablation_studies.dataset.pipeline_creator import PipelineCreator, ROIInputMap, PatchInputMap
from research.ablation_studies.model.create_keypoint_models import create_model, get_preprocess_func
from research.ablation_studies.train.arch_train_scripts import train_arch

###############################################################################
# Let training run for all configs
from research.ablation_studies.train.train_config import TrainConfig


def run_stage_trains_for_arch(arch, stage, tfrecord_files):
    if arch not in TrainConfig.configs:
        print("Could not find architecture {} in training list. Skipping.".format(arch))
        return

    if stage not in TrainConfig.configs[arch]["stages"]:
        print("Could not find stage {} for arch {} in training list. Skipping.".format(stage, arch))
        return

    stage_runs = TrainConfig.configs[arch]["stages"][stage]

    for entry in stage_runs["runs"]:
        run_train_entry(arch, stage, entry, tfrecord_files)
    return


def run_train_entry(arch_key, stage, entry, tfrecord_files):
    stage_entry = TrainConfig.configs[arch_key]["stages"][stage]
    model_input_shape = stage_entry["input_shape"]
    model_output_shape = entry["output_shape"]
    ra_magnitude = stage_entry["ra_magnitude"]
    downscale_ratio = stage_entry["downscale_ratio"]
    batch_size = stage_entry["batch_size"]
    epochs = stage_entry["epochs"]

    if stage == "roi":
        input_mapper = ROIInputMap(
            model_input_shape,
            model_output_shape,
            ra_magnitude,
            downscale_ratio=downscale_ratio,
            input_prep_func=get_preprocess_func(arch_key))
    elif stage == "detail":
        input_mapper = PatchInputMap(
            model_input_shape,
            model_output_shape,
            ra_magnitude,
            downscale_ratio=downscale_ratio,
            input_prep_func=get_preprocess_func(arch_key))
    else:
        input_mapper = None

    train_ds, test_ds = PipelineCreator(
        input_mapper,
        batch_size,
        tfrecord_files,
        entry["id"]).get_datasets()

    current_train_id = arch_key + "_" + entry['id']
    upsampling_layers = entry["upsampling_layers"]
    freeze = entry["freeze"]

    print("Creating model for train id {} ...".format(current_train_id))
    model = create_model(arch_key,
                         model_input_shape,
                         model_output_shape,
                         upsampling_layers,
                         freeze)

    print("Starting training ...")
    train_arch(arch_key, train_ds, test_ds, model, current_train_id, batch_size, epochs)


if __name__ == '__main__':
    large_tfrecord_dir = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/dataset/"

    tfrecords = gather_tfrecords(tfrecord_dir=large_tfrecord_dir)
    if tfrecords is not None:
        run_stage_trains_for_arch("nasnetmobile", "detail", tfrecords)
