import os

from tensorflow.python.keras.models import load_model

from research.ablation_studies.dataset.dataset_utils import gather_tfrecords, sequence_image_feature_description
from research.ablation_studies.dataset.pipeline_creator import ClassificationSequenceInputMap, SequencePipelineCreator
from research.ablation_studies.model.create_classifier_models import compile_created_model_classifier, \
    compile_created_model_classifier_sgd, create_model_detail_effnetB0_classifier_donahue, \
    create_effnetb0_convlstm_encoder, create_full_conv_lstm
from research.ablation_studies.model.create_keypoint_models import get_preprocess_func
from research.ablation_studies.train.model_classification_trainer import ModelClassificationTrainer

if __name__ == '__main__':
    # tf.debugging.experimental.enable_dump_debug_info(
    #    dump_root="/home/florin/tb_db_logs",
    #    tensor_debug_mode="FULL_HEALTH",
    #    circular_buffer_size=-1)

    contine_training = False
    continue_file = None  # "/home/florin/diss_models/effnetb0_classify_rnn_diss_fconv/rnn_ud0_ts3_shift3_lstm64_fconv_adam/1_chkpt/rnn_ud0_ts3_shift3_lstm64_fconv_adam_prauc.046.h5"

    arch = "effnetb0"
    config_id = "{}_classify_rnn_diss".format(arch)

    use_conv = True
    conv_lstm_size = 512
    if use_conv:
        config_id = config_id + "_conv"

    use_full_conv = False
    if use_full_conv:
        config_id = config_id + "_fconv"

    constant_lr = None  # 1e-4#0.2*1e-3
    initial_lr = 1e-3
    batch_size = 8
    time_steps = 30
    shift = 30
    sequence_length = 60
    use_augmentation = True

    lstm_size = 64
    if use_conv:
        lstm_size = conv_lstm_size

    unfreeze_blocks = (7,)
    dense_trainable = True

    freeze_bns = True

    do_basic_imagenet = False

    use_sgd = False
    use_gru = False

    if use_gru:
        cell_string = "gru"
    else:
        cell_string = "lstm"

    input_shape = (224, 224, 3)
    # train_prefix_id = "rnn_ab2_ts{}_shift{}_{}{}_7dtbnf_normlr3_conv_aug".format(time_steps, shift, cell_string, lstm_size)
    train_prefix_id = "rnn_ab2_ts{}_shift{}_convlstm{}_7dtbnf_normlr3_aug".format(time_steps, shift, lstm_size)

    if use_sgd:
        train_prefix_id = train_prefix_id + "_sgd"
    else:
        train_prefix_id = train_prefix_id + "_adam"

    if shift == 1:  # one-off-shift leads to a lot more mini-sequences
        train_item_size = (11160 / sequence_length) * (sequence_length - time_steps + 1)
        test_item_size = (2880 / sequence_length) * (sequence_length - time_steps + 1)
    elif shift == time_steps:  # separating available frames by time_steps is normal case
        train_item_size = 11160 / time_steps
        test_item_size = 2880 / time_steps
    elif shift % time_steps == 0 and shift > time_steps:  # leaving out every other mini-sequence
        shift_steps = shift // time_steps
        train_item_size = 11160 / time_steps / shift_steps
        test_item_size = 2880 / time_steps / shift_steps
        print("Using shift step split.")
    elif shift == time_steps / 2:
        mini_sequences = sequence_length / time_steps
        mini_sequences_per_full = mini_sequences + mini_sequences - 1
        train_item_size = 11160 / sequence_length * mini_sequences_per_full
        test_item_size = 2880 / sequence_length * mini_sequences_per_full
        print("Using shift step split.")

    ##########################################################################

    if do_basic_imagenet:
        pretrained_weights = None
    else:
        pretrained_weights = "/home/florin/diss_models/effnetb0_classify_basic/simple_full_set_weighted_ls0.1/1_chkpt/simple_full_set_weighted_ls0.1_best.094.h5"

    tfrecord_dir = "/home/florin/data/sequence/"
    tfrecords = gather_tfrecords(tfrecord_dir=tfrecord_dir)

    # get datasets train/test
    input_mapper = ClassificationSequenceInputMap(
        input_shape,
        (0, 0),
        6,
        downscale_ratio=1,
        input_prep_func=get_preprocess_func(arch),
        use_augmentation=use_augmentation)

    pipeline = SequencePipelineCreator(
        input_mapper,
        batch_size,
        tfrecords,
        image_features=sequence_image_feature_description,
        timesteps=time_steps,
        shift=shift)
    ds_train, ds_test = pipeline.get_datasets()

    if not contine_training:
        if use_conv:
            my_model = create_effnetb0_convlstm_encoder(
                conv_lstm_size,
                time_steps,
                input_shape,
                config_id,
                model_path=pretrained_weights,
                unfreeze_blocks=unfreeze_blocks,
                dense_trainable=dense_trainable,
                freeze_bns=freeze_bns)
        elif not use_full_conv:
            my_model = create_model_detail_effnetB0_classifier_donahue(
                lstm_size,
                time_steps,
                input_shape,
                config_id,
                model_path=pretrained_weights,
                unfreeze_blocks=unfreeze_blocks,
                dense_trainable=dense_trainable,
                freeze_bns=freeze_bns,
                use_gru=use_gru)
        elif use_full_conv:
            my_model = create_full_conv_lstm(time_steps)
    else:
        my_model = load_model(continue_file)

    if use_sgd:
        my_model = compile_created_model_classifier_sgd(my_model)
    else:
        my_model = compile_created_model_classifier(my_model)

    if use_full_conv:
        my_model.summary()

    ##########################################################################

    model_save_path = "/home/florin/diss_models/{}/{}/1/".format(config_id, train_prefix_id)
    os.makedirs(model_save_path, exist_ok=True)
    print("Storing model in {}".format(model_save_path))

    checkpoint_save_path = "/home/florin/diss_models/{}/{}/1_chkpt/".format(config_id, train_prefix_id)
    os.makedirs(checkpoint_save_path, exist_ok=True)
    print("Storing checkpoints in {}".format(checkpoint_save_path))

    ##########################################################################

    trainer = ModelClassificationTrainer(
        ds_train, ds_test,
        batch_size,
        my_model,
        train_prefix_id,
        checkpoint_save_path,
        train_item_size=train_item_size,
        test_item_size=test_item_size,
        train_neg_weight=0.7492,
        constant_lr=constant_lr,
        initial_lr=initial_lr)

    print("Starting training!")
    trainer.train(200)

    print("Saving model ...")

    my_model.save(model_save_path, save_format='tf')
