import os

import numpy as np
import tensorflow as tf


# from https://github.com/tensorflow/models/blob/master/research/slim/datasets/dataset_utils.py
def bytes_feature(value):
    """Returns a TF-Feature of bytes.
    Args:
      values: A string.
    Returns:
      A TF-Feature.
    """
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


# from https://github.com/tensorflow/models/blob/master/research/slim/datasets/dataset_utils.py
def float_feature(values):
    """Returns a TF-Feature of floats.
    Args:
      values: A scalar of list of values.
    Returns:
      A TF-Feature.
    """
    if not isinstance(values, (tuple, list)):
        values = [values]
    return tf.train.Feature(float_list=tf.train.FloatList(value=values))


# from https://github.com/tensorflow/models/blob/master/research/slim/datasets/dataset_utils.py
def int64_feature(values):
    """Returns a TF-Feature of int64s.
    Args:
      values: A scalar or list of values.
    Returns:
      A TF-Feature.
    """
    if not isinstance(values, (tuple, list)):
        values = [values]
    return tf.train.Feature(int64_list=tf.train.Int64List(value=values))


image_feature_description = {
    'image/encoded': tf.io.FixedLenFeature([], tf.string),
    'image/format': tf.io.FixedLenFeature([], tf.string),
    'image/subject_id': tf.io.FixedLenFeature([], tf.int64),
    'image/task_id': tf.io.FixedLenFeature([], tf.int64),
    'image/frame_id': tf.io.FixedLenFeature([], tf.int64),
    'image/ptip_x': tf.io.FixedLenFeature([], tf.float32),
    'image/ptip_y': tf.io.FixedLenFeature([], tf.float32),
    'image/ptail_x': tf.io.FixedLenFeature([], tf.float32),
    'image/ptail_y': tf.io.FixedLenFeature([], tf.float32),
    'image/height': tf.io.FixedLenFeature([], tf.int64),
    'image/width': tf.io.FixedLenFeature([], tf.int64),
}

sequence_image_feature_description = {
    'image/encoded': tf.io.FixedLenFeature([], tf.string),
    'image/format': tf.io.FixedLenFeature([], tf.string),
    'image/subject_id': tf.io.FixedLenFeature([], tf.int64),
    'image/task_id': tf.io.FixedLenFeature([], tf.int64),
    'image/frame_id': tf.io.FixedLenFeature([], tf.int64),
    'image/sequence_id': tf.io.FixedLenFeature([], tf.int64),
    'image/ptip_x': tf.io.FixedLenFeature([], tf.float32),
    'image/ptip_y': tf.io.FixedLenFeature([], tf.float32),
    'image/ptail_x': tf.io.FixedLenFeature([], tf.float32),
    'image/ptail_y': tf.io.FixedLenFeature([], tf.float32),
    'image/pen_down': tf.io.FixedLenFeature([], tf.int64),
    'image/height': tf.io.FixedLenFeature([], tf.int64),
    'image/width': tf.io.FixedLenFeature([], tf.int64),
}

eval_feature_description = {
    'image/subject_id': tf.io.FixedLenFeature([], tf.int64),
    'image/task_id': tf.io.FixedLenFeature([], tf.int64),
    'image/frame_id': tf.io.FixedLenFeature([], tf.int64),
    'image/ptip_x': tf.io.FixedLenFeature([], tf.float32),
    'image/ptip_y': tf.io.FixedLenFeature([], tf.float32),
    'image/ptail_x': tf.io.FixedLenFeature([], tf.float32),
    'image/ptail_y': tf.io.FixedLenFeature([], tf.float32),
    'image/height': tf.io.FixedLenFeature([], tf.int64),
    'image/width': tf.io.FixedLenFeature([], tf.int64),
    'prediction/height': tf.io.FixedLenFeature([], tf.int64),
    'prediction/width': tf.io.FixedLenFeature([], tf.int64),
    'prediction/ptip_median': tf.io.FixedLenFeature([], tf.float32),
    'prediction/ptail_median': tf.io.FixedLenFeature([], tf.float32),
    'prediction/ptip_heatmap_encoded': tf.io.FixedLenFeature([], tf.string),
    'prediction/ptail_heatmap_encoded': tf.io.FixedLenFeature([], tf.string)
}

bin_set_pair_feature_description = {
    'image/frame_id': tf.io.FixedLenFeature([], tf.int64),
    'image/subject_id': tf.io.FixedLenFeature([], tf.int64),
    'image/encoded': tf.io.FixedLenFeature([], tf.string),
    'bin/encoded': tf.io.FixedLenFeature([], tf.string),
    'image/height': tf.io.FixedLenFeature([], tf.int64),
    'image/width': tf.io.FixedLenFeature([], tf.int64)
}


def image_to_tfexample(image_data,
                       image_format,
                       subject_id,
                       frame_id,
                       task_id,
                       height,
                       width,
                       ptip_x,
                       ptip_y,
                       ptail_x,
                       ptail_y):
    return tf.train.Example(features=tf.train.Features(feature={
        'image/encoded': bytes_feature(image_data),
        'image/format': bytes_feature(image_format),
        'image/subject_id': int64_feature(subject_id),
        'image/task_id': int64_feature(task_id),
        'image/frame_id': int64_feature(frame_id),
        'image/ptip_x': float_feature(ptip_x),
        'image/ptip_y': float_feature(ptip_y),
        'image/ptail_x': float_feature(ptail_x),
        'image/ptail_y': float_feature(ptail_y),
        'image/height': int64_feature(height),
        'image/width': int64_feature(width),
    }))


def image_to_sequence_tfexample(image_data,
                                image_format,
                                subject_id,
                                frame_id,
                                task_id,
                                sequence_id,
                                height,
                                width,
                                ptip_x,
                                ptip_y,
                                ptail_x,
                                ptail_y,
                                pen_down):
    return tf.train.Example(features=tf.train.Features(feature={
        'image/encoded': bytes_feature(image_data),
        'image/format': bytes_feature(image_format),
        'image/subject_id': int64_feature(subject_id),
        'image/task_id': int64_feature(task_id),
        'image/frame_id': int64_feature(frame_id),
        'image/sequence_id': int64_feature(sequence_id),
        'image/ptip_x': float_feature(ptip_x),
        'image/ptip_y': float_feature(ptip_y),
        'image/ptail_x': float_feature(ptail_x),
        'image/ptail_y': float_feature(ptail_y),
        'image/pen_down': int64_feature(pen_down),
        'image/height': int64_feature(height),
        'image/width': int64_feature(width),
    }))


def image_to_eval_tfexample(
        subject_id,
        frame_id,
        task_id,
        ptip_x,
        ptip_y,
        ptail_x,
        ptail_y,
        height,
        width,
        pred_height,
        pred_width,
        ptip_median,
        ptail_median,
        pred_heatmap_ptip_bytes,
        pred_heatmap_ptail_bytes):
    return tf.train.Example(features=tf.train.Features(feature={
        'image/subject_id': int64_feature(subject_id),
        'image/task_id': int64_feature(task_id),
        'image/frame_id': int64_feature(frame_id),
        'image/ptip_x': float_feature(ptip_x),
        'image/ptip_y': float_feature(ptip_y),
        'image/ptail_x': float_feature(ptail_x),
        'image/ptail_y': float_feature(ptail_y),
        'image/height': int64_feature(height),
        'image/width': int64_feature(width),
        'prediction/height': int64_feature(pred_height),
        'prediction/width': int64_feature(pred_width),
        'prediction/ptip_median': float_feature(ptip_median),
        'prediction/ptail_median': float_feature(ptail_median),
        'prediction/ptip_heatmap_encoded': bytes_feature(pred_heatmap_ptip_bytes),
        'prediction/ptail_heatmap_encoded': bytes_feature(pred_heatmap_ptail_bytes)
    }))


def bin_set_pair_to_tfexample(frame_id,
                              subject_id,
                              image_bytes,
                              bin_bytes,
                              height,
                              width):
    return tf.train.Example(features=tf.train.Features(feature={
        'image/frame_id': int64_feature(frame_id),
        'image/subject_id': int64_feature(subject_id),
        'image/encoded': bytes_feature(image_bytes),
        'bin/encoded': bytes_feature(bin_bytes),
        'image/height': int64_feature(height),
        'image/width': int64_feature(width),
    }))


def write_tfrecord_with_examples(filename, examples):
    with tf.io.TFRecordWriter(filename) as writer:
        for example in examples:
            writer.write(example.SerializeToString())
    return


def read_tfrecords(filenames):
    raw_dataset = tf.data.TFRecordDataset(filenames)
    return raw_dataset


def get_tfrecord_eval_dataset(filenames):
    eval_ds = get_tfrecord_dataset(filenames, eval_feature_description)
    return eval_ds


def get_tfrecord_dataset(filenames, feature_description):
    parsed_dataset = read_tfrecords(filenames)
    eval_ds = parsed_dataset.map(
        lambda x: tf.io.parse_single_example(x, feature_description))
    return eval_ds


def load_jpg(image_path):
    img = tf.io.read_file(image_path)
    img = tf.io.decode_jpeg(img, channels=3)
    img = tf.image.convert_image_dtype(img, tf.float32)
    return img


def load_and_resize_jpg(image_path, downscaling_ratio, target_shape):
    img = tf.io.read_file(image_path)
    img = tf.io.decode_jpeg(img, channels=3, ratio=downscaling_ratio)
    img = tf.image.convert_image_dtype(img, tf.float32)

    tf.debugging.assert_greater_equal(tf.shape(img)[0], target_shape[0], name='load_and_resize_jpg_height_check')
    tf.debugging.assert_greater_equal(tf.shape(img)[1], target_shape[1], name='load_and_resize_jpg_width_check')

    img = tf.image.resize(img, target_shape)
    tf.debugging.assert_type(img, tf.float32, name='load_and_resize_jpg_float32_check')
    return img


def merge_images(image_batch, size):
    h, w = image_batch.shape[1], image_batch.shape[2]
    c = image_batch.shape[3]
    img = np.zeros((int(h * size[0]), w * size[1], c))
    for idx, im in enumerate(image_batch):
        i = idx % size[1]
        j = idx // size[1]
        img[j * h:j * h + h, i * w:i * w + w, :] = im
    return img


def gather_tfrecords(tfrecord_dir):
    print("Gathering tfrecords ...")
    tfrecord_files = [os.path.join(tfrecord_dir, file) for file in os.listdir(tfrecord_dir) if '.tfrecord' in file]
    if len(tfrecord_files) == 0:
        print("No tfrecord files found. Exiting.")
        return None
    return tfrecord_files


def determine_downscale_ratio(img_h, target_h):
    test_h = img_h
    ratio = 1
    while (test_h // 2) > target_h:
        test_h = test_h // 2
        ratio *= 2
    return ratio


def get_max_of_heatmap(heatmap):
    values, indices = tf.nn.top_k(tf.reshape(heatmap, (-1,)), k=1)
    indices_2d = tf.unravel_index(indices, tf.shape(heatmap)[:2])
    return values, indices_2d


def prepare_dataset(tfrecords, feature_description):
    raw_dataset = read_tfrecords(tfrecords)
    parsed_dataset = raw_dataset.map(lambda x: tf.io.parse_single_example(x, feature_description))
    return parsed_dataset


def clone_sequence_example_for_disk(example):
    example = image_to_sequence_tfexample(example['image/encoded'],
                                          b'jpg',
                                          example['image/subject_id'],
                                          example['image/frame_id'],
                                          example['image/task_id'],
                                          example['image/sequence_id'],
                                          example['image/height'],
                                          example['image/width'],
                                          example['image/ptip_x'],
                                          example['image/ptip_y'],
                                          example['image/ptail_x'],
                                          example['image/ptail_y'],
                                          example['image/pen_down'])
    return example
