import numpy as np
import tensorflow as tf

from misc_util.heatmap import create_heatmaps
from research.ablation_studies.dataset.dataset_utils import read_tfrecords, image_feature_description
from research.ablation_studies.train.augmenter import Augmenter
from research.ablation_studies.train.keypoint_tf import KeypointHelper


class ROIInputMap:
    def __init__(self,
                 model_input_shape,
                 model_output_shape,
                 ra_magnitude,
                 downscale_ratio,
                 std_dev=3.0,
                 input_prep_func=None):
        self.downscale_ratio = downscale_ratio
        self.model_input_shape = model_input_shape
        self.model_output_shape = model_output_shape
        self.ra_magnitude = ra_magnitude
        self.std_dev = std_dev
        self.input_prep_func = input_prep_func
        self.num_keypoints = model_output_shape[2]
        return

    def _load_img(self, element):
        image_bytes = element['image/encoded']
        ptip_x = element['image/ptip_x']
        ptail_x = element['image/ptail_x']
        ptip_y = element['image/ptip_y']
        ptail_y = element['image/ptail_y']
        img_width = element['image/width']
        img_height = element['image/height']

        img = tf.io.decode_jpeg(image_bytes, ratio=self.downscale_ratio, channels=3)
        img = tf.image.convert_image_dtype(img, tf.float32)

        tf.debugging.assert_less_equal(tf.reduce_max(img), 1.0, name='roi_input_img_max_check')
        tf.debugging.assert_greater_equal(tf.reduce_max(img), 0.0, name='roi_input_img_min_check')

        return {'image': img,
                'ptip': [ptip_x, ptip_y],
                'ptail': [ptail_x, ptail_y],
                'width': img_width,
                'height': img_height}

    def _preprocess_example(self, element, kp_helper):
        img = element['image']
        tip_keypoint = kp_helper.from_list(element['ptip'])
        tail_keypoint = kp_helper.from_list(element['ptail'])

        t_keypoints = tf.cond(tf.greater(self.num_keypoints, 1),
                              lambda: tf.stack([tip_keypoint, tail_keypoint]),
                              lambda: tf.stack([tip_keypoint]), name='augment_img_keypoint_count')

        orig_width = tf.cast(element['width'], dtype=tf.float32, name='cast_orig_img_dim_width')
        orig_height = tf.cast(element['height'], dtype=tf.float32, name='cast_orig_img_dim_height')

        img_h = tf.cast(tf.shape(img)[0], dtype=tf.float32, name='cast_img_dim_height')
        img_w = tf.cast(tf.shape(img)[1], dtype=tf.float32, name='cast_img_dim_width')

        # since we probably opened a downscaled version of the image,
        # we need to rescale the keypoints to the current image size
        t_keypoints = tf.map_fn(
            lambda kp: kp_helper.rescale(kp, orig_width, orig_height, img_w, img_h),
            t_keypoints, name='preprocessing_keypoint_rescale')

        return (img, t_keypoints)

    def _augment_and_get_heatmaps(self, img, t_keypoints, kp_helper):
        choices = tf.random.shuffle(tf.range(12))[:2]
        augment_seed = np.random.randint(low=0, high=10000)
        augmenter = Augmenter(
            num_layers=2,
            magnitude=self.ra_magnitude,
            image_dtype=img.dtype,
            image_shape=img.shape,
            keypoint_helper=kp_helper,
            augment_seed=augment_seed,
            op_choices=choices)

        augmented_image, augmented_keypoints = augmenter.transform(image=img, keypoints=t_keypoints)
        tf.debugging.assert_less_equal(tf.reduce_max(augmented_image), 1.0, name='roi_augment_img_max_check')
        tf.debugging.assert_greater_equal(tf.reduce_max(augmented_image), 0.0, name='roi_augment_img_min_check')

        img_h = tf.cast(tf.shape(augmented_image)[0], dtype=tf.float32)
        img_w = tf.cast(tf.shape(augmented_image)[1], dtype=tf.float32)

        heatmaps = create_heatmaps(img_w, img_h,
                                   self.model_output_shape[1], self.model_output_shape[0],
                                   self.std_dev,
                                   augmented_keypoints,
                                   kp_helper)

        tf.debugging.assert_less_equal(tf.reduce_max(heatmaps), 1.0, name='roi_augment_heatmap_max_check')
        tf.debugging.assert_greater_equal(tf.reduce_max(heatmaps), 0.0, name='roi_augment_heatmap_min_check')

        return augmented_image, heatmaps

    def _augment_img(self, element):
        kp_helper = KeypointHelper(row_major=False)
        img, t_keypoints = self._preprocess_example(element, kp_helper)
        augmented_img, heatmaps = self._augment_and_get_heatmaps(img, t_keypoints, kp_helper)

        tf.assert_equal(tf.shape(img), tf.shape(augmented_img), name='roi_augment_img_shape_check')
        tf.assert_equal(tf.shape(heatmaps)[:2], self.model_output_shape[:2], name='roi_augment_heatmap_shape_check')

        resized_img = tf.image.resize(augmented_img, self.model_input_shape[:2],
                                      name='augment_img_resize_to_model_specs')
        resized_img = self.input_prep_func(resized_img)
        return (resized_img, heatmaps)

        # map_fn: creating heatmaps for test dataset
        # named dict output does not work with keras fit; keras fails to connect the y input to the model
        # resulting in no usable gradients

    def _produce_test_heatmaps(self, element):
        kp_helper = KeypointHelper(row_major=False)
        img, t_keypoints = self._preprocess_example(element, kp_helper)
        tf.debugging.assert_less_equal(tf.reduce_max(img), 1.0, name='roi_test_img_max_check')
        tf.debugging.assert_greater_equal(tf.reduce_max(img), 0.0, name='roi_test_img_min_check')

        img_h = tf.cast(tf.shape(img)[0], dtype=tf.float32)
        img_w = tf.cast(tf.shape(img)[1], dtype=tf.float32)

        heatmaps = create_heatmaps(img_w, img_h,
                                   self.model_output_shape[1], self.model_output_shape[0],
                                   self.std_dev,
                                   t_keypoints,
                                   kp_helper)
        tf.debugging.assert_less_equal(tf.reduce_max(heatmaps), 1.0, name='roi_test_heatmap_max_check')
        tf.debugging.assert_greater_equal(tf.reduce_max(heatmaps), 0.0, name='roi_test_heatmap_min_check')

        resized_img = tf.image.resize(img, self.model_input_shape[:2],
                                      name='test_img_resize_to_model_specs')
        resized_img = self.input_prep_func(resized_img)
        return (resized_img, heatmaps)


class PatchInputMap:
    def __init__(self,
                 model_input_shape,
                 model_output_shape,
                 ra_magnitude,
                 downscale_ratio,
                 use_tail=False,
                 std_dev=3.0,
                 input_prep_func=None,
                 random_offset_divisor=10):
        self.downscale_ratio = downscale_ratio
        self.model_input_shape = model_input_shape
        self.model_output_shape = model_output_shape
        self.ra_magnitude = ra_magnitude
        self.std_dev = std_dev
        self.input_prep_func = input_prep_func
        self.random_offset_divisor = random_offset_divisor
        self.use_tail = tf.constant(use_tail, dtype=tf.bool)
        return

    def _get_crop_window(self, x, y, img_width, img_height):
        '''returns a crop box containing the coordinates x, y with a random offset

        the random offset is applied in every direction, with a safe border of
        5% of the patch size to avoid shifting keypoints at the border outside
        if small augmentation operations are applied.
        '''
        patch_h, patch_w = self.model_input_shape[:2]

        coord_x = tf.cast(tf.floor(x), dtype=tf.int32)
        coord_y = tf.cast(tf.floor(y), dtype=tf.int32)

        rand_border_x = patch_w / self.random_offset_divisor
        rand_border_y = patch_h / self.random_offset_divisor
        rand_x = (tf.random.uniform([]) - 0.5) * (patch_w - rand_border_x)
        rand_y = (tf.random.uniform([]) - 0.5) * (patch_h - rand_border_y)

        rand_x = tf.cast(tf.floor(rand_x), dtype=tf.int32)
        rand_y = tf.cast(tf.floor(rand_y), dtype=tf.int32)

        coord_x = coord_x + rand_x
        coord_y = coord_y + rand_y

        img_width = tf.cast(img_width, dtype=tf.int32)
        img_height = tf.cast(img_height, dtype=tf.int32)

        patch_w_half = tf.cast(patch_w // 2, dtype=tf.int32)
        patch_h_half = tf.cast(patch_h // 2, dtype=tf.int32)
        patch_left = coord_x - patch_w_half
        patch_top = coord_y - patch_h_half
        right_band = img_width - patch_w_half
        top_band = img_height - patch_h_half

        if tf.less(coord_x, patch_w_half):
            crop_x = 0
            crop_w = patch_w - abs(patch_left)
        elif tf.greater(coord_x, right_band):
            crop_x = patch_left
            crop_w = patch_w - (coord_x - (right_band))
        else:
            # inside
            crop_x = patch_left
            crop_w = patch_w

        if coord_y < patch_h_half:
            crop_y = 0
            crop_h = patch_h - abs(patch_top)
        elif coord_y > (top_band):
            crop_y = patch_top
            crop_h = patch_h - (coord_y - (top_band))
        else:
            # inside
            crop_y = patch_top
            crop_h = patch_h

        tf.assert_greater(crop_w, 0, name='crop_window_width_check')
        tf.assert_greater(crop_h, 0, name='crop_window_height_check')
        tf.debugging.assert_greater_equal(crop_x, 0, name='crop_window_offset_x_check')
        tf.debugging.assert_greater_equal(crop_y, 0, name='crop_window_offset_y_check')
        return [crop_y, crop_x, crop_h, crop_w]

    def _get_random_crop_window(self, img_width, img_height):
        patch_h, patch_w = self.model_input_shape[:2]
        x_band = img_width - patch_w
        y_band = img_height - patch_h

        crop_x = tf.cast(tf.random.uniform([1], minval=0, maxval=x_band, dtype=tf.int64), dtype=tf.int32)[0]
        crop_y = tf.cast(tf.random.uniform([1], minval=0, maxval=y_band, dtype=tf.int64), dtype=tf.int32)[0]

        tf.assert_greater(patch_w, 0, name='random_crop_patch_width_check')
        tf.assert_greater(patch_h, 0, name='random_crop_patch_height_check')
        tf.debugging.assert_greater_equal(crop_x, 0, name='random_crop_x_offset_check')
        tf.debugging.assert_greater_equal(crop_y, 0, name='random_crop_y_offset_check')

        return [crop_y, crop_x, patch_h, patch_w]

    def _get_img_patch(self, img_bytes, keypoint, img_width, img_height, kp_helper):
        patch_h, patch_w = self.model_input_shape[:2]
        kp_is_none = KeypointHelper.is_none_keypoint(keypoint)

        # keypoint = tf.print(keypoint, [keypoint], "Keypoint: ")

        crop_window = tf.cond(kp_is_none,
                              lambda: self._get_random_crop_window(img_width, img_height),
                              lambda: self._get_crop_window(
                                  kp_helper.get_x(keypoint),
                                  kp_helper.get_y(keypoint),
                                  img_width, img_height))

        coord_x = tf.cast(tf.floor(kp_helper.get_x(keypoint)), dtype=tf.int32)
        coord_y = tf.cast(tf.floor(kp_helper.get_y(keypoint)), dtype=tf.int32)
        new_kp_x = coord_x - crop_window[1]
        new_kp_y = coord_y - crop_window[0]

        new_kp = tf.cond(kp_is_none,
                         lambda: KeypointHelper.make_none_keypoint(),
                         lambda: kp_helper.make(
                             x=new_kp_x,  # crop_window[3]/2,
                             y=new_kp_y))  # crop_window[2]/2))

        # tf.print(keypoint, crop_window)
        crop_window = tf.ensure_shape(crop_window, shape=(4,))
        patch = tf.io.decode_and_crop_jpeg(contents=img_bytes, crop_window=crop_window)
        patch = tf.image.pad_to_bounding_box(patch, 0, 0, patch_h, patch_w)
        return patch, new_kp

    # map_fn: loading images from file path in a dataset
    def _load_img(self, element):
        image_bytes = element['image/encoded']
        ptip_x = element['image/ptip_x']
        ptail_x = element['image/ptail_x']
        ptip_y = element['image/ptip_y']
        ptail_y = element['image/ptail_y']
        img_width = element['image/width']
        img_height = element['image/height']

        kp_helper = KeypointHelper(row_major=False)

        keypoint = tf.cond(self.use_tail,
                           lambda: kp_helper.from_list([ptail_x, ptail_y]),
                           lambda: kp_helper.from_list([ptip_x, ptip_y]))

        patch, new_kp = self._get_img_patch(image_bytes, keypoint, img_width, img_height, kp_helper)
        patch = tf.image.convert_image_dtype(patch, tf.float32)
        patch_h, patch_w = self.model_input_shape[:2]
        patch = tf.ensure_shape(patch, shape=(patch_h, patch_w, 3))

        return {'patch': patch,
                'keypoint': kp_helper.to_list(new_kp)}

    def _preprocess_example(self, element, kp_helper):
        patch = element['patch']
        keypoint = kp_helper.from_list(element['keypoint'])

        t_keypoints = tf.stack([keypoint])
        return patch, t_keypoints

    def _augment_and_get_heatmaps(self, patch, t_keypoints, kp_helper):
        augmenter = Augmenter(
            num_layers=2,
            magnitude=self.ra_magnitude,
            image_dtype=patch.dtype,
            image_shape=patch.shape,
            keypoint_helper=kp_helper)

        augmented_image, augmented_keypoints = augmenter.transform(image=patch, keypoints=t_keypoints)

        img_h = tf.cast(tf.shape(augmented_image)[0], dtype=tf.float32)
        img_w = tf.cast(tf.shape(augmented_image)[1], dtype=tf.float32)

        heatmaps = create_heatmaps(img_w, img_h,
                                   self.model_output_shape[1], self.model_output_shape[0],
                                   self.std_dev,
                                   augmented_keypoints,
                                   kp_helper)
        return augmented_image, heatmaps

    def _augment_img(self, element):
        kp_helper = KeypointHelper(row_major=False)
        patch, t_keypoints = self._preprocess_example(element, kp_helper)
        augmented_img, heatmaps = self._augment_and_get_heatmaps(patch, t_keypoints, kp_helper)

        augmented_img = self.input_prep_func(augmented_img)
        return (augmented_img, heatmaps)

    # map_fn: creating heatmaps for test dataset
    # named dict output does not work with keras fit; keras fails to connect the y input to the model
    # resulting in no usable gradients
    def _produce_test_heatmaps(self, element):
        kp_helper = KeypointHelper(row_major=False)
        patch, t_keypoints = self._preprocess_example(element, kp_helper)

        img_h = tf.cast(tf.shape(patch)[0], dtype=tf.float32)
        img_w = tf.cast(tf.shape(patch)[1], dtype=tf.float32)

        heatmaps = create_heatmaps(img_w, img_h,
                                   self.model_output_shape[1], self.model_output_shape[0],
                                   self.std_dev,
                                   t_keypoints,
                                   kp_helper)

        patch = self.input_prep_func(patch)
        return (patch, heatmaps)


class ClassifcationInputMap(PatchInputMap):
    def __init__(self,
                 model_input_shape,
                 model_output_shape,
                 ra_magnitude,
                 downscale_ratio,
                 input_prep_func=None):
        super().__init__(model_input_shape, model_output_shape, ra_magnitude, downscale_ratio, use_tail=False,
                         input_prep_func=input_prep_func)
        return

    def _load_img(self, element):
        prepared_x = super()._load_img(element)
        prepared_x['pen_down'] = element['image/pen_down']
        return prepared_x

    def _augment_img(self, element):
        kp_helper = KeypointHelper(row_major=False)
        patch, t_keypoints = self._preprocess_example(element, kp_helper)

        augmenter = Augmenter(
            num_layers=2,
            magnitude=self.ra_magnitude,
            image_dtype=patch.dtype,
            image_shape=patch.shape,
            keypoint_helper=kp_helper)

        augmented_image, augmented_keypoints = augmenter.transform(image=patch, keypoints=t_keypoints)

        augmented_image = self.input_prep_func(augmented_image)
        return (augmented_image, element['pen_down'])

    def _produce_test_heatmaps(self, element):
        kp_helper = KeypointHelper(row_major=False)
        patch, t_keypoints = self._preprocess_example(element, kp_helper)
        patch = self.input_prep_func(patch)
        return (patch, element['pen_down'])


class ClassificationSequenceInputMap(PatchInputMap):
    def __init__(self,
                 model_input_shape,
                 model_output_shape,
                 ra_magnitude,
                 downscale_ratio,
                 input_prep_func=None,
                 use_augmentation=False):
        super().__init__(
            model_input_shape,
            model_output_shape,
            ra_magnitude,
            downscale_ratio,
            use_tail=False,
            input_prep_func=input_prep_func,
            random_offset_divisor=2)
        self.use_augmentation = use_augmentation
        return

    def _get_random_crop_window(self, img_width, img_height):
        patch_h, patch_w = self.model_input_shape[:2]
        x_min = img_width // 2 - patch_w // 2
        x_max = img_width // 2
        y_min = img_height // 2 - patch_h // 2
        y_max = img_height // 2

        crop_x = tf.cast(tf.random.uniform([1], minval=x_min, maxval=x_max, dtype=tf.int64), dtype=tf.int32)[0]
        crop_y = tf.cast(tf.random.uniform([1], minval=y_min, maxval=y_max, dtype=tf.int64), dtype=tf.int32)[0]

        tf.assert_greater(patch_w, 0, name='random_crop_patch_width_check')
        tf.assert_greater(patch_h, 0, name='random_crop_patch_height_check')
        tf.debugging.assert_greater_equal(crop_x, 0, name='random_crop_x_offset_check')
        tf.debugging.assert_greater_equal(crop_y, 0, name='random_crop_y_offset_check')

        return [crop_y, crop_x, patch_h, patch_w]

    def _load_img(self, element):
        prepared_x = super()._load_img(element)
        # if we dont augment we need to format the input here
        patch = self.input_prep_func(prepared_x['patch'])
        updown = element['image/pen_down']
        # check for annotations where we cannot extract a patch
        # we might know it is pen down, but for patch classification it is
        # highly confusing for our system
        kp_helper = KeypointHelper(row_major=False)
        current_keypoint = kp_helper.from_list([element['image/ptip_x'], element['image/ptip_y']])
        if KeypointHelper.is_none_keypoint(current_keypoint):
            updown = tf.cast(0, tf.int64)

        return (patch, element['image/subject_id'], element['image/task_id'], element['image/sequence_id'], updown)

    def _augment_img(self, element):
        # augmenting expects sequences as elements, and up/down annotations. We just return
        # the annotations as is.
        if not self.use_augmentation:
            return None
        else:
            sequence = element[0]

            choices = tf.random.shuffle(tf.range(12))[:2]
            augment_seed = tf.random.uniform([], minval=0, maxval=100000)  # np.random.randint(low=0, high=10000)
            # tf.print(augment_seed)
            augmenter = Augmenter(
                num_layers=2,
                magnitude=self.ra_magnitude,
                image_dtype=sequence[0].dtype,
                image_shape=sequence[0].shape,
                keypoint_helper=KeypointHelper(row_major=False),
                augment_seed=augment_seed,
                op_choices=choices)

            def augment_patch_in_sequence(x):
                augmented_patch, _ = augmenter.transform(image=x,
                                                         keypoints=tf.stack([KeypointHelper.make_none_keypoint()]))
                return augmented_patch

            augmented_patches = tf.map_fn(augment_patch_in_sequence, sequence)
            return (augmented_patches, element[1])

    def _produce_test_heatmaps(self, element):
        return element


class ClassificationKeypointSequenceInputMap(PatchInputMap):
    def __init__(self,
                 model_input_shape,
                 model_output_shape,
                 ra_magnitude,
                 downscale_ratio,
                 input_prep_func=None,
                 use_augmentation=False):
        super().__init__(
            model_input_shape,
            model_output_shape,
            ra_magnitude,
            downscale_ratio,
            input_prep_func=input_prep_func)
        self.use_augmentation = use_augmentation
        return

    def _get_random_crop_window(self, img_width, img_height):
        patch_h, patch_w = self.model_input_shape[:2]
        x_min = img_width // 2 - patch_w // 2
        x_max = img_width // 2
        y_min = img_height // 2 - patch_h // 2
        y_max = img_height // 2

        crop_x = tf.cast(tf.random.uniform([1], minval=x_min, maxval=x_max, dtype=tf.int64), dtype=tf.int32)[0]
        crop_y = tf.cast(tf.random.uniform([1], minval=y_min, maxval=y_max, dtype=tf.int64), dtype=tf.int32)[0]

        tf.assert_greater(patch_w, 0, name='random_crop_patch_width_check')
        tf.assert_greater(patch_h, 0, name='random_crop_patch_height_check')
        tf.debugging.assert_greater_equal(crop_x, 0, name='random_crop_x_offset_check')
        tf.debugging.assert_greater_equal(crop_y, 0, name='random_crop_y_offset_check')

        return [crop_y, crop_x, patch_h, patch_w]

    def _load_img(self, element):
        prepared_x = super()._load_img(element)
        # if we dont augment we need to format the input here
        patch = self.input_prep_func(prepared_x['patch'])
        prepped_keypoint = prepared_x['keypoint']

        updown = element['image/pen_down']

        # check for annotations where we cannot extract a patch
        # we might know it is pen down, but for patch classification it is
        # highly confusing for our system
        kp_helper = KeypointHelper(row_major=False)
        annotated_keypoint = kp_helper.from_list([element['image/ptip_x'], element['image/ptip_y']])
        if KeypointHelper.is_none_keypoint(annotated_keypoint):
            updown = tf.cast(0, tf.int64)

        return (
            patch,
            element['image/subject_id'],
            element['image/task_id'],
            element['image/sequence_id'],
            (prepped_keypoint, updown))

    def _augment_img(self, element):
        # augmenting expects sequences as elements, and up/down annotations. We just return
        # the annotations as is.
        if not self.use_augmentation:
            return self._produce_test_heatmaps(element)
        else:
            kp_helper = KeypointHelper(row_major=False)

            # augment_seed = tf.cast(tf.random.uniform([1], minval=0, maxval=10000, dtype=tf.int32), tf.int32)
            choices = tf.random.shuffle(tf.range(12))[:2]
            augment_seed = np.random.randint(low=0, high=10000)
            # tf.random.set_seed(augment_seed)
            augmenter = Augmenter(
                num_layers=2,
                magnitude=self.ra_magnitude,
                image_dtype=element[0][0].dtype,
                image_shape=element[0][0].shape,
                keypoint_helper=kp_helper,
                augment_seed=augment_seed,
                op_choices=choices)

            output_shape = self.model_output_shape
            std_dev = self.std_dev

            def augment_patch_in_sequence(x):
                patch = x[0]
                keypoint_ud = x[1]
                current_keypoints = tf.stack([keypoint_ud[0]])
                updown = keypoint_ud[1]

                augmented_patch, augmented_keypoint = augmenter.transform(image=patch,
                                                                          keypoints=current_keypoints)

                heatmap_keypoints = tf.cond(tf.equal(updown, 0),
                                            lambda: tf.stack(
                                                [augmented_keypoint[0], KeypointHelper.make_none_keypoint()]),
                                            lambda: tf.stack(
                                                [KeypointHelper.make_none_keypoint(), augmented_keypoint[0]])
                                            )

                img_h = tf.cast(tf.shape(augmented_patch)[0], dtype=tf.float32)
                img_w = tf.cast(tf.shape(augmented_patch)[1], dtype=tf.float32)

                heatmaps = create_heatmaps(img_w, img_h,
                                           output_shape[1], output_shape[0],
                                           std_dev,
                                           heatmap_keypoints,
                                           kp_helper)

                # heatmaps = tf.cond(tf.equal(updown, 0),
                #                   lambda:tf.concat([tf.expand_dims(heatmaps[:, :, 0], -1), tf.expand_dims(tf.math.negative(heatmaps[:, :, 0]), -1)], -1),
                #                   lambda:tf.concat([tf.expand_dims(tf.math.negative(heatmaps[:, :, 1]), -1), tf.expand_dims(heatmaps[:, :, 1], -1)], -1))
                return augmented_patch, heatmaps

            result = tf.map_fn(augment_patch_in_sequence, element, fn_output_signature=(tf.float32, tf.float32))

            return result

    def _produce_test_heatmaps(self, element):
        kp_helper = KeypointHelper(row_major=False)

        output_shape = self.model_output_shape
        std_dev = self.std_dev

        def create_heatmaps_for_patch_in_sequence(x):
            patch = x[0]
            keypoint_ud = x[1]
            current_keypoints = tf.stack([keypoint_ud[0]])
            updown = keypoint_ud[1]

            heatmap_keypoints = tf.cond(tf.equal(updown, 0),
                                        lambda: tf.stack([current_keypoints[0], KeypointHelper.make_none_keypoint()]),
                                        lambda: tf.stack([KeypointHelper.make_none_keypoint(), current_keypoints[0]])
                                        )

            img_h = tf.cast(tf.shape(patch)[0], dtype=tf.float32)
            img_w = tf.cast(tf.shape(patch)[1], dtype=tf.float32)

            heatmaps = create_heatmaps(img_w, img_h,
                                       output_shape[1], output_shape[0],
                                       std_dev,
                                       heatmap_keypoints,
                                       kp_helper)
            return patch, heatmaps

        result = tf.map_fn(create_heatmaps_for_patch_in_sequence, element, fn_output_signature=(tf.float32, tf.float32))

        return result


class PipelineCreator:
    AUTOTUNE = tf.data.experimental.AUTOTUNE

    def __init__(self,
                 input_mapper,
                 batch_size,
                 tfrecord_files,
                 cache_prefix_id="",
                 test_subjects=None,
                 image_features=image_feature_description):
        if test_subjects is None:
            test_subjects = [0, 6, 10, 14, 20, 23, 38, 39]
        self.batch_size = batch_size
        self.input_mapper = input_mapper
        self.tfrecord_files = tfrecord_files
        self.test_subjects = tf.constant(test_subjects, dtype=tf.int64)
        self.train_cache_file = "./{}_train.tfcache".format(cache_prefix_id)
        self.test_cache_file = "./{}_test.tfcache".format(cache_prefix_id)
        self.image_features = image_features

        self.train_ds, self.test_ds = self._load_and_split_dataset()
        self.train_ds = self._prepare_train_dataset(self.train_ds)
        self.test_ds = self._prepare_test_dataset(self.test_ds)
        return

    def _parse_tfrecord_entry(self, example_proto):
        return tf.io.parse_single_example(example_proto, self.image_features)

    @tf.function
    def _check_if_test_subject(self, entry):
        subject_id = entry['image/subject_id']
        check_id_tensor = tf.equal(self.test_subjects, subject_id)

        result = tf.cond(tf.equal(tf.math.count_nonzero(check_id_tensor), True),
                         lambda: tf.constant(True, dtype=tf.bool),
                         lambda: tf.constant(False, dtype=tf.bool), name='check_if_test_subject')
        return result

    def _load_and_split_dataset(self):
        raw_dataset = read_tfrecords(self.tfrecord_files)
        parsed_dataset = raw_dataset.map(lambda x: self._parse_tfrecord_entry(x))
        train_dataset = parsed_dataset.filter(lambda x: not self._check_if_test_subject(x))
        test_dataset = parsed_dataset.filter(lambda x: self._check_if_test_subject(x))
        return train_dataset, test_dataset

    def _prepare_train_dataset(self, train_ds):
        tf.print("Preparing train dataset ...")
        # decode images and cache the dataset
        loaded_dataset = train_ds.map(lambda x: self.input_mapper._load_img(x),
                                      num_parallel_calls=PipelineCreator.AUTOTUNE)
        # loaded_dataset = loaded_dataset.cache(self.train_cache_file)
        # now shuffle and do augmentation
        loaded_dataset = loaded_dataset.shuffle(buffer_size=100)
        loaded_dataset = loaded_dataset.map(lambda x: self.input_mapper._augment_img(x),
                                            num_parallel_calls=PipelineCreator.AUTOTUNE)
        # prepare dataset aspects: repeat for repeating augmentation step, batch, prefetch
        loaded_dataset = loaded_dataset.repeat()
        loaded_dataset = loaded_dataset.batch(self.batch_size)
        loaded_dataset = loaded_dataset.prefetch(buffer_size=self.AUTOTUNE)
        return loaded_dataset

    def _prepare_test_dataset(self, test_ds):
        tf.print("Preparing test dataset ...")
        # decode images, create the output heatmaps and cache the dataset
        loaded_dataset = test_ds.map(lambda x: self.input_mapper._load_img(x),
                                     num_parallel_calls=PipelineCreator.AUTOTUNE)
        loaded_dataset = loaded_dataset.map(lambda x: self.input_mapper._produce_test_heatmaps(x),
                                            num_parallel_calls=PipelineCreator.AUTOTUNE)
        # loaded_dataset = loaded_dataset.cache(self.test_cache_file)
        # ds_test = ds_test.repeat() repeat would only be necessary if we do test time augmentation
        loaded_dataset = loaded_dataset.batch(self.batch_size)
        loaded_dataset = loaded_dataset.prefetch(buffer_size=PipelineCreator.AUTOTUNE)
        # loaded_dataset = loaded_dataset.cache(self.train_cache_file)
        return loaded_dataset

    def get_datasets(self):
        return self.train_ds, self.test_ds

    def get_train_dataset(self):
        return self.train_ds

    def get_test_dataset(self):
        return self.test_ds

    def get_batch_size(self):
        return self.batch_size


class SequencePipelineCreator(PipelineCreator):
    def __init__(self,
                 input_mapper,
                 batch_size,
                 tfrecord_files,
                 cache_prefix_id="",
                 test_subjects=None,
                 image_features=image_feature_description,
                 timesteps=3,
                 shift=3):
        self.timesteps = timesteps
        self.shift = shift
        super().__init__(input_mapper, batch_size, tfrecord_files, cache_prefix_id, test_subjects, image_features)
        return

    def _prepare_train_dataset(self, train_ds):
        tf.print("Preparing train dataset ...")
        # decode images and create tuples with (decoded_img, subject, task, sequence, updown)
        loaded_dataset = train_ds.map(lambda x: self.input_mapper._load_img(x),
                                      num_parallel_calls=PipelineCreator.AUTOTUNE)
        windowed_dataset = loaded_dataset.window(self.timesteps, self.shift, drop_remainder=True)
        # create one dataset from the window result, which is a dataset of datasets
        windowed_dataset = windowed_dataset.flat_map(lambda v, w, x, y, z: tf.data.Dataset.zip((v, w, x, y, z)))
        windowed_dataset = windowed_dataset.batch(self.timesteps)

        # now filter out sequences/batches where subject, task or sequence ids are not homogenous
        filtered_dataset = windowed_dataset.filter(
            lambda v, w, x, y, z: tf.equal(tf.size(tf.unique(w)[0]), 1) and tf.equal(tf.size(tf.unique(x)[0]),
                                                                                     1) and tf.equal(
                tf.size(tf.unique(y)[0]), 1))

        # we can discard the elements that helped us filtering
        # and leave only patches and up/down annotations
        filtered_dataset = filtered_dataset.map(lambda v, w, x, y, z: (v, z),
                                                num_parallel_calls=PipelineCreator.AUTOTUNE)

        if isinstance(self.input_mapper, ClassificationKeypointSequenceInputMap) or self.input_mapper.use_augmentation:
            filtered_dataset = filtered_dataset.map(lambda x, y: self.input_mapper._augment_img((x, y)),
                                                    num_parallel_calls=PipelineCreator.AUTOTUNE)

        # now we shuffle the sequences
        filtered_dataset = filtered_dataset.shuffle(buffer_size=30)

        # prepare dataset aspects: repeat for repeating augmentation step, batch, prefetch
        filtered_dataset = filtered_dataset.repeat()
        filtered_dataset = filtered_dataset.batch(self.batch_size)

        filtered_dataset = filtered_dataset.prefetch(buffer_size=self.AUTOTUNE)
        return filtered_dataset

    def _prepare_test_dataset(self, test_ds):
        tf.print("Preparing test dataset ...")
        if isinstance(self.input_mapper, ClassificationKeypointSequenceInputMap) or self.input_mapper.use_augmentation:
            # decode images and create tuples with (decoded_img, subject, task, sequence, updown)
            loaded_dataset = test_ds.map(lambda x: self.input_mapper._load_img(x),
                                         num_parallel_calls=PipelineCreator.AUTOTUNE)
            windowed_dataset = loaded_dataset.window(self.timesteps, self.shift, drop_remainder=True)
            # create one dataset from the window result, which is a dataset of datasets
            windowed_dataset = windowed_dataset.flat_map(lambda v, w, x, y, z: tf.data.Dataset.zip((v, w, x, y, z)))
            windowed_dataset = windowed_dataset.batch(self.timesteps)

            # now filter out sequences/batches where subject, task or sequence ids are not homogenous
            filtered_dataset = windowed_dataset.filter(
                lambda v, w, x, y, z: tf.equal(tf.size(tf.unique(w)[0]), 1) and tf.equal(tf.size(tf.unique(x)[0]),
                                                                                         1) and tf.equal(
                    tf.size(tf.unique(y)[0]), 1))

            # we can discard the elements that helped us filtering
            # and leave only patches and up/down annotations
            filtered_dataset = filtered_dataset.map(lambda v, w, x, y, z: (v, z),
                                                    num_parallel_calls=PipelineCreator.AUTOTUNE)

            filtered_dataset = filtered_dataset.map(lambda x, y: self.input_mapper._produce_test_heatmaps((x, y)),
                                                    num_parallel_calls=PipelineCreator.AUTOTUNE)

            # now we shuffle the sequences
            filtered_dataset = filtered_dataset.shuffle(buffer_size=100)

            # prepare dataset aspects: repeat for repeating augmentation step, batch, prefetch
            filtered_dataset = filtered_dataset.batch(self.batch_size)

            loaded_dataset = filtered_dataset.prefetch(buffer_size=self.AUTOTUNE)
        else:
            loaded_dataset = self._prepare_train_dataset(test_ds)
        return loaded_dataset

    def get_timesteps(self):
        return self.timesteps
