import math
import os
import pprint

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.python.keras.models import load_model

from research.ablation_studies.dataset.dataset_utils import get_max_of_heatmap
from research.ablation_studies.gestures.adapted_recognizer import recognize_gesture
from research.ablation_studies.gestures.dollar import Recognizer, Result
from research.ablation_studies.sequence.video_streamer import VideoFrameStreamer
from research.ablation_studies.train.keypoint_tf import KeypointHelper


# crop 1 to lower 2/3
# crop 2 to lower 2/3
# center crop 3 to 1280/720


def preprocess_video_frame(frame, video_file):
    my_video_file = os.path.split(video_file)[1]
    video_info = get_info_from_video_filename(my_video_file)
    current_frame = np.array(frame, np.float32)

    if video_info['subject_id'] == 1:
        current_frame = tf.image.crop_to_bounding_box(frame, 140, 0, 568, 480)
    if video_info['subject_id'] == 2:
        current_frame = tf.image.crop_to_bounding_box(frame, 283, 0, 565, 480)
    if video_info['subject_id'] == 3:
        current_frame = tf.image.central_crop(frame, 0.667)

    orig_height, orig_width = current_frame.shape[:2]

    current_frame = tf.image.resize(current_frame, (224, 224))
    current_frame = current_frame / 255.
    return current_frame, orig_height, orig_width


def get_paths_from_video(model, video_file, csv_base_path):
    streamer = VideoFrameStreamer(video_file)
    kp_helper = KeypointHelper(row_major=False)

    raw_tip_path = []
    raw_tail_path = []

    csv_dict = {
        'tip_max': [],
        'tail_max': [],
        'tip_x': [],
        'tip_y': [],
        'tail_x': [],
        'tail_y': []
    }

    current_frame = streamer.get()
    while current_frame is not None:
        current_frame, orig_height, orig_width = preprocess_video_frame(current_frame, video_file)
        # current_frame = np.array(current_frame, np.float32) / 255.
        # current_frame = tf.image.resize(current_frame, (224, 224))

        inferences = model([current_frame])
        tip_heatmap = inferences[0, :, :, 0]
        tail_heatmap = inferences[0, :, :, 1]
        tip_max, tip_indices = get_max_of_heatmap(tip_heatmap)
        tail_max, tail_indices = get_max_of_heatmap(tail_heatmap)

        tip_kp = kp_helper.make(x=tip_indices[1], y=tip_indices[0])
        tail_kp = kp_helper.make(x=tail_indices[1], y=tail_indices[0])

        tip_kp = kp_helper.rescale(tip_kp, 224, 224, orig_width, orig_height)
        tail_kp = kp_helper.rescale(tail_kp, 224, 224, orig_width, orig_height)

        # if tip_max > 1e-1:
        raw_tip_path.append(kp_helper.to_list(tip_kp))
        # if tail_max > 1e-1:
        raw_tail_path.append(kp_helper.to_list(tail_kp))

        csv_dict['tip_max'].append(tip_max.numpy()[0])
        csv_dict['tip_x'].append(kp_helper.get_x(tip_kp).numpy()[0])
        csv_dict['tip_y'].append(kp_helper.get_y(tip_kp).numpy()[0])
        csv_dict['tail_max'].append(tail_max.numpy()[0])
        csv_dict['tail_x'].append(kp_helper.get_x(tail_kp).numpy()[0])
        csv_dict['tail_y'].append(kp_helper.get_y(tail_kp).numpy()[0])
        current_frame = streamer.get()

    video_df = pd.DataFrame(csv_dict)
    video_csv_path = os.path.join(csv_base_path, video_file.split('.')[0] + '.csv')
    video_df.to_csv(video_csv_path)
    return raw_tip_path, raw_tail_path


def get_info_from_video_filename(video_file):
    splits = video_file.split('_')
    subject = int(splits[0])
    keypoint_type = splits[1]
    background_type = splits[2]
    gesture = splits[3]
    gesture_id = int(splits[4].split('.')[0])
    info_dict = {
        'subject_id': subject,
        'kp_type': keypoint_type,
        'bg_type': background_type,
        'gesture_type': gesture,
        'gesture_id': gesture_id,
        'pred_tip': "",
        'pred_tail': "",
        'correct_tip': 0,
        'correct_tail': 0
    }
    return info_dict


def get_stats_for_subject_gesture_type(gesture_type_df, kp_type, subject, result_dict):
    subject_gesture_type_df = gesture_type_df[gesture_type_df['subject_id'] == subject]
    correct_subject_gesture_type_df = subject_gesture_type_df[subject_gesture_type_df['correct_' + kp_type] == 1]

    all_blank_gestures = len(subject_gesture_type_df[subject_gesture_type_df['bg_type'] == "blank"])
    correct_blank_gestures = len(correct_subject_gesture_type_df[correct_subject_gesture_type_df['bg_type'] == "blank"])
    all_print_gestures = len(subject_gesture_type_df[subject_gesture_type_df['bg_type'] == "print"])
    correct_print_gestures = len(correct_subject_gesture_type_df[correct_subject_gesture_type_df['bg_type'] == "print"])

    # result_dict['all_{}_blank'.format(subject)] = all_blank_gestures
    result_dict['correct_{}_blank'.format(subject)] = correct_blank_gestures
    # result_dict['all_{}_print'.format(subject)] = all_print_gestures
    result_dict['correct_{}_print'.format(subject)] = correct_print_gestures
    return


def get_stats_for_gesture_type(kp_type_df, kp_type, gesture_type):
    gesture_type_df = kp_type_df[kp_type_df['gesture_type'] == gesture_type]
    correct_gesture_df = gesture_type_df[gesture_type_df['correct_' + kp_type] == 1]

    all_gestures = len(gesture_type_df)
    correct_gestures = len(correct_gesture_df)
    all_blank_gestures = len(gesture_type_df[gesture_type_df['bg_type'] == "blank"])
    correct_blank_gestures = len(correct_gesture_df[correct_gesture_df['bg_type'] == "blank"])
    all_print_gestures = len(gesture_type_df[gesture_type_df['bg_type'] == "print"])
    correct_print_gestures = len(correct_gesture_df[correct_gesture_df['bg_type'] == "print"])

    result_dict = {
        'kp_type': kp_type,
        'gesture_type': gesture_type,
        'count': all_gestures,
        'correct': correct_gestures,
        'count_blank': all_blank_gestures,
        'correct_blank': correct_blank_gestures,
        'count_print': all_print_gestures,
        'correct_print': correct_print_gestures
    }

    get_stats_for_subject_gesture_type(gesture_type_df, kp_type, 0, result_dict)
    get_stats_for_subject_gesture_type(gesture_type_df, kp_type, 1, result_dict)
    get_stats_for_subject_gesture_type(gesture_type_df, kp_type, 2, result_dict)
    get_stats_for_subject_gesture_type(gesture_type_df, kp_type, 3, result_dict)

    return result_dict


def get_stats_for_kp_type(gesture_df, kp_type):
    kp_type_df = gesture_df[gesture_df['kp_type'] == kp_type]
    all = len(kp_type_df)
    all_correct = len(gesture_df[gesture_df['correct_' + kp_type] == 1])
    proportion_correct = all_correct / all
    print("Proportion correct {}: {} ({}/{})".format(kp_type, proportion_correct, all_correct, all))

    # lines
    lines_result_dict = get_stats_for_gesture_type(kp_type_df, kp_type, "line")
    pprint.pprint(lines_result_dict)
    circle_result_dict = get_stats_for_gesture_type(kp_type_df, kp_type, "circle")
    pprint.pprint(circle_result_dict)
    square_result_dict = get_stats_for_gesture_type(kp_type_df, kp_type, "square")
    pprint.pprint(square_result_dict)
    triangle_result_dict = get_stats_for_gesture_type(kp_type_df, kp_type, "triangle")
    pprint.pprint(triangle_result_dict)
    return all, all_correct


def evaluate_dataframe(gesture_df):
    all_gestures = len(gesture_df['gesture_id'])

    all_tip, correct_tip = get_stats_for_kp_type(gesture_df, "tip")
    all_tail, correct_tail = get_stats_for_kp_type(gesture_df, "tail")

    part_tip_correct = correct_tip / all_tip
    part_tail_correct = correct_tail / all_tail

    print("{} video with {} tip gestures and {} tail gestures".format(all_gestures, all_tip, all_tail))
    print("Correct tip gestures: {} | {}".format(correct_tip, part_tip_correct))
    print("Correct tail gestures: {} | {}".format(correct_tail, part_tail_correct))

    # get stats for single gestures
    return


def do_video_eval():
    video_dir = "/home/florin/data/gestures/"
    video_csv_dir = "/home/florin/data/gestures/cropped_csv/"
    model_weights = "/home/florin/diss_models/final_weights/kp/roi/effnet-b0/effnetb0_roi_u4_ff_ra9_best.178.h5"

    video_files = os.listdir(video_dir)
    video_files = [video for video in video_files if '.mp4' in video or '.mov' in video or '.mkv' in video]

    my_model = load_model(model_weights)
    recognizer = Recognizer()
    all_infos = []

    for video_file in video_files:
        full_video_path = os.path.join(video_dir, video_file)
        raw_tip_path, raw_tail_path = get_paths_from_video(my_model, full_video_path, video_csv_dir)
        video_info = get_info_from_video_filename(video_file)

        if len(raw_tip_path) > 1:
            tip_class = recognizer.Recognize(raw_tip_path)  #
        else:
            tip_class = Result("None", 0)
        if len(raw_tail_path) > 1:
            tail_class = recognizer.Recognize(raw_tail_path)
        else:
            tail_class = Result("None", 0)

        video_info['pred_tip'] = tip_class.Name
        video_info['pred_tip_score'] = tip_class.Score
        video_info['pred_tail'] = tail_class.Name
        video_info['pred_tail_score'] = tail_class.Score

        video_info['correct_tip'] = 1 if (tip_class.Name == video_info['gesture_type']) and (
                    video_info['kp_type'] == "tip") else 0
        video_info['correct_tail'] = 1 if (tail_class.Name == video_info['gesture_type']) and (
                    video_info['kp_type'] == "tail") else 0

        print("Video {} annotation: {}".format(video_file, video_info['gesture_type']))
        print("Tip Path Pred: {} | Score: {}".format(tip_class.Name, tip_class.Score))
        print("Tail Path Pred: {} | Score: {}".format(tail_class.Name, tail_class.Score))

        all_infos.append(video_info)

    kp_df = pd.DataFrame(all_infos)
    # kp_df.to_csv("gesture_classes_1e-1.csv")


def get_path_from_video_csv(video_csv_file, kp_type, zero_threshold):
    my_df = pd.read_csv(video_csv_file)
    my_df = my_df[my_df[kp_type + '_max'] >= zero_threshold]
    xs = my_df[kp_type + '_x'].to_numpy()
    ys = my_df[kp_type + '_y'].to_numpy()
    return list(zip(xs, ys))


def filter_outliers_from_stream(points, threshold):
    my_coords = np.array(points)
    prev_coord = None
    recorded_indices = []

    for index, coord in enumerate(my_coords):
        if index > 0:
            dist = math.sqrt((coord[0] - prev_coord[0]) ** 2 + (coord[1] - prev_coord[1]) ** 2)
            # dist = np.linalg.norm(np.array(my_coords)-np.array(prev_coord))
            if dist > threshold:
                recorded_indices.append(index)
            else:
                prev_coord = coord
        else:
            prev_coord = coord

    result = np.delete(points, recorded_indices, axis=0).tolist()
    return result


def do_video_csv_eval(video_csv_dir):
    video_csv_files = os.listdir(video_csv_dir)
    video_csv_files = [video for video in video_csv_files if '.csv' in video]

    recognizer = Recognizer()
    all_infos = []

    for video_csv_file in video_csv_files:
        full_video_csv_path = os.path.join(video_csv_dir, video_csv_file)

        raw_tip_path = get_path_from_video_csv(full_video_csv_path, "tip", 0.01)
        raw_tail_path = get_path_from_video_csv(full_video_csv_path, "tail", 0.01)

        video_info = get_info_from_video_filename(video_csv_file)

        if len(raw_tip_path) > 1:
            tip_class = recognize_gesture(raw_tip_path)  # recognizer.Recognize(filtered_tip_path)  #
        else:
            tip_class = Result("None", 0)
        if len(raw_tail_path) > 1:
            tail_class = recognize_gesture(raw_tail_path)
        else:
            tail_class = Result("None", 0)

        video_info['pred_tip'] = tip_class.Name
        video_info['pred_tip_score'] = tip_class.Score
        video_info['pred_tail'] = tail_class.Name
        video_info['pred_tail_score'] = tail_class.Score

        video_info['correct_tip'] = 1 if (tip_class.Name == video_info['gesture_type']) and (
                video_info['kp_type'] == "tip") else 0
        video_info['correct_tail'] = 1 if (tail_class.Name == video_info['gesture_type']) and (
                video_info['kp_type'] == "tail") else 0

        print("Video {} annotation: {}".format(video_csv_file, video_info['gesture_type']))
        print("Tip Path Pred: {} | Score: {}".format(tip_class.Name, tip_class.Score))
        print("Tail Path Pred: {} | Score: {}".format(tail_class.Name, tail_class.Score))

        all_infos.append(video_info)

    kp_df = pd.DataFrame(all_infos)
    evaluate_dataframe(kp_df)


if __name__ == '__main__':
    do_video_csv_eval("/home/florin/data/gestures/raw_csv/")
    # do_video_eval()

    # eval_csv = "/home/florin/glowing-keypoint-firefly-dep/gesture_classes.csv"
    # eval_df = pd.read_csv(eval_csv)
    # print(eval_df.head())

    # evaluate_dataframe(eval_df)
