import tensorflow as tf
from tensorflow.keras.layers import Conv2DTranspose, BatchNormalization, Activation, Conv2D, Dense, LSTM
from tensorflow.keras.models import Model
from tensorflow.keras.models import load_model
from tensorflow.python.keras.layers import GlobalAveragePooling2D

from research.ablation_studies.helpers import create_model_from_weights


def create_detail_effnetB0_classifier_no_memory(input_shape, effnet_trainable=True, use_extra_dense1=True,
                                                use_extra_dense2=True, fine_tune_weights=None):
    if fine_tune_weights is not None:
        effnet_model = load_model(fine_tune_weights)
    else:
        effnet_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False,
                                                            input_shape=input_shape)

    last_layer = effnet_model.get_layer('top_activation')

    if not effnet_trainable:
        effnet_model.trainable = False
    else:
        effnet_model.trainable = True

    # freeze BN layers as per https://keras.io/guides/transfer_learning/
    # for layer in effnet_model.layers:
    #    if isinstance(layer, tf.keras.layers.BatchNormalization):
    #        layer.trainable = False

    pooled_conv_output = GlobalAveragePooling2D()(last_layer.output)
    flattened_output = tf.keras.layers.Dropout(0.2)(pooled_conv_output)
    if use_extra_dense1:
        flattened_output = Dense(1024, activation='relu')(flattened_output)
    if use_extra_dense2:
        flattened_output = Dense(512, activation='relu')(flattened_output)

    # either sigmoid and binary crossentropy or softmax ant cat. crossentropy
    # predictions = Dense(2, activation='softmax')(flattened_output)
    predictions = Dense(1, activation='sigmoid')(flattened_output)

    return effnet_model.input, predictions


def create_effnetb0_with_frozen_layers(input_shape, unfreeze_blocks=()):
    effnet_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False, input_shape=input_shape)

    effnet_model.trainable = False

    trainable_layers = []
    for layer_num in unfreeze_blocks:
        block_name = "block" + str(layer_num)
        trainable_layers.append(block_name)

    for layer in effnet_model.layers:
        for trainable_layer in trainable_layers:
            if trainable_layer in layer.name:
                layer.trainable = True

    return effnet_model


def create_model_detail_effnetB0_classifier_simple_lstm(lstm_size, time_steps, input_shape, model_name, model_path=None,
                                                        unfreeze_blocks=()):
    if model_path is None:
        effnet_model = create_effnetb0_with_frozen_layers(input_shape, unfreeze_blocks)
    else:
        effnet_model = load_model(model_path)

    last_layer = effnet_model.get_layer('top_activation')
    effnet_flatten = tf.keras.layers.Flatten()(last_layer.output)

    input_cnn = Model(effnet_model.input, effnet_flatten)

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3))
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn)(input_sequence)
    lstm_layer = LSTM(lstm_size, return_sequences=True)(time_distributed_effnet)
    predictions = tf.keras.layers.TimeDistributed(Dense(1, activation='sigmoid'))(lstm_layer)
    lstm_effnet = Model(input_sequence, predictions, name=model_name)

    return lstm_effnet


# approach: place lstm after first dense layer, like in 2015-donahue
def create_model_detail_effnetB0_classifier_donahue(
        lstm_size,
        time_steps,
        input_shape,
        model_name,
        model_path=None,
        unfreeze_blocks=(),
        dense_trainable=False,
        freeze_bns=False,
        use_gru=False):
    if model_path is None:
        effnet_model = create_effnetb0_with_frozen_layers(input_shape, unfreeze_blocks)
    else:
        effnet_model = load_model(model_path)

        effnet_model.trainable = False

        trainable_layers = []
        for layer_num in unfreeze_blocks:
            block_name = "block" + str(layer_num)
            trainable_layers.append(block_name)

        for layer in effnet_model.layers:
            for trainable_layer in trainable_layers:
                if trainable_layer in layer.name:
                    layer.trainable = True
                    print("made {} trainable".format(layer.name))

    if dense_trainable:
        effnet_model.get_layer('dense').trainable = True

    if freeze_bns:
        for layer in effnet_model.layers:
            if isinstance(layer, tf.keras.layers.BatchNormalization):
                layer.trainable = False

    last_layer = effnet_model.get_layer('dense')

    input_cnn = Model(effnet_model.input, last_layer.output)

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3))
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn)(input_sequence)
    if not use_gru:
        lstm_layer = LSTM(lstm_size, return_sequences=True)(time_distributed_effnet)
    else:
        lstm_layer = tf.keras.layers.GRU(lstm_size, return_sequences=True)(time_distributed_effnet)
    predictions = tf.keras.layers.TimeDistributed(Dense(1, activation='sigmoid'))(lstm_layer)
    lstm_effnet = Model(input_sequence, predictions, name=model_name)

    return lstm_effnet


def create_model_detail_effnetB0_classifier_nonlinear_lstm(time_steps, input_shape, model_name):
    # effnet_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False, input_shape=input_shape)
    effnet_model = load_model(
        "/home/florin/diss_models/effnetb0_classify_basic/simple_full_set_weighted/1_chkpt/simple_full_set_weighted_best.107.h5")
    effnet_model.trainable = False

    last_layer = effnet_model.get_layer('top_activation')
    effnet_flatten = tf.keras.layers.Flatten()(last_layer.output)

    input_cnn = Model(effnet_model.input, effnet_flatten)

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3))
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn)(input_sequence)
    lstm_layer = LSTM(256, return_sequences=True)(time_distributed_effnet)
    lstm_layer_2 = LSTM(128, return_sequences=True)(lstm_layer)
    predictions = tf.keras.layers.TimeDistributed(Dense(1, activation='sigmoid'))(lstm_layer_2)
    lstm_effnet = Model(input_sequence, predictions, name=model_name)

    return lstm_effnet


def create_effnetb0_convlstm_encoder(lstm_size, time_steps, input_shape, model_name, model_path=None,
                                     unfreeze_blocks=(), dense_trainable=False, freeze_bns=False):
    if model_path is None:
        effnet_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False,
                                                            input_shape=input_shape)
    else:
        effnet_model = load_model(model_path)

    effnet_model.trainable = False

    trainable_layers = []
    for layer_num in unfreeze_blocks:
        block_name = "block" + str(layer_num)
        trainable_layers.append(block_name)

    for layer in effnet_model.layers:
        for trainable_layer in trainable_layers:
            if trainable_layer in layer.name:
                layer.trainable = True

    if freeze_bns:
        for layer in effnet_model.layers:
            if isinstance(layer, tf.keras.layers.BatchNormalization):
                layer.trainable = False

    last_layer = effnet_model.get_layer('top_activation')
    input_cnn = Model(effnet_model.input, last_layer.output)

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3))
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn)(input_sequence)
    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=lstm_size, kernel_size=(3, 3), return_sequences=True)(
        time_distributed_effnet)

    flattened_conv = tf.keras.layers.TimeDistributed(tf.keras.layers.Flatten())(conv_lstm_layer)
    predictions = tf.keras.layers.TimeDistributed(Dense(1, activation='sigmoid'))(flattened_conv)
    lstm_effnet = Model(input_sequence, predictions, name=model_name)

    return lstm_effnet


def create_full_conv_lstm(time_steps, input_shape=(224, 224, 3)):
    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3))

    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=32, kernel_size=(3, 3), strides=(1, 1), return_sequences=True)(
        input_sequence)
    # conv_lstm_layer_bn = tf.keras.layers.TimeDistributed(tf.keras.layers.BatchNormalization())(conv_lstm_layer)
    conv_lstm_layer_bn = tf.keras.layers.BatchNormalization()(conv_lstm_layer)

    conv_lstm_layer_2 = tf.keras.layers.ConvLSTM2D(filters=64, kernel_size=(3, 3), strides=(2, 2),
                                                   return_sequences=True)(
        conv_lstm_layer_bn)
    conv_lstm_layer_bn_2 = tf.keras.layers.BatchNormalization()(conv_lstm_layer_2)

    conv_lstm_layer_3 = tf.keras.layers.ConvLSTM2D(filters=128, kernel_size=(3, 3), strides=(1, 1),
                                                   return_sequences=True)(
        conv_lstm_layer_bn_2)
    conv_lstm_layer_bn_3 = tf.keras.layers.BatchNormalization()(conv_lstm_layer_3)

    conv_lstm_layer_4 = tf.keras.layers.ConvLSTM2D(filters=256, kernel_size=(3, 3), strides=(2, 2),
                                                   return_sequences=True)(
        conv_lstm_layer_bn_3)
    conv_lstm_layer_bn_4 = tf.keras.layers.BatchNormalization()(conv_lstm_layer_4)

    conv2d_1 = tf.keras.layers.TimeDistributed(tf.keras.layers.Conv2D(filters=128, kernel_size=(3, 3), strides=(2, 2)))(
        conv_lstm_layer_bn_4)
    conv2d_1_bn = tf.keras.layers.BatchNormalization()(conv2d_1)
    conv2d_1_ac = tf.keras.layers.LeakyReLU()(conv2d_1_bn)

    conv2d_2 = tf.keras.layers.TimeDistributed(tf.keras.layers.Conv2D(filters=128, kernel_size=(3, 3), strides=(2, 2)))(
        conv2d_1_ac)
    conv2d_2_bn = tf.keras.layers.BatchNormalization()(conv2d_2)
    conv2d_1_ac = tf.keras.layers.LeakyReLU()(conv2d_2_bn)

    conv2d_3 = tf.keras.layers.TimeDistributed(tf.keras.layers.Conv2D(filters=1, kernel_size=(1, 1), strides=(1, 1)))(
        conv2d_1_ac)
    # conv2d_3_gap = tf.keras.layers.GlobalAveragePooling2D()(conv2d_3)
    flattened_conv = tf.keras.layers.TimeDistributed(tf.keras.layers.Flatten())(conv2d_3)

    predictions = tf.keras.layers.TimeDistributed(Dense(1, activation='sigmoid'))(flattened_conv)
    lstm_full_conv = Model(input_sequence, predictions, name='FullConvLSTM')
    return lstm_full_conv


def create_effnetb0_keypoint_classifier(
        time_steps,
        conv_lstm_size,
        input_shape,
        effnet_trainable=True,
        fine_tune_weights=None,
        freeze_bns=False):
    if fine_tune_weights is not None:
        effnet_model = load_model(fine_tune_weights)
    else:
        effnet_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False,
                                                            input_shape=input_shape)

    last_layer = effnet_model.get_layer('top_activation')

    if not effnet_trainable:
        effnet_model.trainable = False
    else:
        effnet_model.trainable = True

    if freeze_bns:
        for layer in effnet_model.layers:
            if isinstance(layer, tf.keras.layers.BatchNormalization):
                layer.trainable = False

    input_cnn = Model(effnet_model.input, last_layer.output)

    # start of lstm model ###############################

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3))
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn)(input_sequence)
    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size, kernel_size=(3, 3), return_sequences=True)(
        time_distributed_effnet)

    # deconv model ######################################

    input_deconv = tf.keras.layers.Input(shape=(5, 5, conv_lstm_size))
    pad_layer = tf.keras.layers.ZeroPadding2D(padding=1)(input_deconv)
    deconv1 = Conv2DTranspose(256, (4, 4), strides=(2, 2))(pad_layer)
    deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
    deconv1_bn = Activation('relu')(deconv1_bn)
    deconv2 = Conv2DTranspose(256, (4, 4), strides=(2, 2))(deconv1_bn)
    deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
    deconv2_bn = Activation('relu')(deconv2_bn)

    deconv3 = Conv2DTranspose(128, (4, 4), strides=(2, 2))(deconv2_bn)
    deconv3_bn = BatchNormalization(momentum=0.9)(deconv3)
    deconv3_bn = Activation('relu')(deconv3_bn)

    deconv4 = Conv2DTranspose(64, (4, 4), strides=(3, 3))(deconv3_bn)
    deconv4_bn = BatchNormalization(momentum=0.9)(deconv4)
    deconv4_bn = Activation('relu')(deconv4_bn)
    deconv_output = Conv2D(2, (1, 1), strides=(1, 1), name='decoder_output')(deconv4_bn)

    deconv_model = Model(input_deconv, deconv_output)

    # full lstm model ###################################

    time_distributed_deconv = tf.keras.layers.TimeDistributed(deconv_model)(conv_lstm_layer)

    lstm_effnet = Model(input_sequence, time_distributed_deconv, name="lstm_keypoint_classifier")

    return lstm_effnet


def create_effnetb0_keypoint_lstm_upsampling_classifier(
        time_steps,
        conv_lstm_size,
        input_shape,
        effnet_trainable=True,
        fine_tune_weights=None,
        freeze_bns=False):
    if fine_tune_weights is not None:
        effnet_model = load_model(fine_tune_weights)
    else:
        effnet_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False,
                                                            input_shape=input_shape)

    last_layer = effnet_model.get_layer('top_activation')

    if not effnet_trainable:
        effnet_model.trainable = False
    else:
        effnet_model.trainable = True

    if freeze_bns:
        for layer in effnet_model.layers:
            if isinstance(layer, tf.keras.layers.BatchNormalization):
                layer.trainable = False

    input_cnn = Model(effnet_model.input, last_layer.output)

    # start of lstm model ###############################

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3))
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn)(input_sequence)
    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size, kernel_size=(3, 3), return_sequences=True)(
        time_distributed_effnet)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D())(conv_lstm_layer)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D())(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size // 2, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D())(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size // 4, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    # deconv model ######################################

    input_deconv = tf.keras.layers.Input(shape=(40, 40, conv_lstm_size // 4))
    # pad_layer = tf.keras.layers.ZeroPadding2D(padding=1)(input_deconv)
    deconv1 = Conv2DTranspose(64, (4, 4), strides=(2, 2))(input_deconv)
    deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
    deconv1_bn = Activation('relu')(deconv1_bn)
    deconv2 = Conv2DTranspose(64, (4, 4), strides=(2, 2))(deconv1_bn)
    deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
    deconv2_bn = Activation('relu')(deconv2_bn)

    deconv_output = Conv2D(2, (1, 1), strides=(1, 1), name='decoder_output')(deconv2_bn)

    deconv_model = Model(input_deconv, deconv_output)

    # full lstm model ###################################

    time_distributed_deconv = tf.keras.layers.TimeDistributed(deconv_model)(time_distributed_lstm_output)

    lstm_effnet = Model(input_sequence, time_distributed_deconv, name="lstm_keypoint_upsamplinglstm_classifier")

    return lstm_effnet


def create_scoped_trunc_effnet(name_scope, weights, unfreeze_7=False, effnet_fully_trainable=False):
    effnet_model = load_model(weights, compile=False)
    if not effnet_fully_trainable:
        effnet_model.trainable = False
    with tf.name_scope(name_scope):
        for layer in effnet_model.layers:
            if unfreeze_7:
                if 'block7' in layer.name:
                    layer.trainable = True
            if isinstance(layer, tf.keras.layers.BatchNormalization):
                layer.trainable = False
        last_classify_layer = effnet_model.get_layer('top_activation')
        trunc_model = Model(effnet_model.input, last_classify_layer.output, name=name_scope)

    return trunc_model


def create_concat_keypoint_lstm_upsampling_classifier(
        time_steps,
        conv_lstm_size,
        input_shape,
        effnet_classify_weights,
        effnet_keypoint_weights,
        effnet_classify_fully_trainable=False,
        effnet_keypoint_fully_trainable=False,
        freeze_bns=True):
    effnet_classify_model = create_scoped_trunc_effnet("classify", effnet_classify_weights, unfreeze_7=True,
                                                       effnet_fully_trainable=effnet_classify_fully_trainable)
    effnet_keypoint_model = create_scoped_trunc_effnet("keypoint", effnet_keypoint_weights, unfreeze_7=False,
                                                       effnet_fully_trainable=effnet_keypoint_fully_trainable)

    input_image_layer = tf.keras.layers.Input(shape=input_shape, name="common_input")
    effnet_classify_model_out = effnet_classify_model(input_image_layer)
    effnet_keypoint_model_out = effnet_keypoint_model(input_image_layer)
    concatenated_cnn_outputs = tf.keras.layers.Concatenate(name="concat_effnets")(
        [effnet_classify_model_out, effnet_keypoint_model_out])

    input_cnn = Model(input_image_layer, concatenated_cnn_outputs, name="combined_classify_keypoint_model")

    # start of lstm model ###############################

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3), name='sequence_input')
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn, name='time_dist_combined_input_cnns')(
        input_sequence)
    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size, kernel_size=(3, 3), return_sequences=True)(
        time_distributed_effnet)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D(),
                                                                   name='time_dist_upsampling_1')(conv_lstm_layer)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D(),
                                                                   name='time_dist_upsampling_2')(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size // 2, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D(),
                                                                   name='time_dist_upsampling_3')(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size // 4, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    # deconv model ######################################

    input_deconv = tf.keras.layers.Input(shape=(40, 40, conv_lstm_size // 4), name='upsampling_input')
    # pad_layer = tf.keras.layers.ZeroPadding2D(padding=1)(input_deconv)
    deconv1 = Conv2DTranspose(64, (4, 4), strides=(2, 2))(input_deconv)
    deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
    deconv1_bn = Activation('relu')(deconv1_bn)
    deconv2 = Conv2DTranspose(64, (4, 4), strides=(2, 2))(deconv1_bn)
    deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
    deconv2_bn = Activation('relu')(deconv2_bn)

    deconv_output = Conv2D(2, (1, 1), strides=(1, 1), name='decoder_output')(deconv2_bn)

    deconv_model = Model(input_deconv, deconv_output)

    # full lstm model ###################################

    time_distributed_deconv = tf.keras.layers.TimeDistributed(deconv_model, name='time_dist_deconv')(
        time_distributed_lstm_output)

    lstm_effnet = Model(input_sequence, time_distributed_deconv, name="lstm_keypoint_classify_upsampling_effnets")

    for i, w in enumerate(lstm_effnet.weights):
        split_name = w.name.split('/')
        new_name = split_name[0] + '_' + str(i) + '/' + split_name[1] + '_' + str(i)
        lstm_effnet.weights[i]._handle_name = new_name

    return lstm_effnet


def create_concat_conv_lstm_classifier(
        time_steps,
        conv_lstm_size,
        input_shape,
        effnet_classify_weights,
        effnet_keypoint_weights,
        freeze_bns=True):
    effnet_classify_model = create_scoped_trunc_effnet("classify", effnet_classify_weights, unfreeze_7=True)
    effnet_keypoint_model = create_scoped_trunc_effnet("keypoint", effnet_keypoint_weights, unfreeze_7=False)

    input_image_layer = tf.keras.layers.Input(shape=input_shape, name="common_input")
    effnet_classify_model_out = effnet_classify_model(input_image_layer)
    effnet_keypoint_model_out = effnet_keypoint_model(input_image_layer)
    concatenated_cnn_outputs = tf.keras.layers.Concatenate(name="concat_effnets")(
        [effnet_classify_model_out, effnet_keypoint_model_out])

    input_cnn = Model(input_image_layer, concatenated_cnn_outputs, name="combined_classify_keypoint_model")

    # start of lstm model ###############################

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3), name='sequence_input')
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn, name='time_dist_combined_input_cnns')(
        input_sequence)
    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_size, kernel_size=(3, 3), return_sequences=True)(
        time_distributed_effnet)
    flattened_conv = tf.keras.layers.TimeDistributed(tf.keras.layers.Flatten())(conv_lstm_layer)
    predictions = tf.keras.layers.TimeDistributed(Dense(1, activation='sigmoid'))(flattened_conv)

    # full lstm model ###################################

    lstm_effnet = Model(input_sequence, predictions, name="lstm_keypoint_classify_upsampling_effnets")

    for i, w in enumerate(lstm_effnet.weights):
        split_name = w.name.split('/')
        new_name = split_name[0] + '_' + str(i) + '/' + split_name[1] + '_' + str(i)
        lstm_effnet.weights[i]._handle_name = new_name

    return lstm_effnet


def compile_created_model_classifier_sgd(created_model, test_neg_weight=1.0):
    created_model.compile(
        optimizer=tf.keras.optimizers.SGD(learning_rate=1e-3),
        loss=tf.keras.losses.BinaryCrossentropy(label_smoothing=0.1),
        metrics=[
            tf.keras.metrics.TruePositives(name='tp'),
            tf.keras.metrics.FalsePositives(name='fp'),
            tf.keras.metrics.TrueNegatives(name='tn'),
            tf.keras.metrics.FalseNegatives(name='fn'),
            tf.keras.metrics.BinaryAccuracy(name='accuracy'),
            tf.keras.metrics.Precision(name='precision'),
            tf.keras.metrics.Recall(name='recall'),
            tf.keras.metrics.AUC(name='pr_auc', curve='PR'),
            tf.keras.metrics.AUC(name='roc_auc', curve='ROC')
        ])

    return created_model


def compile_created_model_classifier_rmsprop(created_model, test_neg_weight=1.0):
    created_model.compile(
        optimizer=tf.keras.optimizers.RMSprop(learning_rate=1e-3),
        loss=tf.keras.losses.BinaryCrossentropy(label_smoothing=0.1),
        metrics=[
            tf.keras.metrics.TruePositives(name='tp'),
            tf.keras.metrics.FalsePositives(name='fp'),
            tf.keras.metrics.TrueNegatives(name='tn'),
            tf.keras.metrics.FalseNegatives(name='fn'),
            tf.keras.metrics.BinaryAccuracy(name='accuracy'),
            tf.keras.metrics.Precision(name='precision'),
            tf.keras.metrics.Recall(name='recall'),
            tf.keras.metrics.AUC(name='pr_auc', curve='PR'),
            tf.keras.metrics.AUC(name='roc_auc', curve='ROC')
        ])

    return created_model


def compile_created_model_classifier(created_model):
    created_model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=1e-3),
        loss=tf.keras.losses.BinaryCrossentropy(label_smoothing=0.1),
        metrics=[
            tf.keras.metrics.TruePositives(name='tp'),
            tf.keras.metrics.FalsePositives(name='fp'),
            tf.keras.metrics.TrueNegatives(name='tn'),
            tf.keras.metrics.FalseNegatives(name='fn'),
            tf.keras.metrics.BinaryAccuracy(name='accuracy'),
            tf.keras.metrics.Precision(name='precision'),
            tf.keras.metrics.Recall(name='recall'),
            tf.keras.metrics.AUC(name='pr_auc', curve='PR'),
            tf.keras.metrics.AUC(name='roc_auc', curve='ROC')
        ])

    return created_model


def create_detail_resnetv2_classifier_pretrained(input_shape, detail_weights):
    trained_model = create_model_from_weights("resnetv2-classification-pretrained", detail_weights)
    last_layer = trained_model.get_layer('post_relu')

    pooled_conv_output = GlobalAveragePooling2D()(last_layer.output)
    flattened_output = Dense(1024, activation='relu')(pooled_conv_output)
    flattened_output = Dense(512, activation='relu')(flattened_output)
    predictions = Dense(1, activation='sigmoid')(flattened_output)

    return trained_model.input, predictions


def load_concat_model(weights):
    my_model = load_model(weights, compile=False)
    for i, w in enumerate(my_model.weights):
        split_name = w.name.split('/')
        new_name = split_name[0] + '_' + str(i) + '/' + split_name[1] + '_' + str(i)
        my_model.weights[i]._handle_name = new_name
    for layer in my_model.get_layer('time_dist_combined_input_cnns').submodules:
        layer.trainable = False
    return my_model


if __name__ == '__main__':
    pretrained_weights = "/home/florin/diss_models/effnetb0_classify_basic/simple_full_set_weighted_ls0.1/1_chkpt/simple_full_set_weighted_ls0.1_best.094.h5"
    pretrained_keypoint_weights = "/home/florin/diss_models/effnetb0/effnetb0_detail_u2_ft_ra9_best.179.h5"
    # my_model = create_concat_keypoint_lstm_upsampling_classifier(
    #    3, 128, (224,224,3), effnet_classify_weights=pretrained_weights,
    #    effnet_keypoint_weights=pretrained_keypoint_weights, freeze_bns=True)

    inference_weights = "/home/florin/diss_models/effnetb0_classify_rnn_kp_diss/rnn_kp_ts3_shift3_lstm128_ctd_bnf_adam_upc/1_chkpt/rnn_kp_ts3_shift3_lstm128_ctd_bnf_adam_upc_best.069.h5"
    # my_model.load_weights(inference_weights)
    my_model = load_model(inference_weights)

    for layer in my_model.get_layer('time_dist_combined_input_cnns').submodules:
        if 'block' in layer.name:
            if not isinstance(layer, tf.keras.layers.BatchNormalization):
                layer.trainable = True

    my_model.summary()
    # for submod in my_model.get_layer('time_dist_combined_input_cnns').submodules[12].layers:
    # print(submod.name)

    # for i, w in enumerate(my_model.weights):
    #    split_name = w.name.split('/')
    #    new_name = split_name[0] + '_' + str(i) + '/' + split_name[1] + '_' + str(i)
    #    my_model.weights[i]._handle_name = tf.Variable(w, name=new_name)
