import tensorflow as tf
from tensorflow.python.keras.layers import Dense, Conv2DTranspose, Activation, Conv2D, BatchNormalization
from tensorflow.python.keras.models import load_model, Model


def get_last_layer_keypoint(my_model):
    return my_model.get_layer('top_activation')


def get_renamed_backbone(my_model, infix, unfreeze_last_block):
    for i, w in enumerate(my_model.weights):
        split_name = w.name.split('/')
        new_name = split_name[0] + '_' + infix + '_' + str(i) + '/' + split_name[1] + '_' + infix + '_' + str(i)
        my_model.weights[i]._handle_name = new_name

    for layer in my_model.layers:
        layer.trainable = False
        if unfreeze_last_block:
            if 'block7' in layer.name:
                layer.trainable = True
        if isinstance(layer, tf.keras.layers.BatchNormalization):
            layer.trainable = False

    last_backbone_layer = my_model.get_layer('top_activation')
    truncated_model = Model(my_model.input, last_backbone_layer.output, name="backbone_{}".format(infix))
    return truncated_model


def create_combo_sequence_keypoint_classifier(
        keypoint_model_path,
        classification_model_path,
        input_shape,
        time_steps,
        conv_lstm_filters,
        unfreeze_last_keypoint_block,
        unfreeze_last_classifier_block):
    keypoint_model = load_model(keypoint_model_path, compile=False)
    keypoint_backbone = get_renamed_backbone(keypoint_model, "kp", unfreeze_last_block=unfreeze_last_keypoint_block)

    classifier_model = load_model(classification_model_path, compile=False)
    classifier_backbone = get_renamed_backbone(classifier_model.get_layer('time_distributed').layer, "cls",
                                               unfreeze_last_block=unfreeze_last_classifier_block)

    input_image_layer = tf.keras.layers.Input(shape=input_shape, name="common_input")
    effnet_classify_model_out = keypoint_backbone(input_image_layer)
    effnet_keypoint_model_out = classifier_backbone(input_image_layer)
    concatenated_cnn_outputs = tf.keras.layers.Concatenate(name="concat_effnets")(
        [effnet_classify_model_out, effnet_keypoint_model_out])

    input_cnn = Model(input_image_layer, concatenated_cnn_outputs, name="combo_sequence_keypoint_classifier_input")

    # start of lstm model ###############################

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3), name='sequence_input')
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn, name='time_dist_combined_input_cnns')(
        input_sequence)

    # upsampling ConvLSTM

    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_filters, kernel_size=(3, 3), return_sequences=True)(
        time_distributed_effnet)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D(),
                                                                   name='time_dist_upsampling_1')(conv_lstm_layer)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_filters, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D(),
                                                                   name='time_dist_upsampling_2')(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_filters // 2, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.TimeDistributed(tf.keras.layers.UpSampling2D(),
                                                                   name='time_dist_upsampling_3')(
        time_distributed_lstm_output)
    time_distributed_lstm_output = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_filters // 4, kernel_size=(3, 3),
                                                              padding='same', return_sequences=True)(
        time_distributed_lstm_output)

    # deconv model ######################################

    input_deconv = tf.keras.layers.Input(shape=(40, 40, conv_lstm_filters // 4), name='upsampling_input')

    deconv1 = Conv2DTranspose(64, (4, 4), strides=(2, 2))(input_deconv)
    deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
    deconv1_bn = Activation('relu')(deconv1_bn)
    deconv2 = Conv2DTranspose(64, (4, 4), strides=(2, 2))(deconv1_bn)
    deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
    deconv2_bn = Activation('relu')(deconv2_bn)

    deconv_output = Conv2D(2, (1, 1), strides=(1, 1), name='decoder_output')(deconv2_bn)

    deconv_model = Model(input_deconv, deconv_output)

    # full lstm model ###################################

    time_distributed_deconv = tf.keras.layers.TimeDistributed(deconv_model, name='time_dist_deconv')(
        time_distributed_lstm_output)

    combo_effnet = Model(input_sequence, time_distributed_deconv, name="combo_sequence_keypoint_classifier")
    return combo_effnet


# Create a combination model from a keypoint detector backbone
# and a sequence classifier
def create_combo_sequence_classifier(
        keypoint_model_path,
        classification_model_path,
        input_shape,
        time_steps,
        conv_lstm_filters,
        unfreeze_last_keypoint_block,
        unfreeze_last_classifier_block):
    # load the model without optimizer weights
    keypoint_model = load_model(keypoint_model_path, compile=False)
    keypoint_backbone = get_renamed_backbone(keypoint_model, "kp", unfreeze_last_block=unfreeze_last_keypoint_block)

    classifier_model = load_model(classification_model_path, compile=False)
    classifier_backbone = get_renamed_backbone(classifier_model.get_layer('time_distributed').layer, "cls",
                                               unfreeze_last_block=unfreeze_last_classifier_block)

    input_image_layer = tf.keras.layers.Input(shape=input_shape, name="common_input")
    effnet_classify_model_out = keypoint_backbone(input_image_layer)
    effnet_keypoint_model_out = classifier_backbone(input_image_layer)
    concatenated_cnn_outputs = tf.keras.layers.Concatenate(name="concat_effnets")(
        [effnet_classify_model_out, effnet_keypoint_model_out])

    input_cnn = Model(input_image_layer, concatenated_cnn_outputs, name="combo_sequence_classifier_input")

    # start of lstm model ###############################

    input_sequence = tf.keras.layers.Input(shape=(time_steps, 224, 224, 3), name='sequence_input')
    time_distributed_effnet = tf.keras.layers.TimeDistributed(input_cnn, name='time_dist_combined_input_cnns')(
        input_sequence)
    conv_lstm_layer = tf.keras.layers.ConvLSTM2D(filters=conv_lstm_filters,
                                                 kernel_size=(3, 3),
                                                 return_sequences=True,
                                                 name="sequence_classifier_convlstm_block")(time_distributed_effnet)
    flattened_conv = tf.keras.layers.TimeDistributed(tf.keras.layers.Flatten())(conv_lstm_layer)
    predictions = tf.keras.layers.TimeDistributed(Dense(1, activation='sigmoid'))(flattened_conv)

    # full lstm model ###################################

    combo_effnet = Model(input_sequence, predictions, name="combo_sequence_classifier")

    return combo_effnet
