import tensorflow as tf
from tensorflow.keras.applications import ResNet50V2, NASNetMobile
from tensorflow.keras.layers import Conv2DTranspose, BatchNormalization, Activation, Conv2D, UpSampling2D
from tensorflow.keras.models import Model
###############################################################################
# backbones
from tensorflow.python.keras.applications.resnet import ResNet50


def create_backbone_effnetB0(input_shape, freeze):
    effnet_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False, input_shape=input_shape)
    last_layer = effnet_model.get_layer('top_activation')

    if freeze:
        for layer in effnet_model.layers:
            if "block1" in layer.name or "block2" in layer.name:
                layer.trainable = False

    return effnet_model.input, last_layer.output


def create_backbone_effnetB5(input_shape, freeze):
    effnet_model = tf.keras.applications.EfficientNetB5(weights='imagenet', include_top=False, input_shape=input_shape)
    last_layer = effnet_model.get_layer('top_activation')

    if freeze:
        for layer in effnet_model.layers:
            if "block1" in layer.name or "block2" in layer.name:
                layer.trainable = False

    return effnet_model.input, last_layer.output


def create_backbone_effnetB7(input_shape, freeze):
    effnet_model = tf.keras.applications.EfficientNetB7(weights='imagenet', include_top=False, input_shape=input_shape)
    last_layer = effnet_model.get_layer('top_activation')

    if freeze:
        for layer in effnet_model.layers:
            if "block1" in layer.name or "block2" in layer.name:
                layer.trainable = False

    return effnet_model.input, last_layer.output


def create_backbone_resnetv2(input_shape, freeze):
    resnet_model = ResNet50V2(weights='imagenet', include_top=False, input_shape=input_shape)
    last_layer = resnet_model.get_layer('post_relu')

    if freeze:
        for layer in resnet_model.layers:
            if "conv2" in layer.name or "conv1" in layer.name:
                layer.trainable = False

    return resnet_model.input, last_layer.output


def create_backbone_resnetv1(input_shape, freeze):
    resnet_model = ResNet50(weights='imagenet', include_top=False, input_shape=input_shape)
    last_layer = resnet_model.get_layer('conv5_block3_out')

    if freeze:
        for layer in resnet_model.layers:
            if "conv2" in layer.name or "conv1" in layer.name:
                layer.trainable = False

    return resnet_model.input, last_layer.output


# nasnet mobile input shape is 224, 224, 3
def create_backbone_nasnetmobile(input_shape, freeze):
    nasnet_model = NASNetMobile(input_shape=input_shape, include_top=False, weights='imagenet')
    last_layer = nasnet_model.get_layer('activation_187')

    if freeze:
        layer_bound = len(nasnet_model.layers) // 5
        for index, layer in enumerate(nasnet_model.layers):
            if index < layer_bound:
                layer.trainable = False

    return nasnet_model.input, last_layer.output


###############################################################################
# upsampling architecture

# fcn
def create_upsampling_part(previous_layer_output, output_shape, upsampling_layers):
    output_depth = output_shape[2]
    if upsampling_layers == 2:
        deconv1 = Conv2DTranspose(256, (4, 4), strides=(2, 2))(previous_layer_output)
        deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
        deconv1_bn = Activation('relu')(deconv1_bn)
        deconv2 = Conv2DTranspose(256, (4, 4), strides=(2, 2))(deconv1_bn)
        deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
        deconv2_bn = Activation('relu')(deconv2_bn)
        last_conv = Conv2D(output_depth, (1, 1), strides=(1, 1), name='decoder_output')(deconv2_bn)
        output = UpSampling2D(size=(6, 6))(last_conv)
    elif upsampling_layers == 4:
        deconv1 = Conv2DTranspose(256, (4, 4), strides=(2, 2))(previous_layer_output)
        deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
        deconv1_bn = Activation('relu')(deconv1_bn)
        deconv2 = Conv2DTranspose(256, (4, 4), strides=(2, 2))(deconv1_bn)
        deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
        deconv2_bn = Activation('relu')(deconv2_bn)

        deconv3 = Conv2DTranspose(128, (4, 4), strides=(2, 2))(deconv2_bn)
        deconv3_bn = BatchNormalization(momentum=0.9)(deconv3)
        deconv3_bn = Activation('relu')(deconv3_bn)

        deconv4 = Conv2DTranspose(64, (4, 4), strides=(3, 3))(deconv3_bn)
        deconv4_bn = BatchNormalization(momentum=0.9)(deconv4)
        deconv4_bn = Activation('relu')(deconv4_bn)
        output = Conv2D(output_depth, (1, 1), strides=(1, 1), name='decoder_output')(deconv4_bn)
        # output = UpSampling2D(size=(output_shape[0], output_shape[1]))(last_conv)
    elif upsampling_layers == 3:
        # case for 456x456 input map of effnetb5
        deconv1 = Conv2DTranspose(256, (6, 6), strides=(3, 3))(previous_layer_output)
        deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
        deconv1_bn = Activation('relu')(deconv1_bn)

        deconv2 = Conv2DTranspose(256, (6, 6), strides=(3, 3))(deconv1_bn)
        deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
        deconv2_bn = Activation('relu')(deconv2_bn)

        deconv3 = Conv2DTranspose(128, (6, 6), strides=(3, 3))(deconv2_bn)
        deconv3_bn = BatchNormalization(momentum=0.9)(deconv3)
        deconv3_bn = Activation('relu')(deconv3_bn)

        output = Conv2D(output_depth, (1, 1), strides=(1, 1), name='decoder_output')(deconv3_bn)
    else:
        print("Unknown number of upsampling layers specified.")
        return None
    return output


###############################################################################
# model creation

def create_model_effnet_b0(input_shape, output_shape, upsampling_layers, freeze):
    effnet_input, effnet_output = create_backbone_effnetB0(input_shape, freeze)
    upsampling_output = create_upsampling_part(effnet_output, output_shape, upsampling_layers)

    my_model = Model(effnet_input, upsampling_output, name='effnetb0_u{}_f{}'.format(upsampling_layers, freeze))

    return my_model


def create_model_effnet_b5(input_shape, output_shape, upsampling_layers, freeze):
    effnet_input, effnet_output = create_backbone_effnetB5(input_shape, freeze)
    upsampling_output = create_upsampling_part(effnet_output, output_shape, upsampling_layers)

    my_model = Model(effnet_input, upsampling_output, name='effnetb5_u{}_f{}'.format(upsampling_layers, freeze))

    return my_model


def create_model_effnet_b7(input_shape, output_shape, upsampling_layers, freeze):
    effnet_input, effnet_output = create_backbone_effnetB7(input_shape, freeze)
    upsampling_output = create_upsampling_part(effnet_output, output_shape, upsampling_layers)

    my_model = Model(effnet_input, upsampling_output, name='effnetb7_u{}_f{}'.format(upsampling_layers, freeze))

    return my_model


def create_model_resnet_v2(input_shape, output_shape, upsampling_layers, freeze):
    resnet_input, resnet_output = create_backbone_resnetv2(input_shape, freeze)
    upsampling_output = create_upsampling_part(resnet_output, output_shape, upsampling_layers)

    my_model = Model(resnet_input, upsampling_output, name='resnetv2_u{}_f{}'.format(upsampling_layers, freeze))

    return my_model


def create_model_resnet_v1(input_shape, output_shape, upsampling_layers, freeze):
    resnet_input, resnet_output = create_backbone_resnetv1(input_shape, freeze)
    upsampling_output = create_upsampling_part(resnet_output, output_shape, upsampling_layers)

    my_model = Model(resnet_input, upsampling_output, name='resnetv1_u{}_f{}'.format(upsampling_layers, freeze))

    return my_model


def create_model_nasnetmobile(input_shape, output_shape, upsampling_layers, freeze):
    nasnetmobile_input, nasnetmobile_output = create_backbone_nasnetmobile(input_shape, freeze)
    upsampling_output = create_upsampling_part(nasnetmobile_output, output_shape, upsampling_layers)

    my_model = Model(nasnetmobile_input, upsampling_output,
                     name='nasnetmobile_u{}_f{}'.format(upsampling_layers, freeze))

    return my_model


def preprocess_resnetv2_input(x):
    tf.debugging.assert_type(x, tf.float32)
    tf.debugging.assert_less_equal(tf.reduce_max(x), 1.0)
    tf.debugging.assert_greater_equal(tf.reduce_max(x), 0.0)
    x = x * 255.0
    x = tf.keras.applications.resnet_v2.preprocess_input(x)
    return x


def preprocess_resnetv1_input(x):
    tf.debugging.assert_type(x, tf.float32)
    tf.debugging.assert_less_equal(tf.reduce_max(x), 1.0)
    tf.debugging.assert_greater_equal(tf.reduce_max(x), 0.0)
    x = x * 255.0
    x = tf.keras.applications.resnet.preprocess_input(x)
    return x


def preprocess_nasnetmobile_input(x):
    tf.debugging.assert_type(x, tf.float32)
    tf.debugging.assert_less_equal(tf.reduce_max(x), 1.0)
    tf.debugging.assert_greater_equal(tf.reduce_max(x), 0.0)
    x = x * 255.0
    x = tf.keras.applications.nasnet.preprocess_input(x)
    return x


def preprocess_effnet_input(x):
    tf.debugging.assert_type(x, tf.float32)
    tf.debugging.assert_less_equal(tf.reduce_max(x), 1.0)
    tf.debugging.assert_greater_equal(tf.reduce_max(x), 0.0)
    # x = x * 255.0
    x = tf.keras.applications.efficientnet.preprocess_input(x)
    return x


model_to_creation_method_map = {
    "resnetv2": create_model_resnet_v2,
    "resnetv1": create_model_resnet_v1,
    "nasnetmobile": create_model_nasnetmobile,
    "effnetb0": create_model_effnet_b0,
    "effnetb5": create_model_effnet_b5,
    "effnetb6": create_model_effnet_b7
}

model_to_preprocess_map = {
    "resnetv2": preprocess_resnetv2_input,
    "resnetv1": preprocess_resnetv1_input,
    "nasnetmobile": preprocess_nasnetmobile_input,
    "effnetb0": preprocess_effnet_input,
    "effnetb5": preprocess_effnet_input,
    "effnetb7": preprocess_effnet_input
}


###############################################################################
# model management


# todo: maybe try some other lr scheme
def compile_created_model_keypoint(created_model):
    # lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(
    #    initial_learning_rate=0.001,
    #    decay_steps=25000,
    #    decay_rate=0.96
    # )

    created_model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=1e-3),
        loss=tf.keras.losses.MeanSquaredError(),
        metrics=[tf.keras.metrics.MeanSquaredError()])

    return created_model


def save_compiled_model(compiled_model, saved_model_dir):
    compiled_model.save(saved_model_dir, save_format='tf')


def get_preprocess_func(arch_key):
    return model_to_preprocess_map[arch_key]


def create_model(arch_key, input_shape, output_shape, upsampling_layers, freeze):
    model_creation_method = model_to_creation_method_map[arch_key]
    my_model = model_creation_method(input_shape, output_shape, upsampling_layers, freeze)

    if freeze:
        tf.print("Using layer freezing:")
        for layer in my_model.layers:
            if layer.trainable is False:
                tf.print("Layer {} is frozen.".format(layer.name))

    my_model = compile_created_model_keypoint(my_model)
    my_model.summary()
    return my_model
