import os

import numpy as np
import tensorflow as tf

from research.ablation_studies.dataset.dataset_utils import image_to_sequence_tfexample, \
    write_tfrecord_with_examples
from research.ablation_studies.sequence.video_streamer import VideoFrameStreamer


def gather_video_files(video_dir, video_format='.mp4'):
    video_files = os.listdir(video_dir)
    video_files = [video_file for video_file in video_files if video_format in video_file]
    return video_files


def split_video_name(file):
    filename = os.path.split(file)[1][:-4]
    split_fn = filename.split('_')
    subject = split_fn[0]
    task = split_fn[1]
    return subject, task


def get_sequence_starts_stops(frame_count, sequence_length, sequence_count):
    interval_length = frame_count // sequence_count
    if interval_length > sequence_length:
        offset_range = (interval_length - sequence_length)
    elif interval_length < sequence_length:
        print("Error: Not enough frames ({}) for {} sequences of length {}".format(frame_count, sequence_count,
                                                                                   sequence_length))
        return None
    else:
        offset_range = 0

    start_points = [(frame_count // sequence_count) * i + np.random.randint(low=0, high=offset_range) for i in
                    range(sequence_count)]
    return start_points


def split_val1_anno_filename(anno_filename):
    split_name = anno_filename.split('_')
    split_subject = int(split_name[0])
    split_frame = int(split_name[1])
    return split_subject, split_frame


def get_val1_pen_down_frames(val1_anno_dir):
    down_dir = os.path.join(val1_anno_dir, "pen_down/")
    up_dir = os.path.join(val1_anno_dir, "pen_up/")
    down_files = os.listdir(down_dir)
    up_files = os.listdir(up_dir)
    down_frames = [file[:-4] for file in down_files if '.jpg' in file]
    # up_frames = [file[:-4] for file in up_files if '.jpg' in file]

    down_subjects = {}
    for frame in down_frames:
        split_name = frame.split('_')
        split_subject = int(split_name[0])
        split_frame = int(split_name[1])
        if split_subject not in down_subjects:
            down_subjects[split_subject] = [split_frame]
        else:
            down_subjects[split_subject].append(split_frame)
    return down_subjects


def get_val1_frame_anno(subject, frame, pen_down_frames):
    subject_offset = {
        0: 24,
        1: 2,
        2: 33,
        3: 46,
        4: 1
    }
    if subject in pen_down_frames:
        offset_frame = frame - subject_offset[subject]
        if offset_frame in pen_down_frames[subject]:
            return 1
    return 0


def create_sequences_from_videos(dataset_prefix, video_files, sequences_per_video, sequence_length, do_val1_anno=False,
                                 val1_anno_down_frames=None):
    for file in video_files:
        print("Reading {}".format(file))
        camera_stream = VideoFrameStreamer(os.path.join(video_dir, file))
        frame_count = camera_stream.get_frame_count()
        fps = camera_stream.get_fps()
        seconds = frame_count / fps
        print("{} has {} frames. At {} fps, length is {}".format(file, frame_count, fps, seconds))
        subject, task = split_video_name(file)
        sequence_start_points = get_sequence_starts_stops(frame_count, sequence_length, sequences_per_video)
        sequence_counter = 0
        sequence_examples = []
        for start_point in sequence_start_points:
            current_frame = camera_stream.seek_to_and_get(start_point)
            frame_counter = start_point
            print("Video frame shape is {}".format(current_frame.shape))
            print("Sequence {} - Start point is {}".format(sequence_counter, start_point))

            while current_frame is not None and frame_counter < (start_point + sequence_length):
                resized_frame = tf.image.convert_image_dtype(current_frame, tf.float32)
                resized_frame = tf.image.resize(resized_frame, (1080, 1920))
                resized_frame = tf.image.convert_image_dtype(resized_frame, tf.uint8)
                resized_img_bytes = tf.io.encode_jpeg(resized_frame, format="rgb", chroma_downsampling=False,
                                                      optimize_size=True)

                pen_down = 0
                if do_val1_anno:
                    pen_down = get_val1_frame_anno(int(subject), frame_counter, val1_anno_down_frames)

                example = image_to_sequence_tfexample(resized_img_bytes,
                                                      b'jpg',
                                                      int(subject),
                                                      int(frame_counter),
                                                      int(task),
                                                      int(sequence_counter),
                                                      1080, 1920,
                                                      -1.0, -1.0,
                                                      -1.0, -1.0,
                                                      pen_down)  # pen_down

                sequence_examples.append(example)
                current_frame = camera_stream.get()
                frame_counter += 1

            print("Created {} examples.".format(len(sequence_examples)))
            sequence_counter += 1

        tfrecord_filename = "{}_s{}_t{}_sequences.tfrecord".format(dataset_prefix, subject, task)
        print("Writing tfrecord {}".format(tfrecord_filename))
        write_tfrecord_with_examples(tfrecord_filename, sequence_examples)
        sequence_examples.clear()
    return


def create_val1_frame_example(image_path, subject_id, frame_id, sequence_id, pen_down_frames):
    frame_bytes = tf.io.read_file(image_path)
    pen_down = get_val1_frame_anno(subject_id, frame_id, pen_down_frames)

    example = image_to_sequence_tfexample(frame_bytes,
                                          b'jpg',
                                          subject_id,
                                          frame_id,
                                          0,
                                          sequence_id,
                                          1080, 1920,
                                          -1.0, -1.0,
                                          -1.0, -1.0,
                                          pen_down)  # pen_down
    return example


if __name__ == '__main__':
    # video_dir = "/media/florin/Transcend/large_dataset_study/cut_videos/"
    video_dir = "/home/florin/data/val1/videos/"
    video_files = gather_video_files(video_dir, video_format='.mp4')

    val1_anno_dir = "/home/florin/data/val1/updown/"
    val1_anno_down_frames = get_val1_pen_down_frames(val1_anno_dir)

    dataset_prefix = "val1"
    sequences_per_video = 6
    sequence_length = 120

    create_sequences_from_videos(
        dataset_prefix,
        video_files,
        sequences_per_video,
        sequence_length,
        do_val1_anno=True,
        val1_anno_down_frames=val1_anno_down_frames)
