import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.python.keras.models import load_model

from research.ablation_studies.dataset.dataset_utils import sequence_image_feature_description, gather_tfrecords
from research.ablation_studies.dataset.pipeline_creator import ClassificationSequenceInputMap, SequencePipelineCreator
from research.ablation_studies.model.create_keypoint_models import get_preprocess_func


def get_inferences_for_model(input_patches, model, timesteps, max_sequence_length):
    if timesteps == max_sequence_length:
        return model(input_patches)
    if timesteps == 1:
        inferences = model(input_patches[0])
        return np.expand_dims(inferences, axis=0)
    if timesteps > max_sequence_length:
        print("Error: timesteps exceed maximum length for inference.")
        return None
    if max_sequence_length % timesteps == 0:
        collected_inferences = []
        for i in range(max_sequence_length // timesteps):
            extracted_patches_batch = input_patches[0]
            extracted_patches_batch = extracted_patches_batch[i * timesteps: (i + 1) * timesteps]
            extracted_patches_batch = np.expand_dims(extracted_patches_batch, axis=0)
            current_inferences = model(extracted_patches_batch)
            collected_inferences.extend(current_inferences[0])
        collected_inferences = np.expand_dims(collected_inferences, axis=0)
        return collected_inferences
    else:
        print("Error: maximum length not divisible by timesteps.")
        return None


def process_labels_into_results_dict(labels, results_dict):
    for label in labels[0]:
        results_dict['pen_down'].append(label.numpy())
    return


def process_sequence_into_results_dict(net_name, inference_sequence, results_dict):
    for prediction in inference_sequence[0]:
        results_dict[net_name + '_pred'].append(prediction[0])
    return


def calc_loss_and_ba_for_ensemble(net_list, eval_df):
    all_net_preds = []
    for net in net_list:
        net_preds = eval_df[net + '_pred'].to_numpy().astype(np.float32)
        all_net_preds.append(net_preds)

    trues = eval_df['pen_down'].to_numpy().astype(np.float32)
    all_net_preds = np.array(all_net_preds)

    if len(net_list) == 1:
        combined_preds = all_net_preds[0]
    else:
        combined_preds = np.mean(all_net_preds, axis=0)

    loss = tf.keras.metrics.BinaryCrossentropy(label_smoothing=0.1)
    loss.update_state(y_true=trues, y_pred=combined_preds)
    result_loss = loss.result().numpy()

    tp = tf.keras.metrics.TruePositives()
    tp.update_state(y_true=trues, y_pred=combined_preds)
    result_tp = tp.result().numpy()

    tn = tf.keras.metrics.TrueNegatives()
    tn.update_state(y_true=trues, y_pred=combined_preds)
    result_tn = tn.result().numpy()

    fp = tf.keras.metrics.FalsePositives()
    fp.update_state(y_true=trues, y_pred=combined_preds)
    result_fp = fp.result().numpy()

    fn = tf.keras.metrics.FalseNegatives()
    fn.update_state(y_true=trues, y_pred=combined_preds)
    result_fn = fn.result().numpy()

    TPR = result_tp / (result_tp + result_fn)
    TNR = result_tn / (result_tn + result_fp)

    ba = 0.5 * (TPR + TNR)

    return result_loss, ba


if __name__ == '__main__':
    print("Ensemble evaluation")

    ab2_model_weights = "/home/florin/diss_models/final_weights/sequence_analysis/ab-2-magnitude-6-label-smoothing-0.1.h5"
    fgru_model_weights = "/home/florin/diss_models/final_weights/sequence_analysis/f-gru-3-64.h5"
    convlstm_model_weights = "/home/florin/diss_models/final_weights/sequence_analysis/convlstm-30-64.h5"
    combo_model_weights = "/home/florin/diss_models/final_weights/sequence_analysis/cls_combo_ts10_shift10_convlstm64.142.h5"

    tfrecord_dir = "/home/florin/data/sequence/"
    test_subjects = [0, 6, 10, 14, 20, 23, 38, 39]

    ab2_model = load_model(ab2_model_weights)
    fgru_model = load_model(fgru_model_weights)
    convlstm_model = load_model(convlstm_model_weights)
    combo_model = load_model(combo_model_weights)

    ab2_timesteps = 1
    fgru_timesteps = 3
    convlstm_timesteps = 30
    combo_timesteps = 10

    max_sequence_length = np.max([fgru_timesteps, convlstm_timesteps, combo_timesteps])

    tfrecord_files = gather_tfrecords(tfrecord_dir=tfrecord_dir)

    results_dict = {
        'pen_down': [],
        'fgru_pred': [],
        'convlstm_pred': [],
        'ab2_pred': [],
        'combo_pred': []
    }

    input_mapper = ClassificationSequenceInputMap(
        (224, 224, 3),
        (0, 0),
        6,
        downscale_ratio=1,
        input_prep_func=get_preprocess_func("effnetb0"),
        use_augmentation=True)

    pipeline = SequencePipelineCreator(
        input_mapper,
        1,
        tfrecord_files,
        image_features=sequence_image_feature_description,
        timesteps=max_sequence_length,
        shift=max_sequence_length,
        test_subjects=test_subjects)

    ds_train, ds_test = pipeline.get_datasets()

    for element in ds_test:
        input_patches = element[0]
        labels = element[1]

        process_labels_into_results_dict(labels, results_dict)

        ab2_inference = get_inferences_for_model(input_patches, ab2_model, ab2_timesteps, max_sequence_length)
        fgru_inference = get_inferences_for_model(input_patches, fgru_model, fgru_timesteps, max_sequence_length)
        convlstm_inference = get_inferences_for_model(input_patches, convlstm_model, convlstm_timesteps,
                                                      max_sequence_length)
        combo_inference = get_inferences_for_model(input_patches, combo_model, combo_timesteps, max_sequence_length)

        process_sequence_into_results_dict("ab2", ab2_inference, results_dict)
        process_sequence_into_results_dict("fgru", fgru_inference, results_dict)
        process_sequence_into_results_dict("convlstm", convlstm_inference, results_dict)
        process_sequence_into_results_dict("combo", combo_inference, results_dict)

    eval_df = pd.DataFrame(results_dict)
    eval_df.to_csv("ensemble_preds.csv")

    ab2_loss, ab2_ba = calc_loss_and_ba_for_ensemble(["ab2"], eval_df)
    fgru_loss, fgru_ba = calc_loss_and_ba_for_ensemble(["fgru"], eval_df)
    convlstm_loss, convlstm_ba = calc_loss_and_ba_for_ensemble(["convlstm"], eval_df)
    combo_loss, combo_ba = calc_loss_and_ba_for_ensemble(["combo"], eval_df)

    print("AB2 Loss: {}".format(ab2_loss))
    print("AB2 BA: {}".format(ab2_ba))
    print("FGRU Loss: {}".format(fgru_loss))
    print("FGRU BA: {}".format(fgru_ba))
    print("CONVLSTM Loss: {}".format(convlstm_loss))
    print("CONVLSTM BA: {}".format(convlstm_ba))
    print("COMBO Loss: {}".format(combo_loss))
    print("COMBO BA: {}".format(combo_ba))

    ab2_fgru_loss, ab2_fgru_ba = calc_loss_and_ba_for_ensemble(["ab2", "fgru"], eval_df)
    print("AB2-FGRU Loss: {}".format(ab2_fgru_loss))
    print("AB2-FGRU BA: {}".format(ab2_fgru_ba))

    ab2_convlstm_loss, ab2_convlstm_ba = calc_loss_and_ba_for_ensemble(["ab2", "convlstm"], eval_df)
    print("AB2-CONVLSTM Loss: {}".format(ab2_convlstm_loss))
    print("AB2-CONVLSTM BA: {}".format(ab2_convlstm_ba))

    ab2_combo_loss, ab2_combo_ba = calc_loss_and_ba_for_ensemble(["ab2", "combo"], eval_df)
    print("AB2-COMBO Loss: {}".format(ab2_combo_loss))
    print("AB2-COMBO BA: {}".format(ab2_combo_ba))

    combo_convlstm_loss, combo_convlstm_ba = calc_loss_and_ba_for_ensemble(["combo", "convlstm"], eval_df)
    print("COMBO-CONVLSTM Loss: {}".format(combo_convlstm_loss))
    print("COMBO-CONVLSTM BA: {}".format(combo_convlstm_ba))

    ab2_combo_convlstm_loss, ab2_combo_convlstm_ba = calc_loss_and_ba_for_ensemble(["ab2", "combo", "convlstm"],
                                                                                   eval_df)
    print("AB2-COMBO-CONVLSTM Loss: {}".format(ab2_combo_convlstm_loss))
    print("AB2-COMBO-CONVLSTM BA: {}".format(ab2_combo_convlstm_ba))
