import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.python.keras.models import load_model

from research.ablation_studies.dataset.dataset_utils import gather_tfrecords, \
    sequence_image_feature_description, get_max_of_heatmap
from research.ablation_studies.dataset.pipeline_creator import ClassificationSequenceInputMap, SequencePipelineCreator
from research.ablation_studies.model.create_keypoint_models import get_preprocess_func


def process_sequence_into_results_dict(inference_sequence, labels, results_dict, bin_inference):
    for index, inf_label_tuple in enumerate(zip(inference_sequence[0], labels[0])):
        heatmaps, label = inf_label_tuple
        pen_up_inference = heatmaps[:, :, 0]
        pen_down_inference = heatmaps[:, :, 1]

        up_max, up_indices = get_max_of_heatmap(pen_up_inference)
        down_max, down_indices = get_max_of_heatmap(pen_down_inference)

        if bin_inference is not None:
            bin_heatmap = bin_inference[index]
            bin_heatmap = tf.image.resize(bin_heatmap, (166, 166))
            # bin_heatmap = tf.image.rgb_to_grayscale(bin_heatmap)
            bin_value = bin_heatmap.numpy()[down_indices[0]][down_indices[1]][0]
            down_max = 0.5 * (bin_value + down_max)
            up_max = 0.5 * ((1.0 - bin_value) + up_max)

        # results_dict['subject_id'].append(element['image/subject_id'])
        # results_dict['frame_id'].append(element['image/frame_id'])
        # results_dict['sequence_id'].append(element['image/sequence_id'])
        results_dict['pen_down'].append(label)
        results_dict['net_down_max'].append(down_max)
        results_dict['net_down_x'].append(down_indices[1])
        results_dict['net_down_y'].append(down_indices[0])
        results_dict['net_up_max'].append(up_max)
        results_dict['net_up_x'].append(up_indices[1])
        results_dict['net_up_y'].append(up_indices[0])
        results_dict['out_size'].append(pen_up_inference.shape[0])
        x = np.array([up_max, down_max])
        softmax = np.exp(x) / sum(np.exp(x))
        results_dict['softmax_prob_down'].append(softmax[1])
        results_dict['naive_prob_down'].append(down_max)

        down_max = down_max[0]
        up_max = up_max[0]
        gt_down = label
        if down_max > up_max and gt_down:
            results_dict['distinct_decision'].append("TP")
        elif down_max < up_max and gt_down:
            results_dict['distinct_decision'].append("FN")
        elif down_max > up_max and not gt_down:
            results_dict['distinct_decision'].append("FP")
        elif down_max < up_max and not gt_down:
            results_dict['distinct_decision'].append("TN")

        if softmax[1] >= 0.5 and gt_down:
            results_dict['softmax_decision'].append("TP")
        elif softmax[1] < 0.5 and gt_down:
            results_dict['softmax_decision'].append("FN")
        elif softmax[1] >= 0.5 and not gt_down:
            results_dict['softmax_decision'].append("FP")
        elif softmax[1] < 0.5 and not gt_down:
            results_dict['softmax_decision'].append("TN")

    return


def get_result_dataframe(model_path, tfrecord_dir, test_subjects, do_binary_ensemble=False):
    if do_binary_ensemble:
        bin_model = load_model(
            "/home/florin/diss_models/effnetb0_text_ex_lorem/correct_pair/1_chkpt/correct_pair_best.156.h5")
    else:
        bin_model = None

    my_model = load_model(model_path)
    for i, w in enumerate(my_model.weights):
        split_name = w.name.split('/')
        new_name = split_name[0] + '_' + str(i) + '/' + split_name[1] + '_' + str(i)
        my_model.weights[i]._handle_name = new_name

    tfrecord_files = gather_tfrecords(tfrecord_dir=tfrecord_dir)

    results_dict = {
        # 'subject_id': [],
        # 'frame_id': [],
        # 'sequence_id': [],
        'pen_down': [],
        'net_down_max': [],
        'net_up_max': [],
        'softmax_prob_down': [],
        'naive_prob_down': [],
        'net_down_x': [],
        'net_down_y': [],
        'net_up_x': [],
        'net_up_y': [],
        'distinct_decision': [],
        'softmax_decision': [],
        'out_size': []
    }

    input_mapper = ClassificationSequenceInputMap(
        (224, 224, 3),
        (0, 0),
        6,
        downscale_ratio=1,
        input_prep_func=get_preprocess_func("effnetb0"),
        use_augmentation=True)

    pipeline = SequencePipelineCreator(
        input_mapper,
        1,
        tfrecord_files,
        image_features=sequence_image_feature_description,
        timesteps=10,
        shift=10,
        test_subjects=test_subjects)
    ds_train, ds_test = pipeline.get_datasets()

    # ds_iter = iter(ds_test)
    # element = next(ds_iter)
    for element in ds_test:
        input_patches = element[0]
        inference = my_model(input_patches)

        if do_binary_ensemble:
            bin_inference = bin_model(input_patches[0])
        else:
            bin_inference = None

        process_sequence_into_results_dict(inference, element[1], results_dict, bin_inference)

    kp_df = pd.DataFrame(results_dict)
    return kp_df


def calc_loss_and_ba(kp_eval_df):
    softmax_preds = kp_eval_df['softmax_prob_down'].to_numpy().astype(np.float32)
    naive_preds = kp_eval_df['naive_prob_down'].to_numpy().astype(np.float32)
    trues = kp_eval_df['pen_down'].to_numpy().astype(np.float32)
    loss = tf.keras.metrics.BinaryCrossentropy(label_smoothing=0.1)
    loss.update_state(y_true=trues, y_pred=softmax_preds)
    result_softmax_loss = loss.result().numpy()

    loss.reset_states()
    loss.update_state(y_true=trues, y_pred=naive_preds)
    result_naive_loss = loss.result().numpy()

    dTP = len(kp_eval_df[kp_eval_df['distinct_decision'] == 'TP'])
    dTN = len(kp_eval_df[kp_eval_df['distinct_decision'] == 'TN'])
    dFP = len(kp_eval_df[kp_eval_df['distinct_decision'] == 'FP'])
    dFN = len(kp_eval_df[kp_eval_df['distinct_decision'] == 'FN'])

    sTP = len(kp_eval_df[kp_eval_df['softmax_decision'] == 'TP'])
    sTN = len(kp_eval_df[kp_eval_df['softmax_decision'] == 'TN'])
    sFP = len(kp_eval_df[kp_eval_df['softmax_decision'] == 'FP'])
    sFN = len(kp_eval_df[kp_eval_df['softmax_decision'] == 'FN'])

    dTPR = dTP / (dTP + dFN)
    dTNR = dTN / (dTN + dFP)

    sTPR = sTP / (sTP + sFN)
    sTNR = sTN / (sTN + sFP)

    dist_ba = 0.5 * (dTPR + dTNR)
    softmax_ba = 0.5 * (sTPR + sTNR)
    return result_softmax_loss, result_naive_loss, dist_ba, softmax_ba


if __name__ == '__main__':
    print("Sequence Keypoint Evaluation")

    keypoint_model_path = "/home/florin/diss_models/final_weights/sequence_analysis/effnetb0_combo_diss/kp_combo_ts30_shift30_convlstm64_7tbnf_normlr3_aug/1_chkpt/kp_combo_ts30_shift30_convlstm64_7tbnf_normlr3_aug_best.164.h5"
    test_set_dir = "/home/florin/data/sequence/"
    test_subjects = [0, 6, 10, 14, 20, 23, 38, 39]

    result_df = get_result_dataframe(keypoint_model_path, test_set_dir, test_subjects, do_binary_ensemble=False)
    result_softmax_loss, result_naive_loss, dist_ba, softmax_ba = calc_loss_and_ba(result_df)

    print("Result Loss (softmax): {}".format(result_softmax_loss))
    print("Result Loss (naive): {}".format(result_naive_loss))
    print("Result BA (distinction): {}".format(dist_ba))
    print("Result BA (softmax): {}".format(softmax_ba))
