import numpy as np
import tensorflow as tf

from research.ablation_studies.train.image_transformer import ImageTransformer
# images: uint8
# image values: uint8
# keypoints: float32
from research.ablation_studies.train.keypoint_tf import KeypointHelper


class ImageTransformerTest(tf.test.TestCase):
    def setUp(self):
        self.none_kp = tf.stack([-1.0, -1.0])
        self.kp_helper = KeypointHelper(row_major=True)
        # seed 1: negate tensor does nothing
        tf.random.set_seed(1)

        # tf.compat.v1.disable_eager_execution()
        return

    def testIdentity(self):
        img = tf.ones(shape=(224, 224, 3), dtype=tf.uint8)
        transformer = ImageTransformer(
            0,
            image_shape=(224, 224, 3),
            keypoint_helper=self.kp_helper)

        augmented_image, augmented_keypoint = transformer._identity(img, [self.none_kp])
        tf.assert_equal(augmented_image.dtype, tf.uint8)
        tf.assert_equal(augmented_image, img)
        tf.assert_equal(augmented_keypoint, self.none_kp)
        return

    ###########################################################################
    # Shear Tests

    def shear_keypoint_full_mag_subtest(self, shear_type, source_target_pair):
        tf.random.set_seed(1)
        src_keypoint = source_target_pair[0]
        target_keypoint = source_target_pair[1]
        img = tf.ones(shape=(224, 224, 3), dtype=tf.uint8) * 255

        transformer = ImageTransformer(
            ImageTransformer.MAX_MAGNITUDE,
            image_shape=(224, 224, 3),
            keypoint_helper=self.kp_helper)

        if shear_type == "x":
            shear_func = transformer._shear_x
        else:
            shear_func = transformer._shear_y

        keypoints = np.array([src_keypoint], dtype='float32')

        augmented_image, augmented_keypoint = shear_func(img, keypoints)
        tf.assert_equal(augmented_keypoint, target_keypoint)
        return

    def testShearXZero(self):
        img = tf.ones(shape=(224, 224, 3), dtype=tf.uint8)
        transformer = ImageTransformer(
            0,
            image_shape=(224, 224, 3),
            keypoint_helper=self.kp_helper)

        keypoints = np.array([self.none_kp], dtype='float32')

        augmented_image, augmented_keypoint = transformer._shear_x(img, keypoints)
        tf.assert_equal(augmented_image.dtype, tf.float32)
        tf.assert_equal(augmented_image, tf.image.convert_image_dtype(img, dtype=tf.float32))
        tf.assert_equal(augmented_keypoint, self.none_kp)
        return

    def testShearXKeypoints(self):
        keypoint_pairs = [(tf.stack([100.0, 100.0]), tf.stack([100.0, 70.0])),  # all inside
                          (tf.stack([100.0, 30.0]), tf.stack([100.0, 0.0])),  # just inside
                          (tf.stack([100.0, 29.0]), self.none_kp),  # just outside
                          (self.none_kp, self.none_kp)]
        for pair in keypoint_pairs:
            with self.subTest(pair):
                self.shear_keypoint_full_mag_subtest("x", pair)
        return

    def testShearYZero(self):
        img = tf.ones(shape=(224, 224, 3), dtype=tf.uint8)
        transformer = ImageTransformer(
            0,
            image_shape=(224, 224, 3),
            keypoint_helper=self.kp_helper)

        keypoints = np.array([self.none_kp], dtype='float32')

        augmented_image, augmented_keypoint = transformer._shear_y(img, keypoints)
        tf.assert_equal(augmented_image.dtype, tf.float32)
        tf.assert_equal(augmented_image, tf.image.convert_image_dtype(img, dtype=tf.float32))
        tf.assert_equal(augmented_keypoint, self.none_kp)
        return

    def testShearYKeypoints(self):
        keypoint_pairs = [(tf.stack([100.0, 100.0]), tf.stack([70.0, 100.0])),  # all inside
                          (tf.stack([30.0, 100.0]), tf.stack([0.0, 100.0])),  # just inside
                          (tf.stack([29.0, 100.0]), self.none_kp),  # just outside
                          (self.none_kp, self.none_kp)]
        for pair in keypoint_pairs:
            with self.subTest(pair):
                self.shear_keypoint_full_mag_subtest("y", pair)
        return

    ###########################################################################
    # Translation Tests

    def translate_keypoint_full_mag_subtest(self, shear_type, source_target_pair, negative):
        if negative:
            tf.random.set_seed(10)
        else:
            tf.random.set_seed(1)
        src_keypoint = source_target_pair[0]
        target_keypoint = source_target_pair[1]
        img = tf.ones(shape=(224, 224, 3), dtype=tf.float32)

        transformer = ImageTransformer(
            ImageTransformer.MAX_MAGNITUDE,
            image_shape=(224, 224, 3),
            keypoint_helper=self.kp_helper)

        if shear_type == "x":
            translate_func = transformer._translate_x
        else:
            translate_func = transformer._translate_y

        keypoints = np.array([src_keypoint], dtype='float32')

        augmented_image, augmented_keypoint = translate_func(img, keypoints)
        tf.assert_equal(augmented_keypoint, target_keypoint)
        return

    def testTranslationXKeypoints(self):
        keypoint_pairs = [(tf.stack([100.0, 100.0]), tf.stack([100.0, 200.0])),  # all inside
                          (tf.stack([100.0, 123.0]), tf.stack([100.0, 223.0])),  # just inside
                          (tf.stack([100.0, 124.0]), self.none_kp),  # just outside
                          (self.none_kp, self.none_kp)]
        for pair in keypoint_pairs:
            with self.subTest(pair):
                self.translate_keypoint_full_mag_subtest("x", pair, False)

        neg_keypoint_pairs = [(tf.stack([100.0, 150.0]), tf.stack([100.0, 50.0])),  # all inside
                              (tf.stack([100.0, 100.0]), tf.stack([100.0, 0.0])),  # just inside
                              (tf.stack([100.0, 99.0]), self.none_kp),  # just outside
                              (self.none_kp, self.none_kp)]
        for pair in neg_keypoint_pairs:
            with self.subTest(pair):
                self.translate_keypoint_full_mag_subtest("x", pair, True)
        return

    def testTranslationYKeypoints(self):
        keypoint_pairs = [(tf.stack([100.0, 100.0]), tf.stack([200.0, 100.0])),  # all inside
                          (tf.stack([123.0, 100.0]), tf.stack([223.0, 100.0])),  # just inside
                          (tf.stack([124.0, 100.0]), self.none_kp),  # just outside
                          (self.none_kp, self.none_kp)]
        for pair in keypoint_pairs:
            with self.subTest(pair):
                self.translate_keypoint_full_mag_subtest("y", pair, False)

        neg_keypoint_pairs = [(tf.stack([150.0, 100.0]), tf.stack([50.0, 100.0])),  # all inside
                              (tf.stack([100.0, 100.0]), tf.stack([0.0, 100.0])),  # just inside
                              (tf.stack([99.0, 100.0]), self.none_kp),  # just outside
                              (self.none_kp, self.none_kp)]
        for pair in neg_keypoint_pairs:
            with self.subTest(pair):
                self.translate_keypoint_full_mag_subtest("y", pair, True)
        return

    ###########################################################################
    # Rotation Tests

    def testRotateNone(self):
        img = tf.ones(shape=(224, 224, 3), dtype=tf.uint8)
        transformer = ImageTransformer(
            ImageTransformer.MAX_MAGNITUDE,
            image_shape=(224, 224, 3),
            keypoint_helper=self.kp_helper)

        keypoints = np.array([self.none_kp], dtype='float32')

        augmented_image, augmented_keypoint = transformer._rotate(img, keypoints)
        tf.assert_equal(augmented_keypoint, [KeypointHelper.make_none_keypoint()])
        return

    def testRotateZero(self):
        img = tf.ones(shape=(224, 224, 3), dtype=tf.uint8)
        transformer = ImageTransformer(
            0,
            image_shape=(224, 224, 3),
            keypoint_helper=self.kp_helper)

        src_keypoints = np.array([[33, 127]], dtype='float32')
        target_keypoints = np.array([[33, 127]], dtype='float32')

        augmented_image, augmented_keypoint = transformer._rotate(img, src_keypoints)
        tf.assert_equal(augmented_keypoint, target_keypoints)
        return

    def testRotateZeroMultiKeypoint(self):
        img = tf.ones(shape=(100, 100, 3), dtype=tf.uint8)
        transformer = ImageTransformer(
            0,
            image_shape=(100, 100, 3),
            keypoint_helper=self.kp_helper)

        src_keypoints = np.array([[50.0, 50.0], [0.0, 0.0]], dtype='float32')
        target_keypoints = np.array([[50.0, 50.0], [0.0, 0.0]], dtype='float32')

        augmented_image, augmented_keypoints = transformer._rotate(img, src_keypoints)
        tf.assert_equal(augmented_keypoints, target_keypoints)
        return

    # unit test for 30 degrees (assuming const rotate max magnitude is 30 degs)
    # we rotate cw around the origin by subtracting the image center,
    # then we translate back to the center (cw because of flipped y axis)
    def testRotateMaxMultiKeypoint(self):
        sidelength = 5
        img = tf.ones(shape=(sidelength, sidelength, 3), dtype=tf.uint8)
        transformer = ImageTransformer(
            ImageTransformer.MAX_MAGNITUDE,
            image_shape=(sidelength, sidelength, 3),
            keypoint_helper=self.kp_helper)

        # 30 degs are pi/6
        # rot_sin = tf.sin(math.pi/6.0)
        # rot_cos = tf.cos(math.pi/6.0)

        # row major is true
        src_keypoints = np.array([
            [2.5, 2.5],
            [0.0, 0.0],
            [2.5, 1.0]],
            dtype='float32')

        target_keypoints = np.array([
            [2.5, 2.5],
            [-1.0, -1.0],
            [1.75, 1.20096]],
            dtype='float32')

        augmented_image, augmented_keypoints = transformer._rotate(img, src_keypoints)
        tf.debugging.assert_near(augmented_keypoints, target_keypoints)
        return
