import tensorflow as tf

from research.ablation_studies.dataset.dataset_utils import image_to_eval_tfexample, write_tfrecord_with_examples, \
    get_tfrecord_eval_dataset
from research.ablation_studies.eval.eval_datasets.dataset_evaluator_pixels import EvalDatasetPixelError
from research.ablation_studies.train.keypoint_tf import KeypointHelper


class EvalDatasetAnalyzerTest(tf.test.TestCase):
    def setUp(self):
        self.none_kp = tf.stack([-1.0, -1.0])
        self.kp_helper = KeypointHelper(row_major=False)

        self.input_img_w = 100
        self.input_img_h = 100
        self.output_img_w = 100
        self.output_img_h = 100

        self.create_none_test_ds()
        # tf.compat.v1.disable_eager_execution()
        return

    def create_test_example(self, ptip, ptail, ptip_median, ptail_median, ptip_bytes, ptail_bytes):
        return image_to_eval_tfexample(
            0,  # input['image/subject_id'],
            0,  # input['image/task_id'],
            0,  # input['image/frame_id'],
            self.kp_helper.get_x(ptip),  # input['image/ptip_x'],
            self.kp_helper.get_y(ptip),  # input['image/ptip_y'],
            self.kp_helper.get_x(ptail),  # input['image/ptail_x'],
            self.kp_helper.get_y(ptail),  # input['image/ptail_y'],
            self.input_img_h,  # input['image/height'],
            self.input_img_w,  # input['image/width'],
            self.output_img_h,  # output.shape[1],
            self.output_img_w,  # output.shape[0],
            ptip_median,  # ptip_heatmap_median,
            ptail_median,  # ptail_heatmap_median,
            ptip_bytes,  # ptip_heatmap_bytes,
            ptail_bytes,  # ptail_heatmap_bytes
        )

    def create_none_test_ds(self):
        black_hm = tf.zeros((self.output_img_h, self.output_img_w, 1), dtype=tf.uint8)
        examples = []
        examples.append(
            self.create_test_example(
                self.kp_helper.make_none_keypoint(),
                self.kp_helper.make_none_keypoint(),
                0,
                0,
                tf.io.encode_jpeg(black_hm, chroma_downsampling=False),
                tf.io.encode_jpeg(black_hm, chroma_downsampling=False)
            )
        )
        write_tfrecord_with_examples("none_keypoint_test.tfrecord", examples)

    def test_get_nones_of_element(self):
        eval_ds = get_tfrecord_eval_dataset("none_keypoint_test.tfrecord")
        analyzer = EvalDatasetPixelError(eval_ds)
        ptip_none, ptail_none = analyzer._get_none_annotations_of_element(next(iter(eval_ds)))

        tf.assert_equal(ptip_none, 1)
        tf.assert_equal(ptail_none, 1)
        return

    def test_map_thresholds_to_cases(self):
        eval_ds = get_tfrecord_eval_dataset("none_keypoint_test.tfrecord")
        analyzer = EvalDatasetPixelError(eval_ds)

        tp, fp, tn, fn = analyzer._map_thresholds_to_cases(0.0, False, tf.constant([1.0]))
        tf.assert_equal(tp, True)
        tf.assert_equal(fp, False)
        tf.assert_equal(tn, False)
        tf.assert_equal(fn, False)

        tp, fp, tn, fn = analyzer._map_thresholds_to_cases(0.0, True, tf.constant([1.0]))
        tf.assert_equal(tp, False)
        tf.assert_equal(fp, True)
        tf.assert_equal(tn, False)
        tf.assert_equal(fn, False)

        tp, fp, tn, fn = analyzer._map_thresholds_to_cases(1.0, True, tf.constant([0.0]))
        tf.assert_equal(tp, False)
        tf.assert_equal(fp, False)
        tf.assert_equal(tn, True)
        tf.assert_equal(fn, False)

        tp, fp, tn, fn = analyzer._map_thresholds_to_cases(1.0, False, tf.constant([0.0]))
        tf.assert_equal(tp, False)
        tf.assert_equal(fp, False)
        tf.assert_equal(tn, False)
        tf.assert_equal(fn, True)
        return

    def test_map_measures_to_subdivisions(self):
        eval_ds = get_tfrecord_eval_dataset("none_keypoint_test.tfrecord")
        analyzer = EvalDatasetPixelError(eval_ds)
        subdivisions = tf.linspace(0.0, 1.0, 2)

        ptip_cases, ptail_cases = analyzer._map_measures_to_subdivisions(subdivisions, next(iter(eval_ds)))
        ptip_tp, ptip_fp, ptip_tn, ptip_fn = ptip_cases
        ptail_tp, ptail_fp, ptail_tn, ptail_fn = ptail_cases

        tf.assert_equal(ptip_tp, [0, 0])
        tf.assert_equal(ptip_fp, [0, 1])
        tf.assert_equal(ptip_tn, [1, 0])
        tf.assert_equal(ptip_fn, [0, 0])
        tf.assert_equal(ptail_tp, [0, 0])
        tf.assert_equal(ptail_fp, [0, 1])
        tf.assert_equal(ptail_tn, [1, 0])
        tf.assert_equal(ptail_fn, [0, 0])

    def test_get_tpr_fpr_for_thresholds(self):
        eval_ds = get_tfrecord_eval_dataset("none_keypoint_test.tfrecord")
        analyzer = EvalDatasetPixelError(eval_ds)

        (ptip_fpr, ptip_tpr), (ptail_fpr, ptail_tpr) = analyzer.get_fpr_tpr_for_roc(2)
        tf.assert_equal(ptip_fpr, [0.0, 1.0])
        tf.assert_equal(ptip_tpr, [0.0, 0.0])
        tf.assert_equal(ptail_fpr, [0.0, 1.0])
        tf.assert_equal(ptail_tpr, [0.0, 0.0])
        return

    def test_calc_pixel_error_both_none(self):
        eval_ds = get_tfrecord_eval_dataset("none_keypoint_test.tfrecord")
        analyzer = EvalDatasetPixelError(eval_ds)

        ptip_distance, ptail_distance = analyzer._calc_pixel_error(next(iter(eval_ds)))
        tf.assert_equal(ptip_distance, 0.0)
        tf.assert_equal(ptail_distance, 0.0)
        return

    def test_pixel_error_both_none(self):
        eval_ds = get_tfrecord_eval_dataset("none_keypoint_test.tfrecord")
        analyzer = EvalDatasetPixelError(eval_ds)

        ptip_avg, ptip_median, ptail_avg, ptail_median = analyzer.get_hd_pixel_error()
        tf.assert_equal(ptip_avg, 0.0)
        tf.assert_equal(ptail_avg, 0.0)
        tf.assert_equal(ptip_median, 0.0)
        tf.assert_equal(ptail_median, 0.0)
        return
