import tensorflow as tf

from misc_util.heatmap import create_heatmaps
from research.ablation_studies.train.keypoint_tf import KeypointHelper


class HeatmapTest(tf.test.TestCase):
    def setUp(self):
        # seed 1: negate tensor does nothing
        tf.random.set_seed(1)
        return

    def test_simple_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=False)
        heatmaps = create_heatmaps(10, 10, 10, 10, 1.0, tf.stack([[5.0, 5.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (10, 10, 1))
        tf.assert_equal(heatmaps[5, 5], 1.0)
        return

    def test_simple_multi_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=False)
        heatmaps = create_heatmaps(10, 10, 10, 10, 1.0, tf.stack([[5.0, 5.0], [1.0, 1.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (10, 10, 2))
        tf.assert_equal(heatmaps[5, 5, 0], 1.0)

        tf.debugging.assert_near(heatmaps[5, 5, 1], 0.0)
        tf.debugging.assert_near(heatmaps[1, 1, 0], 0.0)
        tf.assert_equal(heatmaps[1, 1, 1], 1.0)
        return

    def test_col_major_simple_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=False)
        heatmaps = create_heatmaps(4, 4, 4, 4, 1.0, tf.stack([[0.0, 1.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (4, 4, 1))
        tf.assert_equal(heatmaps[1, 0], 1.0)  # because no row majority when creating keypoints
        return

    def test_row_major_simple_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=True)
        heatmaps = create_heatmaps(4, 4, 4, 4, 1.0, tf.stack([[0.0, 1.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (4, 4, 1))
        tf.assert_equal(heatmaps[0, 1], 1.0)
        return

    def test_col_major_multi_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=False)
        heatmaps = create_heatmaps(4, 4, 4, 4, 1.0, tf.stack([[0.0, 1.0], [1.0, 1.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (4, 4, 2))
        tf.assert_equal(heatmaps[1, 0, 0], 1.0)  # because no row majority when creating keypoints

        tf.assert_equal(heatmaps[1, 1, 1], 1.0)  # because no row majority when creating keypoints
        return

    def test_row_major_multi_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=True)
        heatmaps = create_heatmaps(4, 4, 4, 4, 1.0, tf.stack([[0.0, 1.0], [1.0, 0.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (4, 4, 2))
        tf.assert_equal(heatmaps[0, 1, 0], 1.0)

        tf.assert_equal(heatmaps[1, 0, 1], 1.0)
        return

    def test_col_major_dims_simple_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=False)
        heatmaps = create_heatmaps(4, 2, 4, 2, 1.0, tf.stack([[0.0, 1.0]]), keypoint_helper)

        print(heatmaps)
        tf.assert_equal(heatmaps.shape, (2, 4, 1))
        tf.assert_equal(heatmaps[1, 0], 1.0)  # because no row majority when creating keypoints
        return

    def test_row_major_dims_simple_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=True)
        heatmaps = create_heatmaps(4, 2, 4, 2, 1.0, tf.stack([[0.0, 1.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (2, 4, 1))
        tf.assert_equal(heatmaps[0, 1], 1.0)
        return

    def test_row_major_dims_multi_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=True)
        heatmaps = create_heatmaps(4, 2, 4, 2, 1.0, tf.stack([[0.0, 1.0], [1.0, 0.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (2, 4, 2))
        tf.assert_equal(heatmaps[0, 1, 0], 1.0)  # because no row majority when creating keypoints
        tf.assert_equal(heatmaps[1, 0, 1], 1.0)  # because no row majority when creating keypoints
        return

    def test_col_major_dims_multi_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=False)
        heatmaps = create_heatmaps(4, 2, 4, 2, 1.0, tf.stack([[0.0, 1.0], [1.0, 0.0]]), keypoint_helper)

        tf.assert_equal(heatmaps.shape, (2, 4, 2))
        tf.assert_equal(heatmaps[1, 0, 0], 1.0)  # because no row majority when creating keypoints
        tf.assert_equal(heatmaps[0, 1, 1], 1.0)  # because no row majority when creating keypoints
        return

    def test_ulti_multi_heatmap(self):
        keypoint_helper = KeypointHelper(row_major=False)
        heatmaps = create_heatmaps(73, 112, 73, 112, 1.0, tf.stack(
            [[0.0, 1.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0],
             [1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [1.0, 0.0]]), keypoint_helper)
        tf.assert_equal(heatmaps.shape, (112, 73, 13))
        return
