import numpy as np
import tensorflow as tf

from misc_util.tf_util import get_random_tf_bool
from research.ablation_studies.train.image_transformer import ImageTransformer
from research.ablation_studies.train.keypoint_tf import KeypointHelper


class BinPairInputMap:
    def __init__(self,
                 model_input_shape,
                 model_output_shape,
                 ra_magnitude,
                 input_prep_func=None):
        self.model_input_shape = model_input_shape
        self.model_output_shape = model_output_shape
        self.ra_magnitude = ra_magnitude
        self.input_prep_func = input_prep_func
        return

    # map_fn: loading images from file path in a dataset
    def _load_img(self, element):
        image_bytes = element['image/encoded']
        bin_bytes = element['bin/encoded']

        img = tf.io.decode_jpeg(image_bytes, channels=3)
        patch = tf.image.convert_image_dtype(img, tf.float32)

        bin = tf.io.decode_jpeg(bin_bytes, channels=1)
        bin_patch = tf.image.convert_image_dtype(bin, tf.float32)
        bin_patch = tf.image.resize(bin_patch, self.model_output_shape[:2])

        return {'patch': patch,
                'bin_patch': bin_patch}

    def _get_augmented_img_opped(self, transformer, op, image, ignore_visual_ops=False):
        visual_ops = tf.constant([6, 7, 8, 9, 10, 11])

        check_ops_tensor = tf.equal(visual_ops, op)
        op_index = tf.cond(
            tf.logical_and(
                tf.equal(tf.math.count_nonzero(check_ops_tensor), True),
                ignore_visual_ops),
            lambda: 0,
            lambda: op, name='check_if_ops_visual')

        keypoint = tf.stack([KeypointHelper.make_none_keypoint()])
        aug_image, aug_keypoint = tf.switch_case(op_index, branch_fns={
            0: lambda: transformer._identity(image, keypoint),
            1: lambda: transformer._shear_x(image, keypoint),
            2: lambda: transformer._shear_y(image, keypoint),
            3: lambda: transformer._translate_x(image, keypoint),
            4: lambda: transformer._translate_y(image, keypoint),
            5: lambda: transformer._rotate(image, keypoint),
            6: lambda: transformer._equalize(image, keypoint),
            7: lambda: transformer._contrast(image, keypoint),
            8: lambda: transformer._color(image, keypoint),
            9: lambda: transformer._brightness(image, keypoint),
            10: lambda: transformer._sharpness(image, keypoint),
            11: lambda: transformer._auto_contrast(image, keypoint)
        }, name='img_switch_ops')
        return aug_image

    def _get_augmented_bin_opped(self, transformer, op, image, ignore_visual_ops=False):
        def dumb_image_function_wrapper(func, in_gray, keypoint):
            result, keypoint = func(tf.image.grayscale_to_rgb(in_gray), keypoint)
            return tf.image.rgb_to_grayscale(result), keypoint

        keypoint = tf.stack([KeypointHelper.make_none_keypoint()])
        aug_image, aug_keypoint = tf.switch_case(op, branch_fns={
            0: lambda: transformer._identity(image, keypoint),
            1: lambda: dumb_image_function_wrapper(transformer._shear_x, image, keypoint),
            2: lambda: dumb_image_function_wrapper(transformer._shear_y, image, keypoint),
            3: lambda: transformer._translate_x(image, keypoint),
            4: lambda: transformer._translate_y(image, keypoint),
            5: lambda: transformer._rotate(image, keypoint),
            6: lambda: transformer._identity(image, keypoint),
            7: lambda: transformer._identity(image, keypoint),
            8: lambda: transformer._identity(image, keypoint),
            9: lambda: transformer._identity(image, keypoint),
            10: lambda: transformer._identity(image, keypoint),
            11: lambda: transformer._identity(image, keypoint)
        }, name='bin_switch_ops')
        return aug_image

    def _get_augmented_image(self, choices, image, img_transformer, augment_seed, ignore_visual_ops=False):
        do_flip = get_random_tf_bool(random_seed=augment_seed)
        flipped_image = tf.image.flip_left_right(image)

        aug_image = tf.cond(do_flip,
                            lambda: flipped_image,
                            lambda: image)

        aug_image = self._get_augmented_img_opped(img_transformer, choices[0], aug_image, ignore_visual_ops)
        aug_image = self._get_augmented_img_opped(img_transformer, choices[1], aug_image, ignore_visual_ops)
        return aug_image

    def _get_augmented_bin(self, choices, image, img_transformer, augment_seed, ignore_visual_ops=False):
        do_flip = get_random_tf_bool(random_seed=augment_seed)
        flipped_image = tf.image.flip_left_right(image)

        tf.debugging.assert_shapes([
            (flipped_image, self.model_output_shape),
        ], name='Assert_flipped_grayscale')

        aug_image = tf.cond(do_flip,
                            lambda: flipped_image,
                            lambda: image)

        aug_image = self._get_augmented_bin_opped(img_transformer, choices[0], aug_image, ignore_visual_ops)
        tf.ensure_shape(aug_image, shape=self.model_output_shape, name='bin_aug_choice_0')
        aug_image = self._get_augmented_bin_opped(img_transformer, choices[1], aug_image, ignore_visual_ops)
        tf.ensure_shape(aug_image, shape=self.model_output_shape, name='bin_aug_choice_1')
        return aug_image

    def _augment_img(self, element):
        img_patch = element['patch']
        bin_patch = element['bin_patch']

        augment_seed = np.random.randint(low=0, high=10000)

        img_transformer = ImageTransformer(
            magnitude=self.ra_magnitude,
            image_shape=self.model_input_shape,
            keypoint_helper=KeypointHelper(row_major=False),
            random_seed=augment_seed)

        bin_transformer = ImageTransformer(
            magnitude=self.ra_magnitude,
            image_shape=self.model_output_shape,
            keypoint_helper=KeypointHelper(row_major=False),
            random_seed=augment_seed)
        bin_transformer.IMG_AVERAGE_VAL = 0

        choices = tf.random.shuffle(tf.range(12))[:2]

        aug_img_patch = self._get_augmented_image(choices, img_patch, img_transformer, augment_seed,
                                                  ignore_visual_ops=False)
        aug_bin_patch = self._get_augmented_bin(choices, bin_patch, bin_transformer, augment_seed,
                                                ignore_visual_ops=True)

        aug_img_patch = self.input_prep_func(aug_img_patch)

        tf.debugging.assert_shapes([
            (aug_img_patch, self.model_input_shape),
            (aug_bin_patch, self.model_output_shape),
        ], name='augment_output_shapes_assert')

        return (aug_img_patch, aug_bin_patch)

    def _produce_test_heatmaps(self, element):
        img_patch = element['patch']
        bin_patch = element['bin_patch']

        img_patch = self.input_prep_func(img_patch)

        tf.debugging.assert_shapes([
            (img_patch, self.model_input_shape),
            (bin_patch, self.model_output_shape),
        ], name='test_output_shapes_assert')

        return (img_patch, bin_patch)
