import tensorflow as tf

from misc_util.tf_util import get_random_tf_bool
from research.ablation_studies.train.image_transformer import ImageTransformer
from research.ablation_studies.train.keypoint_tf import KeypointHelper


class Augmenter:
    basic_transformation_list = [
        'FlipLR', 'RandomCrop'
    ]

    def __init__(self,
                 num_layers,
                 magnitude,
                 image_dtype,
                 image_shape=(224, 224, 3),
                 keypoint_helper=None,
                 augment_seed=None,
                 op_choices=None):
        self.kp_helper = keypoint_helper
        self.img_dtype = image_dtype
        self.magnitude = magnitude
        self.num_layers = num_layers
        self.image_shape = image_shape
        self.augment_seed = augment_seed
        self.op_choices = op_choices
        self.transformer = ImageTransformer(
            magnitude=magnitude,
            image_shape=image_shape,
            keypoint_helper=self.kp_helper,
            random_seed=self.augment_seed)
        return

    def _op_mapper(self, op, image, keypoint):
        aug_image, aug_keypoint = tf.switch_case(op, branch_fns={
            0: lambda: self.transformer._identity(image, keypoint),
            1: lambda: self.transformer._shear_x(image, keypoint),
            2: lambda: self.transformer._shear_y(image, keypoint),
            3: lambda: self.transformer._translate_x(image, keypoint),
            4: lambda: self.transformer._translate_y(image, keypoint),
            5: lambda: self.transformer._rotate(image, keypoint),
            6: lambda: self.transformer._equalize(image, keypoint),
            7: lambda: self.transformer._contrast(image, keypoint),
            8: lambda: self.transformer._color(image, keypoint),
            9: lambda: self.transformer._brightness(image, keypoint),
            10: lambda: self.transformer._sharpness(image, keypoint),
            11: lambda: self.transformer._auto_contrast(image, keypoint),
            12: lambda: self.transformer._posterize(image, keypoint),
            13: lambda: self.transformer._solarize(image, keypoint)
        })
        return aug_image, aug_keypoint

    possible_randaugment_transformations = 12

    # sample ops without replacement
    def _sample_ops(self):
        if self.augment_seed is not None:
            choices = self.op_choices
        else:
            choices = tf.random.shuffle(tf.range(self.possible_randaugment_transformations))[:self.num_layers]
        return choices

    ###########################################################################
    # public interface

    # dtype in map_fn is deprecated in 2.3
    def _check_none_keypoints(self, t_keypoints):
        none_keypoints = tf.map_fn(lambda kp: KeypointHelper.is_none_keypoint(kp), t_keypoints, dtype=tf.bool)
        return none_keypoints

    def _keep_none_keypoints(self, none_keypoints_before, t_keypoints):
        none_keypoints_after = self._check_none_keypoints(t_keypoints)
        none_keypoints = none_keypoints_before | none_keypoints_after
        return none_keypoints

    def _do_preprocessing_transforms(self, my_image, my_t_keypoints, my_none_keypoints):
        # random x flip
        do_flip = get_random_tf_bool(random_seed=self.augment_seed)
        flipped_image, flipped_keypoints = self.transformer._flip_lr(my_image, my_t_keypoints)

        my_aug_image, my_aug_keypoints = tf.cond(do_flip,
                                                 lambda: (flipped_image, flipped_keypoints),
                                                 lambda: (my_image, my_t_keypoints))

        # random crop
        # my_aug_image, my_aug_keypoints = self.transformer._random_crop(my_aug_image, my_aug_keypoints)
        new_none_keypoints = self._keep_none_keypoints(my_none_keypoints, my_aug_keypoints)
        return my_aug_image, my_aug_keypoints, new_none_keypoints

    # todo use tf2 graph for sampled_ops loop
    # always input float32 for image, please
    # keypoints are expected to be tensors
    def transform(self, image=None, keypoints=None, do_preprocessing=True):

        tf.debugging.assert_type(image, tf.float32, "Input image tensor is not float32.")
        tf.debugging.assert_type(keypoints, tf.float32, "Input keypoint tensor is not float32.")

        t_keypoints = keypoints

        none_keypoints = self._check_none_keypoints(t_keypoints)

        aug_image, aug_keypoints, none_keypoints = tf.cond(
            tf.constant(do_preprocessing, dtype=tf.bool),
            lambda: self._do_preprocessing_transforms(image, t_keypoints, none_keypoints),
            lambda: (image, t_keypoints, none_keypoints),
        )

        # now standard ops are done
        # we now do randaugment ops
        sampled_ops = self._sample_ops()

        aug_image = tf.ensure_shape(aug_image, self.image_shape)
        # aug_keypoints = tf.squeeze(aug_keypoints)
        # aug_keypoints = tf.ensure_shape(aug_keypoints, (2, 2, 1))

        for op in sampled_ops:
            aug_image, aug_keypoints = self._op_mapper(op, aug_image, aug_keypoints)
            none_keypoints = self._keep_none_keypoints(none_keypoints, aug_keypoints)
            aug_image = tf.ensure_shape(aug_image, self.image_shape)
            # aug_keypoints = tf.squeeze(aug_keypoints)
            # aug_keypoints = tf.ensure_shape(aug_keypoints, (2, 2, 1))
            # tf.debugging.check_numerics(aug_image, "Augmentation op lead to nans.")

        aug_keypoints = self.kp_helper.filter_keypoints(none_keypoints, aug_keypoints)
        return aug_image, aug_keypoints
