import math

import tensorflow as tf
import tensorflow_addons as tfa

# from misc_util.tf_util import randomly_negate_tensor
from misc_util.tf_util import randomly_negate_tensor


class ImageTransformer:
    MAX_MAGNITUDE = 30

    IMG_AVERAGE_VAL = tf.constant(127, dtype=tf.uint8)

    level_const = {
        'Cutout': 40,
        'Posterize': 4,
        'ShearX': 0.3,
        'TranslateX': 100,
        # from https://github.com/tensorflow/tpu/blob/06ac8c2af4c196161260d45ba5aa0be923c15a3c/models/official/efficientnet/autoaugment.py#L689
        'TranslateY': 100,
        'ShearY': 0.3,
        'Rotate': 30,
        'Solarize': 256
    }

    def __init__(self,
                 magnitude=9,
                 image_shape=(224, 224, 3),
                 keypoint_helper=None,
                 random_seed=None):
        self.magnitude = magnitude
        self.keypoint_helper = keypoint_helper
        self.image_shape = image_shape
        self.random_seed = random_seed
        return

    ###########################################################################
    # parameter transformations from
    # https://github.com/tensorflow/tpu/blob/06ac8c2af4c196161260d45ba5aa0be923c15a3c/models/official/efficientnet/autoaugment.py#L467

    def _enhance_parameter(self):
        return tf.cast(self.magnitude, tf.float32) / tf.cast(self.MAX_MAGNITUDE, tf.float32) * 1.8 + 0.1

    def _brightness_parameter(self):
        level = self._enhance_parameter() - 1.0
        return level

    def _calc_operation_value(self, value):
        val = (float(self.magnitude) / self.MAX_MAGNITUDE) * value
        return val

    def _shear_parameter(self):
        level = self._calc_operation_value(self.level_const['ShearX'])
        level = randomly_negate_tensor(level, random_seed=self.random_seed)
        return level

    def _translate_parameter(self):
        level = self._calc_operation_value(self.level_const['TranslateX'])
        level = randomly_negate_tensor(level, random_seed=self.random_seed)
        return level

    def _rotate_parameter(self):
        level = self._calc_operation_value(self.level_const['Rotate'])
        level = randomly_negate_tensor(level, random_seed=self.random_seed)
        return level

    def _posterize_parameter(self):
        return tf.cast(self._calc_operation_value(self.level_const['Posterize']), dtype=tf.int32)

    ###########################################################################
    # geometric transformation implementations

    def _identity(self, image, keypoints):
        return image, keypoints

    # affects keypoints
    # expects keypoints as float32 tensors
    def _flip_lr(self, image, keypoints):
        flipped_image = tf.image.flip_left_right(image)
        width = tf.cast(tf.shape(image)[1], dtype=tf.float32)

        flipped_keypoints = tf.map_fn(
            lambda kp:
            self.keypoint_helper.set_x(kp, width - self.keypoint_helper.get_x(kp)), keypoints)
        return flipped_image, flipped_keypoints

    # affects keypoints
    def _random_crop(self, image, keypoints):
        bbox = tf.constant([0.0, 0.0, 1.0, 1.0], dtype=tf.float32, shape=[1, 1, 4])
        height = tf.shape(image)[0]
        width = tf.shape(image)[1]

        # idea and parameters from
        # https://github.com/tensorflow/tpu/blob/06ac8c2af4c196161260d45ba5aa0be923c15a3c/models/official/efficientnet/preprocessing.py#L96
        bbox_begin, bbox_size, bbox_for_draw = tf.image.sample_distorted_bounding_box(
            tf.shape(image),
            bounding_boxes=bbox,  # one bounding box
            min_object_covered=0.1,
            aspect_ratio_range=(3. / 4, 4. / 3.),
            area_range=(0.08, 1.0),
            max_attempts=10,
            use_image_if_no_bounding_boxes=True)

        distorted_image = tf.slice(image, bbox_begin, bbox_size)

        offset_y, offset_x, _ = tf.unstack(bbox_begin)
        target_height, target_width, _ = tf.unstack(bbox_size)

        offset_x = tf.cast(offset_x, dtype=tf.float32)
        offset_y = tf.cast(offset_y, dtype=tf.float32)
        target_height = tf.cast(target_height, dtype=tf.float32)
        target_width = tf.cast(target_width, dtype=tf.float32)

        def check_outside_bounds(keypoint):
            kp_x = kph.get_x(keypoint)
            kp_y = kph.get_y(keypoint)
            return kp_x <= offset_x or kp_x >= offset_x + target_width or \
                   kp_y <= offset_y or kp_y >= offset_y + target_height

        kph = self.keypoint_helper
        keypoints_outside_bounds_mask = tf.map_fn(lambda kp: check_outside_bounds(kp), keypoints, dtype=tf.bool)

        # the augmented keypoints, only valid if original keypoint falls within bbox
        aug_keypoints = tf.map_fn(lambda kp: kph.make(
            x=(kph.get_x(kp) - offset_x) / target_width * tf.cast(width, dtype=tf.float32),
            y=(kph.get_y(kp) - offset_y) / target_height * tf.cast(height, dtype=tf.float32)
        ), keypoints)

        final_keypoints = kph.filter_keypoints(keypoints_outside_bounds_mask, aug_keypoints)
        augmented_image = tf.image.resize(distorted_image, [height, width])
        return augmented_image, final_keypoints

    # affects keypoints
    def _shear_x(self, image, keypoints):
        shear_magnitude = self._shear_parameter()
        img = tf.image.convert_image_dtype(image, tf.uint8)
        augmented_image = tfa.image.transform_ops.shear_x(img, shear_magnitude, self.IMG_AVERAGE_VAL)

        @tf.function
        def calc_shear_coord(keypoint):
            sign = tf.cond(tf.greater_equal(tf.sign(shear_magnitude), 0.0),
                           lambda: 1.0,
                           lambda: -1.0)

            multiplier = tf.abs(shear_magnitude * kph.get_y(keypoint))
            new_keypoint = kph.set_x(keypoint,
                                     kph.get_x(keypoint) - sign * multiplier)
            return new_keypoint

        kph = self.keypoint_helper
        new_keypoints = tf.map_fn(lambda kp: calc_shear_coord(kp), keypoints)

        new_keypoints = tf.map_fn(lambda kp: kph.check_outside_and_set(kp, augmented_image), new_keypoints)
        return tf.image.convert_image_dtype(augmented_image, tf.float32), new_keypoints

    # affects keypoints
    def _shear_y(self, image, keypoints):
        shear_magnitude = self._shear_parameter()
        img = tf.image.convert_image_dtype(image, tf.uint8)
        augmented_image = tfa.image.transform_ops.shear_y(img, shear_magnitude, self.IMG_AVERAGE_VAL)

        @tf.function
        def calc_shear_coord(keypoint):
            sign = tf.cond(tf.greater_equal(tf.sign(shear_magnitude), 0.0),
                           lambda: 1.0,
                           lambda: -1.0)

            multiplier = tf.abs(shear_magnitude * kph.get_x(keypoint))
            new_keypoint = kph.set_y(keypoint,
                                     kph.get_y(keypoint) - sign * multiplier)
            return new_keypoint

        kph = self.keypoint_helper
        new_keypoints = tf.map_fn(lambda kp: calc_shear_coord(kp), keypoints)

        new_keypoints = tf.map_fn(lambda kp: kph.check_outside_and_set(kp, augmented_image), new_keypoints)
        return tf.image.convert_image_dtype(augmented_image, tf.float32), new_keypoints

    # affects keypoints
    def _translate_x(self, image, keypoints):
        translate_val = self._translate_parameter()
        translate_val_int = tf.cast(translate_val, dtype=tf.int32)
        abs_translate_val_int = tf.abs(translate_val_int)

        height = tf.shape(image)[0]
        width = tf.shape(image)[1]

        crop_offset_x, pad_offset_x = tf.cond(tf.math.sign(translate_val_int) < 0,
                                              lambda: (abs_translate_val_int, 0),
                                              lambda: (0, abs_translate_val_int))

        augmented_image = tf.image.crop_to_bounding_box(image, 0, crop_offset_x, height, width - abs_translate_val_int)
        augmented_image = tf.image.pad_to_bounding_box(augmented_image, 0, pad_offset_x, height, width)

        mask_value = tf.cast(self.IMG_AVERAGE_VAL, tf.float32) / 255.0
        masked_bg = tf.ones_like(augmented_image) * mask_value
        masked_bg = tf.image.crop_to_bounding_box(masked_bg, 0, 0, height, abs_translate_val_int)
        pad_mask_offset = tf.cond(tf.math.sign(translate_val_int) < 0,
                                  lambda: width - crop_offset_x,
                                  lambda: 0)
        masked_bg = tf.image.pad_to_bounding_box(masked_bg, 0, pad_mask_offset, height, width)
        black_replaced_img = augmented_image + masked_bg

        kph = self.keypoint_helper
        new_keypoints = tf.map_fn(lambda kp:
                                  kph.set_x(kp, kph.get_x(kp) + translate_val), keypoints)

        augmented_keypoints = tf.map_fn(lambda kp: kph.check_outside_and_set(kp, black_replaced_img), new_keypoints)
        return tf.image.convert_image_dtype(black_replaced_img, tf.float32), augmented_keypoints

    def _mask_zeros_with_half(self, image):
        mask = tf.equal(image, (0.0, 0.0, 0.0))
        targets = tf.where(mask, (0.5, 0.5, 0.5), image)
        return targets

    # affects keypoints
    def _translate_y(self, image, keypoints):
        translate_val = self._translate_parameter()
        translate_val_int = tf.cast(translate_val, dtype=tf.int32)
        abs_translate_val_int = tf.abs(translate_val_int)

        height = tf.shape(image)[0]
        width = tf.shape(image)[1]

        crop_offset_y, pad_offset_y = tf.cond(tf.math.sign(translate_val_int) < 0,
                                              lambda: (abs_translate_val_int, 0),
                                              lambda: (0, abs_translate_val_int))

        augmented_image = tf.image.crop_to_bounding_box(image, crop_offset_y, 0, height - tf.abs(translate_val_int),
                                                        width)
        augmented_image = tf.image.pad_to_bounding_box(augmented_image, pad_offset_y, 0, height, width)

        mask_value = tf.cast(self.IMG_AVERAGE_VAL, tf.float32) / 255.0
        masked_bg = tf.ones_like(augmented_image) * mask_value
        masked_bg = tf.image.crop_to_bounding_box(masked_bg, 0, 0, abs_translate_val_int, width)
        pad_mask_offset = tf.cond(tf.math.sign(translate_val_int) < 0,
                                  lambda: height - crop_offset_y,
                                  lambda: 0)
        masked_bg = tf.image.pad_to_bounding_box(masked_bg, pad_mask_offset, 0, height, width)
        black_replaced_img = augmented_image + masked_bg

        kph = self.keypoint_helper
        new_keypoints = tf.map_fn(lambda kp:
                                  kph.set_y(kp, kph.get_y(kp) + translate_val), keypoints)

        augmented_keypoints = tf.map_fn(lambda kp: kph.check_outside_and_set(kp, black_replaced_img), new_keypoints)
        return black_replaced_img, augmented_keypoints

    def _fill_rotated_image_background(self, image, angle):
        background = tf.ones_like(image)
        background_rotated = tfa.image.rotate(background, angle, interpolation='BILINEAR')
        mask = tf.cast(background_rotated, tf.bool)
        mask_value = tf.cast(self.IMG_AVERAGE_VAL, tf.float32) / 255.0
        background *= mask_value
        combined_image = tf.where(mask, image, background)
        return combined_image

    # affects keypoints
    def _rotate(self, image, keypoints):
        rotate_val = self._rotate_parameter() / 180.0 * math.pi

        height = tf.cast(tf.shape(image)[0], dtype=tf.float32)
        width = tf.cast(tf.shape(image)[1], dtype=tf.float32)

        augmented_image = self._fill_rotated_image_background(image, rotate_val)
        augmented_image = tfa.image.rotate(image, rotate_val, interpolation='BILINEAR')

        kph = self.keypoint_helper
        rot_sin = tf.sin(rotate_val)
        rot_cos = tf.cos(rotate_val)

        def rotate_keypoint(keypoint):
            # move to origin
            origin_keypoint = kph.set_x(keypoint, kph.get_x(keypoint) - width * 0.5)
            origin_keypoint = kph.set_y(origin_keypoint, kph.get_y(origin_keypoint) - height * 0.5)

            rotx_origin_keypoint = kph.set_x(
                origin_keypoint,
                kph.get_x(origin_keypoint) * rot_cos + kph.get_y(origin_keypoint) * rot_sin)

            rotxy_origin_keypoint = kph.set_y(
                rotx_origin_keypoint,
                - kph.get_x(origin_keypoint) * rot_sin + kph.get_y(origin_keypoint) * rot_cos)

            # move back
            new_keypoint = kph.set_x(rotxy_origin_keypoint, kph.get_x(rotxy_origin_keypoint) + width * 0.5)
            new_keypoint = kph.set_y(new_keypoint, kph.get_y(new_keypoint) + height * 0.5)

            return new_keypoint

        rotated_keypoints = tf.map_fn(lambda kp: rotate_keypoint(kp), keypoints)
        augmented_keypoints = tf.map_fn(lambda kp: kph.check_outside_and_set(kp, augmented_image), rotated_keypoints)
        return augmented_image, augmented_keypoints

    ###########################################################################
    # other transformation implementations

    # from https://github.com/tensorflow/addons/pull/1349
    # since then it was decided to move randaugment operations to keras-cv,
    # which is not yet publicly available
    def _auto_contrast(self, image, keypoint):

        def scale_channel(image_channel):
            """Scale the 2D image using the autocontrast rule."""
            # A possibly cheaper version can be done using cumsum/unique_with_counts
            # over the histogram values, rather than iterating over the entire image.
            # to compute mins and maxes.
            lo = tf.reduce_min(image_channel)
            hi = tf.reduce_max(image_channel)

            if hi > lo:
                image_channel = (image_channel - lo) / (hi - lo)
            return image_channel

        # Assumes RGB for now.  Scales each channel independently
        # and then stacks the result.
        image = tf.ensure_shape(image, self.image_shape)
        ss = tf.unstack(image, axis=-1)
        for i in range(len(ss)):
            ss[i] = scale_channel(ss[i])
        image = tf.stack(ss, -1)

        image = tf.clip_by_value(image, 0.0, 1.0)

        return image, keypoint

    # equalize adapted from color_ops in tf addons, because shape access not possible
    # in graph mode
    def _own_equalize(self, image):
        @tf.function
        def scale_channel(my_image, channel):
            image_dtype = my_image.dtype

            my_image = tf.cast(my_image[:, :, channel], tf.int32)
            histo = tf.histogram_fixed_width(my_image, [0, 255], nbins=256)
            # For the purposes of computing the step, filter out the nonzeros.
            nonzero = tf.where(tf.not_equal(histo, 0))
            nonzero_histo = tf.reshape(tf.gather(histo, nonzero), [-1])
            step = (tf.reduce_sum(nonzero_histo) - nonzero_histo[-1]) // 255

            def build_lut(my_histo, my_step):
                # Compute the cumulative sum, shifting by step // 2
                # and then normalization by step.
                lut = (tf.cumsum(my_histo) + (my_step // 2)) // my_step
                # Shift lut, prepending with 0.
                lut = tf.concat([[0], lut[:-1]], 0)
                # Clip the counts to be in range.  This is done
                # in the C code for image.point.
                return tf.clip_by_value(lut, 0, 255)

            # If step is zero, return the original image.  Otherwise, build
            # lut from the full histogram and step and then index from it.

            if step == 0:
                result = my_image
            else:
                result = tf.gather(build_lut(histo, step), my_image)

            return tf.cast(result, image_dtype)

        # image = tf.stack([scale_channel(image, c) for c in range(tf.shape(image)[-1])], 2)

        image = tf.ensure_shape(image, self.image_shape)
        ss = tf.unstack(image, axis=-1)
        for i in range(len(ss)):
            ss[i] = scale_channel(image, i)
        image = tf.stack(ss, -1)
        return image

    # tonwertangleichung
    def _equalize(self, image, keypoint):
        equa_image = tf.image.convert_image_dtype(image, tf.uint8)
        result = self._own_equalize(equa_image)
        converted = tf.image.convert_image_dtype(result, tf.float32)
        return converted, keypoint

    # needs float32 0-1 image
    # adapted from tfa sharpness
    def _sharpness(self, image, keypoint):
        """Implements Sharpness function from PIL using TF ops."""
        orig_image = image
        # Make image 4D for conv operation.
        image = tf.expand_dims(image, 0)
        # SMOOTH PIL Kernel.
        kernel = (
                tf.constant(
                    [[1, 1, 1], [1, 5, 1], [1, 1, 1]], dtype=tf.float32, shape=[3, 3, 1, 1]
                )
                / 13.0
        )
        # Tile across channel dimension.
        kernel = tf.tile(kernel, [1, 1, 3, 1])
        strides = [1, 1, 1, 1]
        degenerate = tf.nn.depthwise_conv2d(
            image, kernel, strides, padding="VALID", dilations=[1, 1]
        )
        degenerate = tf.clip_by_value(degenerate, 0.0, 1.0)
        degenerate = tf.squeeze(tf.cast(degenerate, tf.float32), [0])

        # For the borders of the resulting image, fill in the values of the
        # original image.
        mask = tf.ones_like(degenerate)
        padded_mask = tf.pad(mask, [[1, 1], [1, 1], [0, 0]])
        padded_degenerate = tf.pad(degenerate, [[1, 1], [1, 1], [0, 0]])
        result = tf.where(tf.equal(padded_mask, 1), padded_degenerate, orig_image)
        # Blend the final result.
        blended = self._blend(result, orig_image, self._enhance_parameter())
        return blended, keypoint

    def _contrast(self, image, keypoint):
        return tf.image.adjust_contrast(image, self._enhance_parameter()), keypoint

    # adapted from tfa implementation
    # we always assume float32 images (0-1)
    def _blend(self, image1, image2, factor):
        with tf.name_scope("blend"):
            if tf.equal(factor, 0.0):
                return image1
            if tf.equal(factor, 1.0):
                return image2

            difference = image2 - image1
            scaled = factor * difference

            # Do addition in float.
            temp = image1 + scaled

            if tf.greater(factor, 0.0) and tf.less(factor, 1.0):
                return temp

            # Extrapolate:
            #
            # We need to clip and then cast.
            temp = tf.clip_by_value(temp, 0.0, 1.0)
            return temp

    # based on
    # https://github.com/tensorflow/tpu/blob/06ac8c2af4c196161260d45ba5aa0be923c15a3c/models/official/efficientnet/autoaugment.py#L193
    def _color(self, image, keypoint):
        """Equivalent of PIL Color."""
        color_val = self._enhance_parameter()
        degenerate = tf.image.grayscale_to_rgb(tf.image.rgb_to_grayscale(image))
        return self._blend(degenerate, image, color_val), keypoint

    def _brightness(self, image, keypoint):
        value = self._brightness_parameter()
        tf.debugging.assert_type(image, tf.float32, "Input image tensor is not float32.")
        result = tf.add(
            image, tf.cast(value, image.dtype))
        result = tf.clip_by_value(result, 0.0, 1.0)
        return result, keypoint

    # from
    # https://github.com/tensorflow/tpu/blob/06ac8c2af4c196161260d45ba5aa0be923c15a3c/models/official/efficientnet/autoaugment.py#L222
    def _posterize(self, image, keypoint):
        """Equivalent of PIL Posterize."""
        shift = 8 - tf.cast(self._posterize_parameter(), dtype=tf.uint8)
        img = tf.image.convert_image_dtype(image, tf.uint8)
        shifted_img = tf.bitwise.left_shift(tf.bitwise.right_shift(img, shift), shift)
        return tf.image.convert_image_dtype(shifted_img, tf.float32), keypoint

    def _solarize(self, image, keypoint):
        # For each pixel in the image, select the pixel
        # if the value is less than the threshold.
        # Otherwise, subtract 255 from the pixel.
        # todo doesnt work
        solarize_val = self._calc_operation_value(1.0)  # self.level_const['Solarize'])
        return tf.where(image < (solarize_val, solarize_val, solarize_val), image, 1.0 - image), keypoint
