import numpy as np
import tensorflow as tf


class KeypointHelper:
    def __init__(self, row_major):
        self.row_major = tf.constant(row_major, dtype=tf.bool)
        return

    # todo: replace np with make_none_keypoint
    # skips all keypoints where condition is true
    def filter_keypoints(self, none_keypoints, t_keypoints):
        none_placeholders = tf.repeat(np.array([[-1, -1]], dtype='float32'), tf.shape(t_keypoints)[0], axis=0)
        # we need the condition for every keypoint element, so we stack them to 2d
        none_condition = tf.stack([none_keypoints, none_keypoints], axis=1)
        new_keypoints = tf.where(none_condition, none_placeholders, t_keypoints)
        return new_keypoints

    def is_row_major(self):
        return self.row_major

    def from_list(self, keypoint):
        x = tf.cond(self.row_major, lambda: keypoint[1], lambda: keypoint[0], name='kp_helper_from_list_x')
        y = tf.cond(self.row_major, lambda: keypoint[0], lambda: keypoint[1], name='kp_helper_from_list_y')
        return self.make(x, y)

    def to_list(self, keypoint):
        kp_x = self.get_x(keypoint)
        kp_y = self.get_y(keypoint)
        result = tf.cond(self.row_major,
                         lambda: [kp_y, kp_x],
                         lambda: [kp_x, kp_y],
                         name='kp_helper_to_list')
        return result

    def make(self, x, y):
        # x = tf.reshape(x, (1,))
        # y = tf.reshape(y, (1,))
        t_keypoint = tf.cond(self.row_major,
                             lambda: tf.stack([y, x]),
                             lambda: tf.stack([x, y]))
        return tf.cast(t_keypoint, dtype=tf.float32)

    def get_x(self, keypoint):
        value = tf.cond(self.row_major,
                        lambda: keypoint[1],
                        lambda: keypoint[0])
        return value

    def set_x(self, keypoint, value):
        result = tf.cond(self.row_major,
                         lambda: tf.stack([keypoint[0], value]),
                         lambda: tf.stack([value, keypoint[1]]))
        return result

    def get_y(self, keypoint):
        value = tf.cond(self.row_major,
                        lambda: keypoint[0],
                        lambda: keypoint[1])
        return value

    def set_y(self, keypoint, value):
        result = tf.cond(self.row_major,
                         lambda: tf.stack([value, keypoint[1]]),
                         lambda: tf.stack([keypoint[0], value]))
        return result

    def is_outside_image_bounds(self, keypoint, image):
        height = tf.cast(tf.shape(image)[0], dtype=tf.float32)
        width = tf.cast(tf.shape(image)[1], dtype=tf.float32)

        kp_x = self.get_x(keypoint)
        kp_y = self.get_y(keypoint)
        outside = tf.cond(kp_x < tf.constant(0, dtype=tf.float32) or kp_x >= width or kp_y < tf.constant(0,
                                                                                                         dtype=tf.float32) or kp_y >= height,
                          lambda: True,
                          lambda: False)
        return outside

    def _isclose_and_set(self, a, b):
        tolerance = 1e-5
        a = tf.cond(tf.less_equal(tf.abs(tf.subtract(a, b)), tolerance),
                    lambda: tf.cast(b, dtype=tf.float32),
                    lambda: a)
        return a

    # checks if kp is outside and returns -1, -1 if it is
    def check_outside_and_set(self, keypoint, image):
        height = tf.cast(tf.shape(image)[0], dtype=tf.float32)
        width = tf.cast(tf.shape(image)[1], dtype=tf.float32)

        kp_x = self.get_x(keypoint)
        kp_x = self._isclose_and_set(kp_x, 0)
        kp_x = self._isclose_and_set(kp_x, width)

        kp_y = self.get_y(keypoint)
        kp_y = self._isclose_and_set(kp_y, 0)
        kp_y = self._isclose_and_set(kp_y, height)

        close_keypoint = self.make(kp_x, kp_y)

        checked_kp = tf.cond(self.is_outside_image_bounds(close_keypoint, image),
                             lambda: KeypointHelper.make_none_keypoint(),
                             lambda: close_keypoint)
        return checked_kp

    # keypoints only get rescaled if they are not denoted as non existent (-1, -1)
    def rescale(self, keypoint, from_width, from_height, to_width, to_height):
        result = tf.cond(KeypointHelper.is_none_keypoint(keypoint),
                         lambda: keypoint,
                         lambda: self.make(
                             x=tf.multiply(tf.divide(self.get_x(keypoint), from_width), to_width),
                             y=tf.multiply(tf.divide(self.get_y(keypoint), from_height), to_height)
                         ))
        return result

    # returns 0 if both are not present
    # and -1 if there is a mismatch regarding none-ity
    @staticmethod
    def distance_between(keypoint_a, keypoint_b):
        distance = tf.norm(tf.subtract(keypoint_a, keypoint_b), ord='euclidean')
        return distance

    @staticmethod
    def make_none_keypoint():
        return tf.stack([-1.0, -1.0])

    @staticmethod
    def is_none_keypoint(keypoint):
        return tf.reduce_all(tf.equal(keypoint, KeypointHelper.make_none_keypoint()))
