import datetime

import pandas as pd
import tensorflow as tf
from tensorflow.keras.callbacks import ModelCheckpoint
###############################################################################
# A trainer class wrapping the fit and checkpoint calls
# expects train and validation batches as tf datasets that are batched and
# ready to input into fit
from tensorflow.python.keras.callbacks import LearningRateScheduler


class ModelClassificationTrainer:
    def __init__(self, train_batches, validation_batches, batch_size, model, name, checkpoint_dir, train_item_size,
                 test_item_size,
                 train_neg_weight=1.0, constant_lr=None, initial_lr=1e-3
                 ):
        self.train_batches = train_batches
        self.validation_batches = validation_batches
        self.batch_size = batch_size
        self.model = model
        self.name = name
        self.checkpoint_dir = checkpoint_dir
        self.train_item_size = train_item_size
        self.test_item_size = test_item_size
        self.train_neg_weight = train_neg_weight
        self.constant_lr = constant_lr
        self.initial_lr = initial_lr
        return

    def _step_lr(self, epoch):
        if self.constant_lr is not None:
            return self.constant_lr

        lr = self.initial_lr

        if epoch < 5:
            lr = (1 / (5 - epoch)) * lr

        if epoch >= 30 and epoch < 60:
            lr = 0.75 * lr
        if epoch >= 60 and epoch < 90:
            lr = 0.5 * lr
        if epoch >= 90 and epoch < 120:
            lr = lr * 1e-1
        if epoch >= 120 and epoch < 150:
            lr = lr * 1e-2
        if epoch >= 150:
            lr = 0.5 * lr * 1e-2

        tf.print("Epoch {} -- Learning rate is now {}".format(epoch, lr))
        return lr

    def train(self, epochs, use_class_weights=False):
        checkpoint_loss = ModelCheckpoint(
            filepath="{}".format(self.checkpoint_dir) + "{}_best".format(self.name) + '.{epoch:03d}.h5',
            monitor='val_loss',
            mode='min',
            verbose=1,
            save_best_only=True)

        train_count = self.train_item_size
        test_count = self.test_item_size
        epoch_steps = tf.floor(train_count / self.batch_size + 1)
        val_steps = tf.floor(test_count / self.batch_size)

        checkpoint_rocauc = ModelCheckpoint(
            filepath="{}".format(self.checkpoint_dir) + "{}_rocauc".format(self.name) + '.{epoch:03d}.h5',
            monitor='val_roc_auc',
            mode='max',
            verbose=1,
            save_best_only=True)

        checkpoint_prauc = ModelCheckpoint(
            filepath="{}".format(self.checkpoint_dir) + "{}_prauc".format(self.name) + '.{epoch:03d}.h5',
            monitor='val_pr_auc',
            mode='max',
            verbose=1,
            save_best_only=True)

        # lr_scheduler = LearningRateScheduler(self._scheduler)
        log_dir = "/home/florin/tb_logs/" + self.name + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1, profile_batch='40, 60')
        lr_callback = LearningRateScheduler(self._step_lr)

        callbacks = [checkpoint_loss, tensorboard_callback, lr_callback]  # checkpoint_rocauc, checkpoint_prauc,

        tf.print("Calculated epoch steps: {} / {} + 1 = {}".format(train_count, self.batch_size, epoch_steps))
        tf.print("Calculated val steps: {} / {} + 1 = {}".format(test_count, self.batch_size, val_steps))

        if not use_class_weights:
            history = self.model.fit(self.train_batches,
                                     validation_data=self.validation_batches,
                                     steps_per_epoch=epoch_steps,  # 14325 -> 32-448
                                     validation_steps=val_steps,  # 3774 -> 32-118
                                     epochs=epochs,
                                     verbose=1,
                                     callbacks=callbacks)
        else:
            history = self.model.fit(self.train_batches,
                                     validation_data=self.validation_batches,
                                     steps_per_epoch=epoch_steps,  # 14325 -> 32-448
                                     validation_steps=val_steps,  # 3774 -> 32-118
                                     epochs=epochs,
                                     verbose=1,
                                     callbacks=callbacks,
                                     class_weight={0: self.train_neg_weight,
                                                   # we are weighting according to set distribution
                                                   1: 1.})

        print("Saving history ...")
        my_time = datetime.datetime.now().strftime("%H_%M_%S-%d_%m_%y")
        history_df = pd.DataFrame(history.history)
        with open('{}history_{}_{}.json'.format(self.checkpoint_dir, self.name, my_time), 'w') as json_file:
            history_df.to_json(json_file)
        json_file.close()
        return
