import datetime

import pandas as pd
import tensorflow as tf
from tensorflow.keras.callbacks import ModelCheckpoint
###############################################################################
# A trainer class wrapping the fit and checkpoint calls
# expects train and validation batches as tf datasets that are batched and
# ready to input into fit
from tensorflow.python.keras.callbacks import LearningRateScheduler


class ModelKeypointTrainer:
    def __init__(self,
                 train_batches,
                 validation_batches,
                 batch_size,
                 model,
                 name,
                 checkpoint_dir,
                 train_count=14325,
                 test_count=3774,
                 save_regular_checkpoints=True):
        self.train_batches = train_batches
        self.validation_batches = validation_batches
        self.batch_size = batch_size
        self.model = model
        self.name = name
        self.checkpoint_dir = checkpoint_dir
        self.train_count = train_count
        self.test_count = test_count
        self.save_regular_checkpoints = save_regular_checkpoints
        return

    def _step_lr(self, epoch):
        lr = 1e-3

        if epoch >= 90 and epoch < 120:
            lr = 1e-4
        if epoch >= 120:
            lr = 1e-5

        tf.print("Epoch {} -- Learning rate is now {}".format(epoch, lr))
        return lr

    # def validate(self, validation_steps):
    #     loss, mse = self.model.evaluate(self.validation_batches,  steps=validation_steps)
    #     print("Loss: {}".format(loss))
    #     print("MSE: {}".format(mse))
    #     return

    def train(self, epochs):
        checkpoint_best = ModelCheckpoint(
            filepath="{}".format(self.checkpoint_dir) + "{}_best".format(self.name) + '.{epoch:03d}.h5',
            monitor='val_loss',
            verbose=1,
            save_best_only=True)

        epoch_steps = tf.floor(self.train_count / self.batch_size + 1)
        val_steps = tf.floor(self.test_count / self.batch_size + 1)

        checkpoint_normal = ModelCheckpoint(
            filepath="{}".format(self.checkpoint_dir) + "{}_reg".format(self.name) + '.{epoch:03d}.h5',
            monitor='val_loss',
            verbose=1,
            save_best_only=False,
            save_freq=int(10 * epoch_steps))

        # lr_scheduler = LearningRateScheduler(self._scheduler)
        log_dir = "/home/florin/tb_logs/" + self.name + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)
        lr_callback = LearningRateScheduler(self._step_lr)

        callbacks = [checkpoint_best, tensorboard_callback, lr_callback]
        if self.save_regular_checkpoints:
            callbacks.append(checkpoint_normal)

        tf.print("Calculated epoch steps: {} / {} + 1 = {}".format(self.train_count, self.batch_size, epoch_steps))
        tf.print("Calculated val steps: {} / {} + 1 = {}".format(self.test_count, self.batch_size, val_steps))

        history = self.model.fit(self.train_batches,
                                 validation_data=self.validation_batches,
                                 steps_per_epoch=epoch_steps,  # 14325 -> 32-448
                                 validation_steps=val_steps,  # 3774 -> 32-118
                                 epochs=epochs,
                                 verbose=1,
                                 callbacks=callbacks)

        print("Saving history ...")
        my_time = datetime.datetime.now().strftime("%H_%M_%S-%d_%m_%y")
        history_df = pd.DataFrame(history.history)
        with open('{}history_{}_{}.json'.format(self.checkpoint_dir, self.name, my_time), 'w') as json_file:
            history_df.to_json(json_file)
        json_file.close()
        return
