import os
from ast import literal_eval

import numpy as np
import pandas as pd
import tensorflow as tf

from research.ablation_studies.train.arch_train_scripts import train_densenet121
from research.ablation_studies.train.image_transformer import ImageTransformer

CACHE_FILE = "./pen_tip_training.tfcache"
TEST_CACHE_FILE = "./pen_tip_test.tfcache"
BATCH_SIZE = 32
IMG_HEIGHT = 224
IMG_WIDTH = 224
HM_WIDTH = 211
HM_HEIGHT = 211
AUTOTUNE = tf.data.experimental.AUTOTUNE


class AnnotationGetter:
    def __init__(self, annotation_csv):
        self.annotation_csv = annotation_csv
        self.anno_df = pd.read_csv(self.annotation_csv, index_col=0,
                                   dtype={'img_name': np.str},
                                   converters={
                                       'ptip': literal_eval})
        return

    def get_df(self):
        return self.anno_df


###############################################################################
# dataset preprocessing and mapping functions


# map_fn: loading images from file path in a dataset
def load_img(element):
    image_path = element['image_path']
    annotation = element['keypoint']

    img = tf.io.read_file(image_path)
    img = tf.io.decode_jpeg(img, channels=3)
    img = tf.image.convert_image_dtype(img, tf.float32)

    return {'image': img, 'keypoint': annotation}


# map_fn: creating heatmaps for test dataset
# named dict output does not work with keras fit; keras fails to connect the y input to the model
# resulting in no usable gradients
def produce_test_heatmaps(element):
    img = element['image']
    annotation = element['keypoint']

    transformer = ImageTransformer(1, 1, img.dtype, keypoint_row_major=False)
    heatmap = transformer.create_heatmap(HM_WIDTH, HM_HEIGHT, annotation, 3.0)
    return (img, heatmap)


# map_fn: preparing images for training by augmenting them and creating
# heatmaps for augmented keypoints
def augment_img(element):
    img = element['image']
    keypoint = element['keypoint']

    transformer = ImageTransformer(2, 9, img.dtype, keypoint_row_major=False)

    augmented_image, augmented_keypoint = transformer.transform(image=img, keypoint=keypoint)

    heatmap = transformer.create_heatmap(HM_WIDTH, HM_HEIGHT, augmented_keypoint, 3.0)

    return (augmented_image, heatmap)


###############################################################################
# main script and file helpers


def strip_image_name(path):
    return path.replace('\\', '/').split('/')[-1]


def filter_anno_df_according_to_filelist(my_anno_df, filelist):
    reduced_df = my_anno_df.loc[my_anno_df['img_name'].isin(filelist)]
    return reduced_df


def extract_filenames_and_annotations(anno_df, dir):
    jpgs = [file for file in os.listdir(dir) if '.jpg' in file]
    df = filter_anno_df_according_to_filelist(anno_df, jpgs)
    img_names = df['img_name'].values
    img_names = [os.path.join(dir, img_path) for img_path in img_names]
    ptips = [ptip for ptip in df['ptip'].values]
    return img_names, ptips


if __name__ == '__main__':
    # prepare dataset
    print("Preparing dataset ...")
    annotation_getter = AnnotationGetter("/home/florin/data/patches_224/patches_224_keypoints.csv")
    anno_df = annotation_getter.get_df()
    anno_df['img_name'] = anno_df['img_name'].apply(strip_image_name)

    train_dir = "/home/florin/data/patches_224/train/"
    test_dir = "/home/florin/data/patches_224/test/"

    test_img_names, test_ptips = extract_filenames_and_annotations(anno_df, test_dir)
    train_img_names, train_ptips = extract_filenames_and_annotations(anno_df, train_dir)

    # TRAIN SET ///////////////////////////////////////////////////////////////

    # tensor slices creates a dataset that slices all given lists horizontally
    # i.e. each dataset entry is one row containing one element of each input list
    ds_train = tf.data.Dataset.from_tensor_slices(
        {
            'image_path': tf.cast(train_img_names, tf.string),
            'keypoint': train_ptips,
        },
    )

    # load jpegs and convert them to float32; cache that
    ds_train = ds_train.map(load_img, num_parallel_calls=AUTOTUNE)
    ds_train = ds_train.cache(CACHE_FILE)

    # now do augmentation
    ds_train = ds_train.shuffle(buffer_size=1000)
    ds_train = ds_train.map(augment_img, num_parallel_calls=AUTOTUNE)

    # prepare dataset aspects: repeat for repeating augmentation step, batch, prefetch
    ds_train = ds_train.repeat()
    ds_train = ds_train.batch(BATCH_SIZE)
    ds_train = ds_train.prefetch(buffer_size=AUTOTUNE)

    # statistics
    # ds_train = ds_train.apply(tf.data.experimental.latency_stats("data_latency"))

    # stats_aggregator = tf.data.experimental.StatsAggregator()
    # options = tf.data.Options()
    # options.experimental_stats.aggregator = stats_aggregator

    # ds_train = ds_train.with_options(options)

    print("Training set output shapes: {}".format(ds_train.element_spec))

    # TEST SET ///////////////////////////////////////////////////////////////

    # init ds_test
    ds_test = tf.data.Dataset.from_tensor_slices(
        {
            'image_path': tf.cast(test_img_names, tf.string),
            'keypoint': test_ptips,
        }
    )
    ds_test = ds_test.map(load_img, num_parallel_calls=AUTOTUNE)
    ds_test = ds_test.map(produce_test_heatmaps, num_parallel_calls=AUTOTUNE)
    ds_test = ds_test.cache(TEST_CACHE_FILE)
    # ds_test = ds_test.repeat() repeat would only be necessary if we do test time augmentation
    ds_test = ds_test.batch(BATCH_SIZE)
    ds_test = ds_test.prefetch(buffer_size=AUTOTUNE)

    print("Test set output shapes: {}".format(ds_test.element_spec))

    # MODEL ///////////////////////////////////////////////////////////////

    train_densenet121(ds_train, ds_test)
