import tensorflow as tf

from research.ablation_studies.train.keypoint_tf import KeypointHelper


class ImagePatchExtractor:
    def __init__(self, patch_shape, use_random_offset=True):
        self.model_input_shape = patch_shape
        self.use_random_offset = use_random_offset
        return

    def _get_crop_window(self, x, y, img_width, img_height):
        '''returns a crop box containing the coordinates x, y with a random offset

        the random offset is applied in every direction, with a safe border of
        5% of the patch size to avoid shifting keypoints at the border outside
        if small augmentation operations are applied.
        '''
        patch_h, patch_w = self.model_input_shape[:2]

        coord_x = tf.cast(tf.floor(x), dtype=tf.int32)
        coord_y = tf.cast(tf.floor(y), dtype=tf.int32)

        if self.use_random_offset:
            rand_border_x = patch_w / 20
            rand_border_y = patch_h / 20
            rand_x = (tf.random.uniform([]) - 0.5) * (patch_w - rand_border_x)
            rand_y = (tf.random.uniform([]) - 0.5) * (patch_h - rand_border_y)
            rand_x = tf.cast(tf.floor(rand_x), dtype=tf.int32)
            rand_y = tf.cast(tf.floor(rand_y), dtype=tf.int32)
            coord_x = coord_x + rand_x
            coord_y = coord_y + rand_y

        img_width = tf.cast(img_width, dtype=tf.int32)
        img_height = tf.cast(img_height, dtype=tf.int32)

        patch_w_half = tf.cast(patch_w // 2, dtype=tf.int32)
        patch_h_half = tf.cast(patch_h // 2, dtype=tf.int32)
        patch_left = coord_x - patch_w_half
        patch_top = coord_y - patch_h_half
        right_band = img_width - patch_w_half
        top_band = img_height - patch_h_half

        if tf.less(coord_x, patch_w_half):
            crop_x = 0
            crop_w = patch_w - abs(patch_left)
        elif tf.greater(coord_x, right_band):
            crop_x = patch_left
            crop_w = patch_w - (coord_x - (right_band))
        else:
            # inside
            crop_x = patch_left
            crop_w = patch_w

        if coord_y < patch_h_half:
            crop_y = 0
            crop_h = patch_h - abs(patch_top)
        elif coord_y > (top_band):
            crop_y = patch_top
            crop_h = patch_h - (coord_y - (top_band))
        else:
            # inside
            crop_y = patch_top
            crop_h = patch_h

        tf.assert_greater(crop_w, 0, name='crop_window_width_check_pe')
        tf.assert_greater(crop_h, 0, name='crop_window_height_check_pe')
        tf.debugging.assert_greater_equal(crop_x, 0, name='crop_window_offset_x_check')
        tf.debugging.assert_greater_equal(crop_y, 0, name='crop_window_offset_y_check')
        return [crop_y, crop_x, crop_h, crop_w]

    def _get_random_crop_window(self, img_width, img_height):
        patch_h, patch_w = self.model_input_shape[:2]
        x_band = img_width - patch_w
        y_band = img_height - patch_h

        crop_x = tf.cast(tf.random.uniform([1], minval=0, maxval=x_band, dtype=tf.int64), dtype=tf.int32)[0]
        crop_y = tf.cast(tf.random.uniform([1], minval=0, maxval=y_band, dtype=tf.int64), dtype=tf.int32)[0]

        tf.assert_greater(patch_w, 0, name='random_crop_patch_width_check')
        tf.assert_greater(patch_h, 0, name='random_crop_patch_height_check')
        tf.debugging.assert_greater_equal(crop_x, 0, name='random_crop_x_offset_check')
        tf.debugging.assert_greater_equal(crop_y, 0, name='random_crop_y_offset_check')

        return [crop_y, crop_x, patch_h, patch_w]

    def get_img_patch(self, img_bytes, keypoint, img_width, img_height, kp_helper):
        patch_h, patch_w = self.model_input_shape[:2]
        kp_is_none = KeypointHelper.is_none_keypoint(keypoint)

        # keypoint = tf.print(keypoint, [keypoint], "Keypoint: ")

        crop_window = tf.cond(kp_is_none,
                              lambda: self._get_random_crop_window(img_width, img_height),
                              lambda: self._get_crop_window(
                                  kp_helper.get_x(keypoint),
                                  kp_helper.get_y(keypoint),
                                  img_width, img_height))

        coord_x = tf.cast(tf.floor(kp_helper.get_x(keypoint)), dtype=tf.int32)
        coord_y = tf.cast(tf.floor(kp_helper.get_y(keypoint)), dtype=tf.int32)
        new_kp_x = coord_x - crop_window[1]
        new_kp_y = coord_y - crop_window[0]

        new_kp = tf.cond(kp_is_none,
                         lambda: KeypointHelper.make_none_keypoint(),
                         lambda: kp_helper.make(
                             x=new_kp_x,
                             y=new_kp_y))

        # tf.print(keypoint, crop_window)
        crop_window = tf.ensure_shape(crop_window, shape=(4,))
        patch = tf.io.decode_and_crop_jpeg(contents=img_bytes, crop_window=crop_window)
        patch = tf.image.pad_to_bounding_box(patch, 0, 0, patch_h, patch_w)
        return patch, new_kp
