import tensorflow as tf

from research.ablation_studies.dataset.dataset_utils import get_max_of_heatmap, gather_tfrecords, prepare_dataset, \
    sequence_image_feature_description
from research.ablation_studies.helpers import create_model_from_weights
from research.ablation_studies.model.create_keypoint_models import preprocess_resnetv2_input
from research.ablation_studies.train.keypoint_tf import KeypointHelper
from research.ablation_studies.util.image_patch_extract import ImagePatchExtractor


class KeypointPredictor:
    def __init__(self, roi_weights, detail_weights, detail_preprocess_func):
        self.roi_model = create_model_from_weights("ROI", roi_weights)
        self.detail_model = create_model_from_weights("Detail", detail_weights)
        self.patch_extractor = ImagePatchExtractor(patch_shape=self.detail_model.input_shape[1:3],
                                                   use_random_offset=False)
        self.keypoint_helper = KeypointHelper(row_major=False)
        self.detail_preprocess_func = detail_preprocess_func
        return

    # expects tf images, preprocessed and ready
    def get_keypoint(self, roi_image, image_bytes, image_width, image_height):
        if roi_image.shape[0] != self.roi_model.input_shape[1] or \
                roi_image.shape[1] != self.roi_model.input_shape[2]:
            roi_image = tf.image.resize(roi_image, self.roi_model.input_shape[1:3])
        roi_pred = self.roi_model(roi_image)
        roi_max_value, roi_indices = get_max_of_heatmap(roi_pred[0, :, :, 0])
        roi_keypoint = self.keypoint_helper.make(x=roi_indices[1], y=roi_indices[0])

        patch = self.patch_extractor.get_img_patch(image_bytes, roi_keypoint, image_width, image_height,
                                                   self.keypoint_helper)
        patch = tf.expand_dims(patch, axis=0)
        patch = self.detail_preprocess_func(patch)

        detail_pred_tip = tf.squeeze(self.detail_model(patch))
        detail_max_value, detail_indices = get_max_of_heatmap(detail_pred_tip[0, :, :, 0])
        detail_keypoint = self.keypoint_helper.make(x=detail_indices[1], y=detail_indices[0])
        full_res_detail_keypoint = self.keypoint_helper.rescale(
            detail_keypoint,
            self.detail_model.output_shape[2],
            self.detail_model.output_shape[1],
            image_width,
            image_height)
        return full_res_detail_keypoint


if __name__ == '__main__':
    large_tfrecord_dir = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/sequence/"
    tfrecords = gather_tfrecords(tfrecord_dir=large_tfrecord_dir)
    loaded_dataset = prepare_dataset(tfrecords, sequence_image_feature_description)

    roi_weights = ""
    detail_weights = ""
    predictor = KeypointPredictor(roi_weights, detail_weights, preprocess_resnetv2_input)


    def output_mapper(element):
        image_bytes = element['image/encoded']
        img_width = element['image/width']
        img_height = element['image/height']

        img = tf.io.decode_jpeg(image_bytes, ratio=4, channels=3)
        img = tf.image.convert_image_dtype(img, tf.float32)

        tf.debugging.assert_less_equal(tf.reduce_max(img), 1.0)
        tf.debugging.assert_greater_equal(tf.reduce_max(img), 0.0)

        roi_image = preprocess_resnetv2_input(img)

        detail_tip_keypoint = predictor.get_keypoint(roi_image, image_bytes, img_width, img_height)
        element['image/ptip_x'] = predictor.keypoint_helper.get_x(detail_tip_keypoint)
        element['image/ptip_y'] = predictor.keypoint_helper.get_y(detail_tip_keypoint)
        return element


    loaded_dataset = loaded_dataset.map(output_mapper)
    tf.data.experimental.save(loaded_dataset, "/home/florin/data/sequence/")

    # now create inferences
    exit()
