import matplotlib
import tensorflow as tf

from research.ablation_studies.dataset.dataset_utils import gather_tfrecords, get_max_of_heatmap
from research.ablation_studies.eval.eval_datasets.dataset_evaluator_pixels import EvalDatasetPixelError
from research.ablation_studies.eval.eval_datasets.eval_dataset import EvalDataset
from research.ablation_studies.train.keypoint_tf import KeypointHelper

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class EvalDatasetInspector:
    def __init__(self, dataset, view_size, subject=0):
        self.dataset = dataset
        self.view_size = view_size

        self.pixel_evaluator = EvalDatasetPixelError(None, 1080, detail_mode=False, zero_threshold=1e-3)
        print(self.pixel_evaluator.zero_threshold)
        self.subject = 0
        self.data_it = self.dataset.get_iterator()

        self.fig, self.axes = plt.subplots(ncols=2, nrows=1, squeeze=False)
        self.fig.tight_layout(pad=2.0)
        self.fig.canvas.mpl_connect('key_press_event', self._on_key)
        return

    def _on_key(self, event):
        if event.key == 'e':
            self._plot_batch()
        elif event.key == 'w':
            self._plot_batch()
        elif event.key in ['1', '2', '3', '4', '5', '6', '7', '0']:
            self._seek_to_subject(int(event.key))
            self._plot_batch()
        self.show()
        return

    def _seek_to_subject(self, subject):
        subject_set = self.dataset.get_filtered_subject_set(subject)
        self.data_it = iter(subject_set)
        return

    def _split_batch(self, element):
        images = element[0]
        heatmaps = element[1]
        keypoints = element[2]
        aug_keypoints = element[3]
        orig_images = element[4]
        frame_info = None
        if len(element) > 5:
            frame_info = {
                'subject': element[5],
                'task': element[6],
                'frame': element[7],
            }
        return orig_images, images, heatmaps, keypoints, aug_keypoints, frame_info

    def _convert_keypoint_to_text(self, keypoint):
        kp_helper = KeypointHelper(row_major=False)
        x = kp_helper.get_x(keypoint)
        y = kp_helper.get_y(keypoint)
        return "({} | {})".format(x, y)

    def _plot_batch(self):
        elem = self.data_it.get_next()

        image = elem['image/decoded']
        self.axes[0][0].imshow(image[0])

        output = self.dataset.get_inference(image)
        ptip_heatmap = tf.image.convert_image_dtype(output[0, :, :, 0], tf.uint8)
        self.axes[0][1].imshow(ptip_heatmap)

        ptip_heatmap = tf.reshape(ptip_heatmap, shape=(ptip_heatmap.shape[0], ptip_heatmap.shape[1], 1))
        elem['prediction/ptip_heatmap_encoded'] = tf.io.encode_jpeg(ptip_heatmap, chroma_downsampling=False)
        elem['prediction/width'] = 444
        elem['prediction/height'] = 444

        direct_max_value, _ = get_max_of_heatmap(output[0, :, :, 0])

        max_value, keypoint_rescaled, annotated_keypoint, ptip_none, ptail_none = self.pixel_evaluator._get_element_peak(
            elem, threshold=0.001, keypoint_type="ptip")
        print("----------------")
        print("Detected: {} (max: {})".format(keypoint_rescaled, max_value))
        print("Annotated: {}".format(annotated_keypoint))
        print("Direct max: {}".format(direct_max_value))
        print("Less/Equal: {}".format(tf.math.less_equal(max_value, self.pixel_evaluator.zero_threshold)))

        return

    def _clear_axes(self):
        for ax in self.axes.flatten():
            ax.cla()
        return

    def show(self):
        self._clear_axes()
        self._plot_batch()
        plt.show()
        return


if __name__ == '__main__':
    large_tfrecord_dir = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/dataset/"
    val1_tfrecord_dir = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/eval/"
    val2_tfrecord_dir = "/home/florin/data/val2/tfrecord/"

    model_file = "/home/florin/diss_models/final_weights/kp/roi/effnet-b5/effnetb5_roi_u3_ft_ra9_best.071.h5"

    eval_dataset = EvalDataset(
        model_file,
        gather_tfrecords(val2_tfrecord_dir),
        (456, 456, 3),
        1,
        test_subjects=None,
        val2_mode=True)

    inspector = EvalDatasetInspector(eval_dataset, 1, 0)
    inspector.show()
