import matplotlib
import tensorflow as tf

from misc_util.heatmap import create_heatmaps
from research.ablation_studies.dataset.dataset_utils import gather_tfrecords
from research.ablation_studies.dataset.pipeline_creator import ROIInputMap, PipelineCreator, PatchInputMap
from research.ablation_studies.train.augmenter import Augmenter
from research.ablation_studies.train.keypoint_tf import KeypointHelper

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class DebugROIInputMap(ROIInputMap):
    def _augment_and_get_heatmaps(self, img, t_keypoints, kp_helper):
        augmenter = Augmenter(
            num_layers=2,
            magnitude=self.ra_magnitude,
            image_dtype=img.dtype,
            image_shape=img.shape,
            keypoint_helper=kp_helper)

        augmented_image, augmented_keypoints = augmenter.transform(image=img,
                                                                   keypoints=t_keypoints,
                                                                   do_preprocessing=False)

        img_h = tf.cast(tf.shape(augmented_image)[0], dtype=tf.float32)
        img_w = tf.cast(tf.shape(augmented_image)[1], dtype=tf.float32)

        heatmaps = create_heatmaps(img_w, img_h,
                                   self.model_output_shape[1], self.model_output_shape[0],
                                   2.5,
                                   augmented_keypoints,
                                   kp_helper)

        # we explicitly rescale the augmented keypoints
        # for debugging reasons, because in production,
        # heatmaps are created with downscaled keypoints
        augmented_keypoints = tf.map_fn(
            lambda kp: kp_helper.rescale(kp, img_w, img_h, self.model_output_shape[1], self.model_output_shape[0]),
            augmented_keypoints, name='debug_preprocessing_keypoint_rescale')

        return augmented_image, heatmaps, augmented_keypoints

    def _augment_img(self, element):
        # tf.print(element)
        kp_helper = KeypointHelper(row_major=False)
        img, t_keypoints = self._preprocess_example(element, kp_helper)
        augmented_img, heatmaps, aug_keypoints = self._augment_and_get_heatmaps(img, t_keypoints, kp_helper)

        tf.assert_equal(tf.shape(img), tf.shape(augmented_img))
        tf.assert_equal(tf.shape(heatmaps)[:2], self.model_output_shape[:2])

        resized_img = tf.image.resize(augmented_img, self.model_input_shape[:2],
                                      name='augment_img_resize_to_model_specs')

        # for debugging, we downscale the original keypoints
        img_h = tf.cast(tf.shape(augmented_img)[0], dtype=tf.float32)
        img_w = tf.cast(tf.shape(augmented_img)[1], dtype=tf.float32)
        t_keypoints = tf.map_fn(
            lambda kp: kp_helper.rescale(kp, img_w, img_h, self.model_input_shape[1], self.model_input_shape[0]),
            t_keypoints, name='debug_orig_keypoint_rescale')

        orig_img = tf.image.resize(img, self.model_input_shape[:2],
                                   name='augment_img_resize_to_model_specs')
        return (resized_img, heatmaps, t_keypoints, aug_keypoints, orig_img)


class DebugPatchInputMap(PatchInputMap):
    def _load_img(self, element):
        image_bytes = element['image/encoded']
        ptip_x = element['image/ptip_x']
        ptail_x = element['image/ptail_x']
        ptip_y = element['image/ptip_y']
        ptail_y = element['image/ptail_y']
        img_width = element['image/width']
        img_height = element['image/height']

        kp_helper = KeypointHelper(row_major=False)

        keypoint = tf.cond(self.use_tail,
                           lambda: kp_helper.from_list([ptail_x, ptail_y]),
                           lambda: kp_helper.from_list([ptip_x, ptip_y]))

        patch, new_kp = self._get_img_patch(image_bytes, keypoint, img_width, img_height, kp_helper)
        patch = tf.image.convert_image_dtype(patch, tf.float32)
        patch_h, patch_w = self.model_input_shape[:2]
        patch = tf.ensure_shape(patch, shape=(patch_h, patch_w, 3))

        return {'patch': patch,
                'keypoint': kp_helper.to_list(new_kp),
                'subject': element['image/subject_id'],
                'task': element['image/task_id'],
                'frame': element['image/frame_id']}

    def _augment_and_get_heatmaps(self, patch, t_keypoints, kp_helper):
        augmenter = Augmenter(
            num_layers=2,
            magnitude=self.ra_magnitude,
            image_dtype=patch.dtype,
            image_shape=patch.shape,
            keypoint_helper=kp_helper)

        augmented_image, augmented_keypoints = augmenter.transform(image=patch, keypoints=t_keypoints)

        img_h = tf.cast(tf.shape(augmented_image)[0], dtype=tf.float32)
        img_w = tf.cast(tf.shape(augmented_image)[1], dtype=tf.float32)

        heatmaps = create_heatmaps(img_w, img_h,
                                   self.model_output_shape[1], self.model_output_shape[0],
                                   self.std_dev,
                                   augmented_keypoints,
                                   kp_helper)
        return augmented_image, heatmaps, augmented_keypoints

    def _augment_img(self, element):
        kp_helper = KeypointHelper(row_major=False)
        patch, t_keypoints = self._preprocess_example(element, kp_helper)
        augmented_img, heatmaps, augmented_keypoints = self._augment_and_get_heatmaps(patch, t_keypoints, kp_helper)

        # augmented_img = self.input_prep_func(augmented_img)
        return (augmented_img,
                heatmaps,
                t_keypoints,
                augmented_keypoints,
                patch,
                element['subject'],
                element['task'],
                element['frame'])


class PipelineInspector:
    def __init__(self, input_mapper, view_size, tfrecord_files):
        self.input_mapper = input_mapper
        self.pipeline_creator = PipelineCreator(self.input_mapper,
                                                batch_size=view_size,
                                                tfrecord_files=tfrecord_files,
                                                cache_prefix_id='inspector_',
                                                test_subjects=[0, 1, 2, 3, 4, 5, 6])
        self.view_size = view_size
        self.train_ds, self.test_ds = self.pipeline_creator.get_datasets()
        self.train_it = iter(self.train_ds)

        self.fig, self.axes = plt.subplots(ncols=4, nrows=view_size, squeeze=False)
        self.fig.tight_layout(pad=2.0)
        self.fig.canvas.mpl_connect('key_press_event', self._on_key)
        return

    def _on_key(self, event):
        if event.key == 'e':
            self._plot_batch()
        elif event.key == 'w':
            self._plot_batch()
        self.show()
        return

    def _split_batch(self, element):
        images = element[0]
        heatmaps = element[1]
        keypoints = element[2]
        aug_keypoints = element[3]
        orig_images = element[4]
        frame_info = None
        if len(element) > 5:
            frame_info = {
                'subject': element[5],
                'task': element[6],
                'frame': element[7],
            }
        return orig_images, images, heatmaps, keypoints, aug_keypoints, frame_info

    def _convert_keypoint_to_text(self, keypoint):
        kp_helper = KeypointHelper(row_major=False)
        x = kp_helper.get_x(keypoint)
        y = kp_helper.get_y(keypoint)
        return "({} | {})".format(x, y)

    def _plot_batch(self):
        elem = self.train_it.get_next()
        orig_images, images, heatmaps, keypoints, aug_keypoints, frame_info = self._split_batch(elem)
        tf.print(keypoints.shape)
        for index, item in enumerate(orig_images):
            self.axes[index][0].imshow(item)
            self.axes[index][0].text(0, -15, self._convert_keypoint_to_text(keypoints[index][0]))
            if frame_info is not None:
                self.axes[index][0].text(0, -32,
                                         "{}_{}_{}".format(
                                             frame_info['subject'][index],
                                             frame_info['task'][index],
                                             frame_info['frame'][index]),
                                         color='red')

        for index, item in enumerate(images):
            self.axes[index][1].imshow(item)
            self.axes[index][1].text(0, 290, self._convert_keypoint_to_text(aug_keypoints[index][0]))

        for index, item in enumerate(heatmaps):
            self.axes[index][1].imshow(item[:, :, 0], alpha=0.4)
            self.axes[index][2].imshow(item[:, :, 0])
            if heatmaps.shape[-1] > 1:
                self.axes[index][3].imshow(item[:, :, 1])
            self.axes[index][2].text(0, -10, self._convert_keypoint_to_text(aug_keypoints[index][0]))
        return

    def _clear_axes(self):
        for ax in self.axes.flatten():
            ax.cla()
        return

    def show(self):
        self._clear_axes()
        self._plot_batch()
        plt.show()
        return


if __name__ == '__main__':
    large_tfrecord_dir = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/dataset/"
    val1_tfrecord_dir = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/eval/"
    val2_tfrecord_dir = "/home/florin/data/val2/tfrecord/"

    input_mapper = DebugROIInputMap(model_input_shape=(224, 224, 3),
                                    model_output_shape=(224, 224, 2),
                                    ra_magnitude=9,
                                    downscale_ratio=4,
                                    std_dev=3.0,
                                    input_prep_func=tf.keras.applications.resnet_v2.preprocess_input)

    # input_mapper = DebugPatchInputMap(model_input_shape=(224,224,3),
    #                            model_output_shape=(224,224,2),
    #                            ra_magnitude=9,
    #                            downscale_ratio=1,
    #                            std_dev=3.0,
    #                            input_prep_func=tf.keras.applications.resnet_v2.preprocess_input)

    inspector = PipelineInspector(input_mapper=input_mapper,
                                  view_size=4,
                                  tfrecord_files=gather_tfrecords(val2_tfrecord_dir))
    inspector.show()
