import matplotlib
import numpy as np
import tensorflow as tf
from tensorflow.python.keras.models import load_model

from misc_util.imgproc_utils import resize_image_array_pil
from research.ablation_studies.dataset.dataset_utils import gather_tfrecords, sequence_image_feature_description
from research.ablation_studies.dataset.pipeline_creator import SequencePipelineCreator, \
    ClassificationKeypointSequenceInputMap
from research.ablation_studies.model.create_keypoint_models import get_preprocess_func

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class SequenceKeypointPipelineInspector:
    def __init__(self, batch_size, timesteps, tfrecord_files, inference_weights=None, show_test=True):
        input_mapper = ClassificationKeypointSequenceInputMap(
            (224, 224, 3),
            (211, 211),
            6,
            downscale_ratio=1,
            input_prep_func=get_preprocess_func("effnetb0"),
            use_augmentation=True)

        pipeline = SequencePipelineCreator(
            input_mapper,
            batch_size,
            tfrecord_files,
            image_features=sequence_image_feature_description,
            timesteps=timesteps,
            shift=timesteps)

        self.train_ds, self.test_ds = pipeline.get_datasets()
        self.batch_size = batch_size
        self.train_it = iter(self.train_ds)
        self.test_it = iter(self.test_ds)
        self.inference_weights = inference_weights
        self.show_test = show_test

        if self.inference_weights is not None:
            self.model = load_model(inference_weights)

        self.fig, self.axes = plt.subplots(ncols=timesteps, nrows=batch_size * 3, squeeze=False, sharex='col',
                                           sharey='row')
        self.fig.tight_layout(pad=2.0)
        self.fig.subplots_adjust(hspace=0, wspace=0)

        self.fig.canvas.mpl_connect('key_press_event', self._on_key)
        return

    def _on_key(self, event):
        if event.key == 'e':
            self._plot_batch()
        elif event.key == 'w':
            self._plot_batch()
        self.show()
        return

    def _count_stats(self, y_true, y_pred):
        def get_clean_ups_downs(heatmaps):
            maxes = tf.reduce_max(heatmaps, axis=-2)
            ups = tf.greater(maxes[:, :, :, 0], maxes[:, :, :, 1])
            downs = tf.greater(maxes[:, :, :, 1], maxes[:, :, :, 0])
            over_delta_ups = tf.greater(maxes[:, :, :, 0], 0.1)
            over_delta_downs = tf.greater(maxes[:, :, :, 1], 0.1)
            clean_ups = tf.logical_and(ups, over_delta_ups)
            clean_downs = tf.logical_and(downs, over_delta_downs)
            return clean_ups, clean_downs

        true_ups, true_downs = get_clean_ups_downs(y_true)
        pred_ups, pred_downs = get_clean_ups_downs(y_pred)

        tp = tf.math.count_nonzero(tf.logical_and(pred_downs, true_downs))
        tn = tf.math.count_nonzero(tf.logical_and(true_ups, pred_ups))
        fp = tf.math.count_nonzero(tf.logical_and(pred_downs, true_ups))
        fn = tf.math.count_nonzero(tf.logical_and(pred_ups, true_downs))

        print(tp)
        print(tn)
        print(fp)
        print(fn)

        # self.true_positives.assign_add(tf.cast(tp, dtype=tf.float32))
        # self.true_negatives.assign_add(tf.cast(tn, dtype=tf.float32))
        # self.false_positives.assign_add(tf.cast(fp, dtype=tf.float32))
        # self.false_negatives.assign_add(tf.cast(fn, dtype=tf.float32))

    def _plot_batch(self):
        if self.show_test:
            elem = self.test_it.get_next()
        else:
            elem = self.train_it.get_next()

        batch_x = elem[0]
        heatmaps = elem[1]

        for batch_index, batch_element in enumerate(batch_x):
            for img_index, image in enumerate(batch_element):
                self.axes[batch_index * 3][img_index].imshow(image)

        if self.inference_weights is not None:
            true_heatmaps = heatmaps
            heatmaps = self.model(get_preprocess_func("effnetb0")(batch_x))
            self._count_stats(true_heatmaps, heatmaps)

        for batch_index, batch_element in enumerate(heatmaps):
            for img_index, heatmap in enumerate(batch_element):
                up_heatmap = resize_image_array_pil(heatmap[:, :, 0].numpy(), (224, 224))
                down_heatmap = resize_image_array_pil(heatmap[:, :, 1].numpy(), (224, 224))
                # self.axes[batch_index * 3][img_index].imshow(up_heatmap, alpha=0.3)
                # self.axes[batch_index * 3][img_index].imshow(down_heatmap, alpha=0.3)
                up_max = np.max(up_heatmap)
                down_max = np.max(down_heatmap)
                if up_max > down_max:
                    pen_down = 0
                else:
                    pen_down = 1
                print("UP max: {} and DOWN max: {}. Result: {}".format(up_max, down_max, pen_down))
                self.axes[batch_index][img_index].plot([0, 224], [224, 224], color=(pen_down, 0, 0),
                                                       linewidth=4)

                self.axes[(batch_index * 3) + 1][img_index].imshow(up_heatmap, vmin=0, vmax=1)
                self.axes[(batch_index * 3) + 2][img_index].imshow(down_heatmap, vmin=0, vmax=1)

        return

    def _clear_axes(self):
        for ax in self.axes.flatten():
            ax.cla()
        return

    def show(self):
        self._clear_axes()
        self._plot_batch()
        plt.show()
        return


if __name__ == '__main__':
    tfrecord_dir = "/home/florin/data/val1/sequence/"

    # inference_weights = "/home/florin/diss_models/effnetb0_classify_rnn_kp_diss/rnn_kp_ts3_shift3_lstm64_adam/1_chkpt/rnn_kp_ts3_shift3_lstm64_adam_best.120.h5"
    # inference_weights = "/home/florin/diss_models/effnetb0_classify_rnn_kp_diss/rnn_kp_ts3_shift3_lstm128_ctd_bnf_adam_upc/1_chkpt/rnn_kp_ts3_shift3_lstm128_ctd_bnf_adam_upc_best.069.h5"
    inference_weights = "/home/florin/diss_models/effnetb0_classify_rnn_kp_diss/rnn_kp_ts10_shift10_lstm128_bnf_adam_upc/1_chkpt/rnn_kp_ts10_shift10_lstm128_bnf_adam_upc_best.099.h5"

    inspector = SequenceKeypointPipelineInspector(
        batch_size=1,
        timesteps=10,
        tfrecord_files=gather_tfrecords(tfrecord_dir),
        inference_weights=inference_weights,
        show_test=True)

    inspector.show()
