import matplotlib
from tensorflow.python.keras.models import load_model

from research.ablation_studies.dataset.dataset_utils import gather_tfrecords, sequence_image_feature_description
from research.ablation_studies.dataset.pipeline_creator import ClassificationSequenceInputMap, SequencePipelineCreator
from research.ablation_studies.model.create_keypoint_models import get_preprocess_func

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class DebugClassificationSequenceInputMap(ClassificationSequenceInputMap):
    def __init__(self,
                 model_input_shape,
                 model_output_shape,
                 ra_magnitude,
                 downscale_ratio,
                 input_prep_func=None,
                 use_augmentation=False):
        super().__init__(model_input_shape, model_output_shape, ra_magnitude, downscale_ratio,
                         input_prep_func=input_prep_func, use_augmentation=use_augmentation)
        return

    def _load_img(self, element):
        normal_output_tuple = super()._load_img(element)
        # if we dont augment we need to format the input here
        new_first_element = (
            normal_output_tuple[0],
            element['image/subject_id'],
            element['image/task_id'],
            element['image/frame_id'],
            element['image/ptip_x'],
            element['image/ptip_y'],
            element['image/pen_down'])
        return (new_first_element,
                normal_output_tuple[1],
                normal_output_tuple[2],
                normal_output_tuple[3],
                normal_output_tuple[4])

    def _augment_img(self, element):
        return None

    def _produce_test_heatmaps(self, element):
        return None


class SequencePipelineInspector:
    def __init__(self, batch_size, timesteps, tfrecord_files, inference_weights=None):
        input_mapper = DebugClassificationSequenceInputMap(
            (224, 224, 3),
            (0, 0),
            9,
            downscale_ratio=1,
            input_prep_func=get_preprocess_func("effnetb0"),
            use_augmentation=True)

        pipeline = SequencePipelineCreator(
            input_mapper,
            batch_size,
            tfrecord_files,
            image_features=sequence_image_feature_description,
            timesteps=timesteps,
            shift=1)

        self.train_ds, self.test_ds = pipeline.get_datasets()
        self.batch_size = batch_size
        self.train_it = iter(self.train_ds)
        self.inference_weights = inference_weights
        if self.inference_weights is not None:
            self.model = load_model(inference_weights)

        self.fig, self.axes = plt.subplots(ncols=timesteps, nrows=batch_size, squeeze=False, sharex='col', sharey='row')
        self.fig.tight_layout(pad=2.0)
        self.fig.subplots_adjust(hspace=0, wspace=0)

        self.fig.canvas.mpl_connect('key_press_event', self._on_key)
        return

    def _on_key(self, event):
        if event.key == 'e':
            self._plot_batch()
        elif event.key == 'w':
            self._plot_batch()
        self.show()
        return

    def _plot_batch(self):
        elem = self.train_it.get_next()

        batch_x = elem[0]
        updowns = elem[1]
        anno_color = 'b-'
        do_print = False

        for batch_index, batch_element in enumerate(batch_x[0]):
            print("###################################")
            print("Sequence #{}".format(batch_index))
            for img_index, image in enumerate(batch_element):
                if do_print:
                    print("###################################")
                    print("Subject {}".format(batch_x[1][batch_index][img_index]))
                    print("Task {}".format(batch_x[2][batch_index][img_index]))
                    print("Frame {}".format(batch_x[3][batch_index][img_index]))
                    print("Ptip XY {} | {}".format(batch_x[4][batch_index][img_index],
                                                   batch_x[5][batch_index][img_index]))
                    print("PenDown {}".format(batch_x[5][batch_index][img_index]))
                    print("PenDown Actual {}".format(updowns[batch_index][img_index]))

                self.axes[batch_index][img_index].imshow(image)
                if updowns[batch_index][img_index] == 1:
                    anno_color = 'r-'
                else:
                    anno_color = 'b-'

                self.axes[batch_index][img_index].plot([0, 224], [0, 0], anno_color, linewidth=4)

        if self.inference_weights is not None:
            y = self.model(get_preprocess_func("effnetb0")(batch_x))
            print(y)
            for batch_index, batch_element in enumerate(y):
                for img_index, image in enumerate(batch_element):
                    print(y[batch_index][img_index])
                    if y[batch_index][img_index] > 0.1:
                        anno_color = 'r-'
                    else:
                        anno_color = 'b-'

                    self.axes[batch_index][img_index].plot([0, 224], [224, 224],
                                                           color=(y[batch_index][img_index][0], 0, 0), linewidth=4)
        return

    def _clear_axes(self):
        for ax in self.axes.flatten():
            ax.cla()
        return

    def show(self):
        self._clear_axes()
        self._plot_batch()
        plt.show()
        return


if __name__ == '__main__':
    tfrecord_dir = "/home/florin/data/sequence/"

    inference_weights = None  # "/home/florin/diss_models/effnetb0_classify_basic/simple_full_set_weighted_ls0.1/1_chkpt/simple_full_set_weighted_ls0.1_best.094.h5"

    inspector = SequencePipelineInspector(
        batch_size=4,
        timesteps=10,
        tfrecord_files=gather_tfrecords(tfrecord_dir),
        inference_weights=inference_weights)

    inspector.show()
