import os

import matplotlib
import tensorflow as tf
from tensorflow.python.keras.models import load_model

from research.ablation_studies.dataset.dataset_utils import bin_set_pair_feature_description, gather_tfrecords, \
    prepare_dataset, image_feature_description
from research.ablation_studies.dataset.pipeline_creator import PipelineCreator
from research.ablation_studies.model.create_keypoint_models import get_preprocess_func
from research.ablation_studies.text_ex.bin_pair_input_map import BinPairInputMap
from research.ablation_studies.train.keypoint_tf import KeypointHelper
from research.ablation_studies.util.image_patch_extract import ImagePatchExtractor

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class TextExViewer:
    def __init__(self, inference_weights, image_dir, batch_size):
        self.image_dir = image_dir
        self.images = os.listdir(image_dir)
        self.images = [image for image in self.images if '.jpg' in image]
        self.image_counter = 0
        self.batch_size = batch_size
        self.model = load_model(inference_weights)

        self.fig, self.axes = plt.subplots(ncols=2, nrows=1, squeeze=False)
        self.fig.tight_layout(pad=2.0)
        self.fig.canvas.mpl_connect('key_press_event', self._on_key)
        return

    def _on_key(self, event):
        if event.key == 'e':
            self._plot_batch()
        elif event.key == 'w':
            self._plot_batch()
        self.show()
        return

    def _load_input_img(self, image_counter):
        path = os.path.join(self.image_dir, self.images[image_counter])
        img = tf.io.read_file(path)
        img = tf.image.decode_jpeg(img, channels=3)
        img = tf.image.convert_image_dtype(img, tf.float32)
        return img

    def _plot_batch(self):
        current_img = self._load_input_img(self.image_counter)
        self.axes[0][0].imshow(current_img)

        inference = self.model([current_img])
        # output_heatmap = tf.image.convert_image_dtype(inference[0, :, :, 0], tf.uint8)
        self.axes[0][1].imshow(inference[0, :, :, 0])

        self.image_counter += 1
        return

    def _clear_axes(self):
        for ax in self.axes.flatten():
            ax.cla()
        return

    def show(self):
        self._clear_axes()
        self._plot_batch()
        plt.show()
        return


class TextExInputInspector:
    def __init__(self, inference_weights, tfrecord_files, batch_size):
        input_mapper = BinPairInputMap(
            (224, 224, 3),
            (211, 211, 1),
            6,
            input_prep_func=get_preprocess_func("effnetb0"))

        #
        pipeline = PipelineCreator(input_mapper, batch_size, tfrecord_files,
                                   image_features=bin_set_pair_feature_description,
                                   test_subjects=[0])

        self.ds_train, self.ds_test = pipeline.get_datasets()
        self.test_it = iter(self.ds_test)
        self.train_it = iter(self.ds_train)

        self.batch_size = batch_size
        self.model = load_model(inference_weights)

        self.fig, self.axes = plt.subplots(ncols=3, nrows=1, squeeze=False)
        self.fig.tight_layout(pad=2.0)
        self.fig.canvas.mpl_connect('key_press_event', self._on_key)
        return

    def _on_key(self, event):
        if event.key == 'e':
            self._plot_batch()
        elif event.key == 'w':
            self._plot_batch()
        self.show()
        return

    def _plot_batch(self):
        elem = self.train_it.get_next()

        current_img = elem[0]
        bin_img = elem[1]
        self.axes[0][0].imshow(current_img[0])
        self.axes[0][1].imshow(bin_img[0, :, :, 0])

        inference = self.model(current_img)
        # output_heatmap = tf.image.convert_image_dtype(inference[0, :, :, 0], tf.uint8)

        self.axes[0][2].imshow(inference[0, :, :, 0])
        return

    def _clear_axes(self):
        for ax in self.axes.flatten():
            ax.cla()
        return

    def show(self):
        self._clear_axes()
        self._plot_batch()
        plt.show()
        return


# display dataset images and infer on user defined patches
class TextExTFRecordViewer:
    def __init__(self, inference_weights, tfrecord_files, batch_size):
        self.dataset = prepare_dataset(tfrecord_files, image_feature_description)
        self.data_iter = iter(self.dataset)

        self.batch_size = batch_size
        self.model = load_model(inference_weights)

        self.fig, self.axes = plt.subplots(ncols=3, nrows=1, squeeze=False)
        self.fig.tight_layout(pad=2.0)
        self.fig.canvas.mpl_connect('key_press_event', self._on_key)
        self.fig.canvas.mpl_connect('button_press_event', self.on_click)

        self.patch_extractor = ImagePatchExtractor(patch_shape=(224, 224, 3),
                                                   use_random_offset=False)

        self.current_patch = tf.zeros(shape=(224, 224, 3))
        self.current_element = next(self.data_iter)
        self.subject_set = None
        return

    def _on_key(self, event):
        if event.key == 'e':
            self.current_element = next(self.data_iter)
            self._plot_batch()
        elif event.key == 'w':
            self._plot_batch()
        elif event.key == 'r':
            self.current_patch = tf.image.rot90(self.current_patch).numpy()
            self._update_patch_and_inference_display()
        elif event.key in ['1', '2', '3', '4', '5', '6', '7', '0']:
            self._seek_to_subject(int(event.key))
            self.current_element = next(self.data_iter)
            self._plot_batch()
        self.show()
        return

    def _seek_to_subject(self, subject):
        print(type(subject))

        @tf.function
        def _check_if_test_subject(entry, search_id):
            subject_id = entry['image/subject_id']
            check_id_tensor = tf.equal(tf.cast(search_id, tf.int64), subject_id)

            result = tf.cond(tf.equal(tf.math.count_nonzero(check_id_tensor), True),
                             lambda: tf.constant(True, dtype=tf.bool),
                             lambda: tf.constant(False, dtype=tf.bool), name='check_if_test_subject')
            return result

        self.subject_set = self.dataset.filter(lambda x: _check_if_test_subject(x, subject))
        self.data_iter = iter(self.subject_set)

        return

    def _update_patch_and_inference_display(self):
        self.axes[0][1].imshow(self.current_patch)
        float_patch = tf.image.convert_image_dtype(self.current_patch, tf.float32)
        inference = self.model([float_patch])
        self.axes[0][2].imshow(inference[0, :, :, 0], vmin=0, vmax=1)
        return

    def _convert_ax_pos_to_img_pos(self, ax_pos, img_width, img_height):
        img_x = ax_pos[0] * img_width
        img_y = (1.0 - ax_pos[1]) * img_height
        return (img_x, img_y)

    def _get_patch(self, example, keypoint):
        image_bytes = example['image/encoded']
        img_width = example['image/width']
        img_height = example['image/height']
        patch, patch_keypoint = self.patch_extractor.get_img_patch(image_bytes, keypoint, img_width, img_height,
                                                                   KeypointHelper(row_major=False))
        return patch, patch_keypoint

    def on_click(self, event):
        if self.axes[0][0].in_axes(event):
            # click on large image
            ax_pos = self.axes[0][0].transAxes.inverted().transform((event.x, event.y))
            img_ax_pos = self._convert_ax_pos_to_img_pos(ax_pos,
                                                         int(self.current_element['image/width']),
                                                         int(self.current_element['image/height']))

            kp_helper = KeypointHelper(row_major=False)
            self.current_patch, _ = self._get_patch(self.current_element,
                                                    kp_helper.make(x=img_ax_pos[0], y=img_ax_pos[1]))
            self._update_patch_and_inference_display()
        return

    def _plot_batch(self):
        img_bytes = self.current_element['image/encoded']
        img = tf.image.decode_jpeg(img_bytes, channels=3)

        self.axes[0][0].imshow(img)

        # inference = self.model([self.current_patch])
        # self.axes[0][1].imshow(inference[0, :, :, 0])
        return

    def _clear_axes(self):
        for ax in self.axes.flatten():
            ax.cla()
        return

    def show(self):
        # self._clear_axes()
        plt.show()
        return


if __name__ == '__main__':
    image_dir = "/home/florin/data/bin_patches/set/"
    model_file = "/home/florin/diss_models/effnetb0_text_ex/correct_pair/1_chkpt/correct_pair_best.099.h5"
    tfrecord_dir = "/home/florin/data/bin_patches/"
    tfrecord_val2_dir = "/home/florin/data/val2/tfrecord/"

    # inspector = TextExInputInspector(model_file, gather_tfrecords(tfrecord_dir), 1)
    # inspector = TextExViewer(model_file, image_dir, 1)
    inspector = TextExTFRecordViewer(model_file, gather_tfrecords(tfrecord_val2_dir), 1)

    inspector.show()
