import os
import time

import matplotlib.pyplot as plt
import tensorflow as tf

from research.ablation_studies.dataset.dataset_utils import prepare_dataset, sequence_image_feature_description, \
    write_tfrecord_with_examples, image_to_sequence_tfexample, get_max_of_heatmap
from research.ablation_studies.model.create_keypoint_models import preprocess_resnetv2_input
from research.ablation_studies.sequence.create_raw_video_sequence_tfrecords import gather_video_files
from research.ablation_studies.sequence.video_streamer import VideoFrameStreamer
from research.ablation_studies.train.keypoint_tf import KeypointHelper
from research.ablation_studies.util.image_patch_extract import ImagePatchExtractor
from research.ablation_studies.util.sequence_annotate import create_model_from_weights


class VideoFrameProvider:
    def __init__(self, subject, task):
        self.video_dir = "/media/florin/Transcend/large_dataset_study/cut_videos/"
        self.video_files = gather_video_files(self.video_dir)
        self.video_streamer = self._get_subject_task_video(self.video_files, subject, task)
        self.current_frame = 0
        return

    def _split_video_name(self, file):
        filename = os.path.split(file)[1][:-4]
        split_fn = filename.split('_')
        subject = split_fn[0]
        task = split_fn[1]
        return int(subject), int(task)

    def _get_subject_task_video(self, video_files, subject, task):
        found_file = None
        for file in video_files:
            file_subject, file_task = self._split_video_name(file)
            if file_subject == subject and file_task == task:
                found_file = file
                break

        if found_file is not None:
            streamer = VideoFrameStreamer(os.path.join(self.video_dir, found_file))
        else:
            streamer = None
            print("Error getting task video of {}_{}".format(subject, task))
        return streamer

    def seek(self, index):
        if self.video_streamer is not None:
            frame = self.video_streamer.seek_to_and_get(index)
            self.current_frame = index
        else:
            frame = None
        return frame

    def next(self):
        if self.video_streamer is not None:
            frame = self.video_streamer.get()
            self.current_frame += 1
        else:
            frame = None
        return frame

    def prev(self):
        if self.video_streamer is not None:
            frame = self.seek(self.current_frame - 1)
        else:
            frame = None
        return frame

    def get_current_frame(self):
        return self.current_frame


class TFRecordAnnotator:
    def __init__(self, tfrecord_files, feature_description, patch_shape, detail_weights, auto_anno=True,
                 show_video_frames=False):
        self.files = tfrecord_files
        self.feature_description = feature_description
        self.patch_shape = patch_shape
        self.current_dataset = None
        self.current_subject = None
        self.current_task = None
        self.current_frame = None
        self.current_patch_get_keypoint = None

        self.auto_anno = auto_anno
        self.detail_model = create_model_from_weights("Detail", detail_weights)

        self.current_file = self.files[0]
        self.current_iter = self.load_current_dataset()
        self.current_example = self.load_current_example()

        self.show_video_frames = show_video_frames

        if show_video_frames:
            self.frame_getter = VideoFrameProvider(
                int(self.current_example['image/subject_id']),
                int(self.current_example['image/task_id']))
            self.frame_fig, self.frame_ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 6))
            self.frame_getter.seek(int(self.current_example['image/frame_id']))

        self.keypoint_helper = KeypointHelper(row_major=False)
        self.patch_extractor = ImagePatchExtractor(patch_shape=patch_shape,
                                                   use_random_offset=False)

        self.fig, self.axes = plt.subplots(nrows=1, ncols=2, squeeze=False)
        self.reset_img_axis()
        self.reset_patch_axis()
        self.fig.canvas.mpl_connect('button_press_event', self.on_click)
        self.fig.canvas.mpl_connect('key_press_event', self.on_keypress)

        self.info_ax = plt.axes([0.25, 0.01, 0.5, 0.02])
        self.info_ax.axis('off')

        self.last_patch_pos = (self.current_example['image/ptip_x'], self.current_example['image/ptip_y'])
        self.examples_for_storage = []
        self.example_changed = False
        self.session_suffix = int(round(time.time() * 1000))
        self.update()
        self.update_image_info()
        return

    def on_keypress(self, event):
        if event.key in ['Q', 'q']:
            self.prev()
        if event.key in ['W', 'w']:
            self.next()
        if event.key in [' ']:
            self.toggle_down_annotation()
        if event.key in ['A', 'a']:
            self.frame_prev()
        if event.key in ['D', 'd']:
            self.frame_next()
        if event.key in ['F', 'f']:
            self.frame_seek()
        if event.key in ['T', 't']:
            self.store_example()
        if event.key in ['R', 'r']:
            self.revert()
        return

    def revert(self):
        self.example_changed = False
        self.update_image_info()
        return

    def store_example(self):
        example = image_to_sequence_tfexample(self.current_example['image/encoded'],
                                              b'jpg',
                                              self.current_example['image/subject_id'],
                                              self.current_example['image/frame_id'],
                                              self.current_example['image/task_id'],
                                              self.current_example['image/sequence_id'],
                                              self.current_example['image/height'],
                                              self.current_example['image/width'],
                                              self.current_example['image/ptip_x'],
                                              self.current_example['image/ptip_y'],
                                              self.current_example['image/ptail_x'],
                                              self.current_example['image/ptail_y'],
                                              self.current_example['image/pen_down'])  # pen_down
        self.examples_for_storage.append(example)
        tfrecord_filename = "tf_annotator_s{}_t{}_sequences_{}.tfrecord".format(self.current_subject, self.current_task,
                                                                                self.session_suffix)
        print("Writing tfrecord {}".format(tfrecord_filename))
        write_tfrecord_with_examples(tfrecord_filename, self.examples_for_storage)

    def toggle_down_annotation(self):
        current_anno = self.current_example['image/pen_down']
        if current_anno == 1:
            self.current_example['image/pen_down'] = 0
        elif current_anno == 0:
            self.current_example['image/pen_down'] = 1

        self.example_changed = True
        self.update_image_info()
        return

    def frame_prev(self):
        if self.show_video_frames:
            frame = self.frame_getter.prev()
            if frame is not None:
                self.frame_ax.imshow(frame)
        return

    def frame_seek(self):
        if self.show_video_frames:
            frame = self.frame_getter.seek(int(self.current_example['image/frame_id']))
            if frame is not None:
                self.frame_ax.imshow(frame)
        return

    def frame_next(self):
        if self.show_video_frames:
            frame = self.frame_getter.next()
            if frame is not None:
                self.frame_ax.imshow(frame)
        return

    def prev(self):
        return

    def next(self):
        if self.example_changed:
            self.store_example()

        self.current_example = self.load_current_example()

        self.example_changed = False
        self.update()
        self.update_image_info()

    def _convert_ax_pos_to_img_pos(self, ax_pos, img_width, img_height):
        img_x = ax_pos[0] * img_width
        img_y = (1.0 - ax_pos[1]) * img_height
        return (img_x, img_y)

    def _set_new_keypoint(self, patch_position, img_ax_pos):
        x = tf.cast(patch_position[0], tf.float32)
        y = tf.cast(patch_position[1], tf.float32)
        crop_window = self.patch_extractor._get_crop_window(
            x, y,
            int(self.current_example['image/width']),
            int(self.current_example['image/height']))
        new_x = crop_window[1] + int(img_ax_pos[0])
        new_y = crop_window[0] + int(img_ax_pos[1])
        if new_x < int(self.current_example['image/width']) and new_y < int(self.current_example['image/height']):
            self.current_example['image/ptip_x'] = new_x
            self.current_example['image/ptip_y'] = new_y
            self.example_changed = True
        return

    def on_click(self, event):
        if self.axes[0][0].in_axes(event):
            # click on large image
            ax_pos = self.axes[0][0].transAxes.inverted().transform((event.x, event.y))
            img_ax_pos = self._convert_ax_pos_to_img_pos(ax_pos,
                                                         int(self.current_example['image/width']),
                                                         int(self.current_example['image/height']))
            self.axes[0][0].plot(img_ax_pos[0] / 4, img_ax_pos[1] / 4, 'gx')
            self.update_patch(img_ax_pos[0], img_ax_pos[1])
            self.last_patch_pos = img_ax_pos
        if self.axes[0][1].in_axes(event):
            # click on patch
            ax_pos = self.axes[0][1].transAxes.inverted().transform((event.x, event.y))
            img_ax_pos = self._convert_ax_pos_to_img_pos(ax_pos,
                                                         self.patch_shape[1],
                                                         self.patch_shape[0])
            self.axes[0][1].plot(img_ax_pos[0], img_ax_pos[1], 'bx')
            self._set_new_keypoint(self.last_patch_pos, img_ax_pos)
            self.update_image_info()
        return

    def load_current_dataset(self):
        self.current_dataset = prepare_dataset(self.current_file, self.feature_description)
        self.current_iter = iter(self.current_dataset)
        return self.current_iter

    def get_patch(self, example, keypoint):
        image_bytes = example['image/encoded']
        img_width = example['image/width']
        img_height = example['image/height']
        patch, patch_keypoint = self.patch_extractor.get_img_patch(image_bytes, keypoint, img_width, img_height,
                                                                   self.keypoint_helper)
        return patch, patch_keypoint

    def get_image(self, example):
        image_bytes = example['image/encoded']
        img_width = tf.cast(example['image/width'], tf.float32)
        img_height = tf.cast(example['image/height'], tf.float32)
        tip_keypoint = self.keypoint_helper.from_list([example['image/ptip_x'], example['image/ptip_y']])
        img = tf.io.decode_jpeg(image_bytes, ratio=4, channels=3)
        loaded_img_height, loaded_img_width = img.shape[:2]
        tip_keypoint = self.keypoint_helper.rescale(tip_keypoint, img_width, img_height, loaded_img_width,
                                                    loaded_img_height)
        return img, tip_keypoint

    def load_current_example(self):
        self.current_example = next(self.current_iter)
        self.current_subject = self.current_example['image/subject_id']
        self.current_task = self.current_example['image/task_id']
        self.current_frame = self.current_example['image/frame_id']
        # self.current_example['image/pen_down'] = 0
        # self.example_changed = True
        return self.current_example

    def reset_img_axis(self):
        self.axes[0][0].clear()
        self.axes[0][0].set_xlim(0, int(self.current_example['image/width']) / 4)
        self.axes[0][0].set_ylim(int(self.current_example['image/height']) / 4, 0)

    def reset_patch_axis(self):
        self.axes[0][1].clear()
        self.axes[0][1].set_xlim(0, self.patch_shape[1])
        self.axes[0][1].set_ylim(self.patch_shape[0], 0)

    def update_image(self):
        image, tip_keypoint = self.get_image(self.current_example)
        self.reset_img_axis()
        self.axes[0][0].imshow(image)
        self.axes[0][0].plot([self.keypoint_helper.get_x(tip_keypoint)],
                             [self.keypoint_helper.get_y(tip_keypoint)],
                             'rx')
        plt.show()
        return

    def get_patch_inference(self, patch):
        my_patch = tf.image.convert_image_dtype(patch, tf.float32)
        my_patch = preprocess_resnetv2_input(my_patch)
        detail_pred_tip = tf.squeeze(self.detail_model(tf.expand_dims(my_patch, axis=0)))
        detail_max_value, detail_indices = get_max_of_heatmap(detail_pred_tip[:, :])
        detail_keypoint = self.keypoint_helper.make(x=detail_indices[1], y=detail_indices[0])
        detail_keypoint = self.keypoint_helper.rescale(detail_keypoint, detail_pred_tip.shape[1],
                                                       detail_pred_tip.shape[0], self.patch_shape[1],
                                                       self.patch_shape[0])
        return detail_keypoint

    def update_patch(self, x, y):
        location = self.keypoint_helper.make(x=x, y=y)
        patch, patch_keypoint = self.get_patch(self.current_example, location)

        self.reset_patch_axis()
        self.axes[0][1].imshow(patch)

        if not self.keypoint_helper.is_none_keypoint(location):
            detail_keypoint = self.get_patch_inference(patch)
            if self.auto_anno:
                self._set_new_keypoint(self.keypoint_helper.to_list(location),
                                       (self.keypoint_helper.get_x(detail_keypoint),
                                        self.keypoint_helper.get_y(detail_keypoint)))

            self.axes[0][1].plot([self.keypoint_helper.get_x(detail_keypoint)],
                                 [self.keypoint_helper.get_y(detail_keypoint)],
                                 'gx')

            self.axes[0][1].plot([self.keypoint_helper.get_x(patch_keypoint)],
                                 [self.keypoint_helper.get_y(patch_keypoint)],
                                 'rx')
        else:
            self.axes[0][1].plot([0, self.patch_shape[1]], [0, 0], 'b')

        plt.show()
        return

    def update(self):
        self.update_image()
        if self.current_example['image/ptip_x'] == -1.0 and \
                self.current_example['image/ptip_y'] == -1.0:
            self.update_patch(self.last_patch_pos[0], self.last_patch_pos[1])
        else:
            self.update_patch(self.current_example['image/ptip_x'], self.current_example['image/ptip_y'])
            self.last_patch_pos = (self.current_example['image/ptip_x'], self.current_example['image/ptip_y'])
        return

    def update_image_info(self):
        self.current_subject = self.current_example['image/subject_id']
        self.current_task = self.current_example['image/task_id']
        self.current_frame = self.current_example['image/frame_id']

        self.info_ax.clear()
        self.info_ax.axis('off')
        self.info_ax.text(0, 8.5, "{}_{}_{}".format(
            self.current_subject, self.current_task, self.current_frame), fontsize=14)

        if self.example_changed:
            self.info_ax.text(-0.05, 8.5, "*", fontsize=18, color="red")

        self.info_ax.text(0.23, 8.5, "[ {} | {} ]".format(
            int(self.current_example['image/ptip_x']),
            int(self.current_example['image/ptip_y'])), fontsize=14)

        if self.current_example['image/pen_down'] == 1:
            updown_text = "DOWN"
            updown_color = "red"
        else:
            updown_text = "UP"
            updown_color = "blue"

        self.info_ax.text(0.4, 8.5, updown_text, fontsize=20, fontweight='bold', color=updown_color)
        return


if __name__ == '__main__':
    # tf_record_file = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/sequence/large_dataset_s35_t2_sequences.tfrecord"
    tf_record_file = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/sequence/val1_s7_t0_sequences.tfrecord"
    # tf_record_file = "/home/florin/data/sequence/todo_s26_t4_sequences.tfrecord"
    # tf_record_file = "/home/florin/glowing-keypoint-firefly-dep/research/ablation_studies/vis/tf_annotator_s32_t0_sequences_1603370124901.tfrecord"
    # tf_record_file = "/home/florin/data/sequence/merged_s16_t0_sequences.tfrecord"

    annotator = TFRecordAnnotator([tf_record_file], sequence_image_feature_description,
                                  patch_shape=(224, 224),
                                  detail_weights="/home/florin/diss_models/best_weights/resnetv2_detail_u4_ft_ra9_best.196.h5",
                                  auto_anno=True)
