import numpy as np

from applications.brokers.inference_broker import InferenceBroker
from applications.network_utils import receive_image
from dataset.dataset_utils import is_none_keypoint, get_ndarray_diagnostics
from imgproc.heatmap_image_binarizer import HeatmapImageBinarizer
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction
from imgproc.imgproc_utils import resize_image_array_pil
from imgproc.patch_extractor import PatchExtractor
from util.timer import log_time


class KeypointBinarizationTipletBroker(InferenceBroker):

    def __init__(self, patch_size=32, validate=True, **kwargs):
        super().__init__(**kwargs)
        self.validate = validate
        self.patch_size = patch_size

    @log_time
    def _run_inner_loop(self):
        loop_ok = True

        msg, image = receive_image(self._get_open_port_socket(), comp_method='pil')
        self.logger.debug("Received image from {}".format(msg))
        self.logger.debug(get_ndarray_diagnostics(image))

        image_resized = resize_image_array_pil(image, (710, 400))
        self.logger.debug(get_ndarray_diagnostics(image_resized))

        self.distributor_hub.distribute_data(msg, image_resized)
        inferences_dict = self.distributor_hub.collect_replies()
        self.logger.debug(get_ndarray_diagnostics(inferences_dict["keypoint"]))
        self.logger.debug(get_ndarray_diagnostics(inferences_dict["binarization"]))

        tip, tail = self._extract_tip_and_tail(inferences_dict["keypoint"])
        fine_tip = self._get_fine_tip(image, tip)

        # binarized_full_size_image = self._binarize_image(image, inferences_dict["binarization"])
        # self.tip_path.add_sample(keypoint=fine_tip)

        self._get_open_port_socket().send_data(np.array(fine_tip), "fine_keypoint", compressed=False)

        return loop_ok

    def _extract_heatmap_keypoint(self, heatmap, radius):
        if self.validate:
            print("MIN {}".format(np.min(heatmap)))
            HeatmapKeypointExtractor.validate_input(heatmap, radius)
        keypoint = HeatmapKeypointExtractor(heatmap, MaxFilterPeakExtraction).run()
        self.logger.debug("Heatmap Tuple: {}".format(keypoint))
        return keypoint

    def _get_fine_tip(self, full_size_image, tip):
        tiplet = self._get_tiplet_from_camera_image(tip, full_size_image)
        if not is_none_keypoint(tip):
            tiplet_socket = self._get_inference_sockets_seq()[0]
            tiplet_socket.send_data(tiplet, msg='tiplet_infer', copy=False, compressed=False)
            msg, tiplet_inference = tiplet_socket.recv_data()

            assert (isinstance(tiplet_inference, np.ndarray))
            self.logger.debug(get_ndarray_diagnostics(tiplet_inference))

            tiplet_shape = tiplet_inference.shape
            tiplet_inference = np.reshape(tiplet_inference, (tiplet_shape[0], tiplet_shape[1]))
            fine_tip = self._extract_heatmap_keypoint(tiplet_inference, 3)
            return fine_tip  # row-major
        else:
            return tip

    def _extract_tip_and_tail(self, merged_keypoint_inferences):
        width = merged_keypoint_inferences.shape[1]

        half_width = int(width / 2)
        tip_map = merged_keypoint_inferences[:, 0:half_width]
        tail_map = merged_keypoint_inferences[:, half_width:]

        tip = self._extract_heatmap_keypoint(tip_map, 5)
        tail = self._extract_heatmap_keypoint(tail_map, 5)
        return tip, tail

    def _get_tiplet_from_camera_image(self, tip, full_size_image):
        patch, new_coord = PatchExtractor(full_size_image, (self.patch_size, self.patch_size), tip, do_pad=True).run()
        return patch

    def _binarize_image(self, image, heatmap):
        binary_image = HeatmapImageBinarizer(heatmap, image).run()
        return binary_image


if __name__ == "__main__":
    inference_servers = ['tcp://127.0.0.1:60000',
                         'tcp://127.0.0.1:60001']

    server_bin_key_tiplet = KeypointBinarizationTipletBroker(
        patch_size=32,
        validate=False,
        open_ports=['tcp://*:19858'],
        inference_servers_parallel=inference_servers,
        inference_servers_seq=['tcp://127.0.0.1:60003'],
        name="kbt_broker",
        do_logging=True)

    server_bin_key_tiplet.run()
