import cv2
import matplotlib

from applications.network_utils import send_image

matplotlib.use('TkAgg')

from imgproc.patch_extractor import PatchExtractor
from prototype.camera_streamer import BRIOCameraStreamer
from applications.clients.inference_client import InferenceClient
from util.timer import log_time
from dataset.dataset_utils import get_ndarray_diagnostics, rescale_keypoint
from imgproc.imgproc_utils import resize_image_array_pil


class KeypointBinarizationTipletClientPatchClient(InferenceClient):
    def __init__(self, stream_provider,
                 server_address,
                 compressed_send=False,
                 no_decoding=False,
                 do_logging=False,
                 display=False,
                 store_artifacts=False):
        super().__init__(server_address, compressed_send, no_decoding=no_decoding, do_logging=do_logging)
        self.stream_provider = stream_provider
        self.display = display
        self.store_artifacts = store_artifacts

    @log_time
    def _acquire_image(self):
        image = self.stream_provider.get()
        return image

    def _display_image(self, window_name, image):
        cv2.imshow(window_name, image)

    def _split_message(self, msg):
        list_inferences = msg.split('_')
        list_inferences = [element for element in list_inferences if element]
        inference_list = []
        for element in list_inferences:
            parts = element.split(':')
            type = parts[0]
            dims = (int(parts[1]), int(parts[2]))
            inference_list.append((type, dims))
        return inference_list

    def _split_inference(self, inference_list, inferences):
        inference_dict = {}
        width_counter = 0
        for elem in inference_list:
            h, w = elem[1]
            current_inference = inferences[0:h, width_counter:width_counter + w]
            inference_dict[elem[0]] = current_inference
            width_counter += w
        return inference_dict

    # in: row-major tip
    def _get_tiplet_from_camera_image(self, tip, full_size_image):
        # full_size_tip = rescale_keypoint(tip, 396, 236, full_size_image.shape[1], full_size_image.shape[0])
        full_size_tip = rescale_keypoint(tip, 236, 396, full_size_image.shape[0], full_size_image.shape[1])
        patch, new_coord = PatchExtractor(full_size_image, (32, 32, 3), full_size_tip, do_pad=True).run()
        return patch, full_size_tip, new_coord

    def _handle_tiplet_request(self, full_size_image):
        msg, tip = self.client_socket.recv_data()
        assert (msg == "tip_keypoint")

        patch, full_size_tip, patch_rough_tip = self._get_tiplet_from_camera_image(tip, full_size_image)
        self.logger.debug(get_ndarray_diagnostics(patch))
        # patch = cv2.cvtColor(patch, cv2.COLOR_BGR2RGB)

        self.client_socket.send_data(patch, "fine_patch", copy=False, compressed=False)
        return patch, full_size_tip, patch_rough_tip

    def _draw_marker_on_image(self, image, coord, mean, tip=True):
        if tip:
            cv2.circle(image, (coord[1], coord[0]), 1, (255, 255, 255 - mean * 20), thickness=2)
        else:
            cv2.circle(image, (coord[1], coord[0]), 1, (mean * 20, 255, 255), thickness=2)
        return image

    def _run_inner_loop(self):
        image = self._acquire_image()

        self.logger.debug(get_ndarray_diagnostics(image))

        if image is None:
            self.logger.error("Couldn't acquire image. Exiting.")
            return False

        resized_image = resize_image_array_pil(image, (710, 400))

        send_image(self.client_socket,
                   resized_image,
                   message="client_image",
                   compressed_send=self.compressed_send)

        patch, full_size_tip, patch_rough_tip = self._handle_tiplet_request(image)
        self.logger.debug("full_size_tip: {}".format(full_size_tip))
        self.logger.debug(get_ndarray_diagnostics(patch))

        msg, fine_tip = self.client_socket.recv_data()  # row-major

        self.logger.info("received inference from {}".format(msg))
        self.logger.debug(fine_tip)

        if self.display:
            self._draw_marker_on_image(patch, fine_tip, 0)
            self._draw_marker_on_image(patch, patch_rough_tip, 0, tip=False)
            self._display_image("camera", cv2.cvtColor(image, cv2.COLOR_RGB2BGR))
            self._display_image("tiplet", cv2.cvtColor(patch, cv2.COLOR_RGB2BGR))
            # imgplot = plt.imshow(patch)
            # plt.show()
            cv2.waitKey(1)
        return True


if __name__ == "__main__":
    no_decoding = False
    camera_stream = BRIOCameraStreamer(no_decoding=no_decoding)
    # camera_stream = DummyCameraStreamer("L:\\large_dataset_study\\first_round_subset\\test\\")

    client = KeypointBinarizationTipletClientPatchClient(camera_stream,
                                                         "tcp://schlaubox.ur.de:19858",
                                                         compressed_send=True,
                                                         display=True,
                                                         no_decoding=no_decoding,
                                                         do_logging=True,
                                                         store_artifacts=False)
    client.run()
