import cv2
from skimage.io import imsave

from applications.clients.inference_client import InferenceClient
from applications.network_utils import send_image
from dataset.dataset_utils import get_ndarray_diagnostics, rescale_keypoint
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction
from imgproc.imgproc_utils import resize_image_array_pil
from imgproc.patch_extractor import PatchExtractor
from prototype.camera_streamer import DummyCameraStreamer
from util.timer import log_time


class TipletInferenceClient(InferenceClient):
    def __init__(self, frame_provider,
                 heatmap_provider,
                 server_address,
                 store_inferences=False,
                 compressed_send=False,
                 no_decoding=False,
                 do_logging=False,
                 display=False):
        super().__init__(server_address, compressed_send, no_decoding=no_decoding, do_logging=do_logging)
        self.frame_provider = frame_provider
        self.heatmap_provider = heatmap_provider
        self.display = display
        self.frame_counter = 0
        self.store_inferences = store_inferences

    @log_time
    def _acquire_image_and_heatmap(self):
        image = self.frame_provider.get()
        heatmap = self.heatmap_provider.get()
        return image, heatmap

    def _display_image(self, window_name, image):
        cv2.imshow(window_name, image)

    def _split_inference(self, inference_list, inferences):
        inference_dict = {}
        width_counter = 0
        for elem in inference_list:
            h, w = elem[1]
            current_inference = inferences[0:h, width_counter:width_counter + w]
            inference_dict[elem[0]] = current_inference
            width_counter += w
        return inference_dict

    def _extract_heatmap_keypoint(self, heatmap, radius):
        keypoint = HeatmapKeypointExtractor(heatmap, MaxFilterPeakExtraction).run()
        self.logger.debug("Heatmap Tuple: {}".format(keypoint))
        return keypoint

    # in: row-major tip
    def _get_tiplet_from_camera_image(self, tip, full_size_image):
        # full_size_tip = rescale_keypoint(tip, 396, 236, full_size_image.shape[1], full_size_image.shape[0])
        full_size_tip = rescale_keypoint(tip, 236, 396, full_size_image.shape[0], full_size_image.shape[1])
        patch, new_coord = PatchExtractor(full_size_image, (48, 48, 3), full_size_tip, do_pad=True).run()
        patch = resize_image_array_pil(patch, (32, 32))
        return patch, full_size_tip, new_coord

    def _extract_tip_and_tail(self, merged_keypoint_inferences):
        width = merged_keypoint_inferences.shape[1]

        half_width = int(width / 2)
        tip_map = merged_keypoint_inferences[:, 0:half_width]
        tail_map = merged_keypoint_inferences[:, half_width:]

        tip = self._extract_heatmap_keypoint(tip_map, 5)
        tail = self._extract_heatmap_keypoint(tail_map, 5)
        return tip, tail

    def _run_inner_loop(self):
        image, merged_heatmap = self._acquire_image_and_heatmap()
        self.logger.debug(get_ndarray_diagnostics(image))
        self.logger.debug(get_ndarray_diagnostics(merged_heatmap))

        if image is None:
            self.logger.error("Couldn't acquire image. Exiting.")
            return False

        tip, tail = self._extract_tip_and_tail(merged_heatmap)
        resized_image = resize_image_array_pil(image, (1920, 1080))
        patch, rescaled_tip, coord_in_patch = self._get_tiplet_from_camera_image(tip, resized_image)

        send_image(self.client_socket,
                   patch,
                   message="client_image",
                   compressed_send=self.compressed_send)

        msg, inferences = self.client_socket.recv_data()
        self.logger.info("received inference from {}".format(msg))

        if self.store_inferences:
            imsave("../tmp_out/{}_tiplet_inference48.jpg".format(self.frame_counter), inferences)
            imsave("../tmp_out/{}_tiplet_patch48.jpg".format(self.frame_counter), patch)

        if self.display:
            # self._draw_marker_on_image(converted_image, fine_tip, 0)
            # self._display_image("camera", np.array(resized_image))
            # self._display_image("inferences", inferences)

            cv2.waitKey(1)
        self.frame_counter += 1
        return True


if __name__ == "__main__":
    no_decoding = False
    # camera_stream = BRIOCameraStreamer(no_decoding=no_decoding)
    camera_stream = DummyCameraStreamer("L:\\video_analysis\\6_0\\frames\\")
    heatmap_stream = DummyCameraStreamer("L:\\video_analysis\\6_0\\keypoints\\")
    # camera_stream = VideoFrameStreamer("L:\\large_dataset_study\\cut_videos\\6_0.mp4")

    client = TipletInferenceClient(camera_stream,
                                   heatmap_stream,
                                   "tcp://schlaubox.ur.de:19858",
                                   store_inferences=True,
                                   compressed_send=True,
                                   display=False,
                                   no_decoding=no_decoding,
                                   do_logging=True)
    client.run()
