import matplotlib
import numpy as np
import pandas as pd
import zmq
from skimage.exposure import rescale_intensity
from skimage.io import imsave

from applications.logged_object import LoggedObject
from dataset.dataset_utils import get_ndarray_diagnostics, rescale_keypoint, \
    get_frame_id_patch_location_from_image_name, make_none_keypoint, is_none_keypoint
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction
from imgproc.imgproc_utils import decompress_image_pil, demean_heatmap
from util.network.bidirectional_imagezmq import ZMQHubFactory
from util.timer import Timer, log_time

matplotlib.use('TkAgg')


class AccuracyTestServer(LoggedObject):
    def __init__(self,
                 patch_publisher,
                 patch_inference_publisher,
                 patch_size=(224, 224),
                 heatmap_size=(211, 211),
                 surface_size=(1080, 1920),
                 do_logging=False,
                 record=False,
                 end_on_id=1000,
                 subject_id=0):
        super().__init__(do_logging=do_logging)

        self.patch_publisher = patch_publisher
        self.patch_inference_publisher = patch_inference_publisher

        self.surface_size = surface_size

        self.current_keypoint_hm = np.zeros((224, 224), dtype='uint8')
        self.current_bin_hm = np.zeros((224, 224), dtype='uint8')
        self.current_patch = np.zeros((224, 224, 3), dtype='uint8')
        self.current_extract_image = np.zeros((224, 224), dtype='uint8')

        self.subject_id = subject_id

        self.patch_size = patch_size
        self.heatmap_size = heatmap_size

        self.end_on_id = end_on_id

        self.inferences = {}
        self.patches = {}
        self.current_work_pair = 0
        self.raw_peaks = {}
        self.on_points = np.array([])
        self.all_filter_points = np.array([])
        self.last_all_filter_points_length = 0
        self.last_on_points_length = 0
        self.last_stroke_count = 0
        self.strokes = np.array([])

        self.patch_sub_socket = ZMQHubFactory.get_SUB(self.patch_publisher)
        self.inference_sub_socket = ZMQHubFactory.get_SUB(self.patch_inference_publisher)

        self.patch_sub_socket.setsockopt(zmq.SUBSCRIBE, b"")
        self.inference_sub_socket.setsockopt(zmq.SUBSCRIBE, b"")

        self.frame_count = 0
        self.interval_counter = 0
        self.all_points_line_collection = None
        self.stroke_line_collection = None
        self.on_points_collection = None
        self.record = record
        return

    def _receive_patch_and_location(self):
        msg, data = self.patch_sub_socket.recv_data()
        patch = decompress_image_pil(data)
        frame_id, location = get_frame_id_patch_location_from_image_name(msg)
        split_msg = msg.split('_')
        frame_id = int(split_msg[0])
        location = (int(float(split_msg[1])), int(float(split_msg[2])))
        return frame_id, patch, location

    def _receive_inferences(self):
        msg, data = self.inference_sub_socket.recv_data(copy=True)
        self.logger.debug("Received subscription with msg: {}".format(msg))
        split_msg = msg.split('_')
        frame_id = int(split_msg[0])
        keypoint_len = int(split_msg[1])

        keypoint_hm = decompress_image_pil(data[:keypoint_len])
        bin_hm = decompress_image_pil(data[keypoint_len:])

        self.logger.debug("Keypoint HM {}".format(get_ndarray_diagnostics(keypoint_hm)))
        self.logger.debug("Bin HM {}".format(get_ndarray_diagnostics(bin_hm)))
        return frame_id, keypoint_hm, bin_hm

    def run(self):
        frame_timer = Timer(verbose=False)

        loop_ok = True
        while loop_ok:
            self.logger.info("Frame {}".format(self.frame_count))
            frame_timer.start("")

            loop_ok = self._run_inner_loop()

            frame_time = frame_timer.stop()
            self.logger.info("Frame time: {}".format(frame_time))

            if self.frame_count >= self.end_on_id - 10:
                save_frame = pd.DataFrame({
                    'frame': list(self.raw_peaks.keys()),
                    'ptip': list(self.raw_peaks.values())
                })
                print("saving csv!")
                save_frame.to_csv("testset_accuracy_final_{}.csv".format(self.subject_id))
            self.frame_count += 1

        return

    @log_time
    def _process_binarization_heatmap(self, bin_hm):
        filter_bin_hm = demean_heatmap(bin_hm.astype('float32'))
        filter_bin_hm = filter_bin_hm / np.std(filter_bin_hm)
        mask_3sigma = np.where(filter_bin_hm < 3.0)
        filter_bin_hm[mask_3sigma] = 0
        return rescale_intensity(filter_bin_hm, in_range='image', out_range='uint8')

    @log_time
    def _run_inner_loop(self):
        # get subscriptions
        frame_id_inf, new_keypoint_heatmap, new_bin_heatmap = self._receive_inferences()
        frame_id_patch, patch, location = self._receive_patch_and_location()

        self.logger.debug("Frame ID PATCH: {}".format(frame_id_patch))
        self.logger.debug("Frame ID INFER: {}".format(frame_id_inf))

        # store them in the frame registry
        self.inferences[frame_id_inf] = [new_keypoint_heatmap, new_bin_heatmap]
        self.patches[frame_id_patch] = [patch, location]

        # check if there is a match in the frame ids
        patch_ids = self.patches.keys()
        inf_ids = self.inferences.keys()
        matches = list(set(patch_ids).intersection(inf_ids))

        # todo check if it is possible that the registry gets larger and never processed
        if len(matches) > 0:
            # we have a match
            matches.sort()
            self.process_working_pair(matches[0])
            del self.inferences[matches[0]]
            del self.patches[matches[0]]
        else:
            return True

        return True

    # wants row major peak and patch location
    # gives column major peak
    def _transform_patch_peak(self, peak, patch_location):
        left_patch_coord = int(patch_location[1] - self.patch_size[1] / 2)
        top_patch_coord = int(patch_location[0] - self.patch_size[0] / 2)

        # column major peak with patch transform
        peak_transformed = (peak[1] + left_patch_coord, peak[0] + top_patch_coord)
        return peak_transformed

    def _handle_recording(self, match_id, keypoint_heatmap, bin_heatmap, patch, location):
        if self.record:
            imsave("../tmp_out/record/testset_study/kp/{}_kp.jpg".format(match_id), keypoint_heatmap, quality=95)
            imsave("../tmp_out/record/testset_study/bin/{}_bin.jpg".format(match_id), bin_heatmap, quality=95)
            imsave(
                "../tmp_out/record/testset_study/patch/{}_patch_{}_{}.jpg".format(match_id, location[0], location[1]),
                patch, quality=95)
        return

    @log_time
    def process_working_pair(self, match_id):
        self.logger.debug("PROCESSING WORKING PAIR {}".format(match_id))

        keypoint_heatmap, bin_heatmap = self.inferences[match_id]
        patch, location = self.patches[match_id]

        if keypoint_heatmap is None or bin_heatmap is None or patch is None:
            print("DISCARD")
            return

        if self.record:
            self._handle_recording(match_id, keypoint_heatmap, bin_heatmap, patch, location)

        self.current_bin_hm = bin_heatmap
        self.current_keypoint_hm = keypoint_heatmap
        self.current_patch = patch

        # if np.mean(bin_heatmap) > 50 or np.std(bin_heatmap) < 7:
        #    self.raw_peaks[match_id] = make_none_keypoint()
        #    return

        # three_sigma_hm = self._process_binarization_heatmap(self.current_keypoint_hm)

        # imsave("../tmp_out/record/val2_study/{}/kp/{}_3s.jpg".format(self.subject_id, match_id), three_sigma_hm,
        #       quality=95)

        # DISCARD
        # if np.mean(keypoint_heatmap) > 60:
        #    self.raw_peaks[match_id] = make_none_keypoint()
        #    return

        # EXTRACT CURRENT PEAK
        peak = HeatmapKeypointExtractor(self.current_keypoint_hm, MaxFilterPeakExtraction).run()
        peak = rescale_keypoint(peak, 211, 211, 224, 224)

        if is_none_keypoint(peak):
            self.raw_peaks[match_id] = make_none_keypoint()
            return

        transformed_peak = self._transform_patch_peak(peak, location)
        if transformed_peak[0] >= 0 and transformed_peak[0] < self.surface_size[1]:
            if transformed_peak[1] >= 0 and transformed_peak[1] < self.surface_size[0]:
                self.raw_peaks[match_id] = transformed_peak
            else:
                self.raw_peaks[match_id] = make_none_keypoint()
                print("LOL")
        else:
            self.raw_peaks[match_id] = make_none_keypoint()
            print("LOLLOL")

        self.logger.debug("PROCESSED WORKING PAIR {}".format(match_id))
        return


if __name__ == '__main__':
    val1_subject_frames = {
        "0": 410,
        "1": 192,
        "2": 235,
        "3": 208,
        "4": 182,
        "5": 172,
        "6": 223,
        "7": 285
    }

    val2_subject_frames = {
        "0": 246,
        "1": 219,
        "2": 227,
        "3": 203,
        "4": 213,
        "5": 242,
        "6": 248,
        "7": 286,
        "8": 209,
        "9": 203,
        "10": 205
    }

    test_set_end_id = 3773

    server = AccuracyTestServer(
        patch_publisher='tcp://132.199.132.34:19859',
        patch_inference_publisher='tcp://132.199.140.68:19859',
        do_logging=True,
        record=True,
        end_on_id=test_set_end_id,
        subject_id=23)
    server.run()
