import os
from ast import literal_eval
from math import sqrt

import numpy as np
import pandas as pd
from scipy import stats

from applications.final_prototype.val2_subjects import val2_subjects
from dataset_utils import is_none_keypoint, rescale_keypoint, make_none_keypoint
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction
from prototype.camera_streamer import DummyCameraStreamer


def get_index_of_min_distance_to_perfect_for_fpr_tpr(fpr_coords, tpr_coords):
    combined = np.array(list(zip(fpr_coords, tpr_coords)))
    top_left_point_multiples = np.ones((len(fpr_coords), 2)) * (0, 1)
    distances = np.sqrt(np.sum((top_left_point_multiples - combined) ** 2, axis=1))
    return np.argmin(distances)


class AnnotationsHeatmapsDistancesSetAnalyzer():
    def __init__(self, annotations_path, heatmaps_path, detections_path, frame_step=1, filename_identifier="tmp"):
        self.annotations_path = annotations_path
        self.heatmaps_path = heatmaps_path
        self.detections_path = detections_path
        self.frame_step = frame_step
        self.filename_identifier = filename_identifier

        self.anno_df = self._load_annotations()
        self.detections_df = self._load_detections()

        self.statistics_df = self._create_mean_df(csv_prefix=self.filename_identifier, save=True)
        self.eval_df = self._create_statistics_and_gt_df(self.statistics_df, self.anno_df, frame_step=self.frame_step)
        return

    def get_annotations(self):
        return self.anno_df

    def get_statistics(self):
        return self.statistics_df

    def get_evaluation(self):
        return self.eval_df

    def _load_annotations(self):
        subject_annos = pd.read_csv(
            os.path.join(self.annotations_path),
            index_col=0,
            dtype={'frame': np.int32},
            converters={
                'ptip': literal_eval,
                'ptail': literal_eval})
        return subject_annos

    def _load_heatmaps(self):

        return

    def _load_detections(self):
        detections = pd.read_csv(self.detections_path, index_col=0,
                                 dtype={'frame': np.int32},
                                 converters={
                                     'ptip': literal_eval})

        detections['frame'] *= self.frame_step
        return detections

    def recalculate_kp_peaks(self, patch_size, x_limit, y_limit):
        streamer = DummyCameraStreamer(os.path.join(self.heatmaps_path, "patch"))
        img = streamer.get()
        patch_dict = {}
        while img is not None:
            split_name = streamer.get_current_image_name().strip('.jpg').split('_')
            frame_id = split_name[0]
            current_patch_location_x = int(split_name[2])
            current_patch_location_y = int(split_name[3])
            patch_dict[frame_id] = (current_patch_location_x, current_patch_location_y)
            img = streamer.get()

        peak_dict = {"frame": [], "ptip": []}
        streamer_kp = DummyCameraStreamer(os.path.join(self.heatmaps_path, "kp"))
        img_kp = streamer_kp.get()
        while img_kp is not None:
            split_name = streamer_kp.get_current_image_name().split('_')
            frame_id = split_name[0]
            peak_dict["frame"].append(frame_id)

            peak = HeatmapKeypointExtractor(img_kp, MaxFilterPeakExtraction).run()
            peak = rescale_keypoint(peak, 211, 211, patch_size[0], patch_size[1])

            if not is_none_keypoint(peak):
                left_patch_coord = patch_dict[frame_id][1] - (patch_size[1] / 2)
                top_patch_coord = patch_dict[frame_id][0] - (patch_size[0] / 2)
                correct_peak_x = left_patch_coord + peak[1]
                correct_peak_y = top_patch_coord + peak[0]
                if (correct_peak_x < 0 or correct_peak_x >= x_limit) or \
                        (correct_peak_y < 0 or correct_peak_y >= y_limit):
                    peak_dict["ptip"].append(make_none_keypoint())
                else:
                    peak_dict["ptip"].append((correct_peak_x, correct_peak_y))
            else:
                peak_dict["ptip"].append(peak)
            img_kp = streamer_kp.get()

        return peak_dict

    def _create_mean_df(self, csv_prefix="tmp", save=False):
        streamer = DummyCameraStreamer(os.path.join(self.heatmaps_path, "kp"))

        img = streamer.get()

        means_dict = {
            "frame": [],
            "kp_mean": [],
            "kp_median": [],
            "kp_std": [],
            "kp_var": []
        }

        while img is not None:
            my_id = streamer.get_current_image_name().split('_')[0]
            means_dict["frame"].append(int(my_id))
            means_dict["kp_mean"].append(np.mean(img))
            means_dict["kp_median"].append(np.median(img))
            means_dict["kp_std"].append(np.std(img))
            means_dict["kp_var"].append(np.var(img))
            img = streamer.get()

        means_df = pd.DataFrame(means_dict)

        if save:
            means_df.to_csv("{}_kp_statistics.csv".format(csv_prefix))

        return means_df

    def _create_statistics_and_gt_df(self, statistics_df, anno_df, frame_step=1):
        statistics_df['frame'] *= frame_step
        # get all rows where we have frames in means df for
        means_frames = pd.Index(statistics_df['frame'])
        eval_annos = anno_df.loc[anno_df.frame.isin(means_frames)]

        eval_annos = eval_annos.reset_index()
        eval_annos = eval_annos.drop(columns=['ptail'])
        eval_annos = eval_annos.rename(columns={"frame": "frame_gt", "ptip": "ptip_gt"})
        eval_annos['ptip_gt_label'] = eval_annos.ptip_gt.apply(lambda x: not is_none_keypoint(x))
        eval_df = eval_annos.join(statistics_df)

        eval_frames = pd.Index(eval_df['frame'])
        eval_in_detection_df = self.detections_df.loc[self.detections_df.frame.isin(eval_frames)]

        eval_in_detection_df = eval_in_detection_df.reset_index()
        eval_in_detection_df = eval_in_detection_df.join(eval_df, lsuffix="bla")

        return eval_in_detection_df

    def get_scores_labels_for_set(self):
        scores_mean = 255 - self.eval_df['kp_mean'].to_numpy()
        scores_median = 255 - self.eval_df['kp_median'].to_numpy()
        scores_std = 255 - self.eval_df['kp_std'].to_numpy()
        scores_var = 255 - self.eval_df['kp_var'].to_numpy()
        labels = self.eval_df['ptip_gt_label'].apply(lambda x: int(x)).to_numpy()

        return scores_mean, scores_median, scores_std, scores_var, labels

    def _calc_distance(self, tip_calc, tip_gt, x_scale, y_scale, coord_transform_func):
        if is_none_keypoint(tip_calc):
            if is_none_keypoint(tip_gt):
                return 0  # this represents a true negative
            else:
                # this represents false negatives
                return -1
        if is_none_keypoint(tip_gt):
            if is_none_keypoint(tip_calc):
                return 0  # this represents a true negative
            else:
                # this represents false positives
                return -2

        new_gt_x, new_gt_y = coord_transform_func(tip_gt)

        dx = new_gt_x - tip_calc[0]
        dy = new_gt_y - tip_calc[1]

        dx /= x_scale
        dy /= y_scale

        distance = sqrt(dx * dx + dy * dy)
        return distance

    def _get_distances_for_set(self, x_scale, y_scale, coord_transform_func, distance_identifier="px"):
        distances_df = self.eval_df.copy(deep=True)
        distances_df['distance_{}'.format(distance_identifier)] = distances_df.apply(
            lambda x: self._calc_distance(x.ptip, x.ptip_gt, x_scale, y_scale, coord_transform_func), axis=1)
        return distances_df

    def get_distances_percentile_for_thresholds(self, error_point, thresholds, x_scale, y_scale, coord_transform_func,
                                                distance_identifier="px"):
        distances_df = self._get_distances_for_set(x_scale, y_scale, coord_transform_func, distance_identifier)

        percentiles = []
        all_distances = []
        for threshold in thresholds:
            detected_distances = distances_df.loc[(distances_df['distance_{}'.format(distance_identifier)] > 0)]
            detected_distances = detected_distances.loc[detected_distances['kp_mean'] <= threshold]
            # calc true positives
            current_tp = detected_distances.size

            # calc percentile for error point
            detected_points_with_distance = detected_distances['distance_{}'.format(distance_identifier)].to_numpy()
            if detected_points_with_distance is None:
                detected_points_with_distance = np.array([])

            percentile = stats.percentileofscore(detected_points_with_distance, error_point, kind="weak")
            percentiles.append(percentile)
            all_distances.append(detected_points_with_distance)

        return percentiles, all_distances


if __name__ == '__main__':
    i = 2
    anno_path = os.path.join(val2_subjects[i]["path"], "val2_study_annos_{}.csv".format(i))
    detections_path = os.path.join(val2_subjects[i]["path"], "val2_accuracy_final_{}.csv".format(i))
    heatmap_path = os.path.join(val2_subjects[i]["path"], "final")
    analyzer = AnnotationsHeatmapsDistancesSetAnalyzer(anno_path,
                                                       heatmap_path,
                                                       detections_path,
                                                       filename_identifier="val2_study_correction",
                                                       frame_step=val2_subjects[i]["step"])

    peak_dict = analyzer.recalculate_kp_peaks((126, 149), 1080, 720)
    new_2_df = pd.DataFrame(peak_dict)
    new_2_df.to_csv("val2_new_2_accuracy.csv")
