import logging
import time
from threading import Thread

import numpy as np
from skimage.io import imsave

from applications.logged_object import LoggedObject
from applications.network_utils import receive_image
from dataset.dataset_utils import is_none_keypoint
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction
from imgproc.imgproc_utils import resize_image_array_pil, compress_image_array_pil, demean_heatmap, \
    filter_three_sigma_events
from imgproc.patch_by_heatmap_extractor import PatchByHeatmapExtractor
from imgproc.path_extraction.path_distances_extractor import PathDistancesExtractor
from prototype.camera_streamer import VideoFrameStreamer
from util.network.bidirectional_imagezmq import ZMQHubFactory
from util.timer import log_time, Timer

video_num = 7


class LocalFasterBroker(LoggedObject):

    def __init__(self, stream_provider, roinet_server, pub_port, roinet_input_size=(224, 224), patch_size=(224, 224),
                 **kwargs):
        super().__init__(**kwargs)
        self.logger.setLevel(logging.DEBUG)

        self.stream_provider = stream_provider

        self.patch_size = patch_size
        self.roinet_input_size = roinet_input_size

        self.roinet_server = roinet_server
        self.roinet_socket = ZMQHubFactory.get_REQ(roinet_server)

        self.history_length = 20
        self.roi_coords = []
        self.frame_id = 0

        self.frame_count = 0
        self.video_mode = True
        self.record_all = True
        self.record_path = "/home/florin/stroke_video/" + str(video_num) + "/"

        # prepare and start the camera thread
        self.cam_image = self.stream_provider.get()
        self.cam_thread = Thread(target=self._cam_thread_func, args=())
        self.cam_thread.daemon = True
        if not self.video_mode:
            self.cam_thread.start()

        self.pub_socket = ZMQHubFactory.get_PUB(pub_port)

    @log_time
    def _is_roi_good(self, inference):
        demeaned_heatmap = demean_heatmap(inference)

        if np.max(demeaned_heatmap) == 0:
            self.logger.debug("Empty Heatmap")
            return False

        filtered_heatmap = filter_three_sigma_events(demeaned_heatmap)
        filtered_heatmap = np.nan_to_num(filtered_heatmap)

        peak = HeatmapKeypointExtractor(filtered_heatmap, MaxFilterPeakExtraction).run()
        if not is_none_keypoint(peak):
            self.roi_coords.append(peak)
        else:
            return False

        if len(self.roi_coords) < self.history_length:
            return True
        else:
            # calculate distances between existing points
            # if new distance is a factor larger, return false
            distances, points = PathDistancesExtractor(np.array(self.roi_coords), cumulative=False,
                                                       normalized=False).run()
            new_distance = distances[-1]
            factor = 5
            average_distance = np.mean(distances[:-1])
            if new_distance > average_distance * factor:
                self.logger.debug("Bad ROI: New point too far away")
                return False
            else:
                self.logger.debug("Good ROI")
                return True

    def _cam_thread_func(self):
        while True:
            self.cam_image = self.stream_provider.get()
            time.sleep(0.001)

    def _handle_recording(self, full_size_image, roinet_inference):
        imsave(self.record_path + "frames/{}_frame.jpg".format(self.frame_count), full_size_image, quality=95)
        imsave(self.record_path + "roinet/{}_roinet.jpg".format(self.frame_count), roinet_inference, quality=95)
        return

    @log_time
    def _recv_roinet_and_get_patch(self, full_size_image):
        msg, roinet_inference = receive_image(self.roinet_socket)

        assert (len(roinet_inference.shape) == 2)

        if self.record_all:
            self._handle_recording(full_size_image, roinet_inference)

        if self._is_roi_good(roinet_inference):
            patch, _, roi_location = PatchByHeatmapExtractor(full_size_image, roinet_inference, self.patch_size).run()
        else:
            patch = None
            roi_location = None

        return patch, roinet_inference, roi_location

    @log_time
    def _send_image_to_roinet(self, full_size_image):
        image_resized = resize_image_array_pil(full_size_image, (self.roinet_input_size[1], self.roinet_input_size[0]))
        self.logger.debug("Resized image.")

        self.logger.debug("Sending image to roinet: {}".format(self.roinet_socket))
        self.roinet_socket.send_data(image_resized, msg='roinet_inference', copy=False, compressed=False)
        return image_resized

    @log_time
    def _publish_patch(self, roi_patch, roi_location):
        compressed_patch = compress_image_array_pil(roi_patch, 95)
        msg_frameid_location = str(self.frame_id) + '_' + str(roi_location[0]) + '_' + str(roi_location[1])
        self.pub_socket.send_data(compressed_patch, msg=msg_frameid_location, copy=False, compressed=True)
        return

    def run(self):
        frame_timer = Timer(verbose=False)

        loop_ok = True
        while loop_ok:
            self.logger.info("Frame {}".format(self.frame_count))
            frame_timer.start("")

            loop_ok = self._run_inner_loop()

            frame_time = frame_timer.stop()
            self.logger.info("Frame time: {}".format(frame_time))
            self.frame_count += 1
        self.logger.warning("Loop halted.")
        return

    @log_time
    def _run_inner_loop(self):
        loop_ok = True

        if self.video_mode:
            self.cam_image = self.stream_provider.get()

        current_image = np.copy(self.cam_image)

        # we need to compare with to equal signs
        # because in case cam_image is None, we create
        # a ndarray(None, dtype=object) and not None
        if current_image[()] is None:
            return False

        # send image to roinet
        self._send_image_to_roinet(current_image)
        # get roi
        roi_patch, roinet_inference, roi_location = self._recv_roinet_and_get_patch(current_image)
        if roi_patch is not None:
            self._publish_patch(roi_patch, roi_location)
            self.frame_id += 1

        return loop_ok


if __name__ == "__main__":
    # camera_stream = BRIOCameraStreamer(no_decoding=False)
    camera_stream = VideoFrameStreamer("/home/florin/stroke_video/stroke_accuracy_study_{}.avi".format(video_num))

    patch_broker_server = LocalFasterBroker(
        camera_stream,
        roinet_server='tcp://127.0.0.1:60003',
        pub_port='tcp://*:19859',
        roinet_input_size=(224, 224),
        patch_size=(224, 224),
        do_logging=True)

    patch_broker_server.run()
