import numpy as np
import zmq
from skimage.exposure import rescale_intensity

from applications.logged_object import LoggedObject
from applications.network_utils import receive_image
from dataset.dataset_utils import get_ndarray_diagnostics
from imgproc.imgproc_utils import compress_image_array_pil
from util.network.bidirectional_imagezmq import ZMQHubFactory, WorkDistributorHub
from util.timer import log_time, Timer


class PatchPabloBroker(LoggedObject):

    def __init__(self, geva_broker, pub_port, parallel_inference_servers, patch_size=(224, 224), do_logging=True):
        super().__init__(do_logging=do_logging)
        self.patch_size = patch_size
        self.frame_id = 0

        self.distributor_hub = WorkDistributorHub(parallel_inference_servers,
                                                  compressed=False,
                                                  do_logging=self.do_logging)

        self.frame_count = 0
        self.pub_socket = ZMQHubFactory.get_PUB(pub_port)
        self.sub_socket = ZMQHubFactory.get_SUB(geva_broker)
        self.sub_socket.setsockopt(zmq.SUBSCRIBE, b"")

    @log_time
    def _receive_and_decompress_image(self):
        msg, image = receive_image(self.sub_socket, comp_method='pil')
        self.frame_id = int(msg.split('_')[0])
        self.logger.debug("recv patch id {}".format(msg))
        self.logger.debug(get_ndarray_diagnostics(image))
        return msg, image

    @log_time
    def _get_patch_inferences(self, roi_patch):
        self.distributor_hub.distribute_data("patch_distribute", roi_patch)
        inferences_dict = self.distributor_hub.collect_replies()
        self.logger.debug(get_ndarray_diagnostics(inferences_dict["keypoint"]))
        self.logger.debug(get_ndarray_diagnostics(inferences_dict["binarization"]))
        return inferences_dict

    @log_time
    def _compress_inferences(self, inferences_dict):
        inferences_dict["keypoint"] = self._rescale_heatmap(inferences_dict["keypoint"])  # * 255).astype('uint8')
        inferences_dict["binarization"] = self._rescale_heatmap(
            inferences_dict["binarization"])  # * 255).astype('uint8')
        keypoints_compressed = compress_image_array_pil(inferences_dict["keypoint"], 95)
        binarization_compressed = compress_image_array_pil(inferences_dict["binarization"], 95)
        return keypoints_compressed, binarization_compressed

    @log_time
    def _rescale_heatmap(self, heatmap):
        uint8_output = rescale_intensity(heatmap, in_range='image', out_range='uint8')
        return np.array(uint8_output, dtype='uint8')

    @log_time
    def _publish_results(self, keypoints_compressed, binarization_compressed):
        self.pub_socket.send_data(keypoints_compressed + binarization_compressed,
                                  "{}_{}".format(self.frame_id, len(keypoints_compressed)), compressed=True)
        return

    def run(self):
        frame_timer = Timer(verbose=False)

        loop_ok = True
        while loop_ok:
            self.logger.info("Frame {}".format(self.frame_count))
            frame_timer.start("")

            loop_ok = self._run_inner_loop()

            frame_time = frame_timer.stop()
            self.logger.info("Frame time: {}".format(frame_time))
            self.frame_count += 1
        self.logger.warning("Loop halted.")
        return

    @log_time
    def _run_inner_loop(self):
        loop_ok = True

        msg, patch = self._receive_and_decompress_image()

        inferences_dict = self._get_patch_inferences(patch)

        # hm_shape = inferences_dict["keypoint"].shape
        # inferences_dict["keypoint"] = inferences_dict["keypoint"][:, :, 0]# np.reshape(inferences_dict["keypoint"], (hm_shape[0], hm_shape[1]))
        # inferences_dict["binarization"] = inferences_dict["binarization"][:, :, 0]# np.reshape(inferences_dict["binarization"], (hm_shape[0], hm_shape[1]))

        # imsave("{}.jpg".format(self.frame_count), inferences_dict["binarization"])
        keypoints_compressed, binarization_compressed = self._compress_inferences(inferences_dict)

        self._publish_results(keypoints_compressed, binarization_compressed)
        return loop_ok


if __name__ == "__main__":
    parallel_inference_servers = ['tcp://127.0.0.1:60001',
                                  'tcp://127.0.0.1:60002']

    patch_broker_server = PatchPabloBroker(
        geva_broker='tcp://132.199.132.34:19859',
        pub_port='tcp://*:19859',
        parallel_inference_servers=parallel_inference_servers,
        patch_size=(224, 224),
        do_logging=True)

    patch_broker_server.run()
