from threading import Thread

import cv2
import matplotlib
import numpy as np
import zmq
from skimage.color import rgb2gray
from skimage.exposure import rescale_intensity
from skimage.filters import gaussian
from skimage.io import imsave, imread

from applications.gl_viewer.gl_objects import LineCollection
from applications.gl_viewer.gl_viewer import GLViewer, VS, FS
from applications.logged_object import LoggedObject
from dataset.dataset_utils import get_ndarray_diagnostics, rescale_keypoint, get_frame_id_patch_location_from_image_name
from imgproc.heatmap_image_text_extractor import HeatmapImageTextExtractor
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction
from imgproc.imgproc_utils import decompress_image_pil, demean_heatmap, \
    copy_patch_on_image_age, merge_patch_and_image_on_value
from imgproc.patch_by_path_masker import PatchByPathMasker
from imgproc.path_extraction.on_off_classifier import OnOffClassifier
from imgproc.path_extraction.path_beautifier import PathBeautifier
from imgproc.path_extraction.path_stroke_splitter import PathStrokeSplitter
from imgproc.path_extraction.path_subdivider import PathSubdivider
from imgproc.path_extraction.point_path import PointPath
from util.network.bidirectional_imagezmq import ZMQHubFactory
from util.timer import Timer, log_time

matplotlib.use('TkAgg')


class PathProcessServer(LoggedObject):
    def __init__(self,
                 patch_publisher,
                 patch_inference_publisher,
                 patch_size=(224, 224),
                 heatmap_size=(211, 211),
                 surface_size=(1080, 1920),
                 do_logging=False,
                 interval=30,
                 record=False,
                 static_aggregate=None):
        super().__init__(do_logging=do_logging)

        self.patch_publisher = patch_publisher
        self.patch_inference_publisher = patch_inference_publisher

        self.surface_size = surface_size
        self.aggregated_bin_image = np.ones(surface_size, dtype='float32') * 255
        self.aggregated_age_image = np.ones(surface_size, dtype='float32') * 15
        self.current_age = 15

        self.current_keypoint_hm = np.zeros((224, 224), dtype='uint8')
        self.current_bin_hm = np.zeros((224, 224), dtype='uint8')
        self.current_patch = np.zeros((224, 224, 3), dtype='uint8')
        self.current_extract_image = np.zeros((224, 224), dtype='uint8')

        self.patch_size = patch_size
        self.heatmap_size = heatmap_size
        self.interval = interval

        self.inferences = {}
        self.patches = {}
        self.current_work_pair = 0
        self.raw_peaks = []
        self.on_points = np.array([])
        self.all_filter_points = np.array([])
        self.last_all_filter_points_length = 0
        self.last_on_points_length = 0
        self.last_stroke_count = 0
        self.strokes = np.array([])

        if static_aggregate is not None:
            self.use_static_aggregate = True
            self.aggregated_bin_image = rgb2gray(imread(static_aggregate)) * 255
        else:
            self.use_static_aggregate = False

        self.patch_sub_socket = ZMQHubFactory.get_SUB(self.patch_publisher)
        self.inference_sub_socket = ZMQHubFactory.get_SUB(self.patch_inference_publisher)

        self.patch_sub_socket.setsockopt(zmq.SUBSCRIBE, b"")
        self.inference_sub_socket.setsockopt(zmq.SUBSCRIBE, b"")

        # prepare and start display thread
        self.display_thread = Thread(target=self._display_thread_func, args=())
        self.display_thread.daemon = True
        self.display_thread.start()

        self.strokes_thread = self._create_path_thread()

        self.frame_count = 0
        self.interval_counter = 0
        self.all_points_line_collection = None
        self.stroke_line_collection = None
        self.on_points_collection = None
        self.record = record
        # self.logger.setLevel(logging.WARNING)

        cv2.namedWindow('keypoint', cv2.WINDOW_NORMAL)
        cv2.namedWindow('binarization', cv2.WINDOW_NORMAL)
        cv2.namedWindow('patch', cv2.WINDOW_NORMAL)
        cv2.resizeWindow("keypoint", 400, 400)
        cv2.resizeWindow("binarization", 400, 400)
        cv2.resizeWindow("patch", 400, 400)
        cv2.imshow("keypoint", self.current_keypoint_hm)
        cv2.imshow("binarization", self.current_bin_hm)
        cv2.imshow("patch", self.current_patch)
        return

    def _receive_patch_and_location(self):
        msg, data = self.patch_sub_socket.recv_data()
        patch = decompress_image_pil(data)
        frame_id, location = get_frame_id_patch_location_from_image_name(msg)
        split_msg = msg.split('_')
        frame_id = int(split_msg[0])
        location = (int(float(split_msg[1])), int(float(split_msg[2])))
        return frame_id, patch, location

    def _receive_inferences(self):
        msg, data = self.inference_sub_socket.recv_data(copy=True)
        self.logger.debug("Received subscription with msg: {}".format(msg))
        split_msg = msg.split('_')
        frame_id = int(split_msg[0])
        keypoint_len = int(split_msg[1])

        keypoint_hm = decompress_image_pil(data[:keypoint_len])
        bin_hm = decompress_image_pil(data[keypoint_len:])

        self.logger.debug("Keypoint HM {}".format(get_ndarray_diagnostics(keypoint_hm)))
        self.logger.debug("Bin HM {}".format(get_ndarray_diagnostics(bin_hm)))
        return frame_id, keypoint_hm, bin_hm

    @log_time
    def _extract_path(self, points, aggregated_bin_image):
        # we create a path from the keypoint heatmaps
        path = PointPath(points)

        # we retrieve our filtered path points
        filtered_points, _ = path.get(filtered=True, threshold=150, coord_order='column')

        subdivided_path = np.array(PathSubdivider([filtered_points]).run()[0])

        blurred_aggregate = gaussian(aggregated_bin_image, sigma=2)
        # imsave("../tmp_out/record/lines2/aggregate5/{}_blur.jpg".format(len(points)), blurred_aggregate)
        # we perform a gradient descent on our nice extracted text to move the
        # path points closer to it
        nicer_filtered_points = PathBeautifier(
            subdivided_path,
            [],
            blurred_aggregate,
            iterations=5).run()

        # then we classify the pen tip positions into ink and non ink
        on_points, on_indices = OnOffClassifier(
            nicer_filtered_points,
            aggregated_bin_image,
            mean_threshold=1000,
            inclusion_threshold=200).run()

        # on_points = np.array([])
        # on_indices = np.array([])

        strokes = PathStrokeSplitter(
            nicer_filtered_points,
            on_indices,
            split_gap_size=3).run()

        # my_img = np.copy(blurred_aggregate)
        # cv2.polylines(my_img, [on_points.astype('int32')], False, (255,0,0))
        # imsave("../tmp_out/record/stroke_study/0/{}_blur.jpg".format(len(points)), (blurred_aggregate*255).astype('uint8'))

        return strokes, on_points, nicer_filtered_points

    def update_func(self):
        current_all_filter_points = self.all_filter_points.copy()

        if current_all_filter_points.shape[0] > self.last_all_filter_points_length:
            self.all_points_line_collection.replace_lines(
                [current_all_filter_points])  # (current_all_filter_points[self.last_all_filter_points_length-2:])
            self.last_all_filter_points_length = self.all_filter_points.shape[0]

        # current_strokes = self.strokes.copy()
        # current_on_points = self.on_points.copy()

        # if len(current_strokes) > self.last_stroke_count:
        # self.stroke_line_collection.add_line(current_strokes[-1])
        # self.stroke_line_collection.replace_lines(current_strokes)
        #    self.last_stroke_count = len(current_strokes)

        # if current_on_points.shape[0] > self.last_on_points_length:
        # self.on_points_collection.add_line(current_on_points[self.last_on_points_length-2:])
        # self.on_points_collection.replace_lines([current_on_points])
        #    self.last_on_points_length = current_on_points.shape[0]
        return

    def _display_thread_func(self):
        my_strokes = self.strokes.copy()
        my_points = self.all_filter_points

        study_window_pos = 5850
        work_window_pos = 0  # 4700

        my_viewer = GLViewer(1920, 1080, self.update_func, window_pos_x=work_window_pos)
        screen_frame = LineCollection([[[10.0, 10.0], [1910.0, 10.0], [1910.0, 1070.0], [10.0, 1070.0], [10.0, 10.0]]],
                                      VS, FS, show_points=False)
        screen_frame.set_line_width(2.0)
        screen_frame.set_color((0.39, 0.39, 0.39, 1.0))
        my_viewer.add_line_collection(screen_frame)

        self.all_points_line_collection = LineCollection([[[]]], VS, FS, show_points=True)

        self.all_points_line_collection.set_point_size(3.0)
        self.all_points_line_collection.set_line_width(6.0)
        self.all_points_line_collection.set_color((0.1, 0.1, 0.4, 1.0))
        my_viewer.add_line_collection(self.all_points_line_collection)

        self.stroke_line_collection = LineCollection([[[]]], VS, FS, show_points=True)
        self.stroke_line_collection.set_point_size(1.0)
        self.stroke_line_collection.set_line_width(4.0)
        self.stroke_line_collection.set_color((0.39, 0.39, 0.39, 1.0))
        # my_viewer.add_line_collection(self.stroke_line_collection)

        self.on_points_collection = LineCollection([[[]]], VS, FS, show_points=True, show_line=True)
        self.on_points_collection.set_point_size(5.0)
        self.on_points_collection.set_line_width(1.0)
        self.on_points_collection.set_color((1.0, 0.3, 0.3, 1.0))
        # my_viewer.add_line_collection(self.on_points_collection)

        my_viewer.start()

    def _create_path_thread(self):
        # prepare strokes extractor thread
        strokes_thread = Thread(target=self._path_extract_thread_func, args=())
        strokes_thread.daemon = True
        return strokes_thread

    def _path_extract_thread_func(self):
        peaks = self.raw_peaks.copy()
        aggregated_image = self.aggregated_bin_image.copy()
        self.strokes, self.on_points, self.all_filter_points = self._extract_path(peaks, aggregated_image)
        return

    def run(self):
        frame_timer = Timer(verbose=False)

        loop_ok = True
        while loop_ok:
            self.logger.info("Frame {}".format(self.frame_count))
            frame_timer.start("")

            loop_ok = self._run_inner_loop()

            frame_time = frame_timer.stop()
            self.logger.info("Frame time: {}".format(frame_time))
            # print(frame_time)
            self.frame_count += 1
        return

    @log_time
    def _process_binarization_heatmap(self, bin_hm):
        filter_bin_hm = demean_heatmap(bin_hm.astype('float32'))
        filter_bin_hm = filter_bin_hm / np.std(filter_bin_hm)
        mask_3sigma = np.where(filter_bin_hm < 3.0)
        filter_bin_hm[mask_3sigma] = 0
        return rescale_intensity(filter_bin_hm, in_range='image', out_range='uint8')

    @log_time
    def _run_inner_loop(self):
        # get subscriptions
        frame_id_inf, new_keypoint_heatmap, new_bin_heatmap = self._receive_inferences()
        frame_id_patch, patch, location = self._receive_patch_and_location()

        self.logger.debug("Frame ID PATCH: {}".format(frame_id_patch))
        self.logger.debug("Frame ID INFER: {}".format(frame_id_inf))

        # store them in the frame registry
        self.inferences[frame_id_inf] = [new_keypoint_heatmap, new_bin_heatmap]
        self.patches[frame_id_patch] = [patch, location]

        # check if there is a match in the frame ids
        patch_ids = self.patches.keys()
        inf_ids = self.inferences.keys()
        matches = list(set(patch_ids).intersection(inf_ids))

        if len(matches) > 0:
            # we have a match
            matches.sort()
            self.process_working_pair(matches[0])
            del self.inferences[matches[0]]
            del self.patches[matches[0]]
            self.interval_counter += 1
            if self.interval_counter == self.interval and not self.record:
                # hack
                # if self.strokes_thread.is_alive():
                #    self.strokes_thread.join()
                # self.strokes_thread = self._create_path_thread()
                # self.strokes_thread.start()
                self.interval_counter = 0
        else:
            return True

        return True

    # wants row major peak and patch location
    # gives column major peak
    def _transform_patch_peak(self, peak, patch_location):
        left_patch_coord = int(patch_location[1] - self.patch_size[1] / 2)
        top_patch_coord = int(patch_location[0] - self.patch_size[0] / 2)

        # column major peak with patch transform
        peak_transformed = (peak[0] + top_patch_coord, peak[1] + left_patch_coord)
        return peak_transformed

    def _handle_recording(self, match_id, keypoint_heatmap, bin_heatmap, patch, location):
        if self.record:
            imsave("../tmp_out/record/stroke_study/a7/kp/{}_kp.jpg".format(match_id), keypoint_heatmap, quality=95)
            imsave("../tmp_out/record/stroke_study/a7/bin/{}_bin.jpg".format(match_id), bin_heatmap, quality=95)
            imsave(
                "../tmp_out/record/stroke_study/a7/patch/{}_patch_{}_{}.jpg".format(match_id, location[0], location[1]),
                patch, quality=95)
        return

    @log_time
    def process_working_pair(self, match_id):
        keypoint_heatmap, bin_heatmap = self.inferences[match_id]
        patch, location = self.patches[match_id]

        if keypoint_heatmap is None or bin_heatmap is None or patch is None:
            return

        if self.record:
            self._handle_recording(match_id, keypoint_heatmap, bin_heatmap, patch, location)
            return

        self.current_bin_hm = bin_heatmap
        self.current_keypoint_hm = keypoint_heatmap
        self.current_patch = patch

        # DISCARD
        if np.mean(keypoint_heatmap) > 60:
            return

        if np.mean(bin_heatmap) > 81:  # or np.std(bin_heatmap) < 7:
            return

        # EXTRACT CURRENT PEAK
        peak = HeatmapKeypointExtractor(self.current_keypoint_hm, MaxFilterPeakExtraction).run()
        peak = rescale_keypoint(peak, 211, 211, 224, 224)
        transformed_peak = self._transform_patch_peak(peak, location)
        if transformed_peak[0] >= 0 and transformed_peak[0] < self.surface_size[0]:
            if transformed_peak[1] >= 0 and transformed_peak[1] < self.surface_size[1]:
                self.raw_peaks.append(transformed_peak)
            else:
                print("LOL")
        else:
            print("LOL LOL")

        # CREATE EXTRACTION IMAGE
        self.current_extract_image = HeatmapImageTextExtractor(
            self.current_bin_hm,
            rgb2gray(self.current_patch),
            black_threshold=10).run()

        # if len(self.raw_peaks) > 255 and len(self.raw_peaks) < 265:
        #    imsave("../tmp_out/record/lines3/aggregate1/{}_extract.jpg".format(len(self.raw_peaks)), self.current_extract_image)

        # GET CURRENT FILTERED ALLPOINTS PATH

        # hack
        self.all_filter_points, _ = PointPath(np.array(self.raw_peaks)).get(filtered=True, threshold=60,
                                                                            coord_order='column')

        cv2.imshow("keypoint", keypoint_heatmap)
        cv2.imshow("binarization", bin_heatmap)
        cv2.imshow("patch", self.current_patch)
        key = cv2.waitKey(1)

        if key == ord("r"):
            self.raw_peaks.clear()
        # hack
        if False:  # not self.use_static_aggregate:
            current_filter_points, _ = PointPath(np.array(self.raw_peaks)).get(filtered=True, threshold=150,
                                                                               coord_order='column')

            # MASK EXTRACTED IMAGE BY PATH
            self.current_extract_image = PatchByPathMasker(
                self.current_extract_image,
                current_filter_points,
                location,
                stroke_width=10).run()

            # if len(self.raw_peaks) > 255 and len(self.raw_peaks) < 265:
            #    imsave("../tmp_out/record/lines3/aggregate1/{}_path_mask.jpg".format(len(self.raw_peaks)), self.current_extract_image)

            # AGGREGATE
            self.aggregated_bin_image = merge_patch_and_image_on_value(
                self.current_extract_image * 255,
                self.aggregated_bin_image, location)

            # if match_id > 988 and match_id < 1040:
            # imsave("L:/validation_study_2/videos/P9/aggregate/merge/{}.jpg".format(self.current_age), self.aggregated_bin_image.astype('uint8'))

            self.aggregated_bin_image = copy_patch_on_image_age(
                self.current_extract_image * 255,
                self.aggregated_bin_image,
                self.aggregated_age_image,
                location,
                self.current_age,
                max_age=25)
            # imsave("L:/validation_study_2/videos/P9/aggregate/aged/{}.jpg".format(self.current_age),
            #       self.aggregated_bin_image.astype('uint8'))

        # increase current age after discarding bad heatmaps
        # so we dont artificially age pixels
        self.current_age += 1

        self.logger.debug("PROCESSED WORKING PAIR {}".format(match_id))
        return


if __name__ == '__main__':
    server = PathProcessServer(
        patch_publisher='tcp://132.199.132.34:19859',
        patch_inference_publisher='tcp://132.199.140.68:19859',
        do_logging=True,
        interval=30,
        record=False)
    server.run()
