import matplotlib

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
from ast import literal_eval
import os
import numpy as np
import pandas as pd
from PIL import Image
from sklearn import metrics
from applications.final_prototype.analysis_utils import AnnotationsHeatmapsDistancesSetAnalyzer, \
    get_index_of_min_distance_to_perfect_for_fpr_tpr
from dataset_utils import rescale_keypoint
from imgproc.imgproc_utils import resize_image_pil


def copy_all_test_files_to(all_anno_df, target_dir, new_width, new_height):
    list_test = [0, 6, 10, 14, 20, 23, 38, 39]
    anno_test_dict = {"frame": [], "full_path": [], "ptip": [], "ptail": []}
    test_index = 0
    for index, row in all_anno_df.iterrows():
        path = row['full_path']
        img_name = row.name

        if row['subject'] in list_test:
            pil_img = Image.open(path)
            pil_img = resize_image_pil(pil_img, (new_width, new_height))
            pil_img.save(os.path.join(target_dir, img_name), format="JPEG", quality=95, subsampling=0)

            ptip_keypoint = row['ptip']
            ptail_keypoint = row['ptail']
            ptip_keypoint = rescale_keypoint(ptip_keypoint, 3840, 2160, new_width, new_height)
            ptail_keypoint = rescale_keypoint(ptail_keypoint, 3840, 2160, new_width, new_height)

            anno_test_dict["frame"].append(test_index)
            anno_test_dict["full_path"].append(img_name)
            anno_test_dict["ptip"].append(ptip_keypoint)
            anno_test_dict["ptail"].append(ptail_keypoint)
            test_index += 1
            print(test_index)

    test_anno_df = pd.DataFrame(anno_test_dict)
    test_anno_df.to_csv(os.path.join(target_dir, "final_test_annos_hd.csv"))
    return


def load_all_anno_df(filename):
    all_annos = pd.read_csv('final_consistent_annos.csv', index_col=0,
                            dtype={'subject': np.int32,
                                   'task': np.int32,
                                   'frame': np.int32},
                            converters={
                                'ptip': literal_eval,
                                'ptail': literal_eval})
    return all_annos


def get_weights_for_subjects_in_df_order(anno_df):
    my_df = anno_df.copy(deep=True)
    my_df['subject'] = my_df['full_path'].apply(lambda x: int(x.split('_')[0]))
    group_df = my_df.groupby(['subject']).agg(
        samples=pd.NamedAgg(column='frame', aggfunc='count'))

    weights = group_df.samples.min() / group_df.samples.to_numpy()
    weights_dict = {subject: weights[index] for index, subject in enumerate(group_df.index.to_list())}

    my_df['weight'] = my_df['subject'].apply(lambda x: weights_dict[x])
    # print(my_df.head(n=20))
    return my_df['weight'].to_numpy()


if __name__ == '__main__':
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)

    # all_anno_df = load_all_anno_df('final_consistent_annos.csv')
    # copy_all_test_files_to(all_anno_df, "L:/large_dataset_study/test_analysis/frames", 1920, 1080)
    study_base_path = "L:/large_dataset_study/test_analysis/"
    anno_path = os.path.join(study_base_path, "final_test_annos_hd.csv")
    detections_path = os.path.join(study_base_path, "final_test_detections_hd.csv")
    heatmap_path = os.path.join(study_base_path, "output")
    analyzer = AnnotationsHeatmapsDistancesSetAnalyzer(anno_path, heatmap_path, detections_path,
                                                       filename_identifier="testset_study")

    evaluation_df = analyzer.get_evaluation()
    weights = get_weights_for_subjects_in_df_order(evaluation_df)

    scores_mean, scores_median, scores_std, scores_var, labels = analyzer.get_scores_labels_for_set()
    fpr_coords_median, tpr_coords_median, thresholds_median = metrics.roc_curve(labels, scores_median,
                                                                                sample_weight=weights)
    fpr_coords_mean, tpr_coords_mean, thresholds_mean = metrics.roc_curve(labels, scores_mean, sample_weight=weights)

    median_auc = metrics.roc_auc_score(labels, scores_median, sample_weight=weights, average='micro')
    #
    mean_auc = metrics.roc_auc_score(labels, scores_mean, sample_weight=weights)
    std_auc = metrics.roc_auc_score(labels, scores_std, sample_weight=weights)
    var_auc = metrics.roc_auc_score(labels, scores_var, sample_weight=weights)

    mean_ap = metrics.average_precision_score(labels, scores_mean, sample_weight=weights)
    median_ap = metrics.average_precision_score(labels, scores_median, sample_weight=weights)
    std_ap = metrics.average_precision_score(labels, scores_std, sample_weight=weights)
    var_ap = metrics.average_precision_score(labels, scores_var, sample_weight=weights)

    mean_opt_index = get_index_of_min_distance_to_perfect_for_fpr_tpr(fpr_coords_mean, tpr_coords_mean)
    mean_opt_point = (fpr_coords_mean[mean_opt_index], tpr_coords_mean[mean_opt_index])

    print("AUC // Mean {} // Median {} // Std {} // Var {} ".format(mean_auc, median_auc, std_auc, var_auc))
    print("AP // Mean {} // Median {} // Std {} // Var {} ".format(mean_ap, median_ap, std_ap, var_ap))

    print('\033[92m' + "Best AUC: {}".format(np.max([mean_auc, median_auc, std_auc, var_auc])))
    print('\033[92m' + "Best AP: {}".format(np.max([mean_ap, median_ap, std_ap, var_ap])))

    #########
    # percentiles
    #########

    thresholds = np.linspace(0, 255, 255)
    error_point = 1e-3
    # normalized error values
    percentiles, all_distances = analyzer.get_distances_percentile_for_thresholds(error_point, thresholds, 1920, 1080,
                                                                                  lambda x: x)

    fig, axes = plt.subplots(ncols=2)

    ###########
    # ROC
    ###########

    axes[0].step(fpr_coords_median, tpr_coords_median, label='Median')
    axes[0].step(fpr_coords_mean, tpr_coords_mean, label='Mean')

    np.save("testset_fpr_coords_mean.npy", fpr_coords_mean)
    np.save("testset_tpr_coords_mean.npy", tpr_coords_mean)
    np.save("testset_scores_mean.npy", scores_mean)

    axes[0].annotate("Mean Optimum Threshold: {}".format(thresholds_mean[mean_opt_index]),
                     mean_opt_point,
                     textcoords="offset points",
                     xytext=(0, 50),
                     arrowprops=dict(facecolor='black', shrink=0.05))
    axes[0].set_aspect('equal', 'box')
    axes[0].set(xlim=(-0.1, 1.1), ylim=(-0.1, 1.1))
    axes[0].legend()

    ###########
    # error percentiles
    ###########

    axes[1].plot(thresholds, percentiles)
    axes[1].axvline(x=255 - thresholds_mean[mean_opt_index])

    # median_vals = np.array([np.median(elem) for elem in distance_norm_collection])
    # norm = plt.Normalize(median_vals.min(), median_vals.max())

    opt_real_threshold = 1  # int(255 - thresholds_mean[mean_opt_index])
    np.save("testset_threshold_{}_distances.npy".format(opt_real_threshold),
            all_distances[opt_real_threshold])
    # colors = plt.cm.coolwarm(norm(median_vals))
    # print("Optimal Mean Threshold {}".format(opt_real_threshold))
    # b = sns.boxplot(data=all_distances[opt_real_threshold], ax=axes[1])
    # b.set(yscale='log')

    # plt.show()
