# class vbo and ibo code from https://gist.github.com/ousttrue/c4ae334fc1505cdf4cd7

from OpenGL.GL import *
from OpenGL.GLUT import *


class VBO:
    bytes_per_float = 4

    def __init__(self, update_mode=GL_STATIC_DRAW) -> None:
        self.vbo = glGenBuffers(1)
        self.vertex_component_count = 0
        self.vertex_count = 0
        self.update_mode = update_mode

    def __del__(self) -> None:
        glDeleteBuffers(1, [self.vbo])

    def bind(self) -> None:
        glBindBuffer(GL_ARRAY_BUFFER, self.vbo)

    def unbind(self) -> None:
        glBindBuffer(GL_ARRAY_BUFFER, 0)

    def create_buffer_data_storage(self, vertex_component_count: int, bytes_needed: int,
                                   data: any) -> None:
        self.vertex_component_count = vertex_component_count
        stride = VBO.bytes_per_float * self.vertex_component_count
        self.vertex_count = bytes_needed // stride

        self.bind()
        glBufferData(GL_ARRAY_BUFFER, bytes_needed, data, self.update_mode)
        self.unbind()

    def update_buffer_data(self, bytes_needed: int, data: any):
        stride = VBO.bytes_per_float * self.vertex_component_count
        offset = self.vertex_count * stride
        self.vertex_count += bytes_needed // stride
        self.bind()
        glBufferSubData(GL_ARRAY_BUFFER, offset, bytes_needed, data)
        self.unbind()
        return

    def set_slot(self, slot: int) -> None:
        self.bind()
        glEnableVertexAttribArray(slot)
        glVertexAttribPointer(slot, self.vertex_component_count, GL_FLOAT, GL_FALSE, 0, None)

    def draw(self, draw_mode=GL_TRIANGLES) -> None:
        glDrawArrays(draw_mode, 0, self.vertex_count)


class IBO:
    def __init__(self, draw_mode=GL_TRIANGLES) -> None:
        self.vbo = glGenBuffers(1)
        self.index_count = 0
        self.index_type = 0
        self.draw_mode = draw_mode

    def __del__(self) -> None:
        glDeleteBuffers(1, [self.vbo])

    def bind(self) -> None:
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.vbo)

    def unbind(self) -> None:
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)

    def set_indices(self, stride: int, bytes_needed: int, data) -> None:
        self.index_count = bytes_needed // stride
        self.bind()
        if stride == 1:
            self.index_type = GL_UNSIGNED_BYTE
        elif stride == 2:
            self.index_type = GL_UNSIGNED_SHORT
        elif stride == 4:
            self.index_type = GL_UNSIGNED_INT
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, bytes_needed, data, GL_STATIC_DRAW)

    def draw(self) -> None:
        glDrawElements(self.draw_mode, self.index_count, self.index_type, None)
