import itertools
import struct

from OpenGL.GL import *

from applications.gl_viewer.buffer_objects import VBO
from applications.gl_viewer.shader import Shader


class LineCollection():
    def __init__(self, lines, vertex_shader, fragment_shader, show_points=False, show_line=True, order='row'):
        self.vertex_component_count = 2
        self.bytes_per_vertex = self.vertex_component_count * VBO.bytes_per_float

        self.shader = Shader()
        self.shader.compile(vertex_shader, fragment_shader)
        self.vbo = VBO()
        self.lines = lines
        self.positions = self._create_position_list(lines)
        self._init_buffers()
        self.enable_points = show_points
        self.enable_lines = show_line
        self.point_size = 10.0
        self.line_color = (1.0, 0.0, 0.0, 1.0)
        self.point_color = (0.0, 0.0, 1.0, 1.0)
        self.line_width = 3.0
        return

    def set_color(self, color):
        self.line_color = color

    def set_point_size(self, point_size):
        self.point_size = point_size

    def set_line_width(self, width):
        self.line_width = width

    def show_points(self):
        if not self.enable_points:
            self.enable_points = True
        return

    def hide_points(self):
        self.enable_points = False
        return

    def add_source(self, source):
        self.source = source
        return

    def _create_position_list(self, lines):
        positions = []
        for line in lines:
            positions.extend(self._create_position_list_for_line(line))
        return positions

    # converts the positions for GL_LINES, i.e. vertices 0 and 1 are
    # a line, then 2 and 3 and so on. This means doubling the coords.
    # the resulting list contains all coordinates at the top level.
    def _create_position_list_for_line(self, line):
        new_positions = list(zip(line, line))
        new_positions = list(itertools.chain(*new_positions))
        new_positions = new_positions[1:-1]
        return list(itertools.chain(*new_positions))

    def add_line(self, line):
        self.lines.append(line)
        self.positions.extend(self._create_position_list_for_line(line))
        self._init_buffers()
        return

    def add_lines(self, lines):
        self.lines.extend(lines)
        self.positions.extend(self._create_position_list(lines))
        self._init_buffers()
        return

    def replace_lines(self, lines):
        self.lines = lines
        self.positions = self._create_position_list(lines)
        self._init_buffers()
        return

    def _init_buffers(self):
        # two vertex components of float32
        # self.positions = [item / 200 for item in self.positions]
        vertex_count = len(self.positions) // 2
        self.vbo.create_buffer_data_storage(self.vertex_component_count,
                                            vertex_count * self.bytes_per_vertex,
                                            struct.pack(
                                                "{}f".format(len(self.positions)),
                                                *self.positions)
                                            )
        return

    def draw(self, zoom, offset) -> None:
        line_color_uniform = self.shader.get_uniform_location("line_color")
        point_color_uniform = self.shader.get_uniform_location("point_color")
        zoom_uniform = self.shader.get_uniform_location("zoom")
        offset_uniform = self.shader.get_uniform_location("offset")

        self.shader.use()
        glUniform1f(zoom_uniform, zoom)
        glUniform2f(offset_uniform, *offset)

        self.vbo.set_slot(0)

        if self.enable_lines:
            glLineWidth(self.line_width)
            glUniform4f(line_color_uniform, *self.line_color)
            self.vbo.draw(GL_LINES)

        if self.enable_points:
            glUniform4f(point_color_uniform, *self.point_color)
            glPointSize(self.point_size)
            self.vbo.draw(GL_POINTS)

        self.vbo.unbind()
        self.shader.unuse()
