# class vbo code from https://gist.github.com/ousttrue/c4ae334fc1505cdf4cd7
import numpy as np
from OpenGL.GL import *
from OpenGL.GLUT import *

from applications.gl_viewer.gl_objects import LineCollection
from imgproc.path_extraction.online_ink_requester import OnlineInkRequester

VS = '''
#version 330
in vec2 aPosition;
uniform vec4 line_color;
uniform vec4 point_color;
uniform float zoom;
uniform vec2 offset;
out vec4 vertexColor;
out vec2 lineCenter;

void main ()
{
    mat4 ortho_matrix;
    float r = 1920;
    float l = 0;
    float t = 0;
    float b = 1080;
    float f = 100;
    float n = -100;
    ortho_matrix = mat4(2/(r-l), 0, 0, -(r+l)/(r-l),
                    0, 2/(t-b), 0, -(t+b)/(t-b),
                    0, 0, -2/(f-n), -(f+n)/(f-n),
                    0, 0, 0, 1);
    vec2 myPosition = vec2(aPosition.x, aPosition.y) + offset;
    vec4 finished_position = transpose(ortho_matrix) * vec4(myPosition, 0.0, 1); 
    gl_Position = vec4(finished_position.xy * zoom, 0.0, 1);
    vertexColor = line_color;
    vec2 vp = vec2(1920, 1080);
    lineCenter = 0.5 * (finished_position.xy + vec2(1, 1))*vp;
}
'''

FS = '''
#version 330
in vec4 vertexColor;
in vec2 lineCenter;
out vec4 FragColor;

void main()
{
    vec4 col = vertexColor;        
    float d = length(lineCenter - gl_FragCoord.xy);
    float w = 0.05;
    
    if (d>w)
      col.w = 0;
    else
      col.w *= pow((w-d)/w, 2.5);
      
    FragColor = col;
}
'''


def cb_dbg_msg(source, msg_type, msg_id, severity, length, raw, user):
    msg = raw[0:length]
    print('debug {} {} {} {} {}'.format(source, msg_type, msg_id, severity, msg))


class GLViewer():
    def __init__(self, window_width, window_height, update_func, window_pos_x=0, window_pos_y=0):
        self.background_images = []
        self.line_collections = []

        self.window_width = window_width
        self.window_height = window_height
        self.window_pos_x = window_pos_x
        self.window_pos_y = window_pos_y
        self.update_func = update_func

        self.clear_color = (0.97, 0.95, 0.91, 1.0)

        # initialize opengl and glut after creating all
        # necessary objects

        self.window = self._init_glut_window()
        self._init_gl()

        self.frames = 0
        self.timebase = 0
        self.zoom = 1
        self.offset_x = 0
        self.offset_y = 0
        self.old_mouse_x = 0
        self.old_mouse_y = 0
        # limit display framerate to 60
        self.timer(0)
        return

    def idle_func(self):
        self.update_func()
        return

    def mouse_wheel_func(self, button, dir, x, y):
        if dir > 0:
            self.zoom += 0.5 * dir
        else:
            self.zoom -= 0.5 * abs(dir)
        return

    def mouse_motion_func(self, x, y):
        if self.old_mouse_x != 0 and self.old_mouse_y != 0:
            x_delta = x - self.old_mouse_x
            y_delta = y - self.old_mouse_y
            self.offset_x += x_delta
            self.offset_y += y_delta
        self.old_mouse_x = x
        self.old_mouse_y = y
        return

    def mouse_motion_passive_func(self, x, y):
        self.old_mouse_x = x
        self.old_mouse_y = y
        return

    def mouse_func(self, button, state, x, y):
        if button == GLUT_RIGHT_BUTTON:
            if state == GLUT_DOWN:
                online_recognizer = OnlineInkRequester().online_ink(self.line_collections[1].lines)
        return

    def start(self):
        glutDisplayFunc(self.display)
        glutReshapeFunc(self.reshape)
        glutIdleFunc(self.idle_func)
        glutMouseFunc(self.mouse_func)
        glutMouseWheelFunc(self.mouse_wheel_func)
        glutMotionFunc(self.mouse_motion_func)
        glutPassiveMotionFunc(self.mouse_motion_passive_func)
        glutMainLoop()
        return

    def _init_gl(self):
        glClearColor(*self.clear_color)
        glClearDepth(1.0)

        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)
        return

    def timer(self, integer):
        glutPostRedisplay()
        glutTimerFunc(1000 // 60, self.timer, 0)

    def _init_glut_window(self):
        glutInit()
        glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH)
        glutInitWindowSize(self.window_width, self.window_height)
        glutInitWindowPosition(self.window_pos_x, self.window_pos_y)
        glutInitContextFlags(GLUT_CORE_PROFILE | GLUT_DEBUG)
        window = glutCreateWindow("GL Viewer")
        # glClearColor(0.12, 0.12, 0.12, 1.0)
        return window

    def display(self):
        elapsed_time = glutGet(GLUT_ELAPSED_TIME)
        if elapsed_time - self.timebase > 1000:
            fps = self.frames * 1000 / (elapsed_time - self.timebase)
            self.timebase = elapsed_time
            self.frames = 0
            glutSetWindowTitle("{}".format(fps))
        self.frames += 1

        error = glGetError()
        if error != GL_NO_ERROR:
            print(glGetError())

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()

        for line in self.line_collections:
            line.draw(self.zoom, (self.offset_x, self.offset_y))

        glFlush()
        glutSwapBuffers()
        # glutPostRedisplay()

    def reshape(self, w, h):
        glViewport(0, 0, w, h)

    def add_line_collection(self, line_collection: LineCollection):
        self.line_collections.append(line_collection)
        return

    def add_points(self):
        return

    def add_background_image(self, image: np.ndarray):
        self.background_images.append(image)
        return


if __name__ == '__main__':
    viewer = GLViewer(800, 800)
