import logging

import numpy as np

from applications.inference_servers.inference_server import InferenceServer
from applications.network_utils import receive_image, send_image
from dataset.dataset_utils import get_ndarray_diagnostics
from util.timer import log_time


class KeypointInferenceServer(InferenceServer):
    def __init__(self,
                 inference_provider=None,
                 open_port=None,
                 name="inference server",
                 compressed_send=False,
                 do_logging=False):

        super().__init__(open_ports=[open_port],
                         name=name,
                         do_logging=do_logging)

        self.compressed_send = compressed_send
        self.inference_provider = inference_provider
        self.logger.setLevel(logging.WARNING)
        return

    @log_time
    def _merge_inference(self, inference):
        inf_shape = inference.shape
        self.logger.debug("Shape of inference to merge: {}".format(inference.shape))
        if len(inf_shape) < 3:
            self.logger.warning("Warning: inference shape is greyscale. Not merging.")
            result = inference
        elif inf_shape[2] == 2:
            result = np.zeros((inf_shape[0], 2 * inf_shape[1]), np.float32)
            result[:inf_shape[0], :inf_shape[1]] = inference[:, :, 0]
            result[:inf_shape[0], inf_shape[1]:2 * inf_shape[1]] = inference[:, :, 1]
        else:
            self.logger.warning("Warning: inference shape is weird. Not merging.")
            result = inference
        return result

    @log_time
    def _predict(self, image):
        return self.inference_provider.predict_single_image(image)

    def _run_inner_loop(self):
        msg, image = receive_image(self._get_open_port_socket())
        self.logger.debug(get_ndarray_diagnostics(image))
        self.logger.debug("Received image from {}".format(msg))

        result = self._predict(image)
        self.logger.debug("Prediction: {}".format(get_ndarray_diagnostics(result)))
        merged_inference = self._merge_inference(result[0])

        send_image(self._get_open_port_socket(),
                   np.array(merged_inference),
                   message=self.name,
                   compressed_send=self.compressed_send)
        return True
