import logging

import numpy as np

from applications.brokers.inference_broker import InferenceBroker
from applications.network_utils import receive_image, send_image
from dataset.dataset_utils import get_ndarray_diagnostics
from imgproc.imgproc_utils import resize_image_array_pil, compress_image_array_pil
from imgproc.patch_by_heatmap_extractor import PatchByHeatmapExtractor
from prototype.camera_streamer import BRIOCameraStreamer
from util.timer import log_time


class LocalBroker(InferenceBroker):

    def __init__(self, stream_provider, roinet_server, path_process_server, patch_broker_server,
                 roinet_input_size=(400, 710), patch_size=(224, 224), debug_sends=False, **kwargs):
        super().__init__(inference_servers_seq=[roinet_server, path_process_server, patch_broker_server],
                         compressed_parallel_distribution=True, **kwargs)
        self.stream_provider = stream_provider
        self.patch_size = patch_size
        self.roinet_input_size = roinet_input_size
        self.roinet_server = roinet_server
        self.path_process_server = path_process_server
        self.patch_broker_server = patch_broker_server
        self.debug_sends = debug_sends
        self.logger.setLevel(logging.DEBUG)
        self.cam_image = self._acquire_image()

    @log_time
    def _receive_and_decompress_image(self):
        msg, image = receive_image(self._get_open_port_socket(), comp_method='pil')
        self.logger.debug("Received image from {}".format(msg))
        self.logger.debug(get_ndarray_diagnostics(image))
        return msg, image

    @log_time
    def _get_roi_patch_from_image(self, image):
        image_resized = resize_image_array_pil(image, (self.roinet_input_size[1], self.roinet_input_size[0]))
        self.logger.debug("Resized image.")

        roinet_socket = self.inference_sockets_seq[self.roinet_server]
        self.logger.debug("Sending image to roinet: {}".format(roinet_socket))
        roinet_socket.send_data(image_resized, msg='roinet_inference', copy=False, compressed=False)

        msg, roinet_inference = roinet_socket.recv_data()
        patch, _ = PatchByHeatmapExtractor(image, roinet_inference[:, :396], self.patch_size).run()
        return patch, roinet_inference

    @log_time
    def _get_patch_inferences(self, roi_patch):
        patch_broker_socket = self.inference_sockets_seq[self.patch_broker_server]
        compressed_patch = compress_image_array_pil(roi_patch, 95)
        patch_broker_socket.send_data(compressed_patch, msg='roinet_inference', copy=False, compressed=True)

        inferences_dict = {}
        msg, first_inference = receive_image(patch_broker_socket, comp_method='pil')
        patch_broker_socket.send(b"another please")
        inferences_dict[msg] = first_inference
        msg, second_inference = receive_image(patch_broker_socket, comp_method='pil')
        inferences_dict[msg] = second_inference

        self.logger.debug(get_ndarray_diagnostics(inferences_dict["keypoint"]))
        self.logger.debug(get_ndarray_diagnostics(inferences_dict["binarization"]))

        inferences_dict["patch"] = roi_patch
        return inferences_dict

    def _get_uint8_heatmap(self, heatmap):
        h, w = heatmap.shape[:2]
        my_heatmap = np.reshape(heatmap, (h, w))
        return (my_heatmap * 255 / np.max(my_heatmap)).astype('uint8')

    def _resize_patch_heatmaps_to_patch_size(self, inferences_dict):
        inferences_dict["keypoint"] = resize_image_array_pil(self._get_uint8_heatmap(inferences_dict["keypoint"]),
                                                             (224, 224))
        inferences_dict["binarization"] = resize_image_array_pil(
            self._get_uint8_heatmap(inferences_dict["binarization"]), (224, 224))
        return inferences_dict

    @log_time
    def _send_to_path_processor(self, inferences_dict):
        inferences_dict = self._resize_patch_heatmaps_to_patch_size(inferences_dict)
        combined_data = np.dstack(
            (inferences_dict["keypoint"], inferences_dict["binarization"], inferences_dict["patch"]))
        path_process_socket = self.inference_sockets_seq[self.path_process_server]
        path_process_socket.send_data(combined_data, msg='keypoint_binarization', copy=False, compressed=False)
        self.logger.debug("Sent inferences to path processor")
        msg, strokes = path_process_socket.recv_data()
        return strokes, combined_data

    def _debug_client_sends(self, roinet_inference, inferences_dict, roi_patch):
        client_socket = self._get_open_port_socket()
        # send roinet inference
        send_image(client_socket, self._get_uint8_heatmap(roinet_inference))
        client_socket.recv()
        # send keypoint heatmap
        send_image(client_socket, self._get_uint8_heatmap(inferences_dict["keypoint"]))
        client_socket.recv()
        # send binarization heatmap
        send_image(client_socket, self._get_uint8_heatmap(inferences_dict["binarization"]))
        client_socket.recv()
        # send patch
        send_image(client_socket, roi_patch)
        client_socket.recv()
        return

    @log_time
    def _acquire_image(self):
        image = self.stream_provider.get()

        if image is None:
            self.logger.error("Couldn't acquire image. Exiting.")
            raise IOError

        self.logger.debug(get_ndarray_diagnostics(image))
        return image

    @log_time
    def _run_inner_loop(self):
        loop_ok = True

        # get image from camera
        self.cam_image = self._acquire_image()

        # get roi
        roi_patch, roinet_inference = self._get_roi_patch_from_image(self.cam_image)

        inferences_dict = self._get_patch_inferences(roi_patch)

        strokes, combined_data = self._send_to_path_processor(inferences_dict)

        if self.debug_sends:
            self._debug_client_sends(roinet_inference, inferences_dict, roi_patch)

        print(strokes)
        # self._get_open_port_socket().send_data(strokes, "NextFrame")
        return loop_ok


if __name__ == "__main__":
    camera_stream = BRIOCameraStreamer(no_decoding=False)

    patch_broker_server = LocalBroker(
        camera_stream,
        roinet_server='tcp://127.0.0.1:60003',
        path_process_server='tcp://127.0.0.1:60004',
        patch_broker_server='tcp://132.199.140.68:19858',
        roinet_input_size=(400, 710),
        patch_size=(224, 224),
        open_ports=['tcp://*:19858'],
        inference_servers_parallel=[],
        name="patch_broker",
        do_logging=True,
        debug_sends=False)

    patch_broker_server.run()
