import cv2
import matplotlib
from skimage.io import imsave

from applications.clients.inference_client import InferenceClient
from applications.network_utils import send_image, receive_image
from dataset.dataset_utils import get_ndarray_diagnostics
from prototype.camera_streamer import BRIOCameraStreamer
from util.timer import log_time

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class PatchInferenceClient(InferenceClient):
    '''
    Client part of prototype with patch networks. Takes image from webcam,
    sends it to inference broker in full resolution.

    Works as follows:
    - 1 Sends image
    - 2 acquires new one
    - 3 waits to receive ok from broker
    - 4 upon ok, goto 1

    Not implemented yet:
    Sometimes, in the payload the pen path arrives, as well as any
    detected gestures. The client can yield this information
    to other applications.
    '''

    def __init__(self, stream_provider,
                 server_address,
                 store_inferences=False,
                 compressed_send=True,
                 no_decoding=False,
                 do_logging=False,
                 display=False,
                 debug_data_receive=False):
        super().__init__(server_address, compressed_send, no_decoding=no_decoding, do_logging=do_logging)
        self.stream_provider = stream_provider
        self.display = display
        self.loop_counter = 0
        self.image_counter = 0
        self.store_inferences = store_inferences
        self.debug_data_receive = debug_data_receive
        self.image = self._acquire_image()

        self.fig, self.axes = plt.subplots(ncols=3, nrows=2, figsize=(14, 7))

    @log_time
    def _acquire_image(self):
        image = self.stream_provider.get()

        if image is None:
            self.logger.error("Couldn't acquire image. Exiting.")
            raise IOError
        self.image_counter += 1

        self.logger.debug("Acquired image {}".format(self.image_counter))
        self.logger.debug(get_ndarray_diagnostics(image))

        return image

    def _display_image(self, window_name, image):
        cv2.imshow(window_name, image)

    def _receive_debug_data(self):
        msg, roinet_inference = receive_image(self.client_socket)
        self.client_socket.send(b"Yes1")
        msg, keypoint_heatmap = receive_image(self.client_socket)
        self.client_socket.send(b"Yes2")
        msg, bin_heatmap = receive_image(self.client_socket)
        self.client_socket.send(b"Yes3")
        msg, patch = receive_image(self.client_socket)
        self.client_socket.send(b"Yes4")
        return roinet_inference, keypoint_heatmap, bin_heatmap, patch

    def _run_inner_loop(self):
        send_image(self.client_socket,
                   self.image,
                   message="client_image",
                   compressed_send=self.compressed_send)
        self.logger.debug("Sent image {}".format(self.image_counter))

        if self.debug_data_receive:
            roinet_inference, keypoint_heatmap, bin_heatmap, patch = self._receive_debug_data()
            imsave("../tmp_out/debug_patchy_alpaca/roinet_keypoints/{}_roinet.jpg".format(self.loop_counter),
                   roinet_inference, quality=100)
            imsave("../tmp_out/debug_patchy_alpaca/patch_keypoints/{}_patchhm.jpg".format(self.loop_counter),
                   keypoint_heatmap, quality=100)
            imsave("../tmp_out/debug_patchy_alpaca/bin/{}_bin.jpg".format(self.loop_counter), bin_heatmap, quality=100)
            imsave("../tmp_out/debug_patchy_alpaca/patches/{}_patch.jpg".format(self.loop_counter), patch, quality=100)
            imsave("../tmp_out/debug_patchy_alpaca/frames/{}_frame.jpg".format(self.loop_counter), self.image,
                   quality=100)

        self.image = self._acquire_image()

        msg, payload = self.client_socket.recv_data()
        self.logger.info("received reply from {}".format(msg))
        self.logger.info("with payload {}".format(payload))

        self.loop_counter += 1

        if self.display:
            self._display_image("frame", self.image)
            self._display_image("roinet", roinet_inference)
        return True


if __name__ == "__main__":
    no_decoding = False
    camera_stream = BRIOCameraStreamer(no_decoding=no_decoding)

    client = PatchInferenceClient(camera_stream,
                                  "tcp://132.199.140.68:19858",
                                  store_inferences=False,
                                  compressed_send=True,
                                  display=False,
                                  no_decoding=no_decoding,
                                  do_logging=True,
                                  debug_data_receive=True)
    client.run()
