import time

import imgaug as ia
import numpy as np
from imgaug import augmenters as iaa
from skimage.io import imread
from skimage.transform import resize
from tensorflow.keras.applications import resnet50
from tensorflow.keras.utils import Sequence

from dataset.binary_gt_processor import BinaryGroundTruthProcessor


# INPUT:
# image_filenames
#   RGB image files
# gt_filenames
#   binary image files


class AugmentedBinarizationGenerator(Sequence):
    def __init__(self, image_filenames, gt_filenames, model_config):
        ia.seed(int(time.time()))

        self.image_filenames = image_filenames
        self.gt_filenames = gt_filenames

        self.model_config = model_config

        self.batch_size = self.model_config.get("batch_size")
        self.input_image_width = self.model_config.get("input_width")
        self.input_image_height = self.model_config.get("input_height")
        self.gt_width = self.model_config.get("heatmap_width")
        self.gt_height = self.model_config.get("heatmap_height")

        self.sequences = self._create_augmentation_sequences()
        self.augmentation_multiplier = len(self.sequences) + 1

        if self.batch_size % self.augmentation_multiplier != 0:
            raise Exception(
                "Batch size should be divisible by augmentation multiplier {}.".format(self.augmentation_multiplier))

        self.real_image_count_for_batch = int(self.batch_size / self.augmentation_multiplier)
        return

    def _create_augmentation_sequences(self):
        sequences = []

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Fliplr(0.5),
                                         iaa.CropAndPad(percent=(-0.3, 0.3),
                                                        pad_mode=["edge"])
                                         ]))

        sequences.append(iaa.Sequential([iaa.Multiply((0.7, 1.3), per_channel=0.5),
                                         iaa.Fliplr(0.5),
                                         iaa.AverageBlur(k=((2, 5), (2, 5))),
                                         iaa.AdditiveGaussianNoise(scale=(0, 0.02 * 255))
                                         ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            shear=(-15, 15),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3)),
            iaa.CropAndPad(percent=(-0.3, 0.3),
                           pad_mode=["edge"]),
            iaa.Flipud(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3)),
            iaa.ContrastNormalization((0.5, 1.5), per_channel=0.5),
            iaa.CropAndPad(percent=(-0.3, 0.3),
                           pad_mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.ContrastNormalization((0.5, 1.5)),
            iaa.Fliplr(0.5)
        ]))

        if self.batch_size < 16:
            return sequences

        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=250,
            height=250,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))
        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=250,
            height=250,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))
        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=250,
            height=250,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))
        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=250,
            height=250,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))
        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=100,
            height=100,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))
        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=100,
            height=100,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))
        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=100,
            height=100,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))
        sequences.append(iaa.Sequential([iaa.CropToFixedSize(
            width=100,
            height=100,
            position='normal'),
            iaa.Fliplr(0.5)
        ]))

        return sequences

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) * self.augmentation_multiplier / float(self.batch_size)))

    # for some versions of keras, this is needed even though we don't use it
    def on_epoch_end(self):
        pass

    def _load_images(self, image_list, index):
        my_slice = image_list[
                   index * self.real_image_count_for_batch: (1 + index) * self.real_image_count_for_batch]
        my_imgs = [imread(img) for img in my_slice]
        return my_imgs

    # getitem must return a whole batch
    # this method returns two lists for images and target heatmaps
    def __getitem__(self, index):
        loaded_images = self._load_images(self.image_filenames, index)
        loaded_gt_images = self._load_images(self.gt_filenames, index)

        augmented_images, augmented_gts = self._augment(loaded_images, loaded_gt_images)
        augmented_gts = [gt.get_arr() for gt in augmented_gts]
        gt_processor = BinaryGroundTruthProcessor(augmented_gts, self.model_config)

        batch_y = gt_processor.process()
        batch_x = self._format_input_images(augmented_images)
        return batch_x, batch_y

    def _augment(self, loaded_images, loaded_gts):
        converted_gts = [ia.HeatmapsOnImage(np.array(gt).astype('float32'), shape=loaded_gts[0].shape, min_value=0.0,
                                            max_value=255.0) for gt in loaded_gts]

        result_images = loaded_images.copy()
        result_gts = converted_gts.copy()

        for seq in self.sequences:
            # seq_det = seq.to_deterministic()
            current_images_augmented, current_gts_augmented = seq(images=loaded_images, heatmaps=converted_gts)

            result_images += current_images_augmented
            result_gts += current_gts_augmented

        return result_images, result_gts

    # returns all images in a list in batch_x format
    def _format_input_images(self, img_list):
        input_arr = np.array(
            [resize(img, (self.input_image_height, self.input_image_width), preserve_range=True) for img in
             img_list]).astype(
            'float32')

        preprocessing = self.model_config.get("preprocessing")
        if preprocessing == "resnet50":
            preprocessed_input = resnet50.preprocess_input(input_arr)
        elif preprocessing == "normalize":
            preprocessed_input = input_arr / 255.
        else:
            preprocessed_input = input_arr
        return preprocessed_input


class AugmentedPatchBinarizationGenerator(AugmentedBinarizationGenerator):
    def _create_augmentation_sequences(self):
        sequences = []

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.9, 1.3),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Fliplr(1.0),
                                         iaa.CropAndPad(percent=(-0.15, 0.15),
                                                        pad_mode=["edge"])
                                         ]))

        sequences.append(iaa.Sequential([iaa.Multiply((0.7, 1.3), per_channel=0.5),
                                         iaa.Fliplr(0.5),
                                         iaa.AverageBlur(k=((2, 5), (2, 5))),
                                         iaa.AdditiveGaussianNoise(scale=(0, 0.02 * 255))
                                         ]))

        sequences.append(iaa.Sequential([iaa.Multiply((0.7, 1.3), per_channel=0.5),
                                         iaa.Flipud(0.5),
                                         iaa.AverageBlur(k=((2, 5), (2, 5))),
                                         iaa.AdditiveGaussianNoise(scale=(0, 0.02 * 255))
                                         ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-30, 30),
            shear=(-15, 15),
            scale=(0.9, 1.15),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-30, 30),
            scale=(0.9, 1.1)),
            iaa.CropAndPad(percent=(-0.2, 0.2),
                           pad_mode=["edge"]),
            iaa.Flipud(1.0)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.85, 1.3)),
            iaa.ContrastNormalization((0.5, 1.5), per_channel=0.5),
            iaa.CropAndPad(percent=(-0.2, 0.2),
                           pad_mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        return sequences
