# https://stackoverflow.com/questions/51843149/loading-batches-of-images-in-keras-from-pandas-dataframe
# https://medium.com/datadriveninvestor/keras-training-on-large-datasets-3e9d9dbc09d4

import time

import imgaug as ia
import numpy as np
from imgaug import augmenters as iaa
from skimage.io import imread
from tensorflow.keras.applications import resnet50

from dataset.dataset_utils import rescale_keypoint, is_none_keypoint, make_none_keypoint
from dataset.heatmap import Heatmap
from dataset.heatmap_sequence import KeypointDataGenerator
from imgproc.imgproc_utils import resize_image_array_pil

""" 
This Keras Sequence implementation delivers image data and corresponding heatmaps
for training. It requires a list of image filenames and the corresponding keypoints as lists:
images: [       'image1.jpg', 'image2.jpg', 'image3.jpg']
keypoints: [    [[100, 100], [-1, -1]], 
                [[200, 100], [120, 90]], 
                [[-1, -1], [20, 10]] ]

The image rgb data gets scaled to 0..1.                
"""


class AugmentedKeypointDataGenerator(KeypointDataGenerator):
    def __init__(self, image_filenames, keypoints, model_config):
        super().__init__(image_filenames, keypoints, model_config)

        ia.seed(int(time.time()))
        self.sequences = self._create_augmentation_sequences()
        self.augmentation_multiplier = len(self.sequences) + 1

        if self.batch_size % self.augmentation_multiplier != 0:
            raise Exception("Batch size should be divisible by augmentation multiplier {}.".format(
                self.augmentation_multiplier))

        self.real_image_count_for_batch = int(self.batch_size / self.augmentation_multiplier)
        return

    def _create_augmentation_sequences(self):
        sequences = []

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Fliplr(0.5),
                                         iaa.CropAndPad(percent=(-0.3, 0.3),
                                                        pad_mode=["edge"])
                                         ]))

        sequences.append(iaa.Sequential([iaa.Multiply((0.7, 1.3), per_channel=0.5),
                                         iaa.Fliplr(0.5),
                                         iaa.AverageBlur(k=((2, 5), (2, 5))),
                                         iaa.AdditiveGaussianNoise(scale=(0, 0.02 * 255))
                                         ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-30, 30),
            shear=(-15, 15),
            scale=(0.9, 1.15),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-30, 30),
            scale=(0.9, 1.1)),
            iaa.CropAndPad(percent=(-0.3, 0.3),
                           pad_mode=["edge"]),
            iaa.Flipud(1.0)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.85, 1.3)),
            iaa.ContrastNormalization((0.5, 1.5), per_channel=0.5),
            iaa.CropAndPad(percent=(-0.3, 0.3),
                           pad_mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.ContrastNormalization((0.5, 1.5)),
            iaa.Fliplr(0.5)
        ]))

        return sequences

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) * self.augmentation_multiplier / float(self.batch_size)))

    # for some versions of keras, this is needed even though we don't use it
    def on_epoch_end(self):
        pass

    def _load_images(self, index):
        my_slice = self.image_filenames[
                   index * self.real_image_count_for_batch: (1 + index) * self.real_image_count_for_batch]
        my_imgs = [imread(img) for img in my_slice]
        return my_imgs

    # returns all keypoints to real images in batch_y format
    def _load_keypoints(self, index):
        return np.array(
            [kp for kp in
             self.keypoints[index * self.real_image_count_for_batch: (1 + index) * self.real_image_count_for_batch]],
            dtype='float64')

    def _augment(self, loaded_images, loaded_keypoints):
        result_images = loaded_images.copy()
        result_keypoints = loaded_keypoints.copy()
        non_keypoints_indices = [(i, j) for i, keypoints_per_img in enumerate(loaded_keypoints) for j, point in
                                 enumerate(keypoints_per_img) if is_none_keypoint(point)]

        for seq in self.sequences:
            seq_det = seq.to_deterministic()
            current_images_augmented = seq_det.augment_images(loaded_images)

            imgaug_keypoints = [ia.KeypointsOnImage([ia.Keypoint(x=point[0], y=point[1])
                                                     for point in keypoints_per_img], shape=image.shape)
                                for image, keypoints_per_img in zip(loaded_images, loaded_keypoints)]

            current_keypoints_augmented = seq_det.augment_keypoints(imgaug_keypoints)
            keypoints_as_array = [keypoints_per_img.to_xy_array() for keypoints_per_img in
                                  current_keypoints_augmented]

            for index in non_keypoints_indices:
                keypoints_as_array[index[0]][index[1]] = make_none_keypoint()

            result_images += current_images_augmented
            result_keypoints = np.concatenate((result_keypoints, keypoints_as_array))
        return result_images, result_keypoints

    def _create_heatmaps_for_image(self, image, keypoints, width, height):
        input_image_h, input_image_w = image.shape[:2]
        keypoints = [rescale_keypoint(point, input_image_w, input_image_h, width, height) for point in keypoints]
        return Heatmap(keypoints, width, height, self.standard_dev).get()

    def _create_heatmaps(self, augmented_images, augmented_keypoints, width, height):
        heatmaps = []
        for datapoint in zip(augmented_images, augmented_keypoints):
            heatmaps.append(self._create_heatmaps_for_image(datapoint[0], datapoint[1], width, height))
        return np.array(heatmaps)

    # returns all images in a list in batch_x format
    def _format_input_images(self, img_list):
        h, w = img_list[0].shape[:2]
        if self.input_img_height == h and \
                self.input_img_width == w:
            input_arr = np.array(
                [img for img in img_list]).astype(
                'float32')
        else:
            input_arr = np.array([resize_image_array_pil(img, (self.input_img_width, self.input_img_height)) for img in
                                  img_list]).astype(
                'float32')

        preprocessing = self.model_config.get("preprocessing")
        if preprocessing == "resnet50":
            preprocessed_input = resnet50.preprocess_input(input_arr)
        elif preprocessing == "normalize":
            preprocessed_input = input_arr / 255.
        else:
            preprocessed_input = input_arr

        return preprocessed_input

    # getitem must return a whole batch
    # this method returns two lists for images and target heatmaps
    def __getitem__(self, index):
        loaded_images = self._load_images(index)
        loaded_keypoints = self._load_keypoints(index)
        augmented_images, augmented_keypoints = self._augment(loaded_images, loaded_keypoints)
        batch_y = self._create_heatmaps(augmented_images, augmented_keypoints, self.heatmap_width, self.heatmap_height)
        batch_x = self._format_input_images(augmented_images)
        return batch_x, batch_y
