import numpy as np
from scipy.ndimage.morphology import distance_transform_edt
from skimage.filters import gaussian
from skimage.transform import resize


class BinaryGroundTruthProcessor:
    def __init__(self, batch_gt, model_config):
        self.batch_gt = batch_gt
        self.model_config = model_config
        return

    def process(self):
        # normalize input
        self.batch_gt = [self._normalize(gt) for gt in self.batch_gt]

        # do filtering
        filter = self._get_filter()
        self.batch_gt = [filter.apply(gt, self.model_config) for gt in self.batch_gt]
        self.batch_gt = [filter.rescale(gt) for gt in self.batch_gt]

        # resize
        self._resize_batch()

        # format for output
        self.batch_gt = self._format()

        assert (isinstance(self.batch_gt, np.ndarray))
        assert (issubclass(self.batch_gt.dtype.type, np.floating))
        assert (self.batch_gt.ndim == 4)

        return self.batch_gt

    def _format(self):
        self.batch_gt = [np.reshape(gt, (gt.shape[0], gt.shape[1], 1)) for gt in self.batch_gt]
        return np.array(self.batch_gt).astype('float32')

    def _get_filter(self):
        if self.model_config.get("gt_filter") == "edt":
            return GroundTruthEDTFilter()
        if self.model_config.get("gt_filter") == "gaussian":
            return GroundTruthGaussianFilter()
        if self.model_config.get("gt_filter") == "none":
            return GroundTruthNoneFilter()

    def _normalize(self, gt):
        return gt / 255.

    def _resize_batch(self):
        heatmap_width = self.model_config.get("heatmap_width")
        heatmap_height = self.model_config.get("heatmap_height")

        self.batch_gt = [resize(gt, (heatmap_height, heatmap_width), preserve_range=True) for gt in self.batch_gt]
        return


# the EDT filter assumes that object pixels are the only nonzero pixels
class GroundTruthEDTFilter:
    def apply(self, gt, model_config):
        # augmented_gts = [self._rescale_gt(gt) for gt in augmented_gts]
        inverted_gt = np.logical_not(gt)
        if 0 not in inverted_gt:
            return np.zeros(gt.shape)
        return distance_transform_edt(np.logical_not(gt))

    # rescaling to 0..1, so that object pixels are 1 and pixel intensity
    # gets smaller the further away the pixels are
    def rescale(self, gt):
        min = np.min(gt)
        max = np.max(gt)
        gt += abs(min)
        if max != 0:
            gt /= max
            gt = 1 - gt
        return gt


class GroundTruthGaussianFilter:
    def apply(self, gt, model_config):
        blur_heatmap = gaussian(gt, sigma=model_config.get("heatmap_standard_dev"))
        max = blur_heatmap.max()
        min = blur_heatmap.min()
        interval = max - min
        factor = 1
        if interval > 0:
            factor = 0.5 / interval
        blur_heatmap += min
        blur_heatmap *= factor

        result = np.maximum(gt, blur_heatmap)
        return result

    def rescale(self, gt):
        return gt


class GroundTruthNoneFilter:
    def apply(self, gt, model_config):
        return gt

    def rescale(self, gt):
        return gt
