# https://stackoverflow.com/questions/51843149/loading-batches-of-images-in-keras-from-pandas-dataframe
# https://medium.com/datadriveninvestor/keras-training-on-large-datasets-3e9d9dbc09d4

import numpy as np
from skimage.io import imread
from skimage.transform import resize
from tensorflow.keras.applications import resnet50
from tensorflow.keras.utils import Sequence

from dataset.dataset_utils import rescale_keypoint
from dataset.heatmap import Heatmap

""" 
This Keras Sequence implementation delivers image data and corresponding heatmaps
for training. It requires a list of image filenames and the corresponding keypoints as lists:
images: [       'image1.jpg', 'image2.jpg', 'image3.jpg']
keypoints: [    [[100, 100], [-1, -1]], 
                [[200, 100], [120, 90]], 
                [[-1, -1], [20, 10]] ]

The image rgb data gets scaled to 0..1.                
"""


class KeypointDataGenerator(Sequence):
    def __init__(self, image_filenames, keypoints, model_config):
        self.model_config = model_config
        self.batch_size = self.model_config.get("batch_size")
        self.input_img_width = self.model_config.get("input_width")
        self.input_img_height = self.model_config.get("input_height")
        self.heatmap_width = self.model_config.get("heatmap_width")
        self.heatmap_height = self.model_config.get("heatmap_height")
        self.standard_dev = self.model_config.get("heatmap_standard_dev")
        self.image_filenames, self.keypoints = image_filenames, keypoints

        return

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    # for some versions of keras, this is needed even though we don't use it
    def on_epoch_end(self):
        pass

    def _load_images(self, index):
        my_slice = self.image_filenames[index * self.batch_size: (1 + index) * self.batch_size]
        my_imgs = [imread(img) for img in my_slice]
        my_shapes = [img.shape[:2] for img in my_imgs]
        result = np.array([resize(img, (self.input_img_height, self.input_img_width), preserve_range=True) for img in
                           my_imgs]).astype('float32')

        preprocessing = self.model_config.get("preprocessing")
        if preprocessing == "resnet50":
            preprocessed_input = resnet50.preprocess_input(result)
        elif preprocessing == "normalize":
            preprocessed_input = result / 255.
        else:
            preprocessed_input = result

        return preprocessed_input, my_shapes

    def _create_heatmaps(self, width, height, index, shapes):
        heatmaps = []
        current_keypoints = np.array(
            [kp for kp in self.keypoints[index * self.batch_size: (1 + index) * self.batch_size]])
        for index, keypoints in enumerate(current_keypoints):
            input_image_h, input_image_w = shapes[index]
            keypoints = [rescale_keypoint(point, input_image_w, input_image_h, width, height) for point in keypoints]
            heatmaps.append(Heatmap(keypoints, width, height, self.standard_dev).get())
        return np.array(heatmaps)

    # getitem must return a whole batch
    # this method returns two lists for images and target heatmaps
    def __getitem__(self, item):
        batch_x, shapes_x = self._load_images(item)
        batch_y = self._create_heatmaps(self.heatmap_width, self.heatmap_height, item, shapes_x)
        return batch_x, batch_y
