# https://stackoverflow.com/questions/51843149/loading-batches-of-images-in-keras-from-pandas-dataframe
# https://medium.com/datadriveninvestor/keras-training-on-large-datasets-3e9d9dbc09d4

import math
import random
import time

import imgaug as ia
import numpy as np
from imgaug import augmenters as iaa
from skimage.io import imread
from skimage.transform import resize
from skimage.util import pad
from tensorflow.keras.applications import resnet50
from tensorflow.keras.utils import Sequence

from dataset.dataset_utils import is_none_keypoint
from dataset.heatmap import Heatmap
from imgproc.patch_extractor import PatchExtractor

""" 
This Keras Sequence implementation delivers image data and corresponding heatmaps
for training. It requires a list of image filenames and the corresponding keypoints as lists:
images: [       'image1.jpg', 'image2.jpg', 'image3.jpg']
keypoints: [    [[100, 100], [-1, -1]], 
                [[200, 100], [120, 90]], 
                [[-1, -1], [20, 10]] ]

The image rgb data gets scaled to 0..1.                
"""


class TipletDataGenerator(Sequence):
    def __init__(self, image_filenames, keypoints, model_config):
        self.model_config = model_config
        self.batch_size = self.model_config.get("batch_size")
        self.input_img_width = self.model_config.get("input_width")
        self.input_img_height = self.model_config.get("input_height")
        self.heatmap_width = self.model_config.get("heatmap_width")
        self.heatmap_height = self.model_config.get("heatmap_height")
        self.standard_dev = self.model_config.get("heatmap_standard_dev")
        self.image_filenames, self.keypoints = image_filenames, keypoints

        ia.seed(int(time.time()))
        self.sequences = self._create_augmentation_sequences()
        self.augmentation_multiplier = len(self.sequences) + 1
        self.nonlet_count = self.augmentation_multiplier

        # virtual tiplets are the image patches per image from disk
        # that arrive at the gpu consisting of augmentations of the
        # real tiplet and of nonlets, i.e. background patches
        self.virtual_tiplet_count = self.nonlet_count + self.augmentation_multiplier

        if self.batch_size % self.virtual_tiplet_count != 0:
            raise Exception("Batch size should be divisible by virtual tiplet count {}.".format(
                self.virtual_tiplet_count))

        self.real_image_count_for_batch = int(self.batch_size / self.virtual_tiplet_count)
        return

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) * self.virtual_tiplet_count / float(self.batch_size)))

    def _create_augmentation_sequences(self):
        sequences = []

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Fliplr(0.5),
                                         iaa.CropAndPad(percent=(-0.3, 0.3),
                                                        pad_mode=["edge"])
                                         ]))

        sequences.append(iaa.Sequential([iaa.Multiply((0.7, 1.3), per_channel=0.5),
                                         iaa.Fliplr(0.5),
                                         iaa.AverageBlur(k=((2, 5), (2, 5))),
                                         iaa.AdditiveGaussianNoise(scale=(0, 0.02 * 255))
                                         ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            shear=(-15, 15),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.CropAndPad(percent=(-0.3, 0.3),
                           pad_mode=["edge"]),
            iaa.Flipud(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.ContrastNormalization((0.5, 1.5), per_channel=0.5),
            iaa.CropAndPad(percent=(-0.3, 0.3),
                           pad_mode=["edge"]),
            iaa.Fliplr(0.5)
        ]))

        sequences.append(iaa.Sequential([iaa.Affine(
            rotate=(-40, 40),
            scale=(0.7, 1.3),
            mode=["edge"]),
            iaa.ContrastNormalization((0.5, 1.5)),
            iaa.Fliplr(0.5)
        ]))

        return sequences

    # for some versions of keras, this is needed even though we don't use it
    def on_epoch_end(self):
        pass

    def _get_image_files_slice(self, index):
        my_slice = self.image_filenames[
                   index * self.real_image_count_for_batch: (1 + index) * self.real_image_count_for_batch]
        return my_slice

    def _load_images(self, index):
        my_slice = self._get_image_files_slice(index)
        my_imgs = [imread(img) for img in my_slice]
        return my_imgs

    # returns all keypoints to real images in batch_y format
    def _load_keypoints(self, index):
        return np.array(
            [kp for kp in
             self.keypoints[index * self.real_image_count_for_batch: (1 + index) * self.real_image_count_for_batch]],
            dtype='float64')

    # returns all images in a list in batch_x format
    def _format_input_images(self, img_list):
        h, w = img_list[0].shape[:2]
        if self.input_img_height == h and \
                self.input_img_width == w:
            input_arr = np.array(
                [img for img in img_list]).astype(
                'float32')
        else:
            input_arr = np.array(
                [resize(img, (self.input_img_height, self.input_img_width), preserve_range=True) for img in
                 img_list]).astype(
                'float32')
        return input_arr

    def _preprocess_input(self, batch_x):
        preprocessing = self.model_config.get("preprocessing")
        if preprocessing == "resnet50":
            preprocessed_input = resnet50.preprocess_input(batch_x)
        elif preprocessing == "normalize":
            preprocessed_input = batch_x / 255.
        else:
            preprocessed_input = batch_x
        return preprocessed_input

    # assumption: we always default to top right center point in case
    # of even patch dimensions.
    def _calc_patch_bounds(self, keypoint):
        half_height = math.floor(self.input_img_height / 2.0)
        half_width = math.floor(self.input_img_width / 2.0)
        y, x = (int(keypoint[0]), int(keypoint[1]))

        uneven_modifier_x = 0
        uneven_modifier_y = 0
        if self.input_img_width % 2 != 0:
            uneven_modifier_x = 1
        if self.input_img_height % 2 != 0:
            uneven_modifier_y = 1

        left_border = x - half_width
        right_border = x + half_width + uneven_modifier_x
        top_border = y - half_height
        bottom_border = y + half_height + uneven_modifier_y
        return int(left_border), int(right_border), int(top_border), int(bottom_border)

    def _create_nonlet_coordinates(self, tip_keypoint, nonlet_count, img_width, img_height):
        left, right, top, bottom = self._calc_patch_bounds(tip_keypoint)
        half_height = self.input_img_height / 2
        half_width = self.input_img_width / 2
        nonlet_coords = []
        for i in range(nonlet_count):
            found = False
            while not found:
                nonlet_x = random.randint(half_width, img_width - half_width)
                nonlet_y = random.randint(half_height, img_height - half_height)
                if (nonlet_x < left - half_width) or (nonlet_x > right + half_width) or \
                        (nonlet_y < top - half_height) or (nonlet_y > bottom + half_height):
                    found = True
                    nonlet_coords.append((nonlet_y, nonlet_x))
        return nonlet_coords

    # If the given image does not have a valid keypoint (not -1, -1)
    # then an additional nonlet is cut instead
    def _cut_patches_from_image(self, image, tip_keypoint):
        half_height = self.input_img_height / 2
        half_width = self.input_img_width / 2
        pad_size = int(max(half_height, half_width))
        img_padded = pad(image, pad_width=[(pad_size, pad_size), (pad_size, pad_size), (0, 0)], mode='constant')
        img_h, img_w = image.shape[:2]
        patch_shape = (self.input_img_height, self.input_img_width)

        if is_none_keypoint(tip_keypoint):
            tiplet = None
            transformed_keypoint = tip_keypoint
        else:
            tiplet, transformed_keypoint = PatchExtractor(img_padded, patch_shape, tip_keypoint).run()

        nonlet_coords = self._create_nonlet_coordinates(tip_keypoint, self.nonlet_count, img_w, img_h)
        nonlets = [PatchExtractor(image, patch_shape, coord).run()[0] for coord in nonlet_coords]

        return tiplet, nonlets, transformed_keypoint

    # we cut the pen tip out of images where it is annotated;
    # we transform they annotation keypoint to fit the patch dimensions
    # and return these pairs in all_tiplets and new_coords
    # we cut nonlets according to nonlet_count out of every real image and return them
    def _cut_all_patches(self, images, keypoints):
        all_tiplets = []
        all_nonlets = []
        new_coords = []
        for image, keypoint in zip(images, keypoints):
            tiplet, nonlets, transformed_keypoint = self._cut_patches_from_image(image, keypoint)
            if tiplet is not None:
                all_tiplets.append(tiplet)
                new_coords.append(transformed_keypoint)
            all_nonlets.extend(nonlets)
        return all_tiplets, all_nonlets, new_coords

    def _augment_tiplets(self, tiplets, keypoints):
        result_tiplets = tiplets.copy()
        result_keypoints = np.array(keypoints.copy())

        non_keypoints_indices = [i for i, point in enumerate(keypoints) if is_none_keypoint(point)]

        if len(result_tiplets) == 0:
            return result_tiplets, result_keypoints

        for seq in self.sequences:
            seq_det = seq.to_deterministic()
            current_images_augmented = seq_det.augment_images(tiplets)

            imgaug_keypoints = [ia.KeypointsOnImage([ia.Keypoint(x=keypoint[1], y=keypoint[0])], shape=tiplet.shape)
                                for tiplet, keypoint in zip(tiplets, keypoints)]

            current_keypoints_augmented = seq_det.augment_keypoints(imgaug_keypoints)
            keypoints_as_array = [keypoints_per_img.to_xy_array()[0] for keypoints_per_img in
                                  current_keypoints_augmented]

            for index in non_keypoints_indices:
                keypoints_as_array[index] = [-1, -1]

            result_tiplets += current_images_augmented
            result_keypoints = np.concatenate((result_keypoints, keypoints_as_array))
        return result_tiplets, result_keypoints

    def _create_heatmaps_dot(self, keypoints):
        heatmaps = []
        for keypoint in keypoints:
            new_map = np.zeros((self.input_img_height, self.input_img_width), dtype='float32')
            if not is_none_keypoint(keypoint):
                new_map[int(keypoint[1]), int(keypoint[0])] = 1
            heatmaps.append(new_map)
        return np.array(heatmaps).astype('float32')

    def _create_heatmaps_gaussian(self, keypoints):
        heatmaps = []
        for keypoint in keypoints:
            heatmaps.append(Heatmap([keypoint], self.input_img_width, self.input_img_height, self.standard_dev).get())
        return heatmaps

    def _substitute_nonlets_for_missing_annotations_in_batch(self, real_image_count, annotated_tiplets_count, nonlets):
        substitute_nonlets = []
        if real_image_count > annotated_tiplets_count:
            none_keypoint_count = real_image_count - annotated_tiplets_count
            for i in range(none_keypoint_count):
                random_nonlet_index = random.randint(0, len(nonlets) - 1)
                substitute_nonlets.append(nonlets[random_nonlet_index])
        return substitute_nonlets

    # getitem must return a whole batch
    # this method returns two lists for images and target heatmaps
    def __getitem__(self, item):
        disk_images = self._load_images(item)
        disk_keypoints = self._load_keypoints(item)
        tip_keypoints = [[keypoint[0][1], keypoint[0][0]] for keypoint in disk_keypoints]

        # from here on we assume keypoint coordinates to be row major
        real_tiplets, nonlets, real_keypoints = self._cut_all_patches(disk_images, tip_keypoints)

        substitute_nonlets = self._substitute_nonlets_for_missing_annotations_in_batch(len(disk_images),
                                                                                       len(real_tiplets), nonlets)

        real_keypoints.extend([[-1, -1] for elem in substitute_nonlets])
        real_tiplets.extend(substitute_nonlets)

        augmented_tiplets, augmented_keypoints = self._augment_tiplets(
            real_tiplets, real_keypoints)

        # imgaug prduces column major keypoint coordinates
        # so from here on keypoint coords are in column major format

        all_patches = augmented_tiplets
        all_patches.add_sample(nonlets)
        all_keypoints = np.concatenate((augmented_keypoints, [[-1, -1] for i in nonlets]))

        batch_x = self._format_input_images(all_patches)
        batch_x = self._preprocess_input(batch_x)
        batch_y = np.array(self._create_heatmaps_gaussian(all_keypoints)).astype('float32')
        batch_y = np.reshape(batch_y, (batch_y.shape[0], self.input_img_height, self.input_img_width, 1))
        return batch_x, batch_y
