import numpy as np
from skimage.transform import estimate_transform

from imgproc.image_processor import ImageProcessor
from util.timer import log_time


class BinaryImageRegisterer(ImageProcessor):
    def __init__(self, first_bin_image: np.ndarray, second_bin_image: np.ndarray):
        super().__init__(processing_strategy=None)
        self.first_bin_image = first_bin_image
        self.second_bin_image = second_bin_image
        return

    @log_time
    def run(self) -> np.ndarray:
        epsilon = 0.1
        max_iterations = 100
        current_transformation = np.identity(3)

        points_source = np.array(np.where(self.first_bin_image > 0))
        points_target = np.array(np.where(self.second_bin_image > 0))

        points_source = self._align_point_centroids(points_source, points_target)
        error = self._calc_error(points_source, points_target)
        iteration_counter = 0

        while error > epsilon and iteration_counter < max_iterations:
            correspondences = self._calc_correspondences(points_source, points_target)
            current_transformation = self._estimate_rigid_transformation(correspondences)
            points_source = self._transform_points(points_source, current_transformation)
            error = self._calc_error(points_source, points_target)
            iteration_counter += 1

        return current_transformation

    def _calc_correspondences(self, points_source, points_target):
        # calc distances between points from source to target
        return

    def _estimate_rigid_transformation(self, correspondences):
        # build matrices with 2 rows and _point_count_ columns
        src = [item[0] for item in correspondences]
        dst = [item[1] for item in correspondences]
        # estimate transform
        tform = estimate_transform('euclidean', src, dst)
        return tform

    def _transform_points(self, points, transformation):
        result = transformation(points)
        return result

    def _calc_error(self, points_source, points_target):
        distances = np.linalg.norm(np.array(points_source) - np.array(points_target), axis=1)
        return distances.mean()

    def _calc_centroid_of_bin(self, points):
        x_centroid = np.mean(points[0])
        y_centroid = np.mean(points[1])
        return np.array((x_centroid, y_centroid))

    def _align_point_centroids(self, points_source, points_target):
        source_centroid = self._calc_centroid_of_bin(points_source)
        target_centroid = self._calc_centroid_of_bin(points_target)
        align_vector = target_centroid - source_centroid
        return points_source - align_vector
