import numpy as np
from skimage import filters
from skimage.color import rgb2gray

from imgproc.image_processor import ImageProcessor
from imgproc.imgproc_utils import resize_image_array_pil
from util.timer import log_time


class HeatmapImageBinarizer(ImageProcessor):
    def __init__(self, heatmap: np.ndarray, image_rgb: np.ndarray, black_threshold=5, otsu=True):
        super().__init__(processing_strategy=None)
        self.heatmap = heatmap
        self.image_rgb = image_rgb
        self.black_threshold = black_threshold
        self.otsu = otsu
        return

    def _paint_black_with_mean_color(self, img):
        cropped_back = np.zeros_like(img)

        mean_r = np.max(img[:, :, 0])
        mean_g = np.max(img[:, :, 1])
        mean_b = np.max(img[:, :, 2])
        cropped_back[:] = (mean_r, mean_g, mean_b)

        all_black_mask = np.all(img < [3, 3, 3], axis=-1)
        non_black_mask = ~all_black_mask
        all_black = np.where(all_black_mask)
        non_black = np.where(non_black_mask)

        mean_color = np.mean(img[non_black])
        img[all_black] = mean_color
        return img

    def _resize_heatmap_to_img(self, img, heatmap):
        img_h, img_w = img.shape[:2]
        resized_heatmap = resize_image_array_pil(heatmap, (img_w, img_h))
        # resized_heatmap = resize(heatmap, (img_h, img_w), order=0, preserve_range=True)
        return resized_heatmap

    def _get_cropped_original_image(self, resized_heatmap, image):
        scaled_heatmap = (resized_heatmap > self.black_threshold)

        non_empty_columns = np.where(scaled_heatmap.max(axis=0) > 0)[0]
        non_empty_rows = np.where(scaled_heatmap.max(axis=1) > 0)[0]

        if len(non_empty_columns) == 0 or len(non_empty_rows) == 0:
            cropBox = (0, scaled_heatmap.shape[0] - 1, 0, scaled_heatmap.shape[1] - 1)
        else:
            cropBox = (min(non_empty_rows), max(non_empty_rows), min(non_empty_columns), max(non_empty_columns))

        scaled_heatmap = np.reshape(scaled_heatmap, (scaled_heatmap.shape[0], scaled_heatmap.shape[1], 1))
        thresholded_image = scaled_heatmap * image

        image_data_new = thresholded_image[cropBox[0]:cropBox[1] + 1, cropBox[2]:cropBox[3] + 1, :]
        return image_data_new, cropBox[0], cropBox[2]

    @log_time
    def run(self) -> np.ndarray:
        resized_heatmap = self._resize_heatmap_to_img(self.image_rgb, self.heatmap)
        cropped_image, top, left = self._get_cropped_original_image(resized_heatmap, self.image_rgb)

        image_with_background_mean = self._paint_black_with_mean_color(cropped_image)

        if self.otsu:
            val = filters.threshold_otsu(image_with_background_mean)
            thresholded_cropped_image = cropped_image < val
            result = np.sum(thresholded_cropped_image, axis=2) > 1
        else:
            result = rgb2gray(image_with_background_mean)

        box_shape = result.shape
        base_img = np.zeros((self.image_rgb.shape[0], self.image_rgb.shape[1]), dtype='float32')
        base_img[top:top + box_shape[0], left:left + box_shape[1]] = np.array(result).astype('float32')

        return base_img  # np.array(result).astype('float32')
