import logging

import numpy as np

from imgproc.image_processor import ImageProcessor
from imgproc.imgproc_utils import resize_image_array_pil
from util.timer import log_time


# assumes black background
# and white foreground for input heatmaps
# black_threshold: removes noise around zero level, so should be kept small
class HeatmapImageTextExtractor(ImageProcessor):
    def __init__(self, heatmap: np.ndarray, image_grayscale: np.ndarray, black_threshold=5):
        super().__init__(processing_strategy=None)
        self.heatmap = heatmap
        self.image_grayscale = image_grayscale
        self.black_threshold = black_threshold
        self.logger.setLevel(logging.WARNING)
        return

    def _paint_black_with_mean_color(self, img):
        all_black = np.where(img < 0.01)
        non_black = np.where(img >= 0.01)

        try:
            max_color = np.max(img[non_black])
        except ValueError:
            max_color = 0

        img[all_black] = max_color
        return img

    def _resize_heatmap_to_img(self, img, heatmap):
        img_h, img_w = img.shape[:2]
        resized_heatmap = resize_image_array_pil(heatmap, (img_w, img_h))
        return resized_heatmap

    def _get_cropped_original_image(self, resized_heatmap, image):
        scaled_heatmap = (resized_heatmap > self.black_threshold)

        non_empty_columns = np.where(scaled_heatmap.max(axis=0) > 0)[0]
        non_empty_rows = np.where(scaled_heatmap.max(axis=1) > 0)[0]

        if len(non_empty_columns) == 0 or len(non_empty_rows) == 0:
            cropBox = (0, scaled_heatmap.shape[0] - 1, 0, scaled_heatmap.shape[1] - 1)
        else:
            cropBox = (min(non_empty_rows), max(non_empty_rows), min(non_empty_columns), max(non_empty_columns))

        scaled_heatmap = np.reshape(scaled_heatmap, (scaled_heatmap.shape[0], scaled_heatmap.shape[1]))
        thresholded_image = scaled_heatmap * image

        image_data_new = thresholded_image[cropBox[0]:cropBox[1] + 1, cropBox[2]:cropBox[3] + 1]
        return image_data_new, cropBox[0], cropBox[2]  ##image_data_new

    def _threshold_to_white(self, image, threshold):
        white_mask = np.where(image > threshold)
        image[white_mask] = 1
        return image

    @log_time
    def run(self) -> np.ndarray:
        resized_heatmap = self._resize_heatmap_to_img(self.image_grayscale, self.heatmap)
        cropped_image, top, left = self._get_cropped_original_image(resized_heatmap, self.image_grayscale)
        # print("in extractor")
        # print(get_ndarray_diagnostics(cropped_image))
        image_with_background_mean = self._paint_black_with_mean_color(cropped_image)
        # print(get_ndarray_diagnostics(image_with_background_mean))
        result = image_with_background_mean

        box_shape = result.shape
        base_img = np.ones((self.image_grayscale.shape[0], self.image_grayscale.shape[1]), dtype='float32')
        base_img[top:top + box_shape[0], left:left + box_shape[1]] = result

        base_img = self._threshold_to_white(base_img, 0.5)
        # print(get_ndarray_diagnostics(base_img))
        return base_img
