import logging
import operator

import numpy as np
from skimage.feature import canny
from skimage.transform import hough_circle, hough_circle_peaks
from skimage.util import pad

from dataset.dataset_utils import make_none_keypoint
from imgproc.image_processor import ImageProcessor
from imgproc.imgproc_utils import demean_heatmap, filter_three_sigma_events
from imgproc.processing_strategy import ProcessingStrategy
from util.timer import log_time


class HeatmapKeypointExtractor(ImageProcessor):
    def __init__(self, heatmap: np.ndarray, processing_strategy: type(ProcessingStrategy)):
        super().__init__(processing_strategy)
        self.heatmap = heatmap
        self.processing_strategy = processing_strategy()
        self.logger.setLevel(logging.WARNING)

    @log_time
    def _extract_peaks(self, image, **kwargs):
        result = self.processing_strategy.process_img(image, **kwargs)
        return result

    def run(self, **kwargs) -> tuple:
        winning_tuple = self._extract_peaks(self.heatmap, **kwargs)
        if np.array_equal(winning_tuple, (0, 0)):
            return make_none_keypoint()
        return winning_tuple

    @staticmethod
    def validate_input(heatmap, radius):
        max = np.max(heatmap)
        min = np.min(heatmap)
        assert (max <= 1.0)
        assert (min >= 0.0)
        shape = heatmap.shape
        assert (len(shape) == 2)
        assert (shape[0] > 1)
        assert (shape[1] > 1)
        assert (radius > 0)
        return


###############################################################################
# Peak extraction strategies
###############################################################################


class WindowedPeakExtraction(ProcessingStrategy):
    def _get_mean_list_of_windows_around_coords(self, image, window_size, coords, sorted=True):

        def get_window(coord, array, windowsize):
            a = coord[0]
            b = coord[1]
            window = array[a - windowsize:a + windowsize,
                     b - windowsize:b + windowsize]
            return window

        window_mean_list = []

        for pair in coords:
            window = get_window(pair, image, window_size)
            window_mean = np.mean(window)
            window_mean_list.append((window_mean, pair))

        if sorted:
            window_mean_list.sort(key=operator.itemgetter(0))
        return window_mean_list

    def process_img(self, image, window_size=2):
        float_heatmap = image.astype('float32')

        demeaned_heatmap = demean_heatmap(float_heatmap)
        filtered_heatmap = filter_three_sigma_events(demeaned_heatmap)

        padded_heatmap = pad(filtered_heatmap, pad_width=[(window_size, window_size), (window_size, window_size)],
                             mode='constant')

        mask_nonzero = np.where(padded_heatmap > 0)
        coords_nonzero = list(zip(mask_nonzero[0], mask_nonzero[1]))

        window_mean_list = self._get_mean_list_of_windows_around_coords(padded_heatmap, window_size, coords_nonzero)
        winning_tuple = window_mean_list[-1][1]
        winning_tuple = (winning_tuple[0] - window_size, winning_tuple[1] - window_size)
        return winning_tuple


class MaxFilterPeakExtraction(ProcessingStrategy):
    def process_img(self, image, radius=5):
        # data_max = filters.maximum_filter(image, radius)
        # maxima = np.where(image == data_max)
        i, j = np.unravel_index(image.argmax(), image.shape)
        return (i, j)


# TODO: take care of correctly identified nonlet
class HoughCirclesPeakExtraction(ProcessingStrategy):
    def process_img(self, image, radius=5):
        edges = canny(image, sigma=1, low_threshold=0.2, high_threshold=1.0)
        hough_radii = [radius]
        hough_res = hough_circle(edges, hough_radii)
        # Select the most prominent 3 circles
        accums, cx, cy, radii = hough_circle_peaks(hough_res, hough_radii,
                                                   total_num_peaks=1)
        if cx and cy:
            result = (cy, cx)
        else:
            result = (0, 0)
        return result


class CentroidPeakExtraction(ProcessingStrategy):
    def process_img(self, image, **kwargs):
        float_heatmap = image.astype('float32')

        demeaned_heatmap = demean_heatmap(float_heatmap)
        filtered_heatmap = filter_three_sigma_events(demeaned_heatmap)

        mask_nonzero = np.where(filtered_heatmap > 0)
        coords_nonzero = list(zip(mask_nonzero[0], mask_nonzero[1]))

        x_mean = 0
        y_mean = 0
        for coord in coords_nonzero:
            y_mean += coord[0] * float_heatmap[coord] / 255.0
            x_mean += coord[1] * float_heatmap[coord] / 255.0

        x_mean /= filtered_heatmap.shape[1]
        y_mean /= filtered_heatmap.shape[0]
        return (y_mean, x_mean)
