import io

import cv2
import numpy as np
from PIL import Image
from skimage.exposure import rescale_intensity


def demean_heatmap(heatmap):
    inf_mean = np.mean(heatmap)
    demeaned_heatmap = heatmap - inf_mean
    return demeaned_heatmap


def filter_three_sigma_events(heatmap):
    heatmap_in_sigmas = heatmap / np.std(heatmap)
    mask_3sigma = np.where(heatmap_in_sigmas < 3.0)
    heatmap_in_sigmas[mask_3sigma] = 0
    return heatmap_in_sigmas


def rescale_heatmap(heatmap):
    inf_max = np.max(heatmap)
    # inf_min = np.min(heatmap)
    rescaled_heatmap = heatmap / inf_max * 255.
    return rescaled_heatmap


def decompress_image_cv2(image) -> np.ndarray:
    image = cv2.imdecode(np.fromstring(image, dtype='uint8'), -1)
    return image


def decompress_image_pil(encoded_image) -> np.ndarray:
    image = Image.open(io.BytesIO(encoded_image))
    return np.array(image)


def compress_image(image, quality, method='pil'):
    image_as_raw = None
    if method == 'pil':
        if type(image) is Image:
            image_as_raw = compress_image_pil(image, quality)
        elif type(image) is np.ndarray:
            image_as_raw = compress_image_array_pil(image, quality)
    elif method == 'cv2':
        image_as_raw = compress_image_cv2(image, quality)
    else:
        raise ValueError("Unknown compression method.")
    return image_as_raw


def compress_image_cv2(image, quality):
    uint8_output = rescale_intensity(image, in_range='image', out_range='uint8')
    uint8_output = np.array(uint8_output, dtype=np.uint8)
    ret_code, image_as_raw = cv2.imencode(
        ".jpg", uint8_output, [int(cv2.IMWRITE_JPEG_QUALITY), quality])
    return image_as_raw


def compress_image_pil(image: Image, quality=95) -> bytes:
    output = io.BytesIO()
    image.save(output, format="JPEG", quality=quality, subsampling=0)
    image_as_raw = output.getvalue()
    return image_as_raw


# from the pil doc:
# quality
# The image quality, on a scale from 1 (worst) to 95 (best).
# The default is 75. Values above 95 should be avoided;
# 100 disables portions of the JPEG compression algorithm,
# and results in large files with hardly any gain in image quality.
def compress_image_array_pil(image: np.ndarray, quality=95) -> bytes:
    if len(image.shape) < 3:
        pil_img = Image.fromarray(image, mode='L')
    else:
        pil_img = Image.fromarray(image, mode='RGB')
    output = io.BytesIO()
    pil_img.save(output, format="JPEG", quality=quality, subsampling=0)
    image_as_raw = output.getvalue()
    return image_as_raw


# needs the shape as column-major
def resize_image_array_pil(image: np.ndarray, target_shape: tuple, algo=Image.LANCZOS) -> np.ndarray:
    pil_img = Image.fromarray(image)
    pil_img = pil_img.resize(target_shape, algo)
    return np.array(pil_img)


def resize_image_pil(image: Image, target_shape: tuple) -> Image:
    pil_img = image.resize(target_shape, Image.LANCZOS)
    return pil_img


# expects coord in row major
def draw_marker_on_image(image, coord, radius=2, thickness=2, color=(255, 255, 255)):
    cv2.circle(image, (coord[1], coord[0]), 2, color, thickness=2)
    return image


# wants row major coordinates
def copy_patch_on_image(patch, image, location):
    p_h, p_w = patch.shape[:2]
    left = int(location[1] - ((p_w / 2) - 1))
    right = int(location[1] + ((p_w / 2) + 1))
    top = int(location[0] - p_h / 2)
    bottom = int(location[0] + p_h / 2)

    p_left = 0
    p_top = 0
    p_right = p_w
    p_bottom = p_h

    if left < 0:
        p_left = abs(left)
        left = 0
    if top < 0:
        p_top = abs(top)
        top = 0
    if right > image.shape[1]:
        p_right = p_w - (right - image.shape[1])
        right = image.shape[1]
    if bottom > image.shape[0]:
        p_bottom = p_h - (bottom - image.shape[0])
        bottom = image.shape[0]

    image[top:bottom, left:right] = patch[p_top:p_bottom, p_left:p_right]
    return image


def merge_patch_and_image_on_value(patch, image, location):
    # copy_image = image.copy()
    # copy_image = copy_patch_on_image(patch, copy_image, location)#

    # mask_image_intensity = np.where(image == intensity)
    # mask_copy_image = np.where(copy_image != intensity)
    # final_mask = np.where((mask_copy_image != intensity) & (mask_image_intensity == intensity))
    # image[final_mask] = copy_image[final_mask]
    p_h, p_w = patch.shape[:2]
    left = int(location[1] - ((p_w / 2) - 1))
    right = int(location[1] + ((p_w / 2) + 1))
    top = int(location[0] - p_h / 2)
    bottom = int(location[0] + p_h / 2)

    p_left = 0
    p_top = 0
    p_right = p_w
    p_bottom = p_h

    if left < 0:
        p_left = abs(left)
        left = 0
    if top < 0:
        p_top = abs(top)
        top = 0
    if right > image.shape[1]:
        p_right = p_w - (right - image.shape[1])
        right = image.shape[1]
    if bottom > image.shape[0]:
        p_bottom = p_h - (bottom - image.shape[0])
        bottom = image.shape[0]

    store = (image[top:bottom, left:right]).copy()
    patch_copy = (patch[p_top:p_bottom, p_left:p_right]).copy()

    # copy only over elements that are white
    # mask = np.where(store > 100)
    mask = np.where(patch_copy < 50)
    store[mask] = patch_copy[mask]

    whitening_mask = np.where(store > 100)
    store[whitening_mask] = 255

    image[top:bottom, left:right] = store

    return image


# wants row major coordinates
def copy_patch_on_image_age(patch, image, image_age, location, current_age, max_age):
    p_h, p_w = patch.shape[:2]
    left = int(location[1] - ((p_w / 2) - 1))
    right = int(location[1] + ((p_w / 2) + 1))
    top = int(location[0] - p_h / 2)
    bottom = int(location[0] + p_h / 2)

    p_left = 0
    p_top = 0
    p_right = p_w
    p_bottom = p_h

    if left < 0:
        p_left = abs(left)
        left = 0
    if top < 0:
        p_top = abs(top)
        top = 0
    if right > image.shape[1]:
        p_right = p_w - (right - image.shape[1])
        right = image.shape[1]
    if bottom > image.shape[0]:
        p_bottom = p_h - (bottom - image.shape[0])
        bottom = image.shape[0]

    my_image_patch = np.copy(image[top:bottom, left:right])
    image_black_mask = my_image_patch < 30

    my_copy_patch = np.copy(patch[p_top:p_bottom, p_left:p_right])
    patch_black_mask = my_copy_patch < 30

    max_age = current_age - max_age
    my_age_patch = np.copy(image_age[top:bottom, left:right])

    # get old black pixels that are not to be changed
    # under any circumstances
    old_no_change = my_age_patch < max_age
    old_black_pixels_no_change = old_no_change & image_black_mask
    new_pixels = np.logical_not(old_black_pixels_no_change)

    # now we have young pixels
    # we only want to update the age when a young black pixel changes color
    # or when a white pixel is written
    update_age_pixels_previous_white = new_pixels & \
                                       (np.logical_not(image_black_mask))
    update_age_pixels_previous_black = new_pixels & image_black_mask

    update_age_pixels = ((update_age_pixels_previous_black & np.logical_not(
        patch_black_mask)) | update_age_pixels_previous_white)

    my_age_patch[update_age_pixels] = current_age
    image_age[top:bottom, left:right] = my_age_patch

    # we update all young pixels
    border = 30
    center_ones = np.zeros_like(my_copy_patch)
    center_ones[border:-border, border:-border] = 1
    center_mask = center_ones == 1

    centered_new_pixels = new_pixels & center_mask

    my_image_patch[centered_new_pixels] = my_copy_patch[centered_new_pixels]
    image[top:bottom, left:right] = my_image_patch
    return image
