import math

import numpy as np
from skimage.util import pad

from dataset.dataset_utils import is_none_keypoint
from imgproc.image_processor import ImageProcessor


class PatchExtractor(ImageProcessor):
    '''
    Extracts a patch with given shape and coordinates from an image.
    and fills parts outside the image boundaries with zeroes.

    Always assumes the coordinate as the center of the patch.

    - Image is assumed to be RGB
    - Coords need to be row-major.
    - patch shape is a tuple (row-major)
    - offset can be used to translate the center of the patch away from the coords

    return the patch and the coordinates relative to the patch upper left corner
    '''

    def __init__(self, image: np.ndarray, patch_shape: tuple, coords: np.ndarray, offset: tuple = (0, 0), channels=3,
                 do_pad=False):
        super().__init__(None)
        self.image = image
        self.patch_shape = patch_shape
        self.offset = offset
        self.coords = coords
        self.do_pad = do_pad
        self.channels = channels

    # assumption: we always default to top right center point in case
    # of even patch dimensions.
    # assumption: offset is negative for values left/top of keypoint
    # and positive for values right/below keypoint
    def _calc_patch_bounds(self, keypoint: np.ndarray):
        half_height = math.floor(self.patch_shape[0] / 2.0)
        half_width = math.floor(self.patch_shape[1] / 2.0)
        y, x = (int(keypoint[0]), int(keypoint[1]))

        uneven_modifier_x = 0
        uneven_modifier_y = 0
        if self.patch_shape[1] % 2 != 0:
            uneven_modifier_x = 1
        if self.patch_shape[0] % 2 != 0:
            uneven_modifier_y = 1

        if self.offset == (0, 0):
            left_border = x - half_width
            right_border = x + half_width + uneven_modifier_x
            top_border = y - half_height
            bottom_border = y + half_height + uneven_modifier_y
        else:
            assert (self.offset[0] < half_height)
            assert (self.offset[1] < half_width)

            left_border = x - half_width - self.offset[1]
            right_border = x + half_width - self.offset[1] + uneven_modifier_x
            top_border = y - half_height - self.offset[0]
            bottom_border = y + half_height - self.offset[0] + uneven_modifier_y
        return int(left_border), int(right_border), int(top_border), int(bottom_border)

    # assumption: our image is padded at the borders, so keypoints
    # at the border do not lead to out of bounds problems
    def _cut_patch_at_coord(self, image, keypoint, pad_size=(0, 0)):
        if self.channels == 1:
            patch = np.zeros((self.patch_shape[0], self.patch_shape[1]), dtype='uint8')
        else:
            patch = np.zeros((self.patch_shape[0], self.patch_shape[1], 3), dtype='uint8')

        if is_none_keypoint(keypoint):
            return patch, keypoint

        left, right, top, bottom = self._calc_patch_bounds(keypoint)

        assert (right - left == self.patch_shape[1])
        assert (bottom - top == self.patch_shape[0])

        if self.channels == 1:
            patch[:, :] = image[top + pad_size[0]:bottom + pad_size[0], left + pad_size[1]:right + pad_size[1]]
        else:
            patch[:, :, :] = image[top + pad_size[0]:bottom + pad_size[0], left + pad_size[1]:right + pad_size[1], :]
        y, x = (int(keypoint[0]), int(keypoint[1]))
        new_coord = [x - left, y - top]
        return patch, new_coord

    # in: row-major tip
    # out: patch and new column-major keypoint coordinate
    def run(self):
        if is_none_keypoint(self.coords):
            return np.zeros(self.patch_shape), self.coords
        if self.do_pad:
            half_width = int(self.patch_shape[1])
            half_height = int(self.patch_shape[0])
            padded_image = pad(
                self.image,
                pad_width=[(half_height, half_height), (half_width, half_width), (0, 0)],
                mode='constant')

            patch, new_coord = self._cut_patch_at_coord(
                padded_image,
                self.coords,
                pad_size=(half_height, half_width))
        else:
            padded_image = self.image
            patch, new_coord = self._cut_patch_at_coord(padded_image, self.coords, pad_size=(0, 0))
        return patch, new_coord
