import math

import numpy as np
from scipy.interpolate import interp1d

from imgproc.path_extraction.path_distances_extractor import PathDistancesExtractor
from imgproc.path_extraction.path_from_keypoint_extractor import PathFromKeypointExtractor


class HeatmapPath():
    '''
    heatmap batch is a batch of uint8 1-channel heatmaps with the shape
    (x, heatmap_height, heatmap_width)

    image_y: flip y coordinates so plots on a regular y axis look correct
    '''

    def __init__(self, heatmap_batch, filter_threshold=20):
        self.path_points = np.array(PathFromKeypointExtractor(heatmap_batch).run())
        self.tips_y, self.tips_x = self._split_into_x_y(*self.path_points.T)
        self.tips_x, self.tips_y = self._filter_outliers_from_tip_stream(self.tips_x, self.tips_y, filter_threshold)
        self.filtered_points = np.array(list(zip(self.tips_x, self.tips_y)))

    ###########################################################################

    def get_interpolated_points(self):
        if not self.interpolated_points:
            self.interpolated_points = self._create_interpolated_points()
        return self.interpolated_points

    def get_filtered_points(self, flip_y=True, row_major=False):
        my_tips_y = self.tips_y
        if flip_y:
            my_tips_y = 224 - self.tips_y
        if row_major:
            return np.array(list(zip(my_tips_y, self.tips_x)))
        return np.array(list(zip(self.tips_x, my_tips_y)))

    ###########################################################################

    def _filter_outliers_from_tip_stream(self, tips_x, tips_y, threshold):
        my_coords = list(zip(tips_x, tips_y))
        prev_coord = None
        recorded_indices = []

        for index, coord in enumerate(my_coords):
            if index > 0:
                dist = math.sqrt((coord[0] - prev_coord[0]) ** 2 + (coord[1] - prev_coord[1]) ** 2)
                # dist = np.linalg.norm(np.array(my_coords)-np.array(prev_coord))
                if dist > threshold:
                    recorded_indices.append(index)
                else:
                    prev_coord = coord
            else:
                prev_coord = coord

        # for elem in sorted(recorded_indices, reverse=True):
        #    del tips_x[elem]
        #    del tips_y[elem]
        tips_x = np.delete(tips_x, recorded_indices)
        tips_y = np.delete(tips_y, recorded_indices)

        return tips_x, tips_y

    def _split_into_x_y(self, tips_x, tips_y):
        return tips_x, tips_y

    def _create_interpolated_points(self):
        distances, unique_points = PathDistancesExtractor(self.filtered_points, cumulative=True, normalized=True).run()
        interpolations_methods = ['slinear']  # 'slinear', 'quadratic', 'cubic'
        alpha = np.linspace(0, 1, 1200)
        interpolated_points = {}
        for method in interpolations_methods:
            interpolator = interp1d(distances, unique_points, kind=method, axis=0)
            interpolated_points[method] = interpolator(alpha)
        return interpolated_points
