import numpy as np

from imgproc.image_processor import ImageProcessor
from imgproc.path_extraction.path_extraction_utils import move_points_by_gradient
from imgproc.path_extraction.point_gradient_calculator import PointGradientCalculator


class PathGapFixer(ImageProcessor):
    '''
    points: column-major

    returns all points modified for gap fixing and new on indices
    '''

    def __init__(self, path_points: np.ndarray, on_indices: np.ndarray, gradient_image: np.ndarray, iterations=2,
                 gap_limit=3, on_threshold=0.5):
        super().__init__(None)
        self.path_points = path_points
        self.on_indices = on_indices
        self.gap_limit = gap_limit
        self.gradient_image = gradient_image
        self.on_threshold = on_threshold
        self.iterations = iterations

    def _move_and_check_gap_points(self, start_point_index, gap_points):
        new_on_indices = []
        # calc gradients for gap points
        gradients = PointGradientCalculator(gap_points, self.gradient_image).run()
        # move points by gradient
        gap_points = move_points_by_gradient(gap_points, gradients)
        for index, point in enumerate(gap_points):
            if self.gradient_image[int(point[1]), int(point[0])] < self.on_threshold:
                new_on_indices.append(start_point_index + index)
        return new_on_indices

    def _update_path_points_with_gap_fillers(self, ):
        return

    def _fix_gaps(self):
        gap_list = np.diff(self.on_indices, axis=0)
        gap_start_indices = np.where((gap_list > 1) & (gap_list <= self.gap_limit))

        new_on_indices = []

        if gap_start_indices is not None:
            for gap_start_index in np.nditer(gap_start_indices):
                start_points_index = self.on_indices[gap_start_index]
                gap_points = self.path_points[start_points_index:start_points_index + gap_list[gap_start_index]]
                new_gap_indices = self._move_and_check_gap_points(start_points_index, gap_points)
                new_on_indices.extend(new_gap_indices)

        new_on_indices.sort()
        return np.array(new_on_indices)

    def run(self):
        for i in range(self.iterations):
            new_indices = self._fix_gaps()
            self.on_indices = np.concatenate((self.on_indices, new_indices))
            self.on_indices.sort()

        return self.path_points, self.on_indices
