import numpy as np
from skimage.color import rgb2gray
from skimage.filters import gaussian
from skimage.io import imread

from dataset_utils import rescale_keypoint
from imgproc.heatmap_image_text_extractor import HeatmapImageTextExtractor
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction
from imgproc.image_processor import ImageProcessor
from imgproc.imgproc_utils import copy_patch_on_image_age, copy_patch_on_image
from imgproc.patch_by_path_masker import PatchByPathMasker
from imgproc.path_extraction.on_off_classifier import OnOffClassifier
from imgproc.path_extraction.path_beautifier import PathBeautifier
from imgproc.path_extraction.path_stroke_splitter import PathStrokeSplitter
from imgproc.path_extraction.path_subdivider import PathSubdivider
from imgproc.path_extraction.point_path import PointPath
from prototype.camera_streamer import PrototypeRecordingStreamer

param_dict = {
    "heatmap_size": (211, 211),
    "patch_size": (224, 224),
    "surface_size": (1080, 1920),

    "max_kp_heatmap_mean": 50,
    "max_bin_heatmap_mean": 50,
    "min_bin_heatmap_std": 7,

    "path_distance_filter_threshold": 150,
    "do_subdivision": True,
    "aggregate_blur_sigma": 2,
    "beautify_path_steps": 5,
    "on_point_max_threshold": 200,
    "min_split_gap": 3,
    "use_blurred_aggregate_for_classification": False,

    "bin_heatmap_black_noise_threshold": 5,
    "aggregate_type": "static",
    "aggregate_path": "",
    "aggregate_path_mask_stroke_width": 10,
    "use_aggregate_aging": True,
    "aggregate_max_age": 5
}


class StreamStrokeExtractor(ImageProcessor):
    def __init__(self, dir, start_id, end_id, bad_ids, param_dict):
        super().__init__(None)
        self.start_id = start_id
        self.end_id = end_id
        self.bad_ids = bad_ids
        self.param_dict = param_dict

        self.patch_size = param_dict["patch_size"]
        self.surface_size = param_dict["surface_size"]
        self.heatmap_size = param_dict["heatmap_size"]

        # column major unfiltered peaks
        self.raw_peaks = []
        self.raw_peak_frame_ids = []

        self.stream = PrototypeRecordingStreamer(start_id, end_id, bad_ids, dir)

        self.use_incremental_aggregate = bool(param_dict["use_incremental_aggregate"])
        self.aggregated_image = self._prepare_aggregate()
        self.aggregated_age_image = np.ones(self.surface_size, dtype='float32')
        self.current_age = 1
        return

    def _prepare_aggregate(self):
        if self.param_dict["aggregate_type"] == "static":
            if "aggregate_path" in self.param_dict:
                return rgb2gray(imread(self.param_dict["aggregate_path"])) * 255
            else:
                raise ValueError
        else:
            return np.ones(self.surface_size, dtype='float32') * 255

    # wants row major peak and patch location
    # gives column major peak
    def _transform_patch_peak(self, peak, patch_location):
        left_patch_coord = int(patch_location[1] - self.patch_size[1] / 2)
        top_patch_coord = int(patch_location[0] - self.patch_size[0] / 2)

        # column major peak with patch transform
        peak_transformed = (peak[0] + top_patch_coord, peak[1] + left_patch_coord)
        return peak_transformed

    def _extract_path(self, points, aggregated_image):
        # we create a path from the keypoint heatmaps
        path = PointPath(points)

        # we retrieve our filtered path points
        filtered_points, filtered_frame_ids = path.get(
            filtered=True,
            threshold=self.param_dict["path_distance_filter_threshold"],
            coord_order='column',
            marks=self.raw_peak_frame_ids,
            filter_duplicate_points=self.param_dict["filter_duplicate_points"])

        if self.param_dict["do_subdivision"]:
            filtered_points = np.array(PathSubdivider([filtered_points]).run()[0])

        blurred_aggregate = gaussian(aggregated_image, sigma=self.param_dict["aggregate_blur_sigma"])

        # we perform a gradient descent on our nice extracted text to move the
        # path points closer to it
        nicer_filtered_points = PathBeautifier(
            filtered_points,
            [],
            blurred_aggregate,
            iterations=self.param_dict["beautify_path_steps"]).run()

        aggregate_for_classification = aggregated_image
        if self.param_dict["use_blurred_aggregate_for_classification"]:
            aggregate_for_classification = blurred_aggregate

        # then we classify the pen tip positions into ink and non ink
        on_points, on_indices = OnOffClassifier(
            nicer_filtered_points,
            aggregate_for_classification,
            mean_threshold=1000,
            inclusion_threshold=self.param_dict["on_point_max_threshold"]).run()

        strokes = PathStrokeSplitter(
            nicer_filtered_points,
            on_indices,
            split_gap_size=self.param_dict["min_split_gap"]).run()

        on_frame_ids = []
        off_frame_ids = []

        all_indices = list(range(0, len(nicer_filtered_points)))
        off_indices = [index for index in all_indices if index not in on_indices]
        off_points = np.take(nicer_filtered_points, off_indices, axis=0)

        if not self.param_dict["do_subdivision"]:
            # if we dont subdivide, we can simply take on_indices into frame_ids
            if len(on_indices) > 0:
                on_frame_ids = np.take(filtered_frame_ids, on_indices, axis=0)
            if len(off_indices) > 0:
                off_frame_ids = np.take(filtered_frame_ids, off_indices, axis=0)

        return strokes, on_points, off_points, on_frame_ids, off_frame_ids

    def _update_dynamic_aggregate(self, aggregate, current_bin, current_patch, location):
        if len(self.raw_peaks) == 0:
            return aggregate

        current_filter_points, _ = PointPath(np.array(self.raw_peaks)).get(
            filtered=True,
            threshold=self.param_dict["path_distance_filter_threshold"],
            coord_order='column')

        current_extract_image = HeatmapImageTextExtractor(
            current_bin,
            rgb2gray(current_patch),
            black_threshold=self.param_dict["bin_heatmap_black_noise_threshold"]).run()

        # MASK EXTRACTED IMAGE BY PATH
        current_extract_image = PatchByPathMasker(
            current_extract_image,
            current_filter_points,
            location,
            stroke_width=self.param_dict["aggregate_path_mask_stroke_width"]).run()

        if self.param_dict["use_aggregate_aging"]:
            aggregated_image = copy_patch_on_image_age(
                current_extract_image * 255,
                aggregate,
                self.aggregated_age_image,
                location,
                self.current_age,
                max_age=self.param_dict["aggregate_max_age"],
                border=0)
            self.current_age += 1
        else:
            aggregated_image = copy_patch_on_image(
                current_extract_image * 255,
                aggregate,
                location)

        return aggregated_image

    def process_working_set(self, keypoint_heatmap, bin_heatmap, patch, location, frame_id):
        if keypoint_heatmap is None or bin_heatmap is None or patch is None:
            return

        # DISCARD
        if np.mean(keypoint_heatmap) > self.param_dict["max_kp_heatmap_mean"]:
            return

        bad_bin_heatmap = False
        if np.mean(bin_heatmap) > self.param_dict["max_bin_heatmap_mean"] \
                or np.std(bin_heatmap) < self.param_dict["min_bin_heatmap_std"]:
            bad_bin_heatmap = True

        # EXTRACT CURRENT PEAK
        peak = HeatmapKeypointExtractor(keypoint_heatmap, MaxFilterPeakExtraction).run()
        peak = rescale_keypoint(peak, *self.heatmap_size, *self.patch_size)

        transformed_peak = self._transform_patch_peak(peak, location)
        if transformed_peak[0] >= 0 and transformed_peak[0] < self.surface_size[0]:
            if transformed_peak[1] >= 0 and transformed_peak[1] < self.surface_size[1]:
                self.raw_peaks.append(transformed_peak)
                self.raw_peak_frame_ids.append(frame_id)
            else:
                print("Skipped Frame {}".format(frame_id))
        else:
            print("Skipped Frame {}".format(frame_id))

        if self.param_dict["aggregate_type"] == "dynamic":
            if not bad_bin_heatmap:
                self.aggregated_image = self._update_dynamic_aggregate(
                    self.aggregated_image,
                    bin_heatmap,
                    patch,
                    location)
        return

    def run(self):
        counter = 0
        interval = int(self.param_dict["incremental_aggregate_interval"])

        def is_in_on_or_off_ids(id):
            return (id in result_on_frame_ids) or (id in result_off_frame_ids)

        # take care of incremental aggregate preparation
        if self.use_incremental_aggregate:
            result_strokes = []
            result_on_points = []
            result_off_points = []
            result_on_frame_ids = []
            result_off_frame_ids = []

        current_kp, current_bin, current_patch, location, frame_id = self.stream.get()
        while current_patch is not None:
            # get raw peaks
            self.process_working_set(current_kp, current_bin, current_patch, location, frame_id)

            if self.use_incremental_aggregate:
                if counter % interval == 0 \
                        and counter > 0:
                    tmp_strokes, tmp_on_points, tmp_off_points, tmp_on_frame_ids, tmp_off_frame_ids = self._extract_path(
                        self.raw_peaks,
                        self.aggregated_image)

                    on_points_to_add = [point for point, id in zip(tmp_on_points, tmp_on_frame_ids) if
                                        not is_in_on_or_off_ids(id)]
                    on_ids_to_add = [id for point, id in zip(tmp_on_points, tmp_on_frame_ids) if
                                     not is_in_on_or_off_ids(id)]
                    off_points_to_add = [point for point, id in zip(tmp_off_points, tmp_off_frame_ids) if
                                         not is_in_on_or_off_ids(id)]
                    off_ids_to_add = [id for point, id in zip(tmp_off_points, tmp_off_frame_ids) if
                                      not is_in_on_or_off_ids(id)]

                    result_on_points.extend(on_points_to_add)
                    result_on_frame_ids.extend(on_ids_to_add)
                    result_off_points.extend(off_points_to_add)
                    result_off_frame_ids.extend(off_ids_to_add)

            current_kp, current_bin, current_patch, location, frame_id = self.stream.get()
            counter += 1

        if self.use_incremental_aggregate:
            return result_strokes, np.array(result_on_points), np.array(
                result_off_points), result_on_frame_ids, result_off_frame_ids, self.aggregated_image
        else:
            # extract strokes from collected peaks once, with the complete aggregate
            strokes, on_points, off_points, on_frame_ids, off_frame_ids = self._extract_path(self.raw_peaks,
                                                                                             self.aggregated_image)
            return strokes, on_points, off_points, on_frame_ids, off_frame_ids, self.aggregated_image
