from model.deconv_resnet import KeypointResnet
from model.depth_to_space_resnet_decoder import DepthToSpaceResnetDecoder
from model.detail_deconv_resnet import DetailKeypointResnet
from model.patch_network import PatchNetwork
from model.tiplet_network.tiplet_network import TipletNetwork


class ModelFactory:
    def create_model(self, model_config):
        model_type = model_config.get("net_type")
        if model_type == "keypoint_resnet":
            return KeypointResnetBuilder().get(model_config)
        elif model_type == "detail_keypoint_resnet":
            return DetailKeypointResnetBuilder().get(model_config)
        elif model_type == "depth_to_space_resnet":
            return DepthToSpaceResnetDecoderBuilder().get(model_config)
        elif model_type == "tiplet":
            return TipletBuilder().get(model_config)
        elif model_type == "patch":
            return PatchNetworkBuilder().get(model_config)
        else:
            raise ValueError


class KeypointResnetBuilder:
    def get(self, model_config):
        include_top = True
        input_shape = None

        if model_config.get("input_height") != 224 or model_config.get("input_width") != 224:
            include_top = False
            input_shape = (model_config.get("input_height"), model_config.get("input_width"), 3)

        my_model = KeypointResnet(model_config.get("keypoint_count"),
                                  model_config.get("multi_gpu_count"),
                                  verbose=model_config.get("verbose"),
                                  freeze_resnet=model_config.get("do_freeze_resnet"),
                                  include_top=include_top,
                                  input_shape=input_shape)
        return my_model


class DetailKeypointResnetBuilder:
    def get(self, model_config):
        include_top = True
        input_shape = None

        if model_config.get("input_height") != 224 or model_config.get("input_width") != 224:
            include_top = False
            input_shape = (model_config.get("input_height"), model_config.get("input_width"), 3)

        my_model = DetailKeypointResnet(model_config.get("keypoint_count"),
                                        model_config.get("multi_gpu_count"),
                                        verbose=model_config.get("verbose"),
                                        freeze_resnet=model_config.get("do_freeze_resnet"),
                                        include_top=include_top,
                                        input_shape=input_shape)
        return my_model


class DepthToSpaceResnetDecoderBuilder:
    def get(self, model_config):
        include_top = True
        input_shape = None

        if model_config.get("input_height") != 224 or model_config.get("input_width") != 224:
            include_top = False
            input_shape = (model_config.get("input_height"), model_config.get("input_width"), 3)

        my_model = DepthToSpaceResnetDecoder(model_config.get("keypoint_count"),
                                             model_config.get("multi_gpu_count"),
                                             verbose=model_config.get("verbose"),
                                             freeze_resnet=model_config.get("do_freeze_resnet"),
                                             include_top=include_top,
                                             input_shape=input_shape)
        return my_model


class TipletBuilder:
    def get(self, model_config):
        include_top = True
        input_shape = None

        input_shape = (model_config.get("input_height"), model_config.get("input_width"), 3)

        my_model = TipletNetwork(model_config.get("keypoint_count"),
                                 model_config.get("multi_gpu_count"),
                                 verbose=model_config.get("verbose"),
                                 freeze_resnet=model_config.get("do_freeze_resnet"),
                                 include_top=True,
                                 input_shape=input_shape)
        return my_model


class PatchNetworkBuilder:
    def get(self, model_config):
        include_top = True
        input_shape = None

        input_shape = (model_config.get("input_height"), model_config.get("input_width"), 3)

        my_model = PatchNetwork(model_config.get("keypoint_count"),
                                model_config.get("multi_gpu_count"),
                                verbose=model_config.get("verbose"),
                                freeze_resnet=model_config.get("do_freeze_resnet"),
                                include_top=True,
                                input_shape=input_shape,
                                use_distributed_strategy=True)
        return my_model
