from tensorflow.keras.layers import Conv2DTranspose, BatchNormalization, Activation, Conv2D
from tensorflow.keras.models import Model
from tensorflow.keras.utils import multi_gpu_model

from model.tiplet_network.resnet_generator import resnet_v1


class TipletNetwork:
    def __init__(self, num_keypoints, multi_gpu, verbose=False, freeze_resnet=False,
                 include_top=True, input_shape=None):
        self.num_keypoints = num_keypoints
        self.multi_gpu = multi_gpu

        self.resnet_model = resnet_v1(input_shape, 20)

        if freeze_resnet:
            for layer in self.resnet_model.layers:
                layer.trainable = False

        self.keypoint_model = self._create_model()
        if verbose:
            self.keypoint_model.summary()
        return

    def _create_model(self):
        c5_layer = self.resnet_model.get_layer('activation_19')

        # TODO: check if residual addition is needed - apparently residual addition
        # is only performed in the resnet part in msra implementation

        # 3 deconv layers mit BN und ReLu, kernel size 4, stride 2 und 256 channels
        # BN layers haben laut Simple Baseline Repo momentum = 0.9, default in keras ist 0.99
        deconv1 = Conv2DTranspose(64, (2, 2), strides=(2, 2))(c5_layer.output)
        deconv1_bn = BatchNormalization(momentum=0.9)(deconv1)
        deconv1_bn = Activation('relu')(deconv1_bn)
        deconv2 = Conv2DTranspose(64, (1, 1), strides=(1, 1))(deconv1_bn)
        deconv2_bn = BatchNormalization(momentum=0.9)(deconv2)
        deconv2_bn = Activation('relu')(deconv2_bn)
        deconv3 = Conv2DTranspose(32, (2, 2), strides=(2, 2))(deconv2_bn)
        deconv3_bn = BatchNormalization(momentum=0.9)(deconv3)
        deconv3_bn = Activation('relu')(deconv3_bn)
        deconv4 = Conv2DTranspose(32, (1, 1), strides=(1, 1))(deconv3_bn)
        deconv4_bn = BatchNormalization(momentum=0.9)(deconv4)
        deconv4_bn = Activation('relu')(deconv4_bn)

        # last conv layer 1x1 with output for keypoint predictions
        final_kernel_size = (1, 1)
        last_conv = Conv2D(self.num_keypoints, final_kernel_size, strides=(1, 1))(deconv4_bn)

        my_model = Model(self.resnet_model.input, last_conv)

        if self.multi_gpu > 0:
            my_model = multi_gpu_model(my_model, gpus=self.multi_gpu)

        my_model.compile(loss='mean_squared_error', optimizer='adam', metrics=['mse'])
        return my_model

    def predict(self, image_batch):
        predictions = self.keypoint_model.predict(image_batch)
        return predictions

    def get_model(self):
        return self.keypoint_model


if __name__ == '__main__':
    tiplet_net = TipletNetwork(1, 0, verbose=True, input_shape=(32, 32, 3))
