# see here: https://towardsdatascience.com/from-raw-images-to-real-time-predictions-with-deep-learning-ddbbda1be0e4
import os

import cv2
import numpy as np

from dataset_utils import get_frame_id_patch_location_from_patch_filename
from visual.image_heatmap_file_data_provider import ImageHeatmapFileDataProvider, SelectiveImageProvider


class CameraStreamer:
    def __init__(self):
        return

    def get_width_height(self):
        return

    def get(self):
        return


class BRIOCameraStreamer(CameraStreamer):
    def __init__(self, no_decoding=False):
        super().__init__()
        self.capture = cv2.VideoCapture()
        self.capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)  # probably storing only one frame? check if it works
        # self.capture.open(1 + cv2.CAP_DSHOW)
        self.capture.open(0)
        self.capture.set(cv2.CAP_PROP_MODE, 1)  # for HW mode
        # self.capture.set(cv2.VIDEOWRITER_PROP_QUALITY, 100)

        if no_decoding:
            self.capture.set(cv2.CAP_PROP_CONVERT_RGB, 0)
        # No conversion because we send the image over a network and then
        # its useless to decode, then encode, then decode again

        self._check_capture()
        self._configure_capture()

        print(self.capture.get(cv2.CAP_PROP_FRAME_WIDTH))
        print(self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        print(self.capture.get(cv2.CAP_PROP_FPS))

    def _check_capture(self):
        if not self.capture.isOpened():
            print("Couldn't open camera capture.")
            raise IOError
        return

    def _configure_capture(self):
        # self._configure_2k_30_mjpg()
        # self._configure_1080p_30_yuyv()
        self._configure_1080p_60_mjpg()
        # self._configure_720p_90_mjpg()
        return

    def _configure_1080p_30_yuyv(self):
        fourcc = cv2.VideoWriter_fourcc('Y', 'U', 'Y', 'V')
        self.capture.set(cv2.CAP_PROP_FOURCC, fourcc)
        self.capture.set(cv2.CAP_PROP_FPS, 30)
        self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, 1920)
        self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 1080)

        return

    def _configure_720p_90_mjpg(self):
        fourcc = cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')
        self.capture.set(cv2.CAP_PROP_FOURCC, fourcc)
        self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, 1280)
        self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 720)
        self.capture.set(cv2.CAP_PROP_FPS, 90)
        return

    def _configure_1080p_60_mjpg(self):
        fourcc = cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')
        self.capture.set(cv2.CAP_PROP_FOURCC, fourcc)
        self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, 1920)
        self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 1080)
        self.capture.set(cv2.CAP_PROP_FPS, 60)
        return

    def _configure_4k_30_mjpg(self):
        fourcc = cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')
        self.capture.set(cv2.CAP_PROP_FOURCC, fourcc)
        self.capture.set(cv2.CAP_PROP_FPS, 30)
        self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, 3840)
        self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 2160)
        return

    def _configure_2k_30_mjpg(self):
        fourcc = cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')
        self.capture.set(cv2.CAP_PROP_FOURCC, fourcc)
        self.capture.set(cv2.CAP_PROP_FPS, 30)
        self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, 2560)
        self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 1440)
        self.capture.set(cv2.CAP_PROP_ZOOM, 150)
        return

    def get(self):
        ret, frame = self.capture.read()
        if not ret:
            print("No frame received")
            return np.zeros_like(frame)
        return cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    def __del__(self):
        self.capture.release()


# a cyclic image file streamer
class DummyCameraStreamer(CameraStreamer):
    def __init__(self, dir, counter=0):
        super().__init__()
        self.provider = ImageHeatmapFileDataProvider(dir)
        self.counter = counter
        return

    def get_current_image_name(self):
        if self.counter - 1 < len(self.provider):
            path = self.provider.image_names[self.counter - 1]
            return os.path.split(path)[1]
        return ""

    def get_width_height(self):
        h, w = self.provider[0][0].shape[1:3]
        return w, h

    def get(self):
        if self.counter < len(self.provider):
            result = self.provider[self.counter]
            self.counter += 1
            if len(result[0].shape) == 4:
                return result[0][0, :, :, :]
            else:
                return result[0][0, :, :]
        else:
            return None
            self.counter = 0
            if len(self.provider[self.counter][0].shape) == 4:
                return self.provider[self.counter][0][0, :, :, :]
            else:
                return self.provider[self.counter][0][0, :, :]


# files need to have the frame id in the name
class SelectedFileImageStreamer(CameraStreamer):
    def __init__(self, dir, start_frame, end_frame, skip_frames):
        super().__init__()
        self.provider = SelectiveImageProvider(dir, start_frame, end_frame, skip_frames)
        self.counter = 0
        self.last_returned_image_name = ""

    def get_last_returned_image_name(self):
        return os.path.split(self.last_returned_image_name)[1]

    def get(self):
        if self.counter < len(self.provider):
            result = self.provider[self.counter]
            self.last_returned_image_name = self.provider.needed_filenames[self.counter]
            self.counter += 1
            if len(result.shape) == 4:
                return result[0][:, :, :]
            else:
                return result[0][:, :]
        else:
            return None


class PrototypeRecordingStreamer():
    def __init__(self, start_frame, end_frame, bad_frames, directory):
        self.kp_stream = SelectedFileImageStreamer(directory + "/kp/", start_frame, end_frame, bad_frames)
        self.bin_stream = SelectedFileImageStreamer(directory + "/bin/", start_frame, end_frame, bad_frames)
        self.patch_stream = SelectedFileImageStreamer(directory + "/patch/", start_frame, end_frame, bad_frames)
        self.counter = 0
        return

    def get(self):
        current_kp = self.kp_stream.get()
        current_bin = self.bin_stream.get()
        current_patch = self.patch_stream.get()

        patch_image_name = self.patch_stream.get_last_returned_image_name()
        frame_id, location = get_frame_id_patch_location_from_patch_filename(patch_image_name)

        bin_image_name = self.bin_stream.get_last_returned_image_name()
        bin_split_name = bin_image_name.split('_')
        bin_frame_id = int(bin_split_name[0])

        kp_image_name = self.kp_stream.get_last_returned_image_name()
        kp_split_name = kp_image_name.split('_')
        kp_frame_id = int(kp_split_name[0])

        assert (frame_id == bin_frame_id == kp_frame_id)

        return current_kp, current_bin, current_patch, location, frame_id


class VideoFrameStreamer(CameraStreamer):
    def __init__(self, filename):
        super().__init__()
        self.filename = filename
        self.capture = cv2.VideoCapture()
        self.capture.open(filename)
        self._check_capture()

    def _check_capture(self):
        if not self.capture.isOpened():
            print("Couldn't open camera capture.")
            raise IOError
        return

    def get(self):
        ret, frame = self.capture.read()
        if not ret:
            print("No frame received")
            return np.zeros_like(frame)
        return cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    def get_idx(self, index):
        self.capture.set(cv2.CAP_PROP_POS_FRAMES, index)
        return self.get()

    def get_frame_count(self):
        return self.capture.get(cv2.CAP_PROP_FRAME_COUNT)

    def __del__(self):
        self.capture.release()
