import socket

from dataset.dataset_utils import get_files_and_keypoints
from dataset.heatmap import Heatmap
from model.model_factory import ModelFactory
from prototype.camera_streamer import DummyCameraStreamer


class InferenceProvider:
    def __init__(self, inference_strategy):
        self.inference_strategy = inference_strategy
        return

    def get(self, image):
        return self.inference_strategy.run_inference(image)


class InferenceStrategy:
    def __init__(self):
        return

    def _prepare_inferrer(self):
        return

    def run_inference(self, input):
        return


class DummyKeypointInferenceStrategy(InferenceStrategy):
    def __init__(self, keypoint_json):
        super().__init__()
        self.files, self.annotations = get_files_and_keypoints(keypoint_json)
        self.keypoint_counter = 0
        return

    def _prepare_inferrer(self):
        return

    def run_inference(self, input):
        if self.keypoint_counter < len(self.annotations):
            current_annotation = self.annotations[self.keypoint_counter]
            self.keypoint_counter += 1
        else:
            self.keypoint_counter = 0
            current_annotation = self.annotations[self.keypoint_counter]

        # create heatmap
        h, w = input.shape[:2]
        heatmap_size = 238
        factor_w = heatmap_size / w
        factor_h = heatmap_size / h
        current_annotation[0][0] *= factor_w
        current_annotation[1][0] *= factor_w
        current_annotation[0][1] *= factor_h
        current_annotation[1][1] *= factor_h

        heatmap = Heatmap(current_annotation, 238, 238, 0.25)
        return heatmap.get()


# Dummy inference strategy returns images from folder sequentially with cycling
class DummyBinarizationInferenceStrategy(InferenceStrategy):
    def __init__(self, dir):
        super().__init__()
        self.provider = DummyCameraStreamer(dir)
        return

    def _prepare_inferrer(self):
        return

    def run_inference(self, input):
        return self.provider.get()


class LocalInferenceStrategy(InferenceStrategy):
    def __init__(self, model_config, model_weights):
        super().__init__()
        self.model_config = model_config
        self.model_weights = model_weights
        self.model = self._prepare_inferrer()
        return

    def _prepare_inferrer(self):
        model_factory = ModelFactory()
        my_resnet = model_factory.create_model(self.model_config)
        model = my_resnet.get_model()
        # print("Loading model weights ...")
        model.load_weights(self.model_weights)
        return model

    def run_inference(self, input):
        output = self.model.predict(input)
        return output


# asynchronous nature of interfacing with remote client_server
# adds large amount of complexity
class RemoteInferenceStrategy(InferenceStrategy):
    def __init__(self, address, model_config, model_weights):
        super().__init__()
        self.address = address
        self.model_config = model_config
        self.model_weights = model_weights
        self.model = self._prepare_inferrer()
        return

    def _prepare_inferrer(self):
        self.host = "localhost"
        self.port = 31337
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((self.host, self.port))
        # connect to address
        # exchange handshake
        return None

    def _listen(self):
        self.socket.listen(1)
        conn, addr = self.socket.accept()
        print("Connection from {}".format(addr))
        while not self.done:
            data = conn.recv(4096)
            pose_data = pickle.loads(data)
            self.pose_buffer.append(pose_data)
            if not data:
                break
            conn.send(data)
        conn.close()

    def run_inference(self, input):
        # upload image
        # wait for response
        # handle callback
        # return inference
        return
