import numpy as np
from skimage import filters
from skimage.transform import resize


class TextBinarizer:
    def __init__(self, heatmap_threshold):
        self.heatmap_threshold = heatmap_threshold
        return

    def _binarize_heatmap(self, heatmap):
        return heatmap > self.heatmap_threshold

    def _resize_heatmap_to_img(self, img, heatmap):
        img_h, img_w = img.shape[:2]
        resized_heatmap = resize(heatmap, (img_h, img_w), order=0, preserve_range=True)
        return resized_heatmap

    def _get_cropped_original_image(self, resized_heatmap, image):
        scaled_heatmap = self._binarize_heatmap(resized_heatmap)

        non_empty_columns = np.where(scaled_heatmap.max(axis=0) > 0)[0]
        non_empty_rows = np.where(scaled_heatmap.max(axis=1) > 0)[0]

        if len(non_empty_columns) == 0 or len(non_empty_rows) == 0:
            cropBox = (0, scaled_heatmap.shape[0] - 1, 0, scaled_heatmap.shape[1] - 1)
        else:
            cropBox = (min(non_empty_rows), max(non_empty_rows), min(non_empty_columns), max(non_empty_columns))

        image_data_new = image[cropBox[0]:cropBox[1] + 1, cropBox[2]:cropBox[3] + 1, :]
        return image_data_new

    def mask_image(self, binary_mask, image):
        binary_mask = np.reshape(binary_mask, (binary_mask.shape[0], binary_mask.shape[1], 1))
        masked_image = binary_mask * image
        return masked_image

    def _fill_background_with_mean_color(self, image):
        all_black_mask = np.all(image < [3, 3, 3], axis=-1)
        non_black_mask = ~all_black_mask
        all_black = np.where(all_black_mask)
        non_black = np.where(non_black_mask)

        mean_color = np.mean(image[non_black])
        image[all_black] = mean_color
        return image

    def binarize(self, image, heatmap):
        resized_heatmap = self._resize_heatmap_to_img(image, heatmap)
        binary_mask = self._binarize_heatmap(resized_heatmap)
        masked_image = self.mask_image(binary_mask, image)
        colored_mask_image = self._fill_background_with_mean_color(masked_image)

        threshold = filters.threshold_otsu(colored_mask_image)

        thresholded_masked_image = colored_mask_image < threshold
        binary_image = np.sum(thresholded_masked_image, axis=2) > 1

        return binary_image
