import datetime
import json

from sklearn.model_selection import train_test_split
from tensorflow.keras.callbacks import ModelCheckpoint, LearningRateScheduler

from model.model_factory import ModelFactory


# For one Training process
class Trainer(object):
    def __init__(self, input_params, train_params):
        self.image_filenames = input_params["image_filenames"]
        self.annotations = input_params["annotations"]
        self.pretrained_weights = input_params["pretrained_weights"]

        self.test_annotations = None
        self.test_image_filenames = None

        self.use_test_folder = False

        if "test_image_filenames" in input_params and \
                "test_annotations" in input_params:
            self.test_image_filenames = input_params["test_image_filenames"]
            self.test_annotations = input_params["test_annotations"]
            self.use_test_folder = True

        self.model_config = train_params
        pass

    # adapted from https://keras.io/examples/cifar10_resnet/
    def _lr_schedule(self, epoch):
        lr = self.model_config.get("lr_schedule")[0]['lr']
        for elem in self.model_config.get("lr_schedule"):
            if epoch > elem['epoch']:
                lr = elem['lr']

        print('Learning rate: ', float(lr))
        return float(lr)

    def _create_sequence(self, x_in, y_out, augmentation=False):
        sequence = None
        return sequence

    def _handle_pretrained_weights(self, model):
        if self.pretrained_weights:
            model.load_weights(self.pretrained_weights)
        return

    def _create_callbacks(self):
        name = self.model_config.get("name")
        # Prepare callbacks for model saving and for learning rate adjustment.
        # from https://keras.io/examples/cifar10_resnet/
        checkpoint = ModelCheckpoint(filepath=name + '.{epoch:03d}.h5',
                                     monitor='val_loss',
                                     verbose=1,
                                     save_best_only=True)

        lr_scheduler = LearningRateScheduler(self._lr_schedule)
        return [checkpoint, lr_scheduler]

    def train(self):
        x_train = self.image_filenames
        y_train = self.annotations
        x_test = None
        y_test = None

        test_sequence = None
        train_sequence = None

        if len(x_train) == 0:
            print("There are no image files available. Exiting.")
            return

        print("Image files available: {} jpgs".format(len(x_train)))
        print("Annotations available: {}".format(len(y_train)))

        if self.model_config.get("validation_split") > 0 and not self.use_test_folder:
            print("Splitting datapoints into train and val ...")
            x_train, x_test, y_train, y_test = train_test_split(
                self.image_filenames,
                self.annotations,
                test_size=self.model_config.get("validation_split"))
            print("Creating Test Sequence ...")
            test_sequence = self._create_sequence(x_test, y_test)

            print("Train size -- x: {}, y: {}".format(len(x_train), len(y_train)))
            print("Test size -- x: {}, y: {}".format(len(x_test), len(y_test)))

        elif self.use_test_folder:
            x_test = self.test_image_filenames
            y_test = self.test_annotations

            print("Creating Test Sequence ...")
            test_sequence = self._create_sequence(x_test, y_test)

            print("Train size -- x: {}, y: {}".format(len(x_train), len(y_train)))
            print("Test size -- x: {}, y: {}".format(len(x_test), len(y_test)))

        print("Creating Training Sequence ...")
        train_sequence = self._create_sequence(x_train, y_train, self.model_config.get("do_augmentation"))

        print("Creating Model ...")
        model_factory = ModelFactory()
        my_resnet = model_factory.create_model(self.model_config)
        model = my_resnet.get_model()
        self._handle_pretrained_weights(model)

        print("Starting Training ...")

        # callbacks = self._create_callbacks()
        checkpoint = ModelCheckpoint(filepath="patchbin" + '.{epoch:03d}.h5',
                                     monitor='val_loss',
                                     verbose=1,
                                     save_best_only=True)

        lr_scheduler = LearningRateScheduler(self._lr_schedule)
        callbacks = [checkpoint, lr_scheduler]

        history = model.fit_generator(train_sequence,
                                      validation_data=test_sequence,
                                      steps_per_epoch=None,
                                      epochs=self.model_config.get("epoch_count"),
                                      use_multiprocessing=False,
                                      verbose=1,
                                      shuffle=True,
                                      callbacks=callbacks)

        self._finish(model, history)
        return

    def _finish(self, model, history):
        print("Saving model weights ...")
        my_time = datetime.datetime.now().strftime("%H_%M_%S-%d_%m_%y")
        model.save_weights("patchbin_ckpt_{}_{}.h5".format(self.model_config.get("name"), my_time))
        model.save("patchbin_sm_{}_{}.h5".format(self.model_config.get("name"), my_time), save_format='tf')

        print("Saving history ...")
        with open('history_{}_{}.json'.format(self.model_config.get("name"), my_time), 'w') as json_file:
            json.dump(history.history, json_file)

        json_file.close()
