import os
import random
from ast import literal_eval

import numpy as np
import pandas as pd
from skimage.io import imsave, imread

from dataset.dataset_utils import get_files_and_keypoints, is_none_keypoint, make_none_keypoint, rescale_keypoint
from dataset.tiplet_sequence import TipletDataGenerator
from imgproc.imgproc_utils import resize_image_array_pil
from imgproc.patch_extractor import PatchExtractor
from model.model_config import ModelConfig


class ExtractorSequence(TipletDataGenerator):
    def _create_augmentation_sequences(self):
        return []


# downscaling_4k: if True, we assume that patches will be cut from 4k images,
# but the network should be trained for HD images. So we cut the patches double size
# and downscale them accordingly, because we want to train as if all training images
# were recorded in HD.
class DatasetPatchExtractor():
    def __init__(self, target_dir, model_config, downscaling_4k=False, batch_size=8, border=0.1):
        self.target_dir = target_dir
        self.model_config = model_config
        self.downscaling_4k = downscaling_4k

        if downscaling_4k:
            self.model_config.train_params["input_width"] *= 2
            self.model_config.train_params["input_height"] *= 2

        self.border_x_in_px = int(self.model_config.get("input_width") * border)
        self.border_y_in_px = int(self.model_config.get("input_height") * border)
        self.image_filenames = None
        self.keypoints = None
        self.batch_size = batch_size
        self.batch_count = 0
        self.extracted_dataframe = None
        return

    def run(self):
        self._load_image_filenames_and_keypoints()
        self.batch_count = int(np.ceil(len(self.image_filenames) / float(self.batch_size)))
        self.extracted_dataframe = self._store_tiplet_sequence()
        return self.extracted_dataframe

    def _load_image_filenames_and_keypoints(self):
        return

    def _get_image_files_slice(self, index):
        my_slice = self.image_filenames[
                   index * self.batch_size: (1 + index) * self.batch_size]
        return my_slice

    def _load_images(self, index):
        my_slice = self._get_image_files_slice(index)
        my_imgs = [imread(img) for img in my_slice]
        return my_imgs, my_slice

    # returns all keypoints to real images in batch_y format
    def _load_keypoints(self, index):
        return np.array(
            [kp for kp in
             self.keypoints[index * self.batch_size: (1 + index) * self.batch_size]],
            dtype='float64')

    def _cut_tip_patch(self, image, keypoint):
        patch_shape = (self.model_config.get("input_height"),
                       self.model_config.get("input_width"))
        x_range = patch_shape[1] / 2 - self.border_x_in_px
        y_range = patch_shape[0] / 2 - self.border_y_in_px
        offset_x = random.uniform(-x_range, x_range)
        offset_y = random.uniform(-y_range, y_range)
        offset = (int(offset_y), int(offset_x))
        patch, new_keypoint = PatchExtractor(
            image,
            patch_shape,
            keypoint,
            offset=offset,
            do_pad=True).run()
        return patch, new_keypoint

    def _cut_none_patch(self, image):
        patch_shape = (self.model_config.get("input_height"),
                       self.model_config.get("input_width"))
        h, w = image.shape[:2]
        new_keypoint = np.array((
            int(random.uniform(patch_shape[1] / 2, h - patch_shape[1] / 2)),
            int(random.uniform(patch_shape[0] / 2, w - patch_shape[0] / 2))
        ))
        patch, new_keypoint = PatchExtractor(
            image,
            patch_shape,
            new_keypoint,
            offset=(0, 0),
            do_pad=True).run()
        return patch, make_none_keypoint()

    def _cut_batch_patches(self, index):
        disk_images, image_names = self._load_images(index)
        disk_keypoints = self._load_keypoints(index)
        row_major_keypoints = [[kp[1], kp[0]] for kp in disk_keypoints]
        patches = []
        new_coords = []

        for image, keypoint in zip(disk_images, row_major_keypoints):
            if not is_none_keypoint(keypoint):
                patch, new_coord = self._cut_tip_patch(image, keypoint)
            else:
                patch, new_coord = self._cut_none_patch(image)
            patches.append(patch)
            new_coords.append(new_coord)
        return patches, image_names, new_coords

    def _store_tiplet_sequence(self):
        half_width = int(self.model_config.get("input_width") / 2)
        half_height = int(self.model_config.get("input_height") / 2)
        dataframe = pd.DataFrame(columns=['img_name', 'ptip'])
        for counter in range(self.batch_count):
            patches, image_names, new_coords = self._cut_batch_patches(counter)
            for patch, name, coord in zip(patches, image_names, new_coords):
                if self.downscaling_4k:
                    patch = resize_image_array_pil(patch, (half_width, half_height))
                name = os.path.split(name)[1]
                if os.path.isfile(os.path.join(self.target_dir, "{}.jpg".format(name))):
                    raise IOError
                imsave(os.path.join(self.target_dir, "{}".format(name)), patch, quality=100)

            if self.downscaling_4k:
                new_coords = [rescale_keypoint(kp, half_width * 2, half_height * 2, half_width, half_height) for kp in
                              new_coords]

            dataframe = dataframe.append(pd.DataFrame(list(zip(image_names, new_coords)), columns=['img_name', 'ptip']),
                                         ignore_index=True)
            print(dataframe.head())
            print(dataframe.tail())
        return dataframe


class JsonDatasetPatchExtractor(DatasetPatchExtractor):
    def __init__(self, target_dir, model_config, annotation_json, dataset_dir, downscaling_4k=False):
        super().__init__(target_dir, model_config, downscaling_4k)
        self.annotation_json = annotation_json
        self.dataset_dir = dataset_dir

    def _load_image_filenames_and_keypoints(self):
        self.image_filenames, self.keypoints = get_files_and_keypoints(self.annotation_json,
                                                                       dataset_dir=self.dataset_dir)
        return


class DataFrameDatasetPatchExtractor(DatasetPatchExtractor):
    def __init__(self, target_dir, model_config, dataframe, downscaling_4k=False):
        super().__init__(target_dir, model_config, downscaling_4k)
        self.dataframe = dataframe

    def _load_image_filenames_and_keypoints(self):
        # select all images that have ptip != [-1, -1]
        # self.dataframe['ptip_yes'] = self.dataframe.ptip.map(lambda x: x != [-1, -1])
        # only_ptips = self.dataframe.loc[self.dataframe['ptip_yes'] == True]
        self.image_filenames = self.dataframe['full_path'].to_list()
        self.keypoints = self.dataframe['ptip'].to_list()
        return


if __name__ == '__main__':
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)

    do_json = False

    # target_dir = "/home/IW_student/work/tiplet_results/first_round_test_no_pretrain/"
    # yaml = "/home/IW_student/stalwart-keypoint-alpaca/saved_models/tiplet_no_pretraining_first_round_1080p_32/tiplet.yaml"
    target_dir = "../tmp_out/patches_224/"
    yaml = "../train/params/tiplet_224.yaml"

    config = ModelConfig(yaml)

    if do_json:
        annotation_json = "/home/IW_student/work/first_round_subset/test_1080p/first_round_test.json"
        dataset_dir = "/home/IW_student/work/first_round_subset/test_1080p/"
        extractor = JsonDatasetPatchExtractor(target_dir, config, annotation_json, dataset_dir)
    else:
        all_annos = pd.read_csv('final_consistent_annos.csv', index_col=0,
                                dtype={'subject': np.int32,
                                       'task': np.int32,
                                       'frame': np.int32},
                                converters={
                                    'ptip': literal_eval,
                                    'ptail': literal_eval})
        extractor = DataFrameDatasetPatchExtractor(target_dir, config, all_annos, downscaling_4k=True)
        dataframe = extractor.run()
        dataframe.to_csv("patches_224_keypoints.csv")
