import numpy as np
import pandas as pd

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
import matplotlib.style as style
from util.select_and_extract_keypoint_annotations import split_image_name
from ast import literal_eval


def get_subject(name):
    subject, task, frame = split_image_name(name)
    return subject


def get_task(name):
    subject, task, frame = split_image_name(name)
    return task


def get_frame(name):
    subject, task, frame = split_image_name(name)
    return frame


all_keypoint_files = ["L:/large_dataset_study/extracted_frames/keypoints/keypoints_0_1_2_3_4_5_9_11.json",
                      "L:/large_dataset_study/extracted_frames/keypoints/json_unit_6_0-2000/keypoints_unit_6_all.json",
                      "L:/large_dataset_study/extracted_frames/keypoints/json_unit_7_0-2000/keypoints_unit_7_all.json",
                      "L:/large_dataset_study/extracted_frames/keypoints/json_unit_10_0-1000/json_unit_10_0-1000.json",
                      "L:/large_dataset_study/extracted_frames/keypoints/keypoints_unit_12_partial.json",
                      "L:/large_dataset_study/first_round_subset/train/first_round_train.json",
                      "L:/large_dataset_study/first_round_subset/test/first_round_test.json",
                      "L:/large_dataset_study/second_round_subset/train/second_round_train.json",
                      "L:/large_dataset_study/second_round_subset/test/second_round_test.json"]


def create_dataframe_from_json_list(json_list):
    result = pd.DataFrame()
    for file in json_list:
        df = pd.read_json(file).T
        result = result.append(df)
    result = result.drop('dataset_dir')
    return result


# why use hexbins?
# https://pro.arcgis.com/de/pro-app/tool-reference/spatial-statistics/h-whyhexagons.htm
# Birch, Colin P.D., Oom, Sander P., and Beecham, Jonathan A.
# Rectangular and hexagonal grids used for observation, experiment, and simulation in ecology.
# Ecological Modelling, Vol. 206, No. 3–4. (August 2007), S. 347–359.
def create_hexbin_coord_plot_from_dataframe(dataframe, column):
    coords_frame = pd.DataFrame(dataframe[column].tolist(), columns=['x', 'y'])
    dff_crit = coords_frame['x'] != -1
    coords_frame = coords_frame[dff_crit]
    ax = coords_frame.plot.hexbin(x='x', y='y', gridsize=20, cmap='viridis', vmax=180)
    return ax


def style_hexbin_ax(ax):
    ax.set_facecolor(plt.cm.viridis(0))
    ax.grid(b=False)
    ax.set_xlim(xmin=0, xmax=3840)
    ax.set_ylim(ymin=0, ymax=2160)
    return ax


def save_fig(fig, filename):
    fig.set_size_inches(8, 6)
    plt.savefig(filename, dpi=300)
    return


def get_subject_task_groups(df):
    # get dataframe consisting only of distinct subject and task pairs
    sf = df[['subject', 'task', 'ptip_none', 'ptail_none', 'ptip_only', 'ptail_only']]

    sf = df.groupby(['subject', 'task']).agg(
        samples=pd.NamedAgg(column='frame', aggfunc='count'),
        none_tips=pd.NamedAgg(column='ptip_none', aggfunc='sum'),
        yes_tips=pd.NamedAgg(column='ptip_yes', aggfunc='sum'),
        only_tips=pd.NamedAgg(column='ptip_only', aggfunc='sum'),
        none_tails=pd.NamedAgg(column='ptail_none', aggfunc='sum'),
        yes_tails=pd.NamedAgg(column='ptail_yes', aggfunc='sum'),
        only_tails=pd.NamedAgg(column='ptail_only', aggfunc='sum'),
        both_none=pd.NamedAgg(column='both_none', aggfunc='sum'),
        both_present=pd.NamedAgg(column='both_present', aggfunc='sum'))
    return sf


def stackplot_datapoints(df):
    sf = get_subject_task_groups(df)
    sum_for_subjects = sf.groupby('subject')[['both_present', 'only_tips', 'only_tails', 'both_none']].sum()
    ax_stacked = sum_for_subjects[['both_present', 'only_tips', 'only_tails', 'both_none']].plot(kind='area')
    return ax_stacked


def create_stats_columns(df):
    df['ptip_yes'] = df.ptip.map(lambda x: x != [-1, -1])
    df['ptip_none'] = df.ptip.map(lambda x: x == [-1, -1])
    df['ptail_yes'] = df.ptail.map(lambda x: x != [-1, -1])
    df['ptail_none'] = df.ptail.map(lambda x: x == [-1, -1])
    df['both_none'] = np.where(df['ptip_none'] & df['ptail_none'], True, False)
    df['both_present'] = np.where(~df['ptip_none'] & ~df['ptail_none'], True, False)
    df['ptip_only'] = np.where(df['ptip_yes'] & df['ptail_none'], True, False)
    df['ptail_only'] = np.where(df['ptip_none'] & df['ptail_yes'], True, False)
    return df


def get_keypoints_dataframe(json_files, statistics=True):
    df = create_dataframe_from_json_list(json_files)

    df.columns = ['ptip', 'ptail']
    df['subject'] = df.index.map(get_subject)
    df['task'] = df.index.map(get_task)
    df['frame'] = df.index.map(get_frame)

    if statistics:
        df = create_stats_columns(df)

    df = df.loc[~df.index.duplicated(keep='first')]
    return df


if __name__ is "__main__":
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)
    style.use('ggplot')

    # df = get_keypoints_dataframe(all_keypoint_files, statistics=False)

    ########################################################################
    # df now ready for use
    ########################################################################

    new_todos_df = pd.read_csv('final_consistent_annos.csv', index_col=0,
                               dtype={'subject': np.int32,
                                      'task': np.int32,
                                      'frame': np.int32},
                               converters={
                                   'ptip': literal_eval,
                                   'ptail': literal_eval})

    # df = df.append(new_todos_df, verify_integrity=True, sort=False)

    new_todos_df = create_stats_columns(new_todos_df)

    ax_ptip = create_hexbin_coord_plot_from_dataframe(new_todos_df, 'ptip')
    ax_ptail = create_hexbin_coord_plot_from_dataframe(new_todos_df, 'ptail')
    ax_ptip = style_hexbin_ax(ax_ptip)
    ax_ptail = style_hexbin_ax(ax_ptail)
    ax_stack = stackplot_datapoints(new_todos_df)

    plt.show()

    # figure = plt.gcf()  # get current figure
    # save_fig(figure, "hexbin.png")
