import argparse
import os

import numpy as np
from skimage.io import imread, imsave
from skimage.transform import resize

from dataset.dataset_utils import get_files_and_keypoints, rescale_keypoint, save_json


def parse_args():
    parser = argparse.ArgumentParser(description="stalwart-keypoint-alpaca datapoint resizer")

    # json file with annotations
    parser.add_argument('--annotation_json', dest='annotation_json', required=False, help='')

    parser.add_argument('--source_dir', dest='source_dir', required=False, help='')
    parser.add_argument('--target_dir', dest='target_dir', required=True, help='')

    parser.add_argument('--w', dest='new_width', type=int, required=True, help='')
    parser.add_argument('--h', dest='new_height', type=int, required=True, help='')
    return parser.parse_args()


def main():
    args = parse_args()

    if args.annotation_json is None:
        if args.source_dir is not None:
            resize_images_only(args)
        else:
            print("Error: No source directory given and not annotation file given.")
        return
    else:
        if args.source_dir is None:
            image_filenames, annotations = get_files_and_keypoints(args.annotation_json)
            new_shape = (args.new_height, args.new_width)

            anno_dict = {image_file: annotations[index] for index, image_file in enumerate(image_filenames)}
            resized_anno_dict = {}

            for image_file in image_filenames:
                split_image_name = os.path.split(image_file)[1]
                new_img, new_keypoints = resize_datapoint(image_file, anno_dict[image_file], new_shape)
                resized_anno_dict[split_image_name] = new_keypoints
                imsave(os.path.join(args.target_dir, split_image_name), new_img, quality=100)

            resized_anno_dict["dataset_dir"] = args.target_dir
            save_json(resized_anno_dict, os.path.join(args.target_dir, os.path.split(args.annotation_json)[1]))
        else:
            print("Error: Source dir as well as annotation file given. Can only be one of both.")


def resize_datapoint(image_file, keypoints, new_shape):
    img = imread(image_file)
    h, w = img.shape[:2]
    nh, nw = new_shape
    if h == nh and w == nw:
        print("Old and new dimensions are the same, skipping.")
        return image_file, keypoints
    else:
        # new_img = cv2.resize(img, dsize=(nw, nh), interpolation=cv2.INTER_LANCZOS4)
        new_img = resize(img, (nh, nw), preserve_range=True)
        new_keypoints = [rescale_keypoint(point, w, h, nw, nh) for point in keypoints]

    return new_img, new_keypoints


def resize_images_only(args):
    source_dir = args.source_dir
    new_shape = (args.new_height, args.new_width)
    img_list = [img for img in os.listdir(source_dir) if '.jpg' in img]

    for image_file in img_list:
        print("Resizing {}".format(image_file))
        img_path = os.path.join(source_dir, image_file)
        img = imread(img_path)
        h, w = img.shape[:2]
        nh, nw = new_shape

        if h == nh and w == nw:
            print("Old and new dimensions are the same, skipping.")
            continue

        if img.max() == img.min():
            new_img = np.ones((400, 710, 1)) * 255
        else:
            new_img = resize(img, (nh, nw), preserve_range=True)
        imsave(os.path.join(args.target_dir, image_file), new_img.astype('uint8'), quality=100)


if __name__ == "__main__":
    main()
