import os
import random
import shutil

import matplotlib
import numpy as np

from dataset.dataset_utils import get_files_and_keypoints, save_json, read_json, make_none_keypoint, is_none_keypoint

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


def split_image_name(image_name):
    if image_name == 'dataset_dir':
        return -1, -1, -1
    image_name = os.path.split(image_name)[1]
    parts = image_name.split('_')
    if parts[0] == '':
        parts[0] = '4'

    if parts[1] == '':
        parts[1] = '4'

    subject = int(parts[0])
    task = int(parts[1])
    frame = int(parts[2].split('.')[0])
    return subject, task, frame


def copy_images(target_dir, images):
    for image in images:
        try:
            if os.path.isfile(image):
                shutil.copy(image, target_dir)
            else:
                print("Couldn't find file.")
        except IOError as e:
            print("Unable to copy file. {}".format(e))
        except:
            print("Unexpected error")


def save_new_set(target_dir, images, annotations):
    anno_dict = {os.path.split(image_file)[1]: annotations[index] for index, image_file in enumerate(images)}
    anno_dict["dataset_dir"] = target_dir
    filename = os.path.join(target_dir, "new_set_7.json")
    save_json(anno_dict, filename)


def create_subject_task_dict_from_names(names):
    new_set = {}
    # sort every image filename into subject and tasks
    for image_name in names:
        subject, task, frame = split_image_name(image_name)
        if not subject in new_set:
            new_set[subject] = {}

        if not task in new_set[subject]:
            new_set[subject][task] = []

        new_set[subject][task].append(image_name)
    return new_set


def sample_tasks(dataset, task_datapoint_amount):
    new_set_images = []

    # now we can sample as much as we need
    for subject in dataset:
        print("Sampling subject {}".format(subject))
        for task in dataset[subject]:
            print("Sampling task {}".format(task))
            sampled_images = random.sample(dataset[subject][task], task_datapoint_amount)
            new_set_images.extend(sampled_images)
            print(sampled_images)
    return new_set_images


def get_all_image_paths_units(dataset_dir):
    folder_prefix = "unit_"

    folder_content = os.listdir(dataset_dir)
    unit_folders = [name for name in folder_content if folder_prefix in name]

    print("Collected unit folders:")
    print(unit_folders)

    all_image_paths = []
    for folder in unit_folders:
        full_folder_path = os.path.join(dataset_dir, folder)
        files = [os.path.join(full_folder_path, image) for image in os.listdir(full_folder_path) if '.jpg' in image]
        all_image_paths.extend(files)
    return all_image_paths


def sample_and_copy(dataset_dir, target_dir, exclusion_dirs=None, task_datapoint_amount=30):
    # list all files in subfolders
    all_image_paths = get_all_image_paths_units(dataset_dir)

    if exclusion_dirs is not None:
        all_exclusion_files = []
        for dir in exclusion_dirs:
            files = [os.path.join(dir, file) for file in os.listdir(dir) if '.jpg' in file]
            all_exclusion_files.extend(files)
        all_image_paths = [path for path in all_image_paths if path not in all_exclusion_files]

    # sort them by subject and task
    large_set_dict = create_subject_task_dict_from_names(all_image_paths)

    # sample the tasks
    sampled_images = sample_tasks(large_set_dict, task_datapoint_amount)

    # after all tasks are sampled, copy the files into a new folder
    copy_images(target_dir, sampled_images)


# collect annotations from a list of json files
# for files in the given directory
def collect_annotations(new_set_dir, json_files):
    all_anno_dict = {}
    # read all relevant jsons
    for json_file in json_files:
        image_filenames, annotations = get_files_and_keypoints(json_file, new_set_dir)
        anno_dict = {os.path.split(image_file)[1]: annotations[index] for index, image_file in
                     enumerate(image_filenames)}
        all_anno_dict.update(anno_dict)

    # compare with the sampled file list
    new_set_files = [os.path.split(image)[1] for image in os.listdir(new_set_dir) if '.jpg' in image]

    # copy all annotations that can be found into a new dict
    new_set_annotations = {img: all_anno_dict[img] for img in new_set_files if img in all_anno_dict}

    # store the json file
    new_set_annotations["dataset_dir"] = new_set_dir
    filename = os.path.join(new_set_dir, "new_set.json")
    save_json(new_set_annotations, filename)


def copy_todos_from_new_set(new_set_dir, todo_dir, new_set_json):
    image_names, annotations = get_files_and_keypoints(new_set_json)
    files = [os.path.join(new_set_dir, image) for image in os.listdir(new_set_dir) if '.jpg' in image]

    todo_files = [name for name in files if not name in image_names]
    copy_images(todo_dir, todo_files)


def remove_annos_from_file_according_to_dir(dir, json_file):
    image_filenames, annotations = get_files_and_keypoints(json_file, dir)
    image_filenames_splitted = [os.path.split(image)[1] for image in image_filenames]
    files = [image for image in os.listdir(dir) if '.jpg' in image]
    anno_dict = {image_file: annotations[index] for index, image_file in enumerate(image_filenames_splitted) if
                 image_file in files}
    save_json(anno_dict, os.path.join(dir, "updated_json.json"))


def sample_single_json():
    annotation_json = "L:/large_dataset_study/extracted_frames/keypoints/json_unit_7_0-2000/keypoints_unit_7.json"
    target_dir = "L:/large_dataset_study/small_subset_7/"
    task_datapoint_amount = 30
    image_filenames, annotations = get_files_and_keypoints(annotation_json)

    new_set_images = []
    new_set_annotations = []

    anno_dict = {image_file: annotations[index] for index, image_file in enumerate(image_filenames)}
    print("Sorting image files...")
    new_set = create_subject_task_dict_from_names(image_filenames)

    # now we can sample as much as we need
    new_set_images = sample_tasks(new_set, task_datapoint_amount)

    for sampled_image in new_set_images:
        new_set_annotations.append(anno_dict[sampled_image])

    print("Copying images...")
    copy_images(target_dir, new_set_images)
    save_new_set(target_dir, new_set_images, new_set_annotations)
    print("Done!")
    pass


def check_if_directories_exclusive(first_dirs, second_dirs):
    all_first_files = []
    all_second_files = []
    for dir in first_dirs:
        files = [file for file in os.listdir(dir) if '.jpg' in file]
        all_first_files.extend(files)
    for dir in second_dirs:
        files = [file for file in os.listdir(dir) if '.jpg' in file]
        all_second_files.extend(files)

    same_list = [file for file in all_first_files if file in all_second_files]
    if len(same_list) > 0:
        print("Found {} duplicate files:".format(len(same_list)))
        print(same_list)
    else:
        print("No duplicates found")
    return


def get_anno_dict_from_json(json_file):
    image_names = []
    annotations = []
    if os.path.isfile(json_file):
        json_data = read_json(json_file)
        for elem in json_data:
            if elem != "dataset_dir":
                image_names.append(elem)
                annotations.append(json_data[elem])

    anno_dict = {image_file: annotations[index] for index, image_file in enumerate(image_names)}
    return anno_dict


def get_statistics_from_anno_dict(anno_dict, already_counted_dict, img_shape=(2160, 3840)):
    all_none_keypoints = [make_none_keypoint(), make_none_keypoint()]

    all_count = 0
    none_count = 0
    tip_count = 0
    tail_count = 0
    both_count = 0

    heatmap = np.zeros((img_shape[0], img_shape[1], 3), dtype='int')

    for item in anno_dict:
        if not already_counted_dict[item]:
            all_count += 1
            value = anno_dict[item]
            if np.array_equal(value, all_none_keypoints):
                none_count += 1
            elif is_none_keypoint(value[1]) and not is_none_keypoint(value[0]):
                heatmap[value[0][1], value[0][0], 0] += 1
                tip_count += 1
            elif is_none_keypoint(value[0]) and not is_none_keypoint(value[1]):
                heatmap[value[1][1], value[1][0], 1] += 1
                tail_count += 1
            elif not is_none_keypoint(value[0]) and not is_none_keypoint(value[1]):
                heatmap[value[0][1], value[0][0], 0] += 1
                heatmap[value[1][1], value[1][0], 1] += 1
                both_count += 1
            already_counted_dict[item] = True
        else:
            print("Duplicate img: {}".format(item))

    return all_count, none_count, tip_count, tail_count, both_count, already_counted_dict, heatmap


def get_keypoint_statistics(dataset_dir, json_files):
    sum_all_count = 0
    sum_none_count = 0
    sum_tip_count = 0
    sum_tail_count = 0
    sum_both_count = 0

    all_image_paths = get_all_image_paths_units(dataset_dir)
    already_counted_dict = {os.path.split(img_path)[1]: False for img_path in all_image_paths}

    all_keypoints_heatmap = None

    for keypoint_file in json_files:
        anno_dict = get_anno_dict_from_json(keypoint_file)
        all_count, none_count, tip_count, tail_count, both_count, already_counted_dict, anno_heatmap = get_statistics_from_anno_dict(
            anno_dict, already_counted_dict)
        if all_keypoints_heatmap is None:
            all_keypoints_heatmap = anno_heatmap
        else:
            all_keypoints_heatmap += anno_heatmap
        sum_all_count += all_count
        sum_none_count += none_count
        sum_tip_count += tip_count
        sum_tail_count += tail_count
        sum_both_count += both_count

    return sum_all_count, sum_none_count, sum_tip_count, sum_tail_count, sum_both_count, all_keypoints_heatmap


def sample_large_dataset():
    # target_dir = "L:/large_dataset_study/binarization_subset_1/"
    target_dir = "L:/large_dataset_study/first_round_subset/train/"

    exclusion_dirs = ["L:/large_dataset_study/first_round_subset/test/"]
    # "L:/large_dataset_study/first_round_subset/train/",
    # "L:/large_dataset_study/second_round_subset/test/"]
    # "L:/large_dataset_study/second_round_subset/train/"]

    # sample_and_copy(dataset_dir, target_dir, exclusion_dirs, 30)
    # sample_and_copy(dataset_dir, target_dir, task_datapoint_amount=2)
    # collect_annotations(target_dir, json_files)
    # new_set_json = os.path.join(target_dir, "new_set.json")
    # copy_todos_from_new_set(target_dir, os.path.join(target_dir, "todo"), new_set_json)
    # remove_annos_from_file_according_to_dir(target_dir, "L:/large_dataset_study/second_round_subset/second_round_all.json")
    check_if_directories_exclusive(exclusion_dirs, [target_dir])


dataset_dir = "L:/large_dataset_study/extracted_frames/"
json_files = ["L:/large_dataset_study/extracted_frames/keypoints/keypoints_0_1_2_3_4_5_9_11.json",
              "L:/large_dataset_study/extracted_frames/keypoints/json_unit_6_0-2000/keypoints_unit_6_all.json",
              "L:/large_dataset_study/extracted_frames/keypoints/json_unit_7_0-2000/keypoints_unit_7_all.json",
              "L:/large_dataset_study/extracted_frames/keypoints/json_unit_10_0-1000/json_unit_10_0-1000.json",
              "L:/large_dataset_study/extracted_frames/keypoints/keypoints_unit_12_partial.json",
              "L:/large_dataset_study/first_round_subset/train/first_round_train.json",
              "L:/large_dataset_study/first_round_subset/test/first_round_test.json",
              "L:/large_dataset_study/second_round_subset/train/second_round_train.json",
              "L:/large_dataset_study/second_round_subset/test/second_round_test.json"]

if __name__ == "__main__":
    # sample_large_dataset()
    sum_all_count, sum_none_count, sum_tip_count, sum_tail_count, sum_both_count, all_keypoints_heatmap = get_keypoint_statistics(
        dataset_dir, json_files)
    max = np.max(all_keypoints_heatmap)
    all_keypoints_heatmap = all_keypoints_heatmap * (255. / max)
    plt.imshow(all_keypoints_heatmap)
    plt.show()
