import os
from shutil import copy

import numpy as np
import pandas as pd
from skimage.io import imsave

from imgproc.stream_stroke_extractor import StreamStrokeExtractor


def flatten_tuple_of_lists(my_tuple):
    result = [item for sublist in list(my_tuple) for item in sublist]
    return result


def parse_bad(content):
    if content == 'X':
        return -1
    if content != '':
        result = []
        splitted = content.split(',')
        for item in splitted:
            if not '-' in item:
                result.append(int(item))
            else:
                start_range, end_range = item.split('-')[:2]
                result.extend(list(range(int(start_range), int(end_range) + 1)))
        return result
    else:
        return []


class StrokeStudyAnalyzer():
    def __init__(self, base_dir, words_to_frames_df, on_off_annotation_df, param_dict, start_index=0, cutoff_index=2):
        self.start_index = start_index
        self.cutoff_index = cutoff_index
        self.base_dir = base_dir
        self.words_to_frames_df = words_to_frames_df
        self.on_off_annotation_df = on_off_annotation_df
        self.on_off_annotation_df = on_off_annotation_df.drop_duplicates()
        self.study_overview_df = self._create_stuy_overview_df(self.words_to_frames_df)

        self.image_ids_and_filenames = self._validate_frames_lists()
        self.param_dict = param_dict
        self.calculated_on_point_dfs = []
        self.calculated_off_point_dfs = []
        self.subject_gt_dfs = []
        return

    def _create_stuy_overview_df(self, words_to_frames_df):
        # the words to frames dataframe contains empty cells when whole words are skipped
        # as well as an X in the bad column; this will be replaced with a -1
        words_to_frames_df['start_frame'] = words_to_frames_df['start_frame'].fillna(value=0).astype('int32')
        words_to_frames_df['end_frame'] = words_to_frames_df['end_frame'].fillna(value=0).astype('int32')
        words_to_frames_df['bad'] = words_to_frames_df['bad'].fillna(value='').astype('str')
        words_to_frames_df['bad'] = words_to_frames_df['bad'].map(parse_bad)

        # we now create a study overview dataframe
        # where each row contains start, end and bad frames for one subject
        # subject - start frame - end frame - skip frames
        good_words_frame = words_to_frames_df.loc[~(words_to_frames_df['bad'] == -1)]
        grouped_good_words = good_words_frame.groupby(['subject'])

        # we want a list of bad frames for each subject
        new_frame = grouped_good_words['bad'].apply(tuple).reset_index()
        new_frame.bad = new_frame.bad.map(flatten_tuple_of_lists)

        subject_start_frames = grouped_good_words['start_frame'].apply(lambda x: x.iloc[0])
        subject_end_frames = grouped_good_words['end_frame'].apply(lambda x: x.iloc[-1])

        new_frame['start_frame'] = subject_start_frames
        new_frame['end_frame'] = subject_end_frames
        return new_frame

    def _validate_frames_lists(self):
        assert ("start_frame" in self.study_overview_df.columns)
        assert ("end_frame" in self.study_overview_df.columns)
        assert ("bad" in self.study_overview_df.columns)

        image_ids_and_filenames = []

        for index, row in self.study_overview_df.iterrows():
            subject_ids_and_filenames = self._validate_frames_list_with_directory(
                index,
                row['start_frame'],
                row['end_frame'],
                row['bad'])
            image_ids_and_filenames.append(subject_ids_and_filenames)
        return image_ids_and_filenames

    def _validate_frames_list_with_directory(self, subject, start_frame, end_frame, skip_frames):
        # we collect all possible ids
        all_image_ids = list(range(start_frame, end_frame + 1))
        assert (len(all_image_ids) == end_frame - start_frame + 1)
        # we remove skip frames, as we dont consider them true negatives
        all_image_ids = [item for item in all_image_ids if item not in skip_frames]

        subject_patch_dir = os.path.join(self.base_dir, str(subject), "patch")
        if not os.path.isdir(subject_patch_dir):
            raise ValueError

        all_filenames = os.listdir(subject_patch_dir)
        image_ids_and_filenames = []
        for frame in all_image_ids:
            possible_names = [filename for filename in all_filenames if filename.startswith(str(frame) + '_')]
            assert (len(possible_names) == 1)
            filename_frame_id = int(possible_names[0].split('_')[0])
            assert (frame == filename_frame_id)
            image_ids_and_filenames.append((frame, possible_names[0]))

        assert (len(all_image_ids) == len(image_ids_and_filenames))
        return image_ids_and_filenames

    def analyze_subjects(self):
        assert ("start_frame" in self.study_overview_df.columns)
        assert ("end_frame" in self.study_overview_df.columns)
        assert ("bad" in self.study_overview_df.columns)

        subject_dfs = []

        for index, row in self.study_overview_df.iterrows():
            if index > self.cutoff_index or index < self.start_index:
                continue

            self._analyze_subject(
                index,
                row['start_frame'],
                row['end_frame'],
                row['bad'])
            subject_df = self._create_subject_stat_df(index)
            subject_dfs.append(subject_df)
        return subject_dfs

    def _analyze_subject(self, subject, start_frame, end_frame, skip_frames):
        print("Evaluating subject {}".format(subject))
        current_dir = os.path.join(base_dir, str(subject))

        strokes, \
        on_points, \
        off_points, \
        on_frame_ids, \
        off_frame_ids, \
        aggregated_image = StreamStrokeExtractor(
            current_dir,
            start_frame,
            end_frame,
            skip_frames,
            self.param_dict).run()

        assert (len(on_points) == len(on_frame_ids))
        assert (len(off_points) == len(off_frame_ids))
        all_calculated_ids_count = len(on_frame_ids) + len(off_frame_ids)

        all_image_ids = [item[0] for item in self.image_ids_and_filenames[subject]]
        image_ids_skipped_by_extractor = [item for item in all_image_ids if
                                          (item not in on_frame_ids and item not in off_frame_ids)]

        assert (all_calculated_ids_count == len(self.image_ids_and_filenames[subject]) - len(
            image_ids_skipped_by_extractor))

        # store last agggregate image in subject dir
        imsave(os.path.join(current_dir, "subject_{}_aggregate.jpg".format(subject)), aggregated_image, quality=99)

        on_points_df = pd.DataFrame(dtype=object)
        on_points_df['coord_x'] = pd.Series(on_points[:, 0], dtype=float)
        on_points_df['coord_y'] = pd.Series(on_points[:, 1], dtype=float)
        on_points_df['frame'] = pd.Series(on_frame_ids, dtype=int)

        off_points_df = pd.DataFrame(dtype=object)
        off_points_df['coord_x'] = pd.Series(off_points[:, 0], dtype=float)
        off_points_df['coord_y'] = pd.Series(off_points[:, 1], dtype=float)
        off_points_df['frame'] = pd.Series(off_frame_ids, dtype=int)

        off_points_skipped_by_extractor = [(item, -1, -1) for item in image_ids_skipped_by_extractor]
        off_points_df = off_points_df.append(
            pd.DataFrame(off_points_skipped_by_extractor, columns=['frame', 'coord_x', 'coord_y']), sort=False)
        off_points_df.reset_index(inplace=True, drop=True)

        all_calculated_dfs_count = len(on_points_df) + len(off_points_df)
        assert (all_calculated_dfs_count == len(self.image_ids_and_filenames[subject]))

        self.calculated_on_point_dfs.append(on_points_df)
        self.calculated_off_point_dfs.append(off_points_df)
        return

    def _create_subject_stat_df(self, subject):
        # we have all frames that can be classified either way
        all_images_df = pd.DataFrame(self.image_ids_and_filenames[subject], columns=['frame', 'filename'])

        # get the annotations for the current subject
        current_subject_annotation_df = self.on_off_annotation_df.loc[self.on_off_annotation_df['subject'] == subject]
        current_subject_annotation_df = current_subject_annotation_df.drop_duplicates(keep='first')

        # get all on annotations
        gt_with_only_positives = current_subject_annotation_df.loc[current_subject_annotation_df['on'] == True]

        on_idx = pd.Index(gt_with_only_positives['frame'])
        all_img_idx = pd.Index(all_images_df['frame'])

        # get all off annotations (all files that are not specifically in the on dataframe)
        gt_with_all_negatives = all_images_df.loc[~all_img_idx.isin(on_idx)].copy()
        gt_with_all_negatives['subject'] = int(subject)
        gt_with_all_negatives['on'] = False

        # create a combined dataframe
        all_gt = gt_with_only_positives.append(gt_with_all_negatives, sort=False)
        all_gt = all_gt.reset_index(drop=True)
        all_gt_idx = pd.Index(all_gt['frame'])

        calculated_on_df = self.calculated_on_point_dfs[subject]
        calculated_on_idx = pd.Index(calculated_on_df['frame'])

        all_gt['calc_on'] = all_gt_idx.isin(calculated_on_idx)
        all_gt = all_gt.rename(columns={"on": "gt_on"})

        # since it is possible that files get annotated that are later on considered skip frames
        # we take care that no gt annotations end up in the results that are not in the all image dataframe
        all_gt = all_gt.loc[all_gt_idx.isin(all_img_idx)]

        assert (len(all_gt) == len(self.image_ids_and_filenames[subject]))
        return all_gt

    def store_calculated_on_off_points(self):
        for index, elem in enumerate(self.calculated_on_point_dfs):
            elem.to_csv(os.path.join(self.base_dir, 'subject_{}_on_points.csv'.format(index)))

        for index, elem in enumerate(self.calculated_off_point_dfs):
            elem.to_csv(os.path.join(self.base_dir, 'subject_{}_off_points.csv'.format(index)))
        return


def analyse_dataframe_of_stroke_study(dataframe):
    good_words_frame = dataframe.loc[~(dataframe['bad'] == -1)]
    # avg samples per word
    samples_per_word = good_words_frame['end_frame'] - good_words_frame['start_frame']
    avg_samples_per_word = samples_per_word.mean()

    # avg samples per letter
    letter_count = good_words_frame['word'].apply(len)
    samples_per_letter = samples_per_word / letter_count
    avg_letter_count = letter_count.mean()
    avg_samples_per_letter = samples_per_letter.mean()

    # proportion of good to bad samples
    all_word_samples = samples_per_word.sum()
    bad_sample_count = good_words_frame['bad'].apply(len)
    all_bad_samples = bad_sample_count.sum()

    # avg pause between words

    result = {
        "avg_letter_count": avg_letter_count,
        "avg_samples_per_word": avg_samples_per_word,
        "avg_samples_per_letter": avg_samples_per_letter,
        "all_word_samples": all_word_samples,
        "all_bad_samples": all_bad_samples,
        "proportion_bad_to_good": all_bad_samples / all_word_samples
    }

    return result


my_param_dict = {
    "heatmap_size": (211, 211),
    "patch_size": (224, 224),
    "surface_size": (1080, 1920),
    "filter_duplicate_points": False,

    "max_kp_heatmap_mean": 90,
    "max_bin_heatmap_mean": 50,
    "min_bin_heatmap_std": 7,

    "path_distance_filter_threshold": 150,
    "do_subdivision": False,
    "aggregate_blur_sigma": 2,
    "beautify_path_steps": 2,
    "on_point_max_threshold": 100,
    "min_split_gap": 3,
    "use_blurred_aggregate_for_classification": False,

    "bin_heatmap_black_noise_threshold": 5,
    "aggregate_type": "dynamic",
    "aggregate_path": "../tmp_out/record/stroke_study/0/manual_agg/manual_agg_0_5670.jpg",
    "aggregate_path_mask_stroke_width": 5,
    "use_aggregate_aging": False,
    "aggregate_max_age": 5,

    "use_incremental_aggregate": True,
    "incremental_aggregate_interval": 50
}

my_dynamic_param_dict = {
    "heatmap_size": (211, 211),
    "patch_size": (224, 224),
    "surface_size": (1080, 1920),

    "max_kp_heatmap_mean": 50,
    "max_bin_heatmap_mean": 50,
    "min_bin_heatmap_std": 7,

    "path_distance_filter_threshold": 150,
    "do_subdivision": False,
    "aggregate_blur_sigma": 2,
    "beautify_path_steps": 5,
    "on_point_max_threshold": 100,
    "min_split_gap": 3,
    "use_blurred_aggregate_for_classification": False,

    "bin_heatmap_black_noise_threshold": 5,
    "aggregate_type": "static",
    "aggregate_path": "../tmp_out/record/stroke_study/0/manual_agg/manual_agg_0_5670.jpg",
    "aggregate_path_mask_stroke_width": 10,
    "use_aggregate_aging": True,
    "aggregate_max_age": 5
}


# tip_accuracy -> tip accuracy wenn nicht on; label only off? Besser eigener Test für Gestures, und dann lieber
# die Gesture Accuracy messen; so vermeiden wir massive Aufwendungen für Annotationen; on/off ist da vielversprechender
# und hat bessere Wirkung; Wir können vergleichen, wie es ist mit perfektem Aggregate und dem generierten;
# So können wir zeigen, dass unsere Methode sinnvoll unbekannte Streams processen kann und dann fast so genau ist wie mit
# Endergebnis; ROC-Kurve für Intervall-Länge wäre hier wohl total geil.
# on_off_accuracy -> ! labelling necessary; alteration to tool, da steckt eigentlich tip accuracy schon mit drin
# weil wenn wir on sind, dann stimmt auch die tip position innerhalb eines fehler radius;

# aggregate_accuracy -> ! easiest; direct pixel comparison
# character accuracy -> only useful after optimization

# Wichtigster Schritt: On-Off-Annotation für Wörter. Wird hart. Aber geht eh nur so gut wie ein Mensch das kann.

# annotation tool? visualization? matplotlib

def load_subject_dfs(base_dir, start_index=0, end_index=7):
    subject_dfs = []
    for i in range(start_index, end_index + 1):
        current_df_filename = 'subject_{}_gt_and_calc.csv'.format(i)
        current_path = os.path.join(base_dir, current_df_filename)
        current_dataframe = pd.read_csv(current_path, index_col=0)
        subject_dfs.append(current_dataframe)
    return subject_dfs


def create_training_dirs(base_dir, subject_dfs, start_index=0, end_index=7, out_dir_prefix=""):
    pen_down_dir = os.path.join(base_dir, out_dir_prefix, "pen_down/")
    pen_up_dir = os.path.join(base_dir, out_dir_prefix, "pen_up/")

    index = start_index

    for df in subject_dfs:
        print("copying subject {}".format(index))
        # get subject patch directory
        subject_patch_dir = os.path.join(base_dir, str(index), "patch")

        # get all ground truth on entries
        gt_on_entries = df.loc[df['gt_on'] == True]
        for on_index, row in gt_on_entries.iterrows():
            src_path = os.path.join(subject_patch_dir, row['filename'])
            new_filename = str(index) + '_' + row['filename']
            trgt_path = os.path.join(pen_down_dir, new_filename)
            if os.path.isfile(src_path):
                copy(src_path, trgt_path)
            else:
                raise ValueError

        gt_off_entries = df.loc[df['gt_on'] == False]
        for off_index, row in gt_off_entries.iterrows():
            src_path = os.path.join(subject_patch_dir, row['filename'])
            new_filename = str(index) + '_' + row['filename']
            trgt_path = os.path.join(pen_up_dir, new_filename)
            if os.path.isfile(src_path):
                copy(src_path, trgt_path)
            else:
                raise ValueError

        if index == end_index:
            break
        else:
            index += 1

    return


def analyze(base_dir):
    stroke_study = pd.read_csv('stroke_study_words_to_frames.csv',
                               dtype={'subject': np.int32,
                                      'phrase': np.int32,
                                      'word': str,
                                      'bad': str}, sep=';')
    current_annos = pd.read_csv('words_to_frames_on_off.csv', index_col=0)

    analyzer = StrokeStudyAnalyzer(base_dir, stroke_study, current_annos, my_param_dict, start_index=0, cutoff_index=4)
    subject_dfs = analyzer.analyze_subjects()

    for index, df in enumerate(subject_dfs):
        df_path = os.path.join(base_dir, 'subject_{}_gt_and_calc.csv'.format(index))
        df.to_csv(df_path)
    analyzer.store_calculated_on_off_points()


if __name__ == '__main__':
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)

    base_dir = "../tmp_out/record/stroke_study/"

    subject_dfs = load_subject_dfs(base_dir, start_index=0, end_index=0)
    create_training_dirs(base_dir, subject_dfs, start_index=0, end_index=0, out_dir_prefix="val/")
