import os

import tensorflow as tf
from tensorflow.keras.models import load_model
from tensorflow.python.framework import graph_io

from model.model_config import ModelConfig
from model.model_factory import ModelFactory

input_names = ['input_1']
output_names = ['conv2d_1/BiasAdd']


# from https://www.dlology.com/blog/how-to-run-keras-model-on-jetson-nano/
def freeze_graph(graph, session, output, save_pb_dir='.', save_pb_name='frozen_model.pb', save_pb_as_text=False):
    with graph.as_default():
        graphdef_inf = tf.graph_util.remove_training_nodes(graph.as_graph_def())
        graphdef_frozen = tf.graph_util.convert_variables_to_constants(session, graphdef_inf, output)
        graph_io.write_graph(graphdef_frozen, save_pb_dir, save_pb_name, as_text=save_pb_as_text)
        return graphdef_frozen


def create_tf_session(gpu):
    keras.backend.clear_session()
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu)  # choose from GPUs 0-3
    # only use as much VRAM as needed
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    session = tf.Session(config=config)
    set_session(session)
    return session


def create_model_graph(model_config_yaml, model_file, graph_dir):
    session = create_tf_session(0)

    keras.backend.set_learning_phase(0)
    keras.backend.set_image_data_format('channels_last')

    model_config = ModelConfig(model_config_yaml)
    model_factory = ModelFactory()
    my_resnet = model_factory.create_model(model_config)
    model = my_resnet.get_model()
    model.load_weights(model_file)

    # model = load_model(model_file)

    input_names = [t.op.name for t in model.inputs]
    output_names = [t.op.name for t in model.outputs]

    # Prints input and output nodes names, take notes of them.
    print(input_names, output_names)

    frozen_graph = freeze_graph(session.graph, session, [out.op.name for out in model.outputs], save_pb_dir=graph_dir)
    print("Freezing graph done.")
    trt_graph = trt.create_inference_graph(
        input_graph_def=frozen_graph,
        outputs=output_names,
        max_batch_size=1,
        max_workspace_size_bytes=1 << 25,
        precision_mode='FP16',
        minimum_segment_size=50
    )

    graph_io.write_graph(trt_graph, graph_dir,
                         "trt_graph.pb", as_text=False)


def get_frozen_graph(graph_file):
    """Read Frozen Graph file from disk."""
    with tf.gfile.FastGFile(graph_file, "rb") as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())
    return graph_def


def create_session_with_frozen_graph(input_names, output_names, graph_file):
    trt_graph = get_frozen_graph(graph_file)

    # Create session and load graph
    tf_config = tf.ConfigProto()
    tf_config.gpu_options.allow_growth = True
    tf_sess = tf.Session(config=tf_config)
    tf.import_graph_def(trt_graph, name='')

    # input and output tensor names.
    input_tensor_name = input_names[0] + ":0"
    output_tensor_name = output_names[0] + ":0"

    output_tensor = tf_sess.graph.get_tensor_by_name(output_tensor_name)
    return tf_sess, output_tensor


def predict_with_trt_session(session, output_tensor, batch_x, input_tensor_name='input_1'):
    feed_dict = {
        input_tensor_name: batch_x
    }
    preds = session.run(output_tensor, feed_dict)
    return preds


def convert_normal_model_to_saved_model(model_config_yaml, model_weights, output_dir):
    tf.keras.backend.set_learning_phase(0)
    # model_config = ModelConfig(model_config_yaml)

    # model_factory = ModelFactory()
    # my_resnet = model_factory.create_model(model_config)

    # model = my_resnet.get_model()

    print("Loading model weights ...")
    # model.load_weights(model_weights)
    model = load_model(model_weights)
    model.save(output_dir, save_format='tf')
    return


def convert_saved_model_tf2(saved_model_dir, output_dir, precision='float32'):
    if precision == 'float32':
        conversion_params = trt.DEFAULT_TRT_CONVERSION_PARAMS._replace(
            precision_mode=trt.TrtPrecisionMode.FP32)
    else:
        conversion_params = trt.DEFAULT_TRT_CONVERSION_PARAMS._replace(
            precision_mode=trt.TrtPrecisionMode.FP16)

    converter = trt.TrtGraphConverterV2(
        input_saved_model_dir=saved_model_dir,
        conversion_params=conversion_params)
    converter.convert()

    converter.save(output_dir)
    return


if __name__ == '__main__':
    convert_normal_model_to_saved_model('../models/roinet/roinet_224.yaml', '../models/roinet/roinet.h5',
                                        '../models/roinet/1/')
    # create_model_graph('../models/patchnet/patchnet_224.yaml', '../models/patchnet/patchnet_224_latest.h5', '../models/patchnet/')
