import matplotlib
import pandas as pd
from skimage.io import imread

from util import keypoint_analysis
from util.keypoint_analysis import get_keypoints_dataframe
from util.sample_extracted_frames import get_img_file_from_full_path
from util.select_and_extract_keypoint_annotations import get_all_image_paths_units

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class AnnotationViewer():
    def __init__(self, dataframe, start_image=None):
        fig, axes = plt.subplots(ncols=1, nrows=1, figsize=(20, 14))
        self.fig = fig
        self.axes = axes
        if dataframe is not None:
            self.dataframe = dataframe
        else:
            self.dataframe = pd.DataFrame(columns=('frame', 'full_path', 'ptip', 'ptail'))
        self.index = 0
        self.marked_images = []
        self.change_ptip = False
        self.change_ptail = False
        current_image_name = ""
        if start_image is not None:
            self.index = self.dataframe.index.get_loc(start_image)
        self._load_current_image_and_annos()
        fig.canvas.mpl_connect('key_press_event', self.on_key)
        fig.canvas.mpl_connect('button_press_event', self.on_click)
        pass

    def _load_current_image_and_annos(self):
        current_row = self.dataframe.iloc[self.index]
        print(current_row)
        self.current_image_name = current_row.name
        self.current_image = imread(current_row['full_path'])
        self.ptip = current_row['ptip']
        self.ptail = current_row['ptail']
        return

    def _clear_axes(self):
        # for ax in self.axes.flatten():
        #    ax.cla()
        self.axes.cla()
        return

    def _view_axes(self):
        self._clear_axes()
        self.axes.imshow(self.current_image)
        self.axes.plot(self.ptip[0], self.ptip[1], 'g+')
        self.axes.plot(self.ptail[0], self.ptail[1], 'rx')

    def on_click(self, event):

        if self.change_ptip:
            if self.change_ptail:
                print("ALERT, double keypoint change!")

            new_x = int(round(event.xdata))
            new_y = int(round(event.ydata))
            print("new ptip: [{}, {}]".format(new_x, new_y))
            # print(self.dataframe.loc[self.index, 'ptip'])
            self.dataframe.iat[self.index, self.dataframe.columns.get_loc('ptip')] = [new_x, new_y]
        if self.change_ptail:
            if self.change_ptip:
                print("ALERT, double keypoint change!")

            new_x = int(round(event.xdata))
            new_y = int(round(event.ydata))
            print("new ptail: [{}, {}]".format(new_x, new_y))
            # print(self.dataframe.loc[self.index, 'ptip'])
            self.dataframe.iat[self.index, self.dataframe.columns.get_loc('ptail')] = [new_x, new_y]
        return

    def on_key(self, event):
        if event.key == 'e':
            self.next()
        elif event.key == 'w':
            self.prev()
        elif event.key == 'm':
            self.marked_images.append(self.current_image_name)
            print("Marked file {}".format(self.current_image_name))
        elif event.key == 'n':
            self.save_marked_files()
        elif event.key == 'b':
            self.change_ptail = False
            self.change_ptip = not self.change_ptip
            # print("PTAIL mode is {}".format(str(self.change_ptail)))
            print("PTIP mode is {}".format(str(self.change_ptip)))
        elif event.key == 'v':
            self.change_ptip = False
            self.change_ptail = not self.change_ptail
            # print("PTIP mode is {}".format(str(self.change_ptip)))
            print("PTAIL mode is {}".format(str(self.change_ptail)))
        elif event.key == 'i':
            self.dataframe.to_csv('val2_study_annos_3_add.csv')
            print("Saved dataframe.")

    def save_marked_files(self):
        if self.marked_images:
            with open('marked_val2_study_annos_3_add.txt', 'w') as mark_file:
                mark_file.write('\n'.join(self.marked_images))
            mark_file.close()
        print("Saved marked files.")
        pass

    def view(self):
        self._view_axes()
        plt.show()
        pass

    def next(self):
        self.change_ptip = False
        self.change_ptail = False
        self.index += 1
        self._load_current_image_and_annos()
        self.view()
        return

    def prev(self):
        self.change_ptip = False
        self.change_ptail = False
        if self.index > 0:
            self.index -= 1
        self._load_current_image_and_annos()
        self.view()
        return


def get_frame_with_paths():
    keypoints_df = get_keypoints_dataframe(keypoint_analysis.all_keypoint_files, statistics=False)
    full_path_file_list = get_all_image_paths_units("L:/large_dataset_study/extracted_frames/")

    files_df = pd.DataFrame(full_path_file_list)
    files_df.columns = ['full_path']
    files_df['img_file'] = files_df.full_path.map(get_img_file_from_full_path)
    keypoints_idx = keypoints_df.index
    files_idx = pd.Index(files_df['img_file'])
    files_df = files_df.loc[files_idx.isin(keypoints_idx)]

    keypoints_df = keypoints_df.join(files_df.set_index('img_file'))
    print(keypoints_df.head())

    keypoints_df['ptip'] = keypoints_df['ptip'].astype(object)
    return keypoints_df


if __name__ == '__main__':
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)

    # current_annos = pd.read_csv('updated2_final_annos.csv', index_col=0,
    #                           dtype={'subject':np.int32,
    #                                  'task':np.int32,
    #                                  'frame':np.int32},
    #                           converters={
    #                                  'ptip': literal_eval,
    #                                  'ptail': literal_eval})

    # print(current_annos.head())
    # viewer = AnnotationViewer(current_annos, start_image="(2, '2_3_1353.jpg')")
    viewer = AnnotationViewer(None, start_image=None)
    viewer.view()
