import matplotlib
import numpy as np
import pandas as pd
from imageio import imsave
from skimage.io import imread

from dataset_utils import get_image_files_in_directory
from imgproc.patch_extractor import PatchExtractor

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class BinImgAnnotator():
    '''
    Show images for which a binarization exists and save a 224x224
    image patch as well as a binarization patch on each click on the image.
    '''

    def __init__(self, img_dir, bin_dir):
        fig, axes = plt.subplots(ncols=1, nrows=1, figsize=(20, 14))
        self.fig = fig
        self.axes = axes
        self.index = 0
        self.patch_count = 0
        self.current_image = None
        self.current_bin = None
        fig.canvas.mpl_connect('key_press_event', self.on_key)
        fig.canvas.mpl_connect('button_press_event', self.on_click)
        self.image_files = get_image_files_in_directory(img_dir, full_path=True)
        self.bin_files = get_image_files_in_directory(bin_dir, full_path=True)
        self._load_current_image_and_bin()
        pass

    def _load_current_image_and_bin(self):
        self.current_image = imread(self.image_files[self.index])
        self.current_bin = imread(self.bin_files[self.index])
        return

    def _clear_axes(self):
        self.axes.cla()
        return

    def _view_axes(self):
        self._clear_axes()
        self.axes.imshow(self.current_image)

    def on_click(self, event):
        new_x = int(round(event.xdata))
        new_y = int(round(event.ydata))
        keypoint = np.array((new_y, new_x))
        img_patch, _ = PatchExtractor(
            self.current_image,
            (224, 224),
            keypoint,
            do_pad=True).run()
        bin_patch, _ = PatchExtractor(
            self.current_bin,
            (224, 224),
            keypoint,
            do_pad=True).run()
        imsave("../tmp_out/bin_patches/{}_{}.jpg".format(self.index, self.patch_count), img_patch, quality=100)
        imsave("../tmp_out/bin_patches/{}_{}_bin.jpg".format(self.index, self.patch_count), bin_patch, quality=100)
        self.patch_count += 1
        print("Saved img and bin patch")
        return

    def on_key(self, event):
        if event.key == 'e':
            self.next()
        elif event.key == 'w':
            self.prev()
        return

    def view(self):
        self._view_axes()
        plt.show()
        pass

    def next(self):
        self.index += 1
        self._load_current_image_and_bin()
        self.view()
        return

    def prev(self):
        if self.index > 0:
            self.index -= 1
        self._load_current_image_and_bin()
        self.view()
        return


if __name__ == '__main__':
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)

    image_dir = "L:/large_dataset_study/binarization_subset_1"
    bin_dir = "L:/large_dataset_study/binarization_subset_1/binarized"

    viewer = BinImgAnnotator(image_dir, bin_dir)
    viewer.view()
