import matplotlib
import numpy as np
from skimage.transform import resize

from dataset_utils import rescale_keypoint
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor, MaxFilterPeakExtraction, \
    WindowedPeakExtraction, HoughCirclesPeakExtraction, CentroidPeakExtraction
from visual.image_heatmap_file_data_provider import ImageHeatmapFileDataProvider
from visual.sequence_viewer import SequenceViewer

matplotlib.use('TkAgg')


class ImageHeatmapFileDataProviderForRecordedPairs(ImageHeatmapFileDataProvider):
    def __init__(self, img_dir, batch_size=1):
        super().__init__(img_dir, img_dir, batch_size)
        self.image_names = [filename for filename in self.image_names if not 'inference' in filename]
        self.heatmap_names = [filename for filename in self.heatmap_names if 'inference' in filename]


class HeatmapPeakExtractorViewer(SequenceViewer):
    def __init__(self, display_data_provider, batch_size, show_inferred_data=False):
        super().__init__(display_data_provider, batch_size, rows=3, cols=2)
        self.index_increment = 2
        self.batch_counter = 3
        self._load_current_data()
        pass

    def _load_current_data(self):
        if len(self.display_data_provider) > self.batch_counter >= 0:
            self.batch_x, self.batch_y = self.display_data_provider[self.batch_counter]
            self.batch_y = [item[:, :396] for item in self.batch_y]
        return

    def _extract_peaks(self, heatmap):
        extracted_peaks = {}
        extracted_peaks["MaxFilter"] = HeatmapKeypointExtractor(heatmap, MaxFilterPeakExtraction).run()
        extracted_peaks["Windowed"] = HeatmapKeypointExtractor(heatmap, WindowedPeakExtraction).run(window_size=5)
        extracted_peaks["HoughCircles"] = HeatmapKeypointExtractor(heatmap, HoughCirclesPeakExtraction).run()
        extracted_peaks["Centroid"] = HeatmapKeypointExtractor(heatmap, CentroidPeakExtraction).run()
        return extracted_peaks

    def _draw_peaks_on_heatmap(self, peaks, fig):
        for key in peaks:
            current_peak = peaks[key]
            color = np.random.rand(3, )
            fig.plot(current_peak[1], current_peak[0], 'x', c=color)
            fig.text(current_peak[1] + 15, current_peak[0], key, bbox={'facecolor': color, 'alpha': 0.5, 'pad': 5})
        return

    def _view_axes(self):
        self._clear_axes()
        img_shape = self.batch_x[self.index_counter].shape
        hm_shape = self.batch_y[self.index_counter].shape

        self.axes[0][0].imshow(self.batch_x[self.index_counter])
        self.axes[0][1].imshow(self.batch_x[self.index_counter + 1])

        self.axes[1][0].imshow(self.batch_y[self.index_counter])
        self.axes[1][1].imshow(self.batch_y[self.index_counter + 1])

        ########### Heatmaps and Peaks

        left_heatmap = self.batch_y[self.index_counter]
        left_extracted_peaks = self._extract_peaks(left_heatmap)
        self._draw_peaks_on_heatmap(left_extracted_peaks, self.axes[1][0])

        right_heatmap = self.batch_y[self.index_counter + 1]
        right_extracted_peaks = self._extract_peaks(right_heatmap)
        self._draw_peaks_on_heatmap(right_extracted_peaks, self.axes[1][1])

        ########### Overlay

        self.axes[2][0].imshow(self.batch_x[self.index_counter])
        self.axes[2][1].imshow(self.batch_x[self.index_counter + 1])

        resized_heatmap = resize(left_heatmap, (img_shape[0], img_shape[1]))
        self.axes[2][0].imshow(resized_heatmap, alpha=0.5)
        resized_heatmap_right = resize(right_heatmap, (img_shape[0], img_shape[1]))
        self.axes[2][1].imshow(resized_heatmap_right, alpha=0.5)

        # resized keypoints on overlay
        for key, val in left_extracted_peaks.items():
            left_extracted_peaks[key] = rescale_keypoint((val[0], val[1]), hm_shape[0], hm_shape[1], img_shape[0],
                                                         img_shape[1])

        for key, val in right_extracted_peaks.items():
            right_extracted_peaks[key] = rescale_keypoint((val[0], val[1]), hm_shape[0], hm_shape[1], img_shape[0],
                                                          img_shape[1])

        self._draw_peaks_on_heatmap(left_extracted_peaks, self.axes[2][0])
        self._draw_peaks_on_heatmap(right_extracted_peaks, self.axes[2][1])

        pass


if __name__ == '__main__':
    image_path = "../tmp_out/"

    data_provider = ImageHeatmapFileDataProviderForRecordedPairs(image_path, 16)
    viewer = HeatmapPeakExtractorViewer(data_provider, 16, show_inferred_data=True)
    viewer.view()
