import os

import numpy as np
from natsort import natsorted
from skimage.io import imread


class ImageHeatmapFileDataProvider:
    def __init__(self, img_dir, heatmap_dir=None, batch_size=1):
        self.img_dir = img_dir
        self.heatmap_dir = heatmap_dir
        self.batch_size = batch_size
        self.image_names = self._get_image_filenames(img_dir)
        self.heatmap_names = self._get_image_filenames(heatmap_dir)
        pass

    def _get_image_filenames(self, directory):
        if directory is not None:
            list_images = os.listdir(directory)
            file_list = [os.path.join(directory, filename) for filename in list_images if 'jpg' in filename]
            file_list = natsorted(file_list)
            return file_list
        # filename_list = [os.path.split(filepath)[1] for filepath in file_list]
        return None

    def _load_batch(self, index):
        image_list = None
        heatmap_list = None
        if self.image_names is not None:
            image_list = np.array([imread(filename) for filename in
                                   self.image_names[index * self.batch_size:(1 + index) * self.batch_size]])
        if self.heatmap_names is not None:
            heatmap_list = np.array([imread(filename) for filename in
                                     self.heatmap_names[index * self.batch_size:(1 + index) * self.batch_size]])
        return image_list, heatmap_list

    def __len__(self):
        return int(np.ceil(len(self.image_names) / float(self.batch_size)))

    def __getitem__(self, index):
        return self._load_batch(index)


# filenames need to contain ids first, separated by underscores
class SelectiveImageProvider(ImageHeatmapFileDataProvider):
    def __init__(self, img_dir, start_id, end_id, skip_ids):
        super().__init__(img_dir, batch_size=1)
        self.start_id = start_id
        self.end_id = end_id
        self.skip_ids = skip_ids
        self.needed_filenames = []

        started = False
        for index, item in enumerate(self.image_names):
            image_name_without_path = os.path.split(item)[1]
            splitted_item = int(image_name_without_path.split('_')[0])
            if start_id == splitted_item:
                self.needed_filenames.append(item)
                started = True
            elif started:
                if splitted_item not in self.skip_ids and \
                        splitted_item <= end_id:
                    self.needed_filenames.append(item)
                if splitted_item == end_id:
                    started = False

    def _load_batch(self, index):
        image_list = None
        if len(self.needed_filenames) > 0:
            image_list = np.array([imread(filename) for filename in
                                   self.needed_filenames[index * self.batch_size:(1 + index) * self.batch_size]])
        return image_list

    def __len__(self):
        return int(np.ceil(len(self.needed_filenames) / float(self.batch_size)))
