import matplotlib

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
import numpy as np
from skimage import draw
from imgproc.data_to_vector_path import FromFileVectorizer
from imgproc.patch_extractor import PenPatchExtractor
from skimage.color import rgb2gray
import os
import json
import cv2
import random


# 1178, 710
class KeypointViewer(object):
    def __init__(self, pen_samples, poses_path, keypoint_path=None):
        self.keypoint_path = keypoint_path
        self.pen_samples = pen_samples
        self.pen_patches = self._create_pen_patches()
        self.keypoint_image = self._create_keypoint_image()
        self.view_index = 0

        fig, ((ax1, ax2, ax3), (ax4, ax5, ax6)) = plt.subplots(ncols=3, nrows=2, figsize=(14, 7))
        self.fig = fig
        self.axes = ((ax1, ax2, ax3), (ax4, ax5, ax6))
        self.ax_count = 6
        self.poses = self._get_json_poses(poses_path)
        fig.canvas.mpl_connect('key_press_event', self.on_key)

    def _create_pen_patches(self):
        pen_patches = []
        patch_maker = PenPatchExtractor()
        for index, elem in enumerate(self.pen_samples):
            current_elem_keypoint = elem["coords"][0]
            img_patch, rows, cols = patch_maker.extract_pen_patch(elem["image"], current_elem_keypoint, (150, 150))
            # if index < 20:
            #    io.imsave("patches/patch{}.jpg".format(index), img_patch)
            pen_patches.append(img_patch)
        return pen_patches

    def _create_keypoint_image(self):
        img = np.zeros((710, 1178, 3), dtype=np.uint8)
        for index, elem in enumerate(self.pen_samples):
            if index < len(self.pen_samples) - 1:
                current_elem_keypoint = elem["coords"][0]
                next_elem_keypoint = self.pen_samples[index + 1]["coords"][0]

                rr, cc, val = draw.line_aa(current_elem_keypoint[1], current_elem_keypoint[0], next_elem_keypoint[1],
                                           next_elem_keypoint[0])
                img[rr, cc] = [(elem * 255, elem * 255, elem * 255) for elem in val]
        return img

    def view(self):
        base_index = self.view_index * self.ax_count
        diff_img1 = rgb2gray(self.pen_patches[base_index + 0])
        diff_img2 = rgb2gray(self.pen_patches[base_index + 1])
        # diff_img = diff_img2 - diff_img1

        self.axes[0][0].imshow(self._get_contours(self.pen_patches[base_index + 0], self.axes[1][0]))
        self.axes[0][1].imshow(self._get_contours(self.pen_patches[base_index + 1], self.axes[1][1]))
        # self.axes[0][2].imshow(diff_img)
        self.axes[0][2].imshow(self._get_contours(self.pen_patches[base_index + 1], self.axes[1][2]))
        # self.axes[1][0].imshow(self._get_contours(self.pen_patches[base_index + 2]))
        # self.axes[1][1].imshow(self._get_contours(self.pen_patches[base_index + 3]))
        # self.axes[1][2].imshow(self._get_contours(self.pen_patches[base_index + 4]))
        plt.show()
        return

    # from https://docs.opencv.org/4.1.0/d1/d32/tutorial_py_contour_properties.html
    def _get_cnt_solidity(self, contour):
        area = cv2.contourArea(contour)
        hull = cv2.convexHull(contour)
        hull_area = cv2.contourArea(hull)
        solidity = 0
        if hull_area != 0:
            solidity = float(area) / hull_area
        return solidity

    def _get_contour_aspect_ratio(self, contour):
        x, y, w, h = cv2.boundingRect(contour)
        aspect_ratio = float(w) / h
        return aspect_ratio

    def _get_contours(self, img, plt_ax):
        color_img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        gray_img = cv2.cvtColor(color_img, cv2.COLOR_BGR2GRAY)
        binary_img = cv2.Canny(gray_img, 30, 200)
        contours, hierarchy = cv2.findContours(binary_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        print(len(contours))
        plt_ax.clear()

        ax1 = []
        filtered_contours = []
        contour_colors = []
        for index, contour in enumerate(contours):
            solidity = self._get_cnt_solidity(contour)

            if solidity > 0.2:
                ax1.append(solidity)
                filtered_contours.append(contour)

                # mask = np.zeros(gray_img.shape, np.uint8)
                # cv2.drawContours(mask, [contour], 0, 255, -1)
                # mean_val = cv2.mean(color_img, mask=mask)
                # ax2.append(mean_val)

                col_r = random.randint(0, 255)
                col_g = random.randint(0, 255)
                col_b = random.randint(0, 255)
                contour_colors.append([col_r / 255.0, col_g / 255.0, col_b / 255.0, 1.0])
                cv2.drawContours(color_img, [contour], -1, (col_r, col_g, col_b), 1)

        plt_ax.scatter(np.arange(0, len(filtered_contours)), ax1, c=np.array(contour_colors))
        # plt_ax.plot([0.5], 'bo')

        # cv2.drawContours(color_img, filtered_contours, -1, (0, 255, 0), -1)

        return color_img

    def on_key(self, event):
        if event.key == 'e':
            self.next()
        elif event.key == 'w':
            self.prev()

    def next(self):
        self.view_index += 1
        self.view()
        pass

    def prev(self):
        self.view_index -= 1
        self.view()
        pass

    def _get_json_poses(self, json_poses):
        with open(json_poses, 'r') as json_file:
            json_data = json.load(json_file)
        json_file.close()

        return {os.path.split(elem)[1]: [json_data[elem]["rvec"], json_data[elem]["tvec"]] for elem in json_data.keys()}


if __name__ == "__main__":
    val_points_path = "L:\\experimental_datasets\\incoming\\test_val\\keypoints_aruco.json"
    val_poses_path = "L:\\experimental_datasets\\incoming\\test_val\\poses_aruco.json"
    vectorizer = FromFileVectorizer(val_points_path)
    pen_tip_keypoints = vectorizer.get_keypoint_path(0)
    pen_samples = vectorizer.get_pen_samples()
    viewer = KeypointViewer(pen_samples, val_poses_path)
    viewer.view()
