import os
from ast import literal_eval

import matplotlib
import pandas as pd
from skimage.io import imread

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


# load all frames/words consecutively
# select start of first word
# to end of last word from one subject
# all bad ones are automatically off
# create dataframe: for each subject, list on frames from start to end of all words
# -> go by subject start/end ids, filter out bad ones


# expects files to be in base_dir/subject_number/patch
class OnOffAnnotator():
    def __init__(self, base_dir, subject_dataframe, on_off_dataframe=None, start_subject=0):
        fig, axes = plt.subplots(ncols=1, nrows=1, figsize=(20, 14))
        self.fig = fig
        self.axes = axes
        self.base_dir = base_dir
        self.subject_dataframe = subject_dataframe
        self.current_subject = start_subject

        if on_off_dataframe is None:
            self.on_off_dataframe = pd.DataFrame()
            self.on_off_dataframe['subject'] = pd.Series(dtype=int)
            self.on_off_dataframe['frame'] = pd.Series(dtype=int)
            self.on_off_dataframe['on'] = pd.Series(dtype=bool)
            self.on_off_dataframe['filename'] = pd.Series(dtype=str)
        else:
            self.on_off_dataframe = on_off_dataframe
            print(self.on_off_dataframe.head())

        self.current_subject_row = self.subject_dataframe.iloc[self.current_subject]
        self.patch_base_path, self.needed_filenames = self._create_current_patch_image_list()
        self.subject_index = 0
        self.current_image = self.load_current_image()
        self.current_marked_frames = []

        fig.canvas.mpl_connect('key_press_event', self.on_key)
        # fig.canvas.mpl_connect('button_press_event', self.on_click)
        pass

    def load_current_image(self):
        return imread(os.path.join(self.patch_base_path, self.needed_filenames[self.subject_index]))

    def _create_current_patch_image_list(self):
        start_frame = self.current_subject_row['start']
        end_frame = self.current_subject_row['end']
        skip_frames = self.current_subject_row['bad']

        base_path = os.path.join(self.base_dir, str(self.current_subject), "patch/")
        all_image_ids = list(range(start_frame, end_frame + 1))
        all_image_ids = [item for item in all_image_ids if item not in skip_frames]

        all_filenames = os.listdir(base_path)
        needed_filenames = []
        for frame in all_image_ids:
            possible_names = [filename for filename in all_filenames if filename.startswith(str(frame) + '_')]

            if len(possible_names) > 1:
                print("Too many possible names!")
            elif len(possible_names) == 0:
                print("Filename not found!!")

            needed_filenames.append(possible_names[0])
        return base_path, needed_filenames

    def next_subject(self):
        if self.current_subject < self.subject_dataframe.subject.count() - 1:
            self.current_subject += 1
            self.current_subject_row = self.subject_dataframe.iloc[self.current_subject]
            self.patch_base_path, self.needed_filenames = self._create_current_patch_image_list()
            self.subject_index = 0
        return

    def prev_subject(self):
        if self.current_subject > 0:
            self.current_subject -= 1
            self.current_subject_row = self.subject_dataframe.iloc[self.current_subject]
            self.patch_base_path, self.needed_filenames = self._create_current_patch_image_list()
            self.subject_index = 0
        return

    def next_image(self):
        if self.subject_index < len(self.needed_filenames) - 1:
            self.subject_index += 1
        else:
            self.next_subject()
            print("Current subject: {}".format(self.subject_index))

        print("Current file: {}".format(self.needed_filenames[self.subject_index]))
        self.current_image = self.load_current_image()
        return

    def prev_image(self):
        if self.subject_index > 0:
            self.subject_index -= 1
        else:
            self.prev_subject()
            print("Current subject: {}".format(self.subject_index))

        print("Current file: {}".format(self.needed_filenames[self.subject_index]))
        self.current_image = self.load_current_image()
        return

    def _clear_axes(self):
        # for ax in self.axes.flatten():
        #    ax.cla()
        self.axes.cla()
        return

    def _view_axes(self):
        self._clear_axes()
        self.axes.imshow(self.current_image)
        # self.axes.plot(self.ptip[0], self.ptip[1], 'g+')
        # self.axes.plot(self.ptail[0], self.ptail[1], 'rx')

    def on_key(self, event):
        if event.key == 'e':
            self.next()
        elif event.key == 'w':
            self.prev()
        elif event.key == 'i':
            self.on_off_dataframe.to_csv('words_to_frames_on_off.csv')
            print("Saved dataframe.")
        elif event.key == '-':
            self.mark_on()
            self.next()
        elif event.key == 'ö':
            self.mark_off()
        elif event.key == 'ä':
            self.mark_on()

    def _get_current_frame_from_filename(self):
        return int(self.needed_filenames[self.subject_index].split('_')[0])

    def _toggle_marking_on_frame(self, marking):
        current_frame = self._get_current_frame_from_filename()

        if ((self.on_off_dataframe['subject'] == self.current_subject) &
            (self.on_off_dataframe['frame'] == current_frame)).any():
            self.on_off_dataframe.loc[(self.on_off_dataframe['subject'] == self.current_subject) & \
                                      (self.on_off_dataframe['frame'] == current_frame), 'on'] = marking
        else:
            self.on_off_dataframe = self.on_off_dataframe.append(
                {'subject': self.current_subject,
                 'frame': current_frame,
                 'on': marking,
                 'filename': self.needed_filenames[self.subject_index]}, ignore_index=True
            )

        print(self.on_off_dataframe.tail())
        return

    def mark_on(self):
        self._toggle_marking_on_frame(True)
        return

    def mark_off(self):
        self._toggle_marking_on_frame(False)
        return

    def view(self):
        self._view_axes()
        plt.show()
        pass

    def next(self):
        self.next_image()
        self.view()
        return

    def prev(self):
        self.prev_image()
        self.view()
        return


if __name__ == '__main__':
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)
    # on_off_frames_stroke_study
    subject_dataframe = pd.read_csv('on_off_frames_stroke_study.csv', index_col=0,
                                    converters={'bad': literal_eval})
    current_annos = pd.read_csv('words_to_frames_on_off.csv', index_col=0)

    viewer = OnOffAnnotator("../tmp_out/record/stroke_study/", subject_dataframe, current_annos, start_subject=5)
    viewer.view()
