import os
import pprint

import matplotlib
import pandas as pd
from skimage.io import imread

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class StrokeStudyVisualizer():
    def __init__(self, result_dataframe, on_points_dataframe, off_points_dataframe, aggregate):
        fig, axes = plt.subplots(ncols=1, nrows=1, figsize=(20, 14))
        self.fig = fig
        self.axes = axes
        self.result_dataframe = result_dataframe
        self.aggregate = aggregate
        self.on_points_df = on_points_dataframe
        self.off_points_df = off_points_dataframe

        self.tp, self.fp, self.tn, self.fn = self._get_series()

        fig.canvas.mpl_connect('key_press_event', self.on_key)
        fig.canvas.mpl_connect('button_press_event', self.on_click)
        pass

    def _get_series(self):
        on_points_idx = pd.Index(self.on_points_df['frame'])
        off_points_idx = pd.Index(self.off_points_df['frame'])

        # first aspect: TP
        correct_on_points = self.result_dataframe.loc[self.result_dataframe['calc_on'] & self.result_dataframe['gt_on']]
        correct_on_points_idx = pd.Index(correct_on_points['frame'])
        correct_on_points_coords = self.on_points_df.loc[on_points_idx.isin(correct_on_points_idx)]

        # second aspect: FP
        false_on_points = self.result_dataframe.loc[self.result_dataframe['calc_on'] & ~self.result_dataframe['gt_on']]
        false_on_points_idx = pd.Index(false_on_points['frame'])
        false_on_points_coords = self.on_points_df.loc[on_points_idx.isin(false_on_points_idx)]

        # third aspect: TN
        correct_off_points = self.result_dataframe.loc[
            ~self.result_dataframe['calc_on'] & ~self.result_dataframe['gt_on']]
        correct_off_points_idx = pd.Index(correct_off_points['frame'])
        correct_off_points_coords = self.off_points_df.loc[off_points_idx.isin(correct_off_points_idx)]

        # fourth aspect: FN
        false_off_points = self.result_dataframe.loc[~self.result_dataframe['calc_on'] & self.result_dataframe['gt_on']]
        false_off_points_idx = pd.Index(false_off_points['frame'])
        false_off_points_coords = self.off_points_df.loc[off_points_idx.isin(false_off_points_idx)]

        tp = (correct_on_points_coords['coord_x'].tolist(), correct_on_points_coords['coord_y'].tolist(),
              correct_on_points_coords['frame'].tolist())
        fp = (false_on_points_coords['coord_x'].tolist(), false_on_points_coords['coord_y'].tolist(),
              false_on_points_coords['frame'].tolist())
        tn = (correct_off_points_coords['coord_x'].tolist(), correct_off_points_coords['coord_y'].tolist(),
              correct_off_points_coords['frame'].tolist())
        fn = (false_off_points_coords['coord_x'].tolist(), false_off_points_coords['coord_y'].tolist(),
              false_off_points_coords['frame'].tolist())

        eval_dict = {
            "true positives": len(tp[2]),
            "false positives": len(fp[2]),
            "true negatives": len(tn[2]),
            "false negatives": len(fn[2]),
            "sum of correct guesses": len(tp[2]) + len(tn[2]),
            "sum of all guesses": len(tp[2]) + len(fp[2]) + len(tn[2]) + len(fn[2]),
            "percentage": (len(tp[2]) + len(tn[2])) / (len(tp[2]) + len(fp[2]) + len(tn[2]) + len(fn[2])) * 100,
            "false negative none points": fn[0].count(-1)
        }

        pprint.pprint(eval_dict)

        fn_none_point_ids = [elem[2] for elem in fn if elem[0] == -1]
        print(fn_none_point_ids)
        # missing points: some points were discarded; we detect them by checking all
        # annotated gt frames against all on/off frames
        combined_on_off = self.on_points_df.append(self.off_points_df)
        combined_idx = pd.Index(combined_on_off['frame'])
        result_idx = pd.Index(self.result_dataframe['frame'])
        missing_points_from_calc = self.result_dataframe.loc[~result_idx.isin(combined_idx)]
        print("Missing points: {}".format(len(missing_points_from_calc)))
        print(missing_points_from_calc.head(n=40))
        return tp, fp, tn, fn

    def _clear_axes(self):
        # for ax in self.axes.flatten():
        #    ax.cla()
        self.axes.cla()
        return

    def _view_axes(self):
        self._clear_axes()
        self.axes.imshow(self.aggregate, cmap='viridis')

        self.axes.scatter(self.tp[0], self.tp[1], label='TP')

        self.axes.scatter(self.fp[0], self.fp[1], label='FP')

        self.axes.scatter(self.tn[0], self.tn[1], label='TN')
        self.axes.scatter(self.fn[0], self.fn[1], label='FN')

        for i, txt in enumerate(self.fn[2]):
            self.axes.annotate(txt, (self.fn[0][i], self.fn[1][i]))

        self.axes.legend()

    def on_click(self, event):
        new_x = int(round(event.xdata))
        new_y = int(round(event.ydata))
        return

    def on_key(self, event):
        if event.key == 'e':
            self.next()
        elif event.key == 'w':
            self.prev()

    def view(self):
        self._view_axes()
        plt.show()
        pass

    def next(self):
        self.view()
        return

    def prev(self):
        self.view()
        return


if __name__ == '__main__':
    subject = 4
    base_dir = "../tmp_out/record/stroke_study/"

    aggregate_img_path = os.path.join(base_dir, str(subject), "subject_{}_aggregate.jpg".format(subject))
    aggregate_img = imread(aggregate_img_path)

    result_dataframe_path = os.path.join(base_dir, 'subject_{}_gt_and_calc.csv'.format(subject))
    result_dataframe = pd.read_csv(result_dataframe_path, index_col=0)

    on_points_path = os.path.join(base_dir, 'subject_{}_on_points.csv'.format(subject))
    off_points_path = os.path.join(base_dir, 'subject_{}_off_points.csv'.format(subject))
    on_points_df = pd.read_csv(on_points_path, index_col=0)
    off_points_df = pd.read_csv(off_points_path, index_col=0)

    visualizer = StrokeStudyVisualizer(result_dataframe, on_points_df, off_points_df, aggregate_img)
    visualizer.view()
