import matplotlib
import numpy as np
from skimage.io import imread
from skimage.transform import resize

from dataset.dataset_utils import get_files_and_keypoints
from dataset.tiplet_sequence import TipletDataGenerator
from dataset_utils import rescale_keypoint
from imgproc.heatmap_keypoint_extractor import HeatmapKeypointExtractor
from model.model_config import ModelConfig
from visual.image_heatmap_file_data_provider import ImageHeatmapFileDataProvider
from visual.sequence_viewer import SequenceViewer

matplotlib.use('TkAgg')


class TestTipletDataGenerator(TipletDataGenerator):
    def _load_images(self, index):
        my_slice = self.image_filenames[
                   index * self.real_image_count_for_batch: (1 + index) * self.real_image_count_for_batch]

        my_imgs = [resize(imread(img), (1080, 1920), preserve_range=True) for img in my_slice]
        return my_imgs

    # returns all keypoints to real images in batch_y format
    def _load_keypoints(self, index):
        keypoints = np.array(
            [kp for kp in
             self.keypoints[index * self.real_image_count_for_batch: (1 + index) * self.real_image_count_for_batch]],
            dtype='float64')
        keypoints = [[rescale_keypoint(kp[0], 3840, 2160, 1920, 1080),
                      rescale_keypoint(kp[1], 3840, 2160, 1920, 1080)] for kp in keypoints]
        return keypoints


class TipletSequenceViewer(SequenceViewer):
    def __init__(self, display_data_provider, batch_size, show_inferred_data=False):
        super().__init__(display_data_provider, batch_size, 8, 4)
        self.index_increment = 1
        self.batch_counter = 0
        self.show_inferred_data = show_inferred_data
        self._load_current_data()
        pass

    def _load_current_data(self):
        if len(self.display_data_provider) > self.batch_counter >= 0:
            self.batch_x, self.batch_y = self.display_data_provider[self.batch_counter]
            if not self.show_inferred_data:
                self.batch_y = np.reshape(self.batch_y,
                                          ((self.batch_y.shape[0], self.batch_y.shape[1], self.batch_y.shape[2])))
        return

    def _view_axes(self):
        self._clear_axes()

        for i in range(4):
            for j in range(4):
                current_batch_index = 4 * i + j

                self.axes[i][j].imshow(self.batch_x[current_batch_index],
                                       interpolation='nearest', alpha=0.99)

                current_image_y = self.batch_y[current_batch_index].astype('float32') / 255.0
                # new_y = np.repeat(current_image_y[:, :, np.newaxis], 4, axis=2)

                tuple = HeatmapKeypointExtractor(current_image_y, radius=3).run()

                overlay = np.zeros((32, 32, 4), dtype="float32")
                overlay[(int(tuple[0]), int(tuple[1]))] = [1, 0, 0, 1]

                # self.axes[i][j].imshow(new_y,
                #                       interpolation='nearest', alpha=1.0, cmap='jet')
                self.axes[i][j].imshow(overlay,
                                       interpolation='nearest', alpha=1.0, cmap='jet')

        for i in range(4):
            for j in range(4):
                self.axes[i + 4][j].imshow(self.batch_y[4 * i + j],
                                           interpolation='nearest')
        pass

    def next(self):
        self.batch_counter += 1
        self._load_current_data()
        self.view()
        return

    def prev(self):
        self.batch_counter -= 1
        self._load_current_data()
        self.view()
        return


if __name__ == '__main__':
    show_inferred_data = True
    if not show_inferred_data:
        image_path = "L:\\large_dataset_study\\first_round_subset\\test\\"
        annotation_json = "L:\\large_dataset_study\\first_round_subset\\test\\first_round_test.json"

        files, keypoints = get_files_and_keypoints(annotation_json, dataset_dir=image_path)

        model_config = ModelConfig('../train/params/tiplet.yaml')

        data_provider = TestTipletDataGenerator(
            files,
            keypoints,
            model_config)
    else:
        image_path = "L:\\large_dataset_study\\training\\tiplet_first_round_subset_training_gauss\\first_round_test_1080p_patches\\"
        heatmap_path = "L:\\large_dataset_study\\training\\tiplet_first_round_subset_training_gauss\\inf_gauss\\"

        data_provider = ImageHeatmapFileDataProvider(image_path, heatmap_path, 16)

    viewer = TipletSequenceViewer(data_provider, 16, show_inferred_data=True)
    viewer.view()
