import numpy as np
from imageio import imsave

from prototype.camera_streamer import DummyCameraStreamer


class BinaryAnalyzer():
    def __init__(self, bin_dir=None, binz_dir=None, start_frame=0, frame_count=0):
        self.bin_dir = bin_dir
        self.binz_dir = binz_dir
        self.start_frame = start_frame
        if frame_count == 0:
            self.frame_count = 1000000
        else:
            self.frame_count = frame_count

        self.bin_frames = None
        self.binz_frames = None
        self.binz_sum = None
        self.bin_sum = None
        self.binz_h, self.binz_w = (0, 0)
        self.bin_h, self.bin_w = (0, 0)
        self.mean_threshold = 0.2

        if self.binz_dir is not None:
            self.binz_frames, self.binz_sum = self._read_dir(self.binz_dir, start_frame, self.frame_count)
            self.binz_h, self.binz_w = self.binz_frames[0].shape[:2]
        if self.bin_dir is not None:
            self.bin_frames, self.bin_sum = self._read_dir(self.bin_dir, start_frame, self.frame_count)
            self.bin_h, self.bin_w = self.bin_frames[0].shape[:2]
        return

    def _read_dir(self, dir, start_frame, frame_count):
        dir_stream = DummyCameraStreamer(dir, counter=start_frame)
        current_img = dir_stream.get()
        frame_sum = np.zeros_like(current_img, dtype='float32')
        frames = []
        counter = 0
        while current_img is not None and counter < frame_count:
            frames.append(current_img)
            if self._has_ok_mean(current_img):
                frame_sum += current_img / 255.
            current_img = dir_stream.get()
            counter += 1
        return frames, frame_sum

    # we check the mean to filter out binarized images which
    # definitely have too many white pixels
    # which means it is a wrong binarization
    def _has_ok_mean(self, frame):
        mean = np.mean(frame)
        if mean > self.mean_threshold * 255:
            return False
        return True

    def _frames_to_float(self, frames):
        frames = np.array(frames, dtype='float32') / 255.
        return frames

    def _get_frame_sum(self, frames):
        frames = self._frames_to_float(frames)
        return np.sum(frames, axis=0)

    def _scale_to_uint8(self, arrays):
        return arrays / np.max(arrays) * 255.

    ######################################################

    def get_binz_frame_sum(self):
        # frames_sum = self._get_frame_sum(self.binz_frames)
        return self._scale_to_uint8(self.binz_sum)

    def get_binz_3d_surface_data(self):
        xx, yy = np.mgrid[0:self.binz_h, 0:self.binz_w]
        z = self.get_binz_frame_sum()
        return xx, yy, z

    def save_binz_frame_sum(self, filepath):
        imsave(filepath, self.get_binz_frame_sum(), quality=100, subsampling=0)
        return


if __name__ == '__main__':
    analyzer = BinaryAnalyzer(binz_dir="L:\\video_analysis\\combined\\binz\\")
    binz_frame_sum = analyzer.get_binz_frame_sum()
    imsave("L:\\video_analysis\\combined\\binz_sum.jpg", binz_frame_sum, quality=100, subsampling=0)
