import os

import matplotlib
import pandas as pd
from skimage.io import imread, imsave

from dataset.dataset_utils import make_none_keypoint
from prototype.camera_streamer import VideoFrameStreamer
from visual.annotation_viewer import AnnotationViewer

matplotlib.use('TkAgg')
import matplotlib.pyplot as plt


class VideoAnnotationViewer(AnnotationViewer):
    def __init__(self, dataframe, video_path, step=1, start_image=None, save_frames=True,
                 save_frames_dir="../tmp_out/"):
        self.video_streamer = VideoFrameStreamer(video_path)
        self.save_frames = save_frames
        self.save_frames_dir = save_frames_dir
        self.video_index = 0
        self.step = step
        super().__init__(dataframe, start_image=start_image)
        pass

    def _load_current_image_and_annos(self):
        print("Current video index: {}".format(self.video_index))
        if self.index >= self.video_streamer.get_frame_count():
            print("Error: Index larger than video frame count.")
            return

        if self.video_index in self.dataframe.frame.values:
            print("no new row")
            current_row = (self.dataframe.loc[self.dataframe['frame'] == self.video_index]).iloc[0]
            print(current_row)

            self.current_image_name = str(self.video_index)
            self.current_image = imread(current_row['full_path'])
            self.ptip = current_row['ptip']
            self.ptail = current_row['ptail']
        else:
            print("new row!")
            self.current_image = self.video_streamer.get_idx(self.video_index)
            self.current_image_name = str(self.video_index)

            full_path = ""
            if self.save_frames:
                full_path = os.path.join(self.save_frames_dir, "{}.jpg".format(self.video_index))
                imsave(full_path, self.current_image, quality=95)

            new_row = {
                'name': self.index,
                'frame': self.video_index,
                'full_path': full_path,
                'ptip': make_none_keypoint(),
                'ptail': make_none_keypoint(),
            }

            line = pd.Series(new_row, name=self.index)
            self.dataframe = self.dataframe.append(line)
            self.dataframe = self.dataframe.sort_index().reset_index(drop=True)

            self.ptip = make_none_keypoint()
            self.ptail = make_none_keypoint()

            print(self.dataframe.tail(n=6))

        return

    def _clear_axes(self):
        self.axes.cla()
        return

    def _view_axes(self):
        self._clear_axes()
        self.axes.imshow(self.current_image)
        self.axes.plot(self.ptip[0], self.ptip[1], 'g+')
        self.axes.plot(self.ptail[0], self.ptail[1], 'rx')

    def on_click(self, event):

        if self.change_ptip:
            new_x = int(round(event.xdata))
            new_y = int(round(event.ydata))
            print("new ptip: [{}, {}]".format(new_x, new_y))
            self.dataframe.iat[self.index, self.dataframe.columns.get_loc('ptip')] = [new_x, new_y]
        if self.change_ptail:
            if self.change_ptip:
                print("ALERT, double keypoint change!")

            new_x = int(round(event.xdata))
            new_y = int(round(event.ydata))
            print("new ptail: [{}, {}]".format(new_x, new_y))
            self.dataframe.iat[self.index, self.dataframe.columns.get_loc('ptail')] = [new_x, new_y]
        return

    def view(self):
        self._view_axes()
        plt.show()
        pass

    def next(self):
        self.change_ptip = False
        self.change_ptail = False
        self.index += 1
        self.video_index += self.step
        self._load_current_image_and_annos()
        self.view()
        return

    def prev(self):
        self.change_ptip = False
        self.change_ptail = False
        if self.index > 0:
            self.index -= 1
            self.video_index -= self.step
        self._load_current_image_and_annos()
        self.view()
        return


if __name__ == '__main__':
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.max_columns', None)

    # current_annos = pd.read_csv('val2_study_annos_10.csv', index_col=0,
    #                           dtype={'frame':np.int32},
    #                           converters={
    #                                  'ptip': literal_eval,
    #                                  'ptail': literal_eval})
    # print(current_annos.head())
    viewer = VideoAnnotationViewer(None, video_path="L:/validation_study_2/videos/P4/p4.mp4", start_image=None, step=1,
                                   save_frames=True, save_frames_dir="L:/validation_study_2/videos/P4/frames/")
    viewer.view()
